#!/bin/bash
# Copyright (C) 2020 Intel Corporation
# For licensing information, see the file 'LICENSE' in the root folder

BUILD=$1
LIB_IRDMA=$2
RDMA_CORE_DIR=$3
UNIFDEF=$4

LIB_RDMA_BLD_TYPE=SI
RDMA_CORE_VER=35.0
LIB_RDMA_PREP_PATCH="irdma-$RDMA_CORE_VER-prep.patch"

if [ "$BUILD" == "MEV_SIMICS" ]; then
  LIB_RDMA_BLD_TYPE=MEV_XEON_SIMICS
  RDMA_CORE_VER=35.0
  LIB_RDMA_PREP_PATCH="irdma-mev-$RDMA_CORE_VER-prep.patch"
fi

if [ "$BUILD" == "MEV_VELOCE" ]; then
  LIB_RDMA_BLD_TYPE=MEV_XEON_VELOCE
  RDMA_CORE_VER=35.0
  LIB_RDMA_PREP_PATCH="irdma-mev-$RDMA_CORE_VER-prep.patch"
fi

if [ "$BUILD" == "MEV_SILICON" ]; then
  LIB_RDMA_BLD_TYPE=MEV_XEON_SILICON
  RDMA_CORE_VER=35.0
  LIB_RDMA_PREP_PATCH="irdma-mev-$RDMA_CORE_VER-prep.patch"
fi

if [ "$BUILD" == "MEV_ACC_SIMICS" ]; then
  LIB_RDMA_BLD_TYPE=MEV_ACC_SIMICS
  RDMA_CORE_VER=35.0
  LIB_RDMA_PREP_PATCH="irdma-mev-acc-$RDMA_CORE_VER-prep.patch"
fi

if [ "$BUILD" == "MEV_ACC_VELOCE" ]; then
  LIB_RDMA_BLD_TYPE=MEV_ACC_VELOCE
  RDMA_CORE_VER=35.0
  LIB_RDMA_PREP_PATCH="irdma-mev-acc-$RDMA_CORE_VER-prep.patch"
fi

if [ "$BUILD" == "MEV_ACC_SILICON" ]; then
  LIB_RDMA_BLD_TYPE=MEV_ACC_SILICON
  RDMA_CORE_VER=35.0
  LIB_RDMA_PREP_PATCH="irdma-mev-acc-$RDMA_CORE_VER-prep.patch"
fi

echo "Patching sources"
cd $RDMA_CORE_DIR
patch -p1 < $LIB_IRDMA/prep-$RDMA_CORE_VER/$LIB_RDMA_PREP_PATCH 
if $? ; then
        echo "ERROR: Could not apply prep-patch!"
        exit 22
fi
mkdir -p $RDMA_CORE_DIR/providers/irdma

echo "Generating additional sources"
cd $LIB_IRDMA/src
make -f build.mk lib-dir BUILD=$LIB_RDMA_BLD_TYPE VERSION=$BUILD LIB_DIR=$RDMA_CORE_DIR/providers/irdma UNIFDEF=$UNIFDEF DRIVER_CORE=$LIB_IRDMA/src/DRIVER_CORE
cp $LIB_IRDMA/prep-$RDMA_CORE_VER/CMakeLists.txt $RDMA_CORE_DIR/providers/irdma/CMakeLists.txt
mv -f $RDMA_CORE_DIR/providers/irdma/irdma-abi.h $RDMA_CORE_DIR/kernel-headers/rdma/irdma-abi.h

echo "List generated sources"
ls -la $RDMA_CORE_DIR/kernel-headers/rdma
ls -la $RDMA_CORE_DIR/providers/irdma
