#!/bin/bash

# Search for only the lines that have the Device IDs and
# then sort by the Device ID...making things MUCH easier!
NETDEV=(`for ID in $(egrep 'VENDOR.*I40E_DEV_ID' $1/src/i40e_main.c | \
	sed 's/.*(\(.*\)).*/\1/') ; do grep -w $ID $1/src/i40e_type.h | \
	awk '{print $3}' | sed 's/0x//g' ; done | sort`)
HW=(`grep I40E_DEV_ID_ $1/src/i40e_type.h |
	sed 's/.*0x\(.*[[:xdigit:]]\).*/\1/' | sort`)
# Do not sort pci.updates (see below)
UPDATES=(`grep "^	[[:xdigit:]]" pci.updates |
	sed 's/^\t\(.[[:xdigit:]]\{,4\}\).*/\1/' | tr a-z A-Z`)
ETHREGS=(`for ID in $(grep I40E_DEV_ID $1/src/ethregs.c |\
	sed 's/.*case \(.*\)/\1/' | sed 's/://') ; do grep -w $ID $1/src/i40e_type.h |\
	awk '{print $3}' | sed 's/0x//g' ; done | sort`)

# Concatenate all entries into a single array;
# there most likely will be repeated entries.
TOTAL[${#TOTAL[*]}]=${NETDEV[*]}
TOTAL[${#TOTAL[*]}]=${HW[*]}
TOTAL[${#TOTAL[*]}]=${UPDATES[*]}
TOTAL[${#TOTAL[*]}]=${ETHREGS[*]}

# Cut the repeated entries and end up with the complete list of device IDs.
TOTAL=(`echo ${TOTAL[*]} | sed 's/\s/\n/g' | sort -u`)
LIMIT=${#TOTAL[*]}

# Indices for all lists (except TOTAL)
J=0
K=0
L=0
M=0

# Compare against the TOTAL list and populate the "missing" lists.
for (( I=0; I<LIMIT; I++ )); do
	if [[ ${TOTAL[$I]} != ${NETDEV[$J]} ]]; then
		NETDEV_MISSING[${#NETDEV_MISSING[*]}]=${TOTAL[$I]}
	else
		(( J++ ))
	fi
	if [[ ${TOTAL[$I]} != ${HW[$K]} ]]; then
		HW_MISSING[${#HW_MISSING[*]}]=${TOTAL[$I]}
	else
		(( K++ ))
	fi
	if [[ ${TOTAL[$I]} != ${UPDATES[$L]} ]]; then
		UPDATES_MISSING[${#UPDATES_MISSING[*]}]=${TOTAL[$I]}
	else
		(( L++ ))
	fi
	if [ -f $1/src/ethregs.c ]; then
		if [[ ${TOTAL[$I]} != ${ETHREGS[$M]} ]]; then
			ETHREGS_MISSING[${#ETHREGS_MISSING[*]}]=${TOTAL[$I]}
		else
			(( M++ ))
		fi
	fi
done

# We're done if all "missing" lists are empty
if [[ ${#NETDEV_MISSING[*]} -eq 0 &&
	${#HW_MISSING[*]} -eq 0 &&
	${#UPDATES_MISSING[*]} -eq 0 &&
	${#ETHREGS_MISSING[*]} -eq 0 ]]; then
	exit 0
fi

# Print out the findings
echo
echo "*******************************************************************************"
echo "* ERROR!!"
echo "*"
echo "* Device ID lists do not all match. "
# Change this list accordingly
echo "i40e_main.c, i40e_type.h, pci.updates, ethregs.c do not all match."

echo "*"

###### Add/remove missing arrays in this area
if [ ${#NETDEV_MISSING[*]} -ne 0 ]; then
	echo "* i40e_main.c is missing ${NETDEV_MISSING[*]}"
fi
if [ ${#HW_MISSING[*]} -ne 0 ]; then
	echo "* i40e_type.h is missing ${HW_MISSING[*]}"
fi
if [ ${#UPDATES_MISSING[*]} -ne 0 ]; then
	if [ ${#UPDATES[*]} -ne ${#TOTAL[*]} ]; then
		echo "* pci.updates is missing ${UPDATES_MISSING[*]}"
	else
		echo "* pci.updates is not sorted in numerical order."
	fi
fi
if [ ${#ETHREGS_MISSING[*]} -ne 0 ]; then
	echo "* ethregs.c is missing ${ETHREGS_MISSING[*]}"
fi
######

echo "*"
echo "*******************************************************************************"
echo
exit 1
