// Copyright (c) 2019 - 2021 Intel Corporation
// For licensing information, see the file 'LICENSE' in the root folder
// IRDMA print-functions handling transforms

@@
expression list E;
@@
(
- irdma_pr_info(E);
+ pr_info(E);
)

@@
expression list E;
@@
(
- irdma_pr_warn(E);
+ pr_warn(E);
)

@@
expression list E;
@@
(
- irdma_pr_err(E);
+ pr_err(E);
)


//ibdev print transformation
@a@
expression mask,str,hw;
expression list E;
@@
(
irdma_debug(hw,mask,str,E);
|
irdma_debug_buf(hw,mask,str,E);
)

@script:python b@
x << a.mask;
y << a.str;
prefix;
z;
@@
if x == "IRDMA_DEBUG_ERR":
        prefix = "ERR"
elif x == "IRDMA_DEBUG_INIT":
        prefix = "INIT"
elif x == "IRDMA_DEBUG_DEV":
        prefix = "DEV"
elif x == "IRDMA_DEBUG_CM":
        prefix = "CM"
elif x == "IRDMA_DEBUG_VERBS":
        prefix = "VERBS"
elif x == "IRDMA_DEBUG_PUDA":
        prefix = "PUDA"
elif x == "IRDMA_DEBUG_ILQ":
        prefix = "ILQ"
elif x == "IRDMA_DEBUG_IEQ":
        prefix = "IEQ"
elif x == "IRDMA_DEBUG_QP":
        prefix = "QP"
elif x == "IRDMA_DEBUG_CQ":
        prefix = "CQ"
elif x == "IRDMA_DEBUG_MR":
        prefix = "MR"
elif x == "IRDMA_DEBUG_PBLE":
        prefix = "PBLE"
elif x == "IRDMA_DEBUG_WQE":
        prefix = "WQE"
elif x == "IRDMA_DEBUG_AEQ":
        prefix = "AEQ"
elif x == "IRDMA_DEBUG_CQP":
	prefix = "CQP"
elif x == "IRDMA_DEBUG_HMC":
        prefix = "HMC"
elif x == "IRDMA_DEBUG_USER":
        prefix = "USER"
elif x == "IRDMA_DEBUG_VIRT":
        prefix = "VIRT"
elif x == "IRDMA_DEBUG_CQE":
	prefix = "CQE"
elif x == "IRDMA_DEBUG_DCB":
	prefix = "DCB"
elif x == "IRDMA_DEBUG_WS":
	prefix = "WS"
elif x == "IRDMA_DEBUG_CLNT":
	prefix = "CLNT"
elif x == "IRDMA_DEBUG_STATS":
	prefix = "STATS"

coccinelle.z = '"' + prefix + ': ' + y.replace('\"','') + '"'

@@
expression a.mask,a.str,a.hw;
expression list E;
identifier b.z;
@@
(
- irdma_debug(hw,mask,str,E);
+ ibdev_dbg(to_ibdev(hw),z,E);
|
- irdma_debug_buf(hw, mask, str, E);
+ print_hex_dump_debug(z,DUMP_PREFIX_OFFSET,16,8,E,false);
)

//IRDMA DMA memory-allocation transforms

@@
expression desc,hw,sz,alignment;
@@
(
- desc.size = sz;
- desc.va = irdma_allocate_dma_mem(&hw,&desc, desc.size, alignment);
+ desc.size = ALIGN(sz, alignment);
+ desc.va  = dma_alloc_coherent(hw.device, desc.size, &desc.pa, GFP_KERNEL);
|
- desc->size = sz;
- desc->va = irdma_allocate_dma_mem(&hw,desc, desc->size, alignment);
+ desc->size = ALIGN(sz, alignment);
+ desc->va  = dma_alloc_coherent(hw.device, desc->size, &desc->pa, GFP_KERNEL);
|
- desc.size = sz;
- desc.va = irdma_allocate_dma_mem(hw,&desc, desc.size, alignment);
+ desc.size = ALIGN(sz, alignment);
+ desc.va  = dma_alloc_coherent(hw->device, desc.size, &desc.pa, GFP_KERNEL);
|
- desc->size = sz;
- desc->va = irdma_allocate_dma_mem(hw,desc, desc->size, alignment);
+ desc->size = ALIGN(sz, alignment);
+ desc->va  = dma_alloc_coherent(hw->device, desc->size, &desc->pa, GFP_KERNEL);
)

//GFP_NOWARN transform - There _must_ be a fallback scheme if the allocation fails to use this
@@
expression desc,hw,sz,alignment;
@@
(
- desc.size = sz;
- desc.va = irdma_allocate_dma_mem_NOWARN(hw,&desc, desc.size, alignment);
+ desc.size = ALIGN(sz, alignment);
+ desc.va  = dma_alloc_coherent(hw->device, desc.size, &desc.pa, GFP_KERNEL | __GFP_NOWARN);
|
- desc->size = sz;
- desc->va = irdma_allocate_dma_mem_NOWARN(hw,desc, desc->size, alignment);
+ desc->size = ALIGN(sz, alignment);
+ desc->va  = dma_alloc_coherent(hw->device, desc->size, &desc->pa, GFP_KERNEL | __GFP_NOWARN);
)

@@
expression hw,desc;
@@
(
- irdma_free_dma_mem(&hw,&desc);
+ dma_free_coherent(hw.device, desc.size, desc.va, desc.pa);
+ desc.va = NULL;
|
- irdma_free_dma_mem(&hw,desc);
+ dma_free_coherent(hw.device, desc->size, desc->va, desc->pa);
+ desc->va = NULL;
|
- irdma_free_dma_mem(hw,&desc);
+ dma_free_coherent(hw->device, desc.size, desc.va, desc.pa);
+ desc.va = NULL;
|
- irdma_free_dma_mem(hw,desc);
+ dma_free_coherent(hw->device, desc->size, desc->va, desc->pa);
+ desc->va = NULL;
)

//Streaming DMA mapping transforms
@@
expression desc,hw,alignment,dir;
identifier label;
@@
(
- desc.va = IRDMA_ALLOC_AND_DMA_MAP_SINGLE(hw, &desc, desc.size, alignment, dir);
- if (!desc.va)
-	goto label;
+ desc.va = kzalloc(desc.size, GFP_KERNEL);
+ if (!desc.va)
+	goto label;
//'alignment' ignored.
+ desc.pa = dma_map_single(hw->device, desc.va, desc.size, dir);
+ if (dma_mapping_error(hw->device, desc.pa)) {
+ 	kfree(desc.va);
+ 	goto label;
+ }
|
- IRDMA_FREE_AND_DMA_UNMAP_SINGLE(hw, &desc, dir);
+ dma_unmap_single(hw->device, desc.pa, desc.size, dir);
+ kfree(desc.va);
|
- IRDMA_DMA_SYNC_SINGLE_FOR_CPU(hw, &desc, dir);
+ dma_sync_single_for_cpu(hw->device, desc.pa, desc.size, dir);
|
- IRDMA_DMA_SYNC_SINGLE_FOR_DEVICE(hw, &desc, dir);
+ dma_sync_single_for_device(hw->device, desc.pa, desc.size, dir);
|
- IRDMA_DMA_MAPPING_ERROR(hw, &desc)
+ dma_mapping_error(hw->device, desc.pa)
)

//IRDMA virtual memory-allocation transforms

@@
expression hw,desc,sz;
@@
(
- desc.va = irdma_allocate_virt_mem(hw,&desc,sz);
+ desc.size = sz;
+ desc.va = kzalloc(desc.size, GFP_KERNEL);
)

@@
expression hw,desc,sz;
@@
(
- desc->va = irdma_allocate_virt_mem(hw,desc,sz);
+ desc->size = sz;
+ desc->va = kzalloc(desc->size, GFP_KERNEL);
)

@@
expression hw,desc,sz;
@@
(
- desc.va = irdma_allocate_virt_mem_atomic(hw,&desc,sz);
+ desc.size = sz;
+ desc.va = kzalloc(desc.size, GFP_ATOMIC);
)

@@
expression hw,desc,sz;
@@
(
- desc->va = irdma_allocate_virt_mem_atomic(hw,desc,sz);
+ desc->size = sz;
+ desc->va = kzalloc(desc->size, GFP_ATOMIC);
)

@@
expression hw,desc;
@@
(
- irdma_free_virt_mem(hw,&desc);
+ kfree(desc.va);
)

@@
expression hw,desc;
@@
(
- irdma_free_virt_mem(hw,desc);
+ kfree(desc->va);
)

@@
expression hw,chunk;
@@
(
- irdma_prm_rem_bitmapmem(hw,chunk);
+ kfree(chunk->bitmapmem.va);
)

@@
expression hw, mem_size, va;
@@
(
- va = irdma_vzalloc(&hw, mem_size);
+ va = vzalloc(mem_size);
)

@@
expression hw, mem_va;
@@
(
- irdma_vfree(&hw, mem_va);
+ vfree(mem_va);
)

@@
expression rf, msix, devid;
@@
(
- irdma_free_irq(rf, msix, devid);
+ free_irq(msix->irq, devid);
)

@@
expression result, rf, msix, handler, flags, name, arg;
@@
(
- result = irdma_irq_request(rf, msix, handler, flags, name, arg);
+ result = request_irq(msix->irq, handler, flags, name, arg);
)

@@
expression rf, msix, status;
@@
(
- status = irdma_irq_set_affinity(rf, msix);
+ cpumask_clear(&msix->mask);
+ cpumask_set_cpu(msix->cpu_affinity, &msix->mask);
+ irq_update_affinity_hint(msix->irq, &msix->mask);
)

@@
expression rf, msix;
@@
(
- irdma_irq_unset_affinity(rf, msix);
+ irq_update_affinity_hint(msix_vec->irq, NULL);
)

//container_of transforms

@@
expression desc,entry;
@@
(
- desc = irdma_node_from_entry(entry);
+ desc = container_of(entry, struct irdma_ws_node, siblings);
)

@@
expression entry;
@@
(
- return irdma_qp_from_entry(entry);
+ return container_of(entry, struct irdma_sc_qp, list);
)

@@
expression entry;
@@
(
- LIST_NEXT(&entry->clist)
+ entry->clist.next
|
- LIST_FIRST(&entry->clist)
+ entry->clist.next
)

@@
expression entry, head;
@@
(
- entry = LIST_NEXT(&head);
+ entry = head.next;
|
- entry = LIST_FIRST(&head);
+ entry = head.next;
)

@@
expression entry, head;
@@
(
- entry = LIST_NEXT(head);
+ entry = head->next;
|
- entry = LIST_FIRST(head);
+ entry = head->next;
)

@@
expression node, parent, type, siblings;
@@
(
- list_for_each_entry(node, &parent, type, siblings)
+ list_for_each_entry(node, &parent, siblings)
)

// Shared code OS primitive transforms. NOPs in linux.
@@
expression E;
@@
(
- U32(E)
+ E
|
- SIZE_T(E)
+ E
)

@r exists@
iterator name LIST_FOR_EACH_ENTRY;
iterator name list_for_each_entry;
iterator name LIST_FOR_EACH_ENTRY_SAFE;
iterator name list_for_each_entry_safe;
statement S;
@@
(
LIST_FOR_EACH_ENTRY(...) S
|
list_for_each_entry(...) S
|
LIST_FOR_EACH_ENTRY_SAFE(...) S
|
list_for_each_entry_safe(...) S
)

@@
expression iter,head,stype,member;
statement S;
@@
(
- LIST_FOR_EACH_ENTRY(iter,head,stype,member)
+ list_for_each_entry(iter,head,member)
S
)

@@
expression iter,tmp,head,stype,member;
statement S;
@@
(
- LIST_FOR_EACH_ENTRY_SAFE(iter,tmp,head,stype,member)
+ list_for_each_entry_safe(iter,tmp,head,member)
S
)

@@
expression addr,val;
@@
(
- wr32v(addr, val)
+ writel(val, addr)
)

@@
expression hw,page,offset,sz,direction;
@@
(
- irdma_dma_map_page(hw, page, offset, sz, direction)
+ dma_map_page(hw->device, page, offset, sz, direction)
)

@@
expression hw,dma_handle,sz,direction;
@@
(
- irdma_dma_unmap_page(hw, dma_handle, sz, direction)
+ dma_unmap_page(hw->device, dma_handle, sz, direction)
)

@@
expression hw,dma_handle;
@@
(
- irdma_dma_mapping_error(hw, dma_handle)
+ dma_mapping_error(hw->device, dma_handle)
)

@@
expression val, shift;
@@
(
- I40E_MASK(val, shift)
+ (val << shift)
)

//RCU transforms
@@
expression lock;
@@
(
- RCU_READ_UNLOCK(lock)
+ rcu_read_unlock()
|
- RCU_READ_LOCK(lock)
+ rcu_read_lock()
)

@@
expression tbl, list;
@@
(
- HASH_DEL_RCU(tbl, &list);
+ hash_del_rcu(&list);
)

@@
expression tbl, list;
@@
(
- HASH_DEL(tbl, &list);
+ hash_del(&list);
)

// argument drop
@@
expression timer, cb, flag, timer_data;
@@
(
- timer_setup(&timer, cb, flag, timer_data);
+ timer_setup(&timer, cb, flag);
)
