#!/usr/bin/perl
#
# Keep one or more C pre-processor defines unchanged in source code,
# stripping all other pre-processor macros and its associated code.
# Correctly handles ifdef and ifndef statements, including else blocks.
#
# Usage:  Reads from stdin, writes to stdout, symbols to check on cmd line.
#          To strip a symbol, but treat it as defined,
#          precede the name with a '-'.
#
# Example: strip.pl KEEP_SYM -STRIP_SYM < input.c > output.c
#		

use strict;

my ( %tags, @stack, @unknowns, $print, $specified );

for ( @ARGV ) { $tags{$_} = s/^-// ? 0 : 1; }
( $specified, $print ) = ( 0, 1 );

LOOP: while ( <STDIN> ) {
	SWITCH: {
		/^\s*?#if/ and do {
			splice @stack, @stack, 0, $print, $specified;
			# Defaulting #if macros to be "specified" because the
			# only time we use them is to specify kernel version...
			# which is an ugly flag by itself, and we'd rather not
			# include them in the build.mk.
			if ( /^\s*?#if\s+?/ and $print ) {
				$specified = 1;
				last SWITCH;
			}
			if ( /^\s*?#if(n??)def\s+(\S+)/ ) {
				if ( exists $tags{$2} ) {
					if ( defined $tags{$2} and $print ) {
						$specified = 1;
						last SWITCH;
					}
				} else {
					push @unknowns, $2;
				}
			}
			$specified = 0;
			$print = $1 ? 1 : 0;
			next LOOP; 
		};

		/^\s*?#else/ and do {
			if ( $specified || $stack[$#stack-1] == 0 ) {
				last SWITCH;
			}
			$print = $print ? 0 : 1;
			next LOOP; 
		};

		/^\s*?#endif/ and do {
			if ( $specified ) {
				( $print, $specified ) = splice @stack, -2, 2;
				last SWITCH;
			}
			( $print, $specified ) = splice @stack, -2, 2;
			next LOOP;
		};
	}
	print if $print;
}

if ( $#unknowns >= 0 ) {
	open (UNKNOWNS, ">>Unknowns");
	foreach ( @unknowns ) { print UNKNOWNS "$_\n"; }
	close (UNKNOWNS);
	`sort Unknowns -u -o THESE_UNKNOWN_FLAGS_NEED_PLACING_IN_BUILD_MK`;
	unlink ("./Unknowns");
}

