#!/usr/bin/perl

# include the definitions of some driver specific strings
require SpecSetup;

print <<EOF
%{?!license: %define license Unknown}
%{?!_dkmsdir: %define _dkmsdir /var/lib/dkms}
%{?!_srcdir: %define _srcdir %_prefix/%{name}}
%{?!install_tree: %define install_tree /lib/modules}
%{?!kernel_test: %define kernel_test \$(uname -r)}

Summary:	$dkms_summary
Name:		$name
Version:	$driver_version
Vendor:		Intel Corporation
URL:		$URL
License:	GPL
Release:	1dkms
BuildArch:	noarch
Group:		System/Kernel
Provides:	%{name}
Requires:	dkms >= 1.95
BuildRequires:	dkms
Source0:	%{name}-%{version}-mktarball.dkms.tgz
Source1:	README
Source2:	COPYING
Source3:	pci.updates
BuildRoot:	%{_tmppath}/%{name}-%{version}-%{release}-root/
Requires: kernel, fileutils, findutils, gawk, bash

%description
$dkms_description

%prep
/usr/sbin/dkms mktarball -m %{module_name} -v %{version} %{mktarball_line} --archive `basename %{SOURCE0}`
cp -f %{_srcdir}/%{name}-%{version}/`basename %{SOURCE1}` %{SOURCE1}
cp -f %{_srcdir}/%{name}-%{version}/`basename %{SOURCE2}` %{SOURCE2}
cp -f %{_srcdir}/%{name}-%{version}/`basename %{SOURCE3}` %{SOURCE3}
cp -af %{_dkmsdir}/%{module_name}/%{version}/tarball/`basename %{SOURCE0}` %{SOURCE0}

%install
if [ "\$RPM_BUILD_ROOT" != "/" ]; then
        rm -rf \$RPM_BUILD_ROOT
fi
mkdir -p \$RPM_BUILD_ROOT/%{_srcdir}/%{module_name}-%{version}/
install -m 644 %{SOURCE1} \$RPM_BUILD_ROOT/%{_srcdir}/%{module_name}-%{version}
install -m 644 %{SOURCE2} \$RPM_BUILD_ROOT/%{_srcdir}/%{module_name}-%{version}
install -m 644 %{SOURCE3} \$RPM_BUILD_ROOT/%{_srcdir}/%{module_name}-%{version}
install -m 644 %{SOURCE0} \$RPM_BUILD_ROOT/%{_srcdir}/%{module_name}-%{version}

%clean
if [ "\$RPM_BUILD_ROOT" != "/" ]; then
        rm -rf \$RPM_BUILD_ROOT
fi

%files
%defattr(-,root,root)
/usr/src/%{module_name}-%{version}/

%post
#Copy driver documents to document directory
DOCSDIR="/usr/share/doc/%{name}-%{version}"
if [ -f /etc/SuSE-release ]; then
	DOCSDIR="/usr/share/doc/packages/%{name}-%{version}"
fi
if [ ! -d "\$DOCSDIR" ]; then
	mkdir -p \$DOCSDIR
fi
cp %{_srcdir}/%{name}-%{version}/README \$DOCSDIR/
cp %{_srcdir}/%{name}-%{version}/COPYING \$DOCSDIR/
cp %{_srcdir}/%{name}-%{version}/pci.updates \$DOCSDIR/

# Determine current arch / kernel
[ `uname -m` == "x86_64" ] && [ `cat /proc/cpuinfo | grep -c "Intel"` -gt 0 ] && [ -e /etc/redhat-release ] && [ `grep -c "Taroon" /etc/redhat-release` -gt 0 ] && c_arch="ia32e" || c_arch=`uname -m`
c_kern=`uname -r`

# Load prebuilt binaries
echo -e ""
echo -e "Loading kernel module source and prebuilt module binaries (if any)"
dkms ldtarball --archive %{_srcdir}/%{module_name}-%{version}/%{module_name}-%{version}-mktarball.dkms.tgz >/dev/null 2>&1

# Install prebuilt binaries
echo -e "Installing prebuilt kernel module binaries (if any)"
IFS='
'
for kern in `dkms status -m %{module_name} -v %{version} -a \$c_arch | grep ": built" | awk {'print \$3'} | sed 's/,\$//'`; do
	dkms install -m %{module_name} -v %{version} -k \$kern -a \$c_arch >/dev/null 2>&1
done
unset IFS

# Define where pci.table is located (if exists)
PCITABLE="/dev/null"
if [ -f /usr/share/hwdata/pcitable ]; then
	PCITABLE="/usr/share/hwdata/pcitable"
fi
# Define where pci.ids is located
PCIIDS="/usr/share/hwdata/pci.ids"
if [ -f /etc/SuSE-release ]; then
	PCIIDS="/usr/share/pci.ids"
fi
# Save original pci.* files before we add our new devices
echo "original pci.ids saved in /usr/local/share/%{name}\$PCIIDS";
if [ "\$PCITABLE" != "/dev/null" ]; then
	echo "original pcitable saved in /usr/local/share/%{name}\$PCITABLE";
fi
mkdir -p /usr/local/share/%{name}
cp --parents \$PCIIDS /usr/local/share/%{name}/
if [ "\$PCITABLE" != "/dev/null" ]; then
	cp --parents \$PCITABLE /usr/local/share/%{name}/
fi

LD="%{_srcdir}/%{name}-%{version}";
#Yes, this really needs bash
bash -s \$PCIIDS \\
	\$PCITABLE \\
	\$LD/pci.updates \\
	\$LD/pci.ids.new \\
	\$LD/pcitable.new \\
	%{name} \\
<<"END"
EOF
;
open(SCRIPT, "scripts/pciids_update.bash") or die;
while (<SCRIPT>) {
	print $_;
}
close(SCRIPT);
print <<EOF
END

mv -f \$LD/pci.ids.new  \$PCIIDS
if [ "\$PCITABLE" != "/dev/null" ]; then
mv -f \$LD/pcitable.new \$PCITABLE
fi

echo -e ""
echo -e "Your DKMS tree now includes:"
dkms status -m %{module_name} -v %{version}

exit 0

%preun
echo -e
echo -e "Uninstalling %{module_name} module (version %{version})..."
dkms remove -m %{module_name} -v %{version} --all --rpm_safe_upgrade
exit 0

%postun
rm -rf %{_srcdir}/%{name}-%{version}
rm -rf %{_docdir}/%{name}-%{version}
uname -r | grep BOOT || /sbin/depmod -a > /dev/null 2>&1 || true

%changelog
* %(date "+%a %b %d %Y") Intel(R) Corporation <linux.nics\@intel.com>
- Automatic build of DKMS rpm %{name}-%{version}-%{release}
EOF
