#!/usr/bin/perl

# use diagnostics;

# include the definitions of some driver specific strings
require SpecSetup;

print <<EOF;
Name: $name
Summary: $summary
Version: $driver_version
Release: 1
Source: %{name}-%{version}.tar.gz
Vendor: Intel Corporation
License: GPL
ExclusiveOS: linux
Group: System Environment/Kernel
Provides: %{name}
URL: $URL
BuildRoot: %{_tmppath}/%{name}-%{version}-root
# do not generate debugging packages by default - newer versions of rpmbuild
# may instead need:
#%define debug_package %{nil}
%debug_package %{nil}
# macros for finding system files to update at install time (pci.ids, pcitable)
%define find() %(for f in %*; do if [ -e \$f ]; then echo \$f; break; fi; done)
%define _pciids   /usr/share/pci.ids        /usr/share/hwdata/pci.ids
%define _pcitable /usr/share/kudzu/pcitable /usr/share/hwdata/pcitable /dev/null
%define pciids    %find %{_pciids}
%define pcitable  %find %{_pcitable}
Requires: kernel, fileutils, findutils, gawk, bash

%description
$description

%prep
%setup

%build
make

%install
make INSTALL_MOD_PATH=%{buildroot} MANDIR=%{_mandir} install
# Append .new to driver name to avoid conflict with kernel RPM
cd %{buildroot}
find lib -name "$name.*o" -exec mv {} {}.new \\; \\
         -fprintf %{_builddir}/%{name}-%{version}/file.list "/%p.new\\n"
$special_install

%clean
rm -rf %{buildroot}

%files -f file.list
%defattr(-,root,root)
%{_mandir}/man7/$name.7.gz
%doc COPYING
%doc README
%doc file.list
%doc pci.updates

%post
FL="%{_docdir}/%{name}-%{version}/file.list
    %{_docdir}/%{name}/file.list"
FL=\$(for d in \$FL ; do if [ -e \$d ]; then echo \$d; break; fi;  done)

if [ -d /usr/local/lib/%{name} ]; then
	rm -rf /usr/local/lib/%{name}
fi
if [ -d /usr/local/share/%{name} ]; then
	rm -rf /usr/local/share/%{name}
fi

# Save old drivers (aka .o and .o.gz)
echo "original pci.ids saved in /usr/local/share/%{name}";
if [ "%{pcitable}" != "/dev/null" ]; then
	echo "original pcitable saved in /usr/local/share/%{name}";
fi
for k in \$(sed 's/\\/lib\\/modules\\/\\([0-9a-zA-Z_\\.\\-]*\\).*/\\1/' \$FL) ; 
do
	d_drivers=/lib/modules/\$k
	d_usr=/usr/local/share/%{name}/\$k
	mkdir -p \$d_usr
	cd \$d_drivers; find . -name %{name}.*o -exec cp --parents {} \$d_usr \\; -exec rm -f {} \\;
	cd \$d_drivers; find . -name %{name}_*.*o -exec cp --parents {} \$d_usr \\; -exec rm -f {} \\;
	cd \$d_drivers; find . -name %{name}.*o.gz -exec cp --parents {} \$d_usr \\; -exec rm -f {} \\;
	cd \$d_drivers; find . -name %{name}_*.*o.gz -exec cp --parents {} \$d_usr \\; -exec rm -f {} \\;
	cp --parents %{pciids} /usr/local/share/%{name}/
	if [ "%{pcitable}" != "/dev/null" ]; then
		cp --parents %{pcitable} /usr/local/share/%{name}/
	fi
done

# Add driver link
for f in \$(sed 's/\\.new\$//' \$FL) ; do
	ln -f \$f.new \$f 
done

# Check if kernel version rpm was built on IS the same as running kernel
BK_LIST=\$(sed 's/\\/lib\\/modules\\/\\([0-9a-zA-Z_\\.\\-]*\\).*/\\1/' \$FL)
MATCH=no
for i in \$BK_LIST
do
	if [ \$(uname -r) == \$i ] ; then
		MATCH=yes
		break
	fi
done
if [ \$MATCH == no ] ; then
	echo -n "WARNING: Running kernel is \$(uname -r).  "
	echo -n "RPM supports kernels (  "
	for i in \$BK_LIST
	do
		echo -n "\$i  "
	done
	echo ")"
fi

LD="%{_docdir}/%{name}";
if [ -d %{_docdir}/%{name}-%{version} ]; then
	LD="%{_docdir}/%{name}-%{version}";
fi

#Yes, this really needs bash
bash -s %{pciids} \\
	%{pcitable} \\
	\$LD/pci.updates \\
	\$LD/pci.ids.new \\
	\$LD/pcitable.new \\
	%{name} \\
<<"END"
EOF
;
open(SCRIPT, "scripts/pciids_update.bash") or die;
while (<SCRIPT>) {
	print $_;
}
close(SCRIPT);
print <<EOF
END

mv -f \$LD/pci.ids.new  %{pciids}
if [ "%{pcitable}" != "/dev/null" ]; then
mv -f \$LD/pcitable.new %{pcitable}
fi

uname -r | grep BOOT || /sbin/depmod -a > /dev/null 2>&1 || true

%preun
# If doing RPM un-install
if [ \$1 -eq 0 ] ; then
	FL="%{_docdir}/%{name}-%{version}/file.list
    		%{_docdir}/%{name}/file.list"
	FL=\$(for d in \$FL ; do if [ -e \$d ]; then echo \$d; break; fi;  done)

	# Remove driver link
	for f in \$(sed 's/\\.new\$//' \$FL) ; do
		rm -f \$f
	done

	# Restore old drivers
	if [ -d /usr/local/share/%{name} ]; then
		cd /usr/local/share/%{name}; find . -name \'%{name}.*o*\' -exec cp --parents {} /lib/modules/ \\;
		cd /usr/local/share/%{name}; find . -name \'%{name}_*.*o*\' -exec cp --parents {} /lib/modules/ \\;
		rm -rf /usr/local/share/%{name}
	fi
fi

%postun
uname -r | grep BOOT || /sbin/depmod -a > /dev/null 2>&1 || true

EOF
