#!/usr/bin/perl
#
# Strip out one or more C pre-processor defines from source code,
# passing all other pre-processor macros through unchanged.
# Correctly handles ifdef and ifndef statements, including else blocks.
#
# Usage:  Reads from stdin, writes to stdout, symbols to check on cmd line.
#          To process a symbol, but treat it as undefined,
#          precede the name with a '-'.
#
# Example: strip.pl DEFINED_SYM -UNDEFINED_SYM < input.c > output.c
#		

use strict;

my ( %tags, @stack, $print, $specified );

for ( @ARGV ) { $tags{$_} = s/^-// ? 0 : 1; }
( $specified, $print ) = ( 0, 1 );

LOOP: while ( <STDIN> ) {
	SWITCH: {
		/^#if/ and do {
			splice @stack, @stack, 0, $print, $specified;
			if ( /#if(n??)def\s+(\S+)/ and defined $tags{$2} and $print ) {
				$print = $1 ? $tags{$2} ? 0 : 1 : $tags{$2};
				$specified = 1;
				next LOOP;
			}
			$specified = 0;
			last SWITCH;
		};

		/^#else/ and do {
			if ( $specified ) {
				$print = $print ? 0 : 1;
				next LOOP; 
			}
			last SWITCH;
		};

		/^#endif/ and do {
			if ( $specified ) {
				( $print, $specified ) = splice @stack, -2, 2;
				next LOOP;
			}
			( $print, $specified ) = splice @stack, -2, 2;
			last SWITCH;
		};
	}
	print if $print;
}

