/* Copyright (c) 2019 Intel Corporation */
#ifndef LICENSE_REPLACED
/* For licensing information, see the file 'LICENSE' in the root folder */
#endif /* LICENSE_REPLACED */
#ifndef IRDMA_DBG_H
#define IRDMA_DBG_H

#include <net/genetlink.h>
#include "irdma_dbg_k_u.h"

#define IRDMADBG_EVENT_CFG_WAITLIST_MAX_SIZE 100
#define IRDMADBG_MESSAGE_RESERVE_BYTES 40

/*
 * List element
 */
struct icrdmadbg_device {
	struct irdma_device *iwdev;
	struct list_head node;
};

/*
 * Holds event subscription information per irdma_device
 */
struct icrdmadbg_evt_ctx {
	int ifindex;
	/* indicates subsciption for given event
	 *  spin_lock are not used for evt_ref_cnt and
	 * client_ref_cnt as they are protected by global
	 * icrdmadbg_evt_ctx_wait_list spinlock
	 */
	int evt_ref_cnt[IRDMADBG_EVENT_TYPE_COUNT];
	int client_ref_cnt;
	struct list_head node;
};

/*
 * CQP event info data to be send to user app
 */
struct icrdmadbg_cqp_event_info {
	u32 op_ret_val;
	u16 maj_err_code;
	u16 min_err_code;
	u8 op_code;
	bool error;
};

/*
 * Holds info until "dump it" callback is finnished
 */
struct icrdmadbg_hmc_dump_ctx {
	struct irdma_device *iwdev;
	u32 ifindex;
	u32 dump_index;
	u32 dump_cnt;
	u32 hmc_obj_type;
	u32 hmc_obj_size;
	u32 max_obj;
	u32 buffer_len;
	int hmc_obj_id_req;
	bool qp_freeze;
	bool qp_raw;
	bool valid_only;
	bool no_more_data_to_dump;
	bool no_cqp;
	unsigned long *resource_status_array;
	void *buffer;
};

int icrdmadbg_init(void);
void icrdmadbg_deinit(void);
int icrdmadbg_register_device(struct irdma_device *iwdev);
void icrdmadbg_unregister_device(struct irdma_device *iwdev);
int icrdmadbg_process_cqp_event(struct irdma_device *iwdev,
			       struct irdma_ccq_cqe_info *cqe_info);
int icrdmadbg_process_adapater_status_event(struct net *net,
					   enum
					   icrdmadbg_event_adapter_status_evt_id
					   event_id, int ifindex);
#endif				/* _IRDMA_DBG_H */
