/* Copyright (c) 2019 Intel Corporation */
#ifndef LICENSE_REPLACED
/* For licensing information, see the file 'LICENSE' in the root folder */
#endif /* LICENSE_REPLACED */
#ifndef IRDMA_DBG_K_U_H
#define IRDMA_DBG_K_U_H

#define IRDMADBG_GENL_NAME "icrdmadbg"
#define IRDMADBG_GENL_VERSION 0x1
#define IRDMADBG_GENL_EVENTS_MC_GROUP_NAME "events"
#define IRDMADBG_QP_CTX_STRUCTERED_SIZE 128
#define IRDMADBG_APBVT_PART_SIZE 2048
#define IRDMADBG_APBVT_PART_CNT 4

/* commmands */
enum {
	IRDMADBG_CMD_REG,
	IRDMADBG_CMD_HMC,
	IRDMADBG_CMD_EVENT_CFG,	/* sent from monitor app at the beginning
				 *                                of monitor session to configure driver
				 *                                for sending event types requested
				 *                                for particular interface,
				 *                                and during session close to inform
				 *                                the driver that it does not need
				 *                                these types of events anymore */
	IRDMADBG_CMD_EVENT,	/* sent asynchronously from driver
				 *               to monitor app */
	IRDMADBG_CMD_COUNT,
};
#define IRDMADBG_CMD_MAX (IRDMADBG_CMD_COUNT - 1)

/*attributes*/
enum {
	IRDMADBG_ATTR_UNSPEC,
	IRDMADBG_ATTR_IFINDEX,	/* u32 */
	IRDMADBG_ATTR_REG,	/* nested */
	IRDMADBG_ATTR_HMC,	/* nested */
	IRDMADBG_ATTR_EVENT_CFG,	/* nested */
	IRDMADBG_ATTR_EVENT,	/* nested */
	IRDMADBG_ATTR_COUNT,
};
#define IRDMADBG_ATTR_MAX (IRDMADBG_ATTR_COUNT - 1)

/* reg operation types */
enum icrdmadbg_reg_op_type {
	IRDMADBG_REG_OP_READ,
	IRDMADBG_REG_OP_WRITE,
	IRDMADBG_REG_OP_SET_BITS,
	IRDMADBG_REG_OP_CLEAR_BITS,
};

/* reg nested attributes */
enum {
	IRDMADBG_ATTR_REG_UNSPEC,	/* u8, value from icrdmadbg_reg_op_type */
	IRDMADBG_ATTR_REG_OP_TYPE,	/* u32 */
	IRDMADBG_ATTR_REG_ADDR,	/* u32, stores either reg data or mask */
	IRDMADBG_ATTR_REG_VALUE,
	IRDMADBG_ATTR_REG_COUNT,
};
#define IRDMADBG_ATTR_REG_MAX (IRDMADBG_ATTR_REG_COUNT - 1)

/* hmc object status values */
enum icrdmadbg_hmc_obj_status {
	IRDMADBG_HMC_OBJ_VALID,
	IRDMADBG_HMC_OBJ_NOT_VALID,
	IRDMADBG_HMC_OBJ_STATUS_UNKNOWN,	/* Used for hmc objects that are not tracked by driver */
};

enum {
	IRDMADBG_ATTR_HMC_UNSPEC,
	/* request */
	IRDMADBG_ATTR_HMC_OBJ_TYPE,	/* u32, values in sync with shared code I40e_hmc_iw_rsrc_type */
	IRDMADBG_ATTR_HMC_OBJ_ID_REQ,	/* u32, object ID requested */
	IRDMADBG_ATTR_HMC_VALID_ONLY,	/* flag */
	IRDMADBG_ATTR_HMC_QP_NO_CQP,	/* flag, driver only HMC dump */
	IRDMADBG_ATTR_HMC_QP_FREEZE,	/* flag */
	IRDMADBG_ATTR_HMC_QP_RAW,	/* flag */
	/* response */
	IRDMADBG_ATTR_HMC_OBJ_SIZE,	/* u32 */
	IRDMADBG_ATTR_HMC_OBJ_CNT,	/* u32,set only if IRDMADBG_ATTR_HMC_OBJ_ID_REQ not specified */
	IRDMADBG_ATTR_HMC_OBJ_START_ID,	/* u32,set only if IRDMADBG_ATTR_HMC_OBJ_ID_REQ not specified */
	IRDMADBG_ATTR_HMC_OBJ_DATA,	/* binary, sotres objects contents and info associated with them */
	IRDMADBG_ATTR_HMC_COUNT,
};
#define IRDMADBG_ATTR_HMC_MAX (IRDMADBG_ATTR_HMC_COUNT - 1)

/* hmc object list nested attributes */
enum {
	IRDMADBG_ATTR_HMC_ITEM_OBJ_UNSPEC,	/* nested, single dumped object item */
	IRDMADBG_ATTR_HMC_ITEM_OBJ,
	IRDMADBG_ATTR_HMC_ITEM_OBJ_COUNT,
};
#define IRDMADBG_ATTR_HMC_ITEM_OBJ_MAX (IRDMADBG_ATTR_HMC_ITEM_OBJ_COUNT - 1)

/* monitor states */
enum icrdmadbg_event_monitor_state {
	IRDMADBG_EVENT_MONITOR_START,	/* set when monitor sends EVENT_CFG
					 *                              at the beginning of its session */
	IRDMADBG_EVENT_MONITOR_STOP,	/* set when monitor sends EVENT_CFG
					 *                              during session close */
};

/* event types */
enum icrdmadbg_event_type {
	IRDMADBG_EVENT_TYPE_AEQ,
	IRDMADBG_EVENT_TYPE_CQ0_ERR,
	IRDMADBG_EVENT_TYPE_CQ0_ALL,
	IRDMADBG_EVENT_TYPE_CEQ_ERR,
	IRDMADBG_EVENT_TYPE_CEQ_ALL,
	IRDMADBG_EVENT_TYPE_AQ,
	IRDMADBG_EVENT_TYPE_COUNT,
	IRDMADBG_EVENT_TYPE_ADAPTER_STATUS,	/* used only in event (driver->app) */
};

/* event specific queue types */
enum icrdmadbg_event_queue_type {
	IRDMADBG_EVENT_QUEUE_TYPE_CEQ,
	IRDMADBG_EVENT_QUEUE_TYPE_CQ,
	IRDMADBG_EVENT_QUEUE_TYPE_QP,
};

/* event config nested attributes */
enum {
	IRDMADBG_ATTR_EVENT_CFG_UNSPEC,
	IRDMADBG_ATTR_EVENT_CFG_MONITOR_STATE,	/* u32, values from
						 *                                      icrdmadbg_event_monitor_state
						 */
	IRDMADBG_ATTR_EVENT_CFG_LIST_EVT_TYPE,	/* nested */
	IRDMADBG_ATTR_EVENT_CFG_QUEUE_TYPE,	/* u32, values from
						 *                              icrdmadbg_event_queue_type,
						 *                              CEQ events specific */
	IRDMADBG_ATTR_EVENT_CFG_QUEUE_ID,	/* u32, CEQ events specific */
	IRDMADBG_ATTR_EVENT_CFG_IF_NOT_FOUND,	/* flag,
						 *                                         set to TRUE by driver
						 *                                         in case no iw_dev for ifindex
						 *                                         found
						 */
	IRDMADBG_ATTR_EVENT_CFG_COUNT,
};
#define IRDMADBG_ATTR_EVENT_CFG_MAX (IRDMADBG_ATTR_EVENT_CFG_COUNT - 1)

/* event config list_evt_type nested attributes */
enum {
	IRDMADBG_ATTR_EVENT_CFG_ITEM_EVT_TYPE_UNSPEC,
	IRDMADBG_ATTR_EVENT_CFG_ITEM_EVT_TYPE,	/* u32, values from
						 *                                      icrdmadbg_event_type */
	IRDMADBG_ATTR_EVENT_CFG_ITEM_EVT_TYPE_COUNT,
};
#define IRDMADBG_ATTR_EVENT_CFG_ITEM_EVT_TYPE_MAX \
	(IRDMADBG_ATTR_EVENT_CFG_ITEM_EVT_TYPE_COUNT - 1)

/* event IDs per type */
enum icrdmadbg_event_aeq_evt_id {
	IRDMADBG_EVENT_AEQ_event1_EVT_ID,
	IRDMADBG_EVENT_AEQ_event2_EVT_ID,
};

enum icrdmadbg_event_cq0_evt_id {
	IRDMADBG_EVENT_CQ0_ERROR_EVT_ID,
	IRDMADBG_EVENT_CQ0_COMPLETED_EVT_ID,
};

enum icrdmadbg_event_ceq_evt_id {
	IRDMADBG_EVENT_CEQ_event1_EVT_ID,
	IRDMADBG_EVENT_CEQ_event2_EVT_ID,
};

enum icrdmadbg_event_aq_evt_id {
	IRDMADBG_EVENT_AQ_event1_EVT_ID,
	IRDMADBG_EVENT_AQ_event2_EVT_ID,
};

enum icrdmadbg_event_adapter_status_evt_id {
	IRDMADBG_EVENT_ADAPTER_STATUS_REGISTER_DEV_EVT_ID,
	IRDMADBG_EVENT_ADAPTER_STATUS_UNREGISTER_DEV_EVT_ID,
};

/* event nested attributes */
enum {
	IRDMADBG_ATTR_EVENT_UNSPEC,
	IRDMADBG_ATTR_EVENT_TYPE,	/* u32, values from icrdmadbg_event_type */
	IRDMADBG_ATTR_EVENT_ID,	/* u32,
				 *                      values from icrdmadbg_event_xxx_evt_id */
	IRDMADBG_ATTR_EVENT_DATA,	/* binary, stores event specific data */
	IRDMADBG_ATTR_EVENT_COUNT,
};
#define IRDMADBG_ATTR_EVENT_MAX (IRDMADBG_ATTR_EVENT_COUNT - 1)

#endif				/* _IRDMA_DBG_K_U_H */
