/* Copyright (c) 2015 - 2023 Intel Corporation */
#ifndef LICENSE_REPLACED
/* For licensing information, see the file 'LICENSE' in the root folder */
#endif /* LICENSE_REPLACED */
#ifndef EXTERNAL_RELEASE
/****************************************************************
* debugfs support
*
* Experiments in exporting data
*
****************************************************************/
#endif
#include <linux/fs.h>
#include <linux/debugfs.h>
#include <linux/list.h>

#include "main.h"

#ifdef DEBUG_SUPPORT
#ifdef CONFIG_DEBUG_FS
#ifdef DEBUG_SUPPORT_INTERNAL
#include "iidc.h"
/* This is available only in 4.16 kernels and beyond. So OFED 4.8 distro wont compile
 * for internal debug build
 */
#include <rdma/restrack.h>
extern void dump_ctx(struct irdma_sc_dev *dev, u32 pf_num, u32 qp_num);
void dumpCSR(struct irdma_sc_dev *dev);
#endif
static const char irdma_driver_name[] = "irdma";
static struct dentry *irdma_dbg_root;
static char cmd_buf[64];

/* QP subcommand */
#define QP_ACTIVE               1
#define QP_INFO                 2
#define QP_Q2                   3
#define QP_HOST_CTX             4
#define QP_CHIP_CTX             5
#define QP_CHIP_RAW             6
#define QP_CHIP_REG             7

#define DUMP_HMC_INFO           1
#define DUMP_HMC_INDEXES        2
#define DUMP_HMC_OBJ_INFO       4
#define DUMP_HMC_SD             8
#define DUMP_HMC_ALL            0xf

#define OFFSET_MASK_4K          0x0000000000000fffl

#ifdef DEBUG_SUPPORT_INTERNAL
enum mr_cmd_type {
	MR_CMD_LIST,
#ifndef EXTERNAL_RELEASE
	MR_CMD_PADDR,
#endif
	MR_CMD_STAG
};

#if 0
enum dump_mr_type {
	DUMP_MR_DATA,
	DUMP_MR_LEVEL1,
	DUMP_MR_ROOT,
	DUMP_MR_LEAF
};

#endif
static const char *const sd_type[] = { "Invalid", "Paged", "Direct" };

static const char *const obj_info_type[] = {
	"QP", "CQ", "SRQ", "HTE", "ARP", "APBVT", "MR", "XF",
	"XFFL", "Q1", "Q1FL", "TIMER", "FSIMC", "FSIAV", "PBLE",
	"RRF", "RRFFL", "HDR", "MD", "OOISC", "OOISCFFL"
};

static char *mr_types[] = { "MEM", "QP", "CQ", "SRQ"};
#endif /* DEBUG_SUPPORT_INTERNAL */
#ifndef EXTERNAL_RELEASE
/**************************************************************
* dump
* The dump entry in debugfs is for getting a data snapshow of
* the driver's current configuration and runtime details.
* When the filesystem entry is written, the command is saved.
* When the entry is read, the most snapshot is taken and dumped.
*
* Syntax is QP nnn type
*   where nnn is either a qp number or the word "all" and
*   where type is empty or "q2"  (add more options later)
**************************************************************/
#endif
static char *irdma_dbg_dump_buf;
static size_t irdma_dbg_dump_data_len;
static size_t irdma_dbg_dump_buf_len;
static u32 cmdnew;
static u32 cmddone;
#define BUF_ADDR_SIZE 17	/* address plus space */
#define BUF_DATA_ASCII_SIZE (32 * 3 + 2 + 32 + 1)	/* data plus ASCII */
#define IRDMA_DUMP_BUF_SIZE             16384
#define IRDMA_DUMP_BUF_HALF_FULL        (IRDMA_DUMP_BUF_SIZE / 2)

struct hmc_find {
	dma_addr_t paddr;
	bool fnd;
	dma_addr_t bp_pa;
	bool in_vf;
	u32 vf_id;
	int sd_indx;
	bool paged;
	bool in_pd_itself;
	int pd_indx;
	u64 hmc_addr;
	bool obj_info_fnd;
	enum irdma_hmc_rsrc_type obj_type;
	int obj_indx;
};

/**
 * dbg_vsnprintf -
 * @fmt: print formatting string
 */
static void dbg_vsnprintf(char *fmt, ...) __attribute__ ((format(gnu_printf, 1, 2)));
static void dbg_vsnprintf(char *fmt, ...)
{
	int cnt;
	va_list argp;

	va_start(argp, fmt);
	cnt = vsnprintf(irdma_dbg_dump_buf + irdma_dbg_dump_data_len,
			irdma_dbg_dump_buf_len - irdma_dbg_dump_data_len,
			fmt, argp);
	va_end(argp);

	irdma_dbg_dump_data_len += cnt;
}

/**
 * dump_help -
 */
static void dump_help(void)
{
	dbg_vsnprintf("Dump commands:\n");
	dbg_vsnprintf(" sw-stats\n");
	dbg_vsnprintf(" hw-stats\n");
	dbg_vsnprintf(" qp <n> chip_ctx\n");
	dbg_vsnprintf(" qp <n> chip_raw\n");
	dbg_vsnprintf(" qp <n> active           <n> is a qp number or 'all'\n");
	dbg_vsnprintf(" qp <n> info\n");
	dbg_vsnprintf(" cq <n>                  <n> is a cq number or 'all'\n");
#ifdef DEBUG_SUPPORT_INTERNAL
	dbg_vsnprintf(" cqp\n");
	dbg_vsnprintf(" hmc info\n");
#if 0
	dbg_vsnprintf(" hmc paddr <addr>\n");
	dbg_vsnprintf(" mem <addr> <length> <groupsize(1, 4 or 8)>\n");
#endif
	dbg_vsnprintf(" mr list\n");
	dbg_vsnprintf(" mr stag\n");
#if 0
	dbg_vsnprintf(" mr paddr <phys_addr>\n");
	dbg_vsnprintf(" pmem <phys_addr> <length> <groupsize(1, 4 or 8)>\n");
	dbg_vsnprintf(" qp <n> host_ctx\n");
#endif
	dbg_vsnprintf(" qp <n> q2\n");
	dbg_vsnprintf(" qp <n> chip_reg		raw context using regs(no cpq)\n");

	dbg_vsnprintf(" hw-regs\n");
	dbg_vsnprintf(" reset\n");
	dbg_vsnprintf(" en filter\n");
	dbg_vsnprintf(" dis filter\n");
	dbg_vsnprintf(" suspend\n");
	dbg_vsnprintf(" resume\n");
#endif /* DEBUG_SUPPORT_INTERNAL */
#if defined(DEBUG_SUPPORT_INTERNAL) || defined(MEV_RCA)
	dbg_vsnprintf(" cqp-nop\n");
#endif /* defined(DEBUG_SUPPORT_INTERNAL) || defined(MEV_RCA) */
#ifdef MEV_RCA
	dbg_vsnprintf(" cqp-rq\n");
#endif /* MEV_RCA */
	cmddone = true;
}

/**
 * irdma_dbg_save_ucontext -
 * @iwdev: device
 * @ucontext: user context
 */
void irdma_dbg_save_ucontext(struct irdma_device *iwdev,
			     struct irdma_ucontext *ucontext)
{
	struct irdma_handler *hdl;
	unsigned long flags;

	hdl = iwdev->hdl;

	spin_lock_irqsave(&hdl->uctx_list_lock, flags);
	list_add_tail(&ucontext->uctx_list, &hdl->ucontext_list);
	spin_unlock_irqrestore(&hdl->uctx_list_lock, flags);
}

/**
 * irdma_dbg_free_ucontext -
 * @ucontext: user context
 */
void irdma_dbg_free_ucontext(struct irdma_ucontext *ucontext)
{
	struct irdma_handler *hdl;
	unsigned long flags;

	hdl = ucontext->iwdev->hdl;

	spin_lock_irqsave(&hdl->uctx_list_lock, flags);
	list_del(&ucontext->uctx_list);
	spin_unlock_irqrestore(&hdl->uctx_list_lock, flags);
}

#ifdef DEBUG_SUPPORT_INTERNAL
/**
 * dump_to_buffer -
 * @dataptr: data pointer
 * @len: length
 * @offset: offset
 * @rowsize: row size
 * @groupsize: group size
 * @ascii: ascii flag
 * @lendumped: length of buffer to dump
 * @return:
 */
static int dump_to_buf(char *dataptr,
			  size_t len,
			  int offset,
			  int rowsize,
			  int groupsize,
			  bool ascii,
			  size_t *lendumped)
{
	char *bufptr;
	size_t bufleft, linelen;
	int dataleft = len;
	int count, thisrow, i;

	bufptr = irdma_dbg_dump_buf + irdma_dbg_dump_data_len;
	bufleft = irdma_dbg_dump_buf_len - irdma_dbg_dump_data_len;

	if (rowsize != 16 && rowsize != 32)
		rowsize = 16;

	*lendumped = 0;
	for (i = 0; i < len; i += rowsize) {
		if (bufleft < (BUF_ADDR_SIZE + BUF_DATA_ASCII_SIZE))
			break;
		count = snprintf(bufptr, bufleft, "  (%04x) ", offset);

		bufleft -= count;
		bufptr += count;
		thisrow = min(dataleft, rowsize);
		hex_dump_to_buffer(dataptr, thisrow, rowsize, groupsize,
				   bufptr, BUF_DATA_ASCII_SIZE, ascii);

		*lendumped += thisrow;
		offset += thisrow;
		dataptr += thisrow;
		dataleft -= thisrow;

		linelen = strlen(bufptr);
		bufptr[linelen++] = '\n';
		bufptr += linelen;
		bufleft -= linelen;
	}

	irdma_dbg_dump_data_len = bufptr - irdma_dbg_dump_buf;
	return 0;
}

/**
 * virt_mem -
 * @vmem: virtual memory pointer
 * @membrname: member name
 * @prefix: prefix
 */
static void virt_mem(struct irdma_virt_mem *vmem, char *membrname, char *prefix)
{
	dbg_vsnprintf("%s%s\n", prefix, membrname);
	dbg_vsnprintf("%s  va: 0x%p\n", prefix, vmem->va);
	dbg_vsnprintf("%s  size: 0x%x\n", prefix, vmem->size);
}

/**
 * dma_mem -
 * @addr: dma memory structure
 * @membrname: member name
 * @prefix: prefix
 */
static void dma_mem(struct irdma_dma_mem *addr, char *membrname, char *prefix)
{
	dbg_vsnprintf("%s%s\n", prefix, membrname);
	dbg_vsnprintf("%s  va: 0x%p\n", prefix, addr->va);
	dbg_vsnprintf("%s  pa: 0x%llx\n", prefix, addr->pa);
	dbg_vsnprintf("%s  size: 0x%x\n", prefix, addr->size);
}

#endif
/**
 * find_next_rsrc - Find next free resource ID
 * @rf: RDMA PCI function
 * @rsrc_array: resource array
 * @max_rsrc: max num resources
 * @next: next id to search
 */
static int find_next_rsrc(struct irdma_pci_f *rf,
			      irdma_bits_t rsrc_array,
			      u32 max_rsrc,
			      u32 next)
{
	u32 rsrc_num;
	unsigned long flags;

	spin_lock_irqsave(&rf->rsrc_lock, flags);
	rsrc_num = find_next_bit(rsrc_array, max_rsrc, next);
	spin_unlock_irqrestore(&rf->rsrc_lock, flags);
	return (rsrc_num < max_rsrc) ? rsrc_num : -2;
}

/**
 * find_next_qp_num -
 * @rf: RDMA PCI function
 * @qp_id: QP ID
 */
static int find_next_qp_num(struct irdma_pci_f *rf, int qp_id)
{
	do {
		++qp_id;
#ifndef EXTERNAL_RELEASE
/* 0 and 1 used for UDA */
#endif
		if ((qp_id == 0) || (qp_id == 1))
			qp_id = 2;
		qp_id = find_next_rsrc(rf, rf->allocated_qps,
					   rf->max_qp, qp_id);
		if (qp_id < 0)
			break;
	} while (!rf->qp_table[qp_id]);

	return qp_id;
}

/**
 * find_next_cq_num -
 * @rf: RDMA PCI function
 * @cq_id: CQ ID
 */
static long find_next_cq_num(struct irdma_pci_f *rf, long cq_id)
{
	do {
		++cq_id;
#ifndef EXTERNAL_RELEASE
/* 0, 1 and 2 CQs are reserved for driver control ops */
#endif
		if (cq_id < 3 ||
		    cq_id >= rf->cqp.sc_cqp.dev->hmc_info->hmc_obj[IRDMA_HMC_IW_CQ].max_cnt)
			cq_id = 3;
		cq_id = find_next_rsrc(rf, rf->allocated_cqs,
				       rf->max_cq, cq_id);
		if (cq_id < 0)
			break;

	} while (!rf->cq_table[cq_id]);

	return cq_id;
}

/**
 * dump_qp_shared -
 * @qp: QP pointer
 */
static void dump_qp_shared(struct irdma_qp *qp)
{
	struct irdma_sc_qp *qpsh = &qp->sc_qp;
	struct irdma_qp_uk *qpuk = &qpsh->qp_uk;

#ifdef DEBUG_SUPPORT_INTERNAL
	dbg_vsnprintf("    qp_uk: 0x%p\n", &qpsh->qp_uk);
	dbg_vsnprintf("      sq_base: 0x%p\n", qpuk->sq_base);
	dbg_vsnprintf("      rq_base: 0x%p\n", qpuk->rq_base);
	dbg_vsnprintf("      wqe_alloc_db: 0x%p\n", qpuk->wqe_alloc_db);
	dbg_vsnprintf("      sq_wrtrk_array: 0x%p\n", qpuk->sq_wrtrk_array);
	dbg_vsnprintf("      rq_wrid_array: 0x%p\n", qpuk->rq_wrid_array);
	dbg_vsnprintf("      shadow_area: 0x%p\n", qpuk->shadow_area);
	dbg_vsnprintf("      push_db: 0x%p\n", qpuk->push_db);
	dbg_vsnprintf("      push_wqe: 0x%p\n", qpuk->push_wqe);
#endif
	dbg_vsnprintf("      sq_ring.head: 0x%x tail: 0x%x size: 0x%x\n",
		      qpuk->sq_ring.head, qpuk->sq_ring.tail,
		      qpuk->sq_ring.size);
	dbg_vsnprintf("      rq_ring.head: 0x%x tail: 0x%x size: 0x%x\n",
		      qpuk->rq_ring.head, qpuk->rq_ring.tail,
		      qpuk->rq_ring.size);
	dbg_vsnprintf("      qp_id: 0x%x\n", qpuk->qp_id);
	dbg_vsnprintf("      sq_size: 0x%x\n", qpuk->sq_size);
	dbg_vsnprintf("      rq_size: 0x%x\n", qpuk->rq_size);
	dbg_vsnprintf("      swqe_polarity: 0x%x\n", qpuk->swqe_polarity);
	dbg_vsnprintf("      rwqe_polarity: 0x%x\n", qpuk->rwqe_polarity);
	dbg_vsnprintf("      rq_wqe_size: 0x%x\n", qpuk->rq_wqe_size);
	dbg_vsnprintf("      rq_wqe_size_multiplier: 0x%x\n",
		      qpuk->rq_wqe_size_multiplier);
	dbg_vsnprintf("      max_sq_frag_cnt: 0x%x\n", qpuk->max_sq_frag_cnt);
	dbg_vsnprintf("      max_rq_frag_cnt: 0x%x\n", qpuk->max_rq_frag_cnt);

#ifdef DEBUG_SUPPORT_INTERNAL
#ifdef UDA_ENABLED
	dbg_vsnprintf("      uda_ops: 0x%p\n", qpuk->uda_ops);
	dbg_vsnprintf("      sq_uda_phb_mem_base: 0x%p\n",
		      qpuk->sq_uda_phb_mem_base);
	dbg_vsnprintf("      rq_uda_phb_mem_base: 0x%p\n",
		      qpuk->rq_uda_phb_mem_base);

#endif				/* UDA_ENABLED */
	dbg_vsnprintf("    sq_pa: 0x%llx\n", qpsh->sq_pa);
	dbg_vsnprintf("    rq_pa: 0x%llx\n", qpsh->rq_pa);
	dbg_vsnprintf("    hw_host_ctx_pa: 0x%llx\n", qpsh->hw_host_ctx_pa);
	dbg_vsnprintf("    shadow_area_pa: 0x%llx\n", qpsh->shadow_area_pa);
	dbg_vsnprintf("    q2_pa: 0x%llx\n", qpsh->q2_pa);
	dbg_vsnprintf("    dev: 0x%p\n", qpsh->dev);
	dbg_vsnprintf("    pd: 0x%p\n", qpsh->pd);
	dbg_vsnprintf("    hw_host_ctx: 0x%p\n", qpsh->hw_host_ctx);
	dbg_vsnprintf("    llp_stream_handle: 0x%p\n", qpsh->llp_stream_handle);
	dbg_vsnprintf("    back_qp: 0x%p\n", qpuk->back_qp);
	/*dbg_vsnprintf("    srq: 0x%p\n", qpsh->srq); */
	dbg_vsnprintf("    q2_buf: 0x%p\n", qpsh->q2_buf);
	dbg_vsnprintf("    qp_compl_ctx: 0x%llx\n", qpsh->qp_compl_ctx);
#endif
	dbg_vsnprintf("    qs_handle: 0x%x\n", qpsh->qs_handle);
	dbg_vsnprintf("    exception_lan_q: 0x%x\n",
		      qp->iwdev->vsi.exception_lan_q);
	dbg_vsnprintf("    sq_tph_val: 0x%x\n", qpsh->sq_tph_val);
	dbg_vsnprintf("    rq_tph_val: 0x%x\n", qpsh->rq_tph_val);
	dbg_vsnprintf("    sq_tph_en: 0x%x\n", qpsh->sq_tph_en);
	dbg_vsnprintf("    rq_tph_en: 0x%x\n", qpsh->rq_tph_en);
	dbg_vsnprintf("    rcv_tph_en: 0x%x\n", qpsh->rcv_tph_en);
	dbg_vsnprintf("    xmit_tph_en: 0x%x\n", qpsh->xmit_tph_en);

#ifndef EXTERNAL_RELEASE
#if 0				/* The driver does not use this member */
	dbg_vsnprintf("    qp_state: 0x%x\n", qpsh->qp_state);
#endif
#endif
	dbg_vsnprintf("    virtual_map: 0x%x\n", qpsh->virtual_map);
	dbg_vsnprintf("    flush_sq: 0x%x\n", qpsh->flush_sq);
	dbg_vsnprintf("    flush_rq: 0x%x\n", qpsh->flush_rq);
	dbg_vsnprintf("    qp_type: 0x%x\n", qpsh->qp_uk.qp_type);
	dbg_vsnprintf("    hw_sq_size: 0x%x\n", qpsh->hw_sq_size);
	dbg_vsnprintf("    hw_rq_size: 0x%x\n", qpsh->hw_rq_size);
}

/**
 * dump_qp_info -
 * @qp: QP pointer
 */
static void dump_qp_info(struct irdma_qp *qp)
{
#ifdef DEBUG_SUPPORT_INTERNAL
	dbg_vsnprintf("%p (struct irdma_qp)\n", qp);
	dbg_vsnprintf("  ib_qp: 0x%p\n", &qp->ibqp);
	dbg_vsnprintf("  sc_qp: 0x%p (struct irdma_sc_qp)\n", &qp->sc_qp);
#else
	dbg_vsnprintf("struct irdma_qp\n");
#endif

	dump_qp_shared(qp);

#ifdef DEBUG_SUPPORT_INTERNAL
	dbg_vsnprintf("  iwdev: 0x%p\n", qp->iwdev);
	dbg_vsnprintf("  iwscq: 0x%p\n", qp->iwscq);
	dbg_vsnprintf("  iwrcq: 0x%p\n", qp->iwrcq);
	dbg_vsnprintf("  iwpd: 0x%p\n", qp->iwpd);
	dbg_vsnprintf("  ctx_info: 0x%p\n", &qp->ctx_info);
	dbg_vsnprintf("  iwarp_info: 0x%p\n", &qp->iwarp_info);
	dbg_vsnprintf("  qp: 0x%p\n", qp);
#endif
	dbg_vsnprintf("  refcount: 0x%x\n", refcount_read(&qp->refcnt));
#ifdef DEBUG_SUPPORT_INTERNAL
	dbg_vsnprintf("  cm_id: 0x%p\n", qp->cm_id);
	dbg_vsnprintf("  cm_node: 0x%p\n", qp->cm_node);
	dbg_vsnprintf("  lsmm_mr: 0x%p\n", qp->lsmm_mr);
#endif
	dbg_vsnprintf("  ibqp_state: 0x%x\n", qp->ibqp_state);
	dbg_vsnprintf("  iwarp_state: 0x%x\n", qp->iwarp_state);
	dbg_vsnprintf("  qp_mem_size: 0x%x\n", qp->qp_mem_size);
	dbg_vsnprintf("  last_aeq: 0x%x\n", qp->last_aeq);
	dbg_vsnprintf("  close_timer_started: 0x%x\n",
		      qp->close_timer_started.counter);
	dbg_vsnprintf("  qp_mem_size: 0x%x\n", qp->qp_mem_size);
#ifdef DEBUG_SUPPORT_INTERNAL
	dbg_vsnprintf("  lock: 0x%p\n", &qp->lock);
	dbg_vsnprintf("  iwqp_context: 0x%p\n", qp->iwqp_context);
	dbg_vsnprintf("  pbl_vbase: 0x%p\n", qp->pbl_vbase);
	dbg_vsnprintf("  pbl_pbase: 0x%llx\n", qp->pbl_pbase);
	dbg_vsnprintf("  page: 0x%p\n", qp->page);
	dbg_vsnprintf("  terminate_timer: 0x%p\n", &qp->terminate_timer);
	dbg_vsnprintf("  terminate_eventtype: 0x%llx\n", qp->pbl_pbase);
#endif
	dbg_vsnprintf("  active_conn: %d\n", qp->active_conn);
	dbg_vsnprintf("  user_mode: %d\n", qp->user_mode);
	dbg_vsnprintf("  hte_added: %d\n", qp->hte_added);
	dbg_vsnprintf("  flush_issued: %d\n", qp->flush_issued);
	dbg_vsnprintf("  destroy_pending: %d\n", qp->sc_qp.qp_uk.destroy_pending);
	dbg_vsnprintf("  sig_all: %d\n", qp->sig_all);
	dbg_vsnprintf("  pau_mode: %d\n", qp->pau_mode);
	dbg_vsnprintf("  term_sq_flush_code: 0x%x\n", qp->term_sq_flush_code);
	dbg_vsnprintf("  term_rq_flush_code: 0x%x\n", qp->term_rq_flush_code);
	dbg_vsnprintf("  hw_iwarp_state: %d\n", qp->hw_iwarp_state);
	dbg_vsnprintf("  hw_tcp_state: %d\n", qp->hw_tcp_state);
#ifdef DEBUG_SUPPORT_INTERNAL
	dbg_vsnprintf("  host_ctx.va: 0x%p pa: 0x%llx size: 0x%x\n",
		      qp->host_ctx.va, qp->host_ctx.pa, qp->host_ctx.size);
	dbg_vsnprintf("  iwpbl: 0x%p\n", qp->iwpbl);
	dbg_vsnprintf("  q2_ctx_mem.va: 0x%p pa: 0x%llx size: 0x%x\n",
		      qp->q2_ctx_mem.va, qp->q2_ctx_mem.pa,
		      qp->q2_ctx_mem.size);
	dbg_vsnprintf("  ietf_mem.va: 0x%p pa: 0x%llx size: 0x%x\n",
		      qp->ietf_mem.va, qp->ietf_mem.pa, qp->ietf_mem.size);
#endif
}

#if 0
/**
 * dump_qp_ctx -
 * @qp: QP pointer
 */
static void dump_qp_ctx(struct irdma_qp *qp)
{
	u64 *hwctx = qp->sc_qp.hw_host_ctx;
	size_t lendumped = 0;
	u32 val;
	u64 val64;
	u8 qp_type = qp->sc_qp.qp_type;

	dump_to_buf((char *)qp->sc_qp.hw_host_ctx, 256, 0, 32, 8,
			     false, &lendumped);

	dbg_vsnprintf("host_context 0x%p\n", qp->sc_qp.hw_host_ctx);
	if (qp_type == IRDMA_CQ_TYPE_IWARP)
		dbg_vsnprintf("  iwarp_rdmap_ver: 0x%llx\n", RS_64(hwctx[0], IRDMAQPC_RDMAP_VER));
	else
		dbg_vsnprintf("  roce_tver_ver: 0x%llx\n", RS_64(hwctx[0], IRDMAQPC_ROCE_TVER));
	dbg_vsnprintf("  push_mode_ena: 0x%llx\n",
		      RS_64(hwctx[0], IRDMAQPC_PMENA));
	dbg_vsnprintf("  push_page_index:  0x%llx\n",
		      RS_64(hwctx[0], IRDMAQPC_PPIDX));
	dbg_vsnprintf("  SQ_THP_en:  0x%llx\n",
		      RS_64(hwctx[0], IRDMAQPC_SQTPHEN));
	dbg_vsnprintf("  RQ_THP_en:  0x%llx\n",
		      RS_64(hwctx[0], IRDMAQPC_RQTPHEN));
	if (qp_type == IRDMA_CQ_TYPE_IWARP) {
		dbg_vsnprintf("  XMIT_THP_en:  0x%llx\n",
			      RS_64(hwctx[0], IRDMAQPC_XMITTPHEN));
		dbg_vsnprintf("  RCV_THP_en:  0x%llx\n",
			      RS_64(hwctx[0], IRDMAQPC_RCVTPHEN));
		dbg_vsnprintf("  dupack_thresh:  0x%llx\n",
			      RS_64(hwctx[0], IRDMAQPC_DUPACK_THRESH));
		dbg_vsnprintf("  drop_out_of_order_seg: 0x%llx\n",
			      RS_64(hwctx[0], IRDMAQPC_DROPOOOSEG));
		dbg_vsnprintf("  ECN_ena: no constants defined\n");
		dbg_vsnprintf("  limit: 0x%llx\n", RS_64(hwctx[0], IRDMAQPC_LIMIT));
	}
	dbg_vsnprintf("  RQ_WQE_Size: 0x%llx\n",
		      RS_64(hwctx[0], IRDMAQPC_RQWQESIZE));
	dbg_vsnprintf("  timestamp: 0x%llx\n",
		      RS_64(hwctx[0], IRDMAQPC_TIMESTAMP));
	dbg_vsnprintf("  use_srq: 0x%llx\n", RS_64(hwctx[0], IRDMAQPC_USESRQ));
	dbg_vsnprintf("  Insert VLAN Tag: 0x%llx\n",
		      RS_64(hwctx[0], IRDMAQPC_INSERTVLANTAG));
	if (qp_type == IRDMA_CQ_TYPE_IWARP) {
		dbg_vsnprintf("  NoNagle: 0x%llx\n", RS_64(hwctx[0], IRDMAQPC_NONAGLE));
		dbg_vsnprintf("  iwarp_ddp_ver: 0x%llx\n",
			      RS_64(hwctx[0], IRDMAQPC_DDP_VER));
	}
	dbg_vsnprintf("  IPv4: 0x%llx\n", RS_64(hwctx[0], IRDMAQPC_IPV4));

	/* Offset 8 */
	dbg_vsnprintf("  SQ_Address: 0x%llx\n",
		      RS_64(hwctx[1], IRDMAQPC_SQADDR));

	/* Offset 16 */
	dbg_vsnprintf("  RQ_Address: 0x%llx\n",
		      RS_64(hwctx[2], IRDMAQPC_RQADDR));

	/* Offset 24 */
	dbg_vsnprintf("  Dest Port Number: 0x%llx (%lld)\n",
		      RS_64(hwctx[3], IRDMAQPC_DESTPORTNUM),
		      RS_64(hwctx[3], IRDMAQPC_DESTPORTNUM));
	dbg_vsnprintf("  Source Port Number: 0x%llx (%lld)\n",
		      RS_64(hwctx[3], IRDMAQPC_SRCPORTNUM),
		      RS_64(hwctx[3], IRDMAQPC_SRCPORTNUM));
	dbg_vsnprintf("  Traffic Class or TOS: 0x%llx\n",
		      RS_64(hwctx[3], IRDMAQPC_TOS));
	dbg_vsnprintf("  avoid_stretch_ack: 0x%llx\n",
		      RS_64(hwctx[3], IRDMAQPC_AVOIDSTRETCHACK));
	dbg_vsnprintf("  SQ Size: 0x%llx\n", RS_64(hwctx[3], IRDMAQPC_SQSIZE));
	dbg_vsnprintf("  RQ Size: 0x%llx\n", RS_64(hwctx[3], IRDMAQPC_RQSIZE));
	dbg_vsnprintf("  Hop Limit or TTL: 0x%llx\n",
		      RS_64(hwctx[3], IRDMAQPC_TTL));

	/* Offset 32 */
	val = be32_to_cpu(RS_64(hwctx[4], IRDMAQPC_DESTIPADDR2));
	dbg_vsnprintf("  Dest_IP_Address_2: %pI4\n", &val);

	val = be32_to_cpu(RS_64(hwctx[4], IRDMAQPC_DESTIPADDR3));
	dbg_vsnprintf("  Dest_IP_Address_3: %pI4\n", &val);

	/* Offset 40 */
	val = be32_to_cpu(RS_64(hwctx[5], IRDMAQPC_DESTIPADDR0));
	dbg_vsnprintf("  Dest_IP_Address_0: %pI4\n", &val);

	val = be32_to_cpu(RS_64(hwctx[5], IRDMAQPC_DESTIPADDR1));
	dbg_vsnprintf("  Dest_IP_Address_1: %pI4\n", &val);

	/* Offset 48 */
	dbg_vsnprintf("  ARP Index: 0x%llx\n",
		      RS_64(hwctx[6], IRDMAQPC_ARPIDX));
	dbg_vsnprintf("  VLAN Tag: 0x%llx\n",
		      RS_64(hwctx[6], IRDMAQPC_VLANTAG));
	dbg_vsnprintf("  snd_mss: 0x%llx\n", RS_64(hwctx[6], IRDMAQPC_SNDMSS));

	/* Offset 56 */
	dbg_vsnprintf("  pd_index: 0x%llx\n",
		      RS_64(hwctx[7], IRDMAQPC_PDIDX) | RS_64(hwctx[0],
			    IRDMAQPC_PDIDXHI));

	if (qp_type == IRDMA_CQ_TYPE_IWARP) {
		dbg_vsnprintf("  Snd_wscale: 0x%llx\n",
			      RS_64(hwctx[7], IRDMAQPC_SNDSCALE));
		dbg_vsnprintf("  Rcv_wscale: 0x%llx\n",
			      RS_64(hwctx[7], IRDMAQPC_RCVSCALE));
		dbg_vsnprintf("  TCP_state: 0x%llx\n",
			      RS_64(hwctx[7], IRDMAQPC_TCPSTATE));
		dbg_vsnprintf("  ignore_tcp_uns_options: 0x%llx\n",
			      RS_64(hwctx[7], IRDMAQPC_IGNORE_TCP_UNS_OPT));
		dbg_vsnprintf("  ignore_tcp_options: 0x%llx\n",
			      RS_64(hwctx[7], IRDMAQPC_IGNORE_TCP_OPT));
		dbg_vsnprintf("  Flow Label: 0x%llx\n",
			      RS_64(hwctx[7], IRDMAQPC_FLOWLABEL));

		/* Offset 72 */
		dbg_vsnprintf("  timestamp_age: 0x%llx\n",
			      RS_64(hwctx[9], IRDMAQPC_TIMESTAMP_AGE));
		dbg_vsnprintf("  timestamp_recent: 0x%llx\n",
			      RS_64(hwctx[9], IRDMAQPC_TIMESTAMP_RECENT));

		/* Offset 80 */
		dbg_vsnprintf("  snd_wnd: 0x%llx\n", RS_64(hwctx[10], IRDMAQPC_SNDWND));
		dbg_vsnprintf("  snd_nxt: 0x%llx\n", RS_64(hwctx[10], IRDMAQPC_SNDNXT));

		/* Offset 88 */
		dbg_vsnprintf("  rcv_wnd: 0x%llx\n", RS_64(hwctx[11], IRDMAQPC_RCVWND));
		dbg_vsnprintf("  rcv_nxt: 0x%llx\n", RS_64(hwctx[11], IRDMAQPC_RCVNXT));

		/* Offset 96 */
		dbg_vsnprintf("  snd_una: 0x%llx\n", RS_64(hwctx[12], IRDMAQPC_SNDUNA));
		dbg_vsnprintf("  snd_max: 0x%llx\n", RS_64(hwctx[12], IRDMAQPC_SNDMAX));

		/* Offset 104 */
		dbg_vsnprintf("  rtt_var: 0x%llx\n", RS_64(hwctx[13], IRDMAQPC_RTTVAR));
		dbg_vsnprintf("  srtt: 0x%llx\n", RS_64(hwctx[13], IRDMAQPC_SRTT));
	} else {
		dbg_vsnprintf("  pkey: 0x%llx\n", RS_64(hwctx[7], IRDMAQPC_PKEY));
		dbg_vsnprintf("  ack_credits: 0x%llx\n", RS_64(hwctx[7], IRDMAQPC_ACKCREDITS));
		dbg_vsnprintf("  qkey: 0x%llx\n", RS_64(hwctx[8], IRDMAQPC_QKEY));
		dbg_vsnprintf("  Dest_QPN: 0x%llx\n", RS_64(hwctx[8], IRDMAQPC_DESTQP));
		dbg_vsnprintf("  Isn: 0x%llx\n", RS_64(hwctx[10], IRDMAQPC_ISN));
		dbg_vsnprintf("  psn_nxt: 0x%llx\n", RS_64(hwctx[10], IRDMAQPC_QKEY));
		dbg_vsnprintf("  epsn: 0x%llx\n", RS_64(hwctx[11], IRDMAQPC_EPSN));
		dbg_vsnprintf("  psn_una: 0x%llx\n", RS_64(hwctx[12], IRDMAQPC_PSNUNA));
		dbg_vsnprintf("  psn_max: 0x%llx\n", RS_64(hwctx[12], IRDMAQPC_PSNMAX));
		dbg_vsnprintf("  ss_thresh: 0x%llx\n", RS_64(hwctx[14], IRDMAQPC_SSTHRESH));
	}
	/* Offset 112 */
	dbg_vsnprintf("  cwnd: 0x%llx\n", RS_64(hwctx[14], IRDMAQPC_CWND));

	if (qp_type == IRDMA_CQ_TYPE_IWARP) {
		/* Offset 120 */
		dbg_vsnprintf("  snd_wl2: 0x%llx\n", RS_64(hwctx[15], IRDMAQPC_SNDWL2));
		dbg_vsnprintf("  snd_wl1: 0x%llx\n", RS_64(hwctx[15], IRDMAQPC_SNDWL1));

		/* Offset 128 */
		dbg_vsnprintf("  rexmit_thresh: 0x%llx\n",
			      RS_64(hwctx[16], IRDMAQPC_REXMIT_THRESH));
		dbg_vsnprintf("  max_snd_window: 0x%llx\n",
			      RS_64(hwctx[16], IRDMAQPC_MAXSNDWND));
	}
	/* Offset 136 */
	dbg_vsnprintf("  RxCmpQueueNum: 0x%llx\n",
		      RS_64(hwctx[17], IRDMAQPC_RXCQNUM));
	dbg_vsnprintf("  TxCmpQueueNum: 0x%llx\n",
		      RS_64(hwctx[17], IRDMAQPC_TXCQNUM));

	/* Offset 144 */
	if (qp_type == IRDMA_CQ_TYPE_IWARP)
		dbg_vsnprintf("  Q2_Address: 0x%llx\n", RS_64(hwctx[18], IRDMAQPC_Q2ADDR);
	dbg_vsnprintf("  Stats_instance_Index: 0x%llx\n", RS_64(hwctx[18], IRDMAQPC_STAT_INDEX));

	/* Offset 152 */
	dbg_vsnprintf("  L2TAG2: no constants defined\n");
	dbg_vsnprintf("  last_byte_sent: 0x%llx\n", RS_64(hwctx[19], IRDMAQPC_LASTBYTESENT));
	val64 = RS_64(hwctx[19], IRDMAQPC_MACADDRESS);
	dbg_vsnprintf("  src_MAC_Address: 0x%pM\n", &val64);

	if (qp_type == IRDMA_CQ_TYPE_IWARP) {
		/* Offset 160 */
		dbg_vsnprintf("  snd_mrk_offset: 0x%llx\n",
			      RS_64(hwctx[20], IRDMAQPC_SNDMARKOFFSET));
		dbg_vsnprintf("  rcv_mrk_offset: 0x%llx\n",
			      RS_64(hwctx[20], IRDMAQPC_RCVMARKOFFSET));

		dbg_vsnprintf("  rcv_no_mpa_crc: 0x%llx\n",
			      RS_64(hwctx[20], IRDMAQPC_RCVNOMPACRC));
		dbg_vsnprintf("  assume_aligned_hdrs: 0x%llx\n",
			      RS_64(hwctx[20], IRDMAQPC_ALIGNHDRS));
		dbg_vsnprintf("  Receive Markers: 0x%llx\n",
			      RS_64(hwctx[20], IRDMAQPC_RCVMARKERS));
		dbg_vsnprintf("  iWARP Mode: 0x%llx\n",
			      RS_64(hwctx[20], IRDMAQPC_IWARPMODE));
	}
	dbg_vsnprintf("  AdjustForLSMM: 0x%llx\n",
		      RS_64(hwctx[20], IRDMAQPC_ADJUSTFORLSMM));
	dbg_vsnprintf("  PrivilegedEnable: 0x%llx\n",
		      RS_64(hwctx[20], IRDMAQPC_PRIVEN));
	dbg_vsnprintf("  FastRegisterEnable: 0x%llx\n",
		      RS_64(hwctx[20], IRDMAQPC_FASTREGEN));
	dbg_vsnprintf("  BindEnable: 0x%llx\n",
		      RS_64(hwctx[20], IRDMAQPC_BINDEN));
	dbg_vsnprintf("  Send Markers: 0x%llx\n",
		      RS_64(hwctx[20], IRDMAQPC_SNDMARKERS));
	dbg_vsnprintf("  rdmard_ok: 0x%llx\n", RS_64(hwctx[20], IRDMAQPC_RDOK));
	dbg_vsnprintf("  rdmawr_rdresp_ok: 0x%llx\n",
		      RS_64(hwctx[20], IRDMAQPC_WRRDRSPOK));
	dbg_vsnprintf("  IRD_Size: 0x%llx\n",
		      RS_64(hwctx[20], IRDMAQPC_IRDSIZE));
	dbg_vsnprintf("  ORD_Size: 0x%llx\n",
		      RS_64(hwctx[20], IRDMAQPC_ORDSIZE));

	/* Offset 168 */
	dbg_vsnprintf("  Queue Pair Completion Context: 0x%llx\n",
		      RS_64(hwctx[21], IRDMAQPC_QPCOMPCTX));

	/* Offset 176 */
	if (qp_type == IRDMA_CQ_TYPE_IWARP)
		dbg_vsnprintf("  Exception_LAN_Queue: 0x%llx\n", RS_64(hwctx[22], IRDMAQPC_EXCEPTION_LAN_QUEUE));
	dbg_vsnprintf("  QS_Handle: 0x%llx\n",
		      RS_64(hwctx[22], IRDMAQPC_QSHANDLE));
	dbg_vsnprintf("  RQ_TPH_val: 0x%llx\n",
		      RS_64(hwctx[22], IRDMAQPC_RQTPHVAL));
	dbg_vsnprintf("  SQ_TPH_val: 0x%llx\n",
		      RS_64(hwctx[22], IRDMAQPC_SQTPHVAL));

	/* Offset 184 */
	val = be32_to_cpu(RS_64(hwctx[23], IRDMAQPC_LOCAL_IPADDR2));
	dbg_vsnprintf("  Local_IP_Address_2: %pI4\n", &val);

	val = be32_to_cpu(RS_64(hwctx[23], IRDMAQPC_LOCAL_IPADDR3));
	dbg_vsnprintf("  Local_IP_Address_3: %pI4\n", &val);

	/* Offset 192 */
	val = be32_to_cpu(RS_64(hwctx[24], IRDMAQPC_LOCAL_IPADDR0));
	dbg_vsnprintf("  Local_IP_Address_0: %pI4\n", &val);

	val = be32_to_cpu(RS_64(hwctx[24], IRDMAQPC_LOCAL_IPADDR1));
	dbg_vsnprintf("  Local_IP_Address_1: %pI4\n", &val);
}
#endif

/**
 * dump_qp -
 * @rf: RDMA PCI function
 * @qp_id: QP ID
 * @subtype: QP info type
 */
static void dump_qp(struct irdma_pci_f *rf, u32 qp_id, u32 subtype)
{
	struct irdma_qp *iwqp = NULL;
#ifdef DEBUG_SUPPORT_INTERNAL
	size_t lendumped = 0;
#endif

	if ((qp_id >= 2) && (qp_id < rf->max_qp))
		iwqp = rf->qp_table[qp_id];

	if (!iwqp) {
		dbg_vsnprintf("QP %d is not valid\n", qp_id);
		return;
	}

	dbg_vsnprintf("QP %d ibqp_state=0x%x warp_state=0x%x\n", qp_id,
		      iwqp->ibqp_state, iwqp->iwarp_state);

	switch (subtype) {
	case QP_INFO:
		dump_qp_info(iwqp);
		break;
#ifdef DEBUG_SUPPORT_INTERNAL
	case QP_Q2:
		dbg_vsnprintf("Q2 area 0x%p\n", iwqp->sc_qp.q2_buf);

		dump_to_buf((char *)iwqp->sc_qp.q2_buf, 256, 0, 16, 1,
				     true, &lendumped);
		break;
#if 0
	case QP_HOST_CTX:
		dump_qp_ctx(iwqp);
		break;
#endif
#endif
	case QP_CHIP_CTX:
		dbg_vsnprintf
		    ("QP %d context will be dumped to /var/log/messages\n",
		     qp_id);
		irdma_upload_qp_context(iwqp, 0, 0);
		break;
	case QP_CHIP_RAW:
		dbg_vsnprintf
		    ("QP %d context will be dumped using regs (no CQP ops) to /var/log/messages\n",
		     qp_id);
		irdma_upload_qp_context(iwqp, 0, 1);
		break;
#ifdef DEBUG_SUPPORT_INTERNAL
	case QP_CHIP_REG:
		dump_ctx(&rf->sc_dev, rf->sc_dev.hmc_fn_id, iwqp->ibqp.qp_num);
		break;
#endif
	case QP_ACTIVE:
		if (!iwqp->user_mode) {
			unsigned long flags;

			spin_lock_irqsave(&iwqp->lock, flags);
			irdma_print_sq_wqes(&iwqp->sc_qp.qp_uk);
			spin_unlock_irqrestore(&iwqp->lock, flags);
		}
		break;
	}
}

/**
 * dump_qp_cmd -
 * @rf: RDMA PCI function
 * @cbuf: character buffer to dump to
 */
static void dump_qp_cmd(struct irdma_pci_f *rf, char *cbuf)
{
	static int qp_id;
	static u32 all;
	static u32 subtype = QP_ACTIVE;
	char stype[9];
	char qpstr[11];
	int offset;
	int rc;

	if (cmdnew) {
		if (sscanf(cbuf, "%10s%n", qpstr, &offset) == 0)
			return;

		if (strncasecmp(qpstr, "all", 3) == 0) {
			all = true;
			qp_id = find_next_qp_num(rf, -1);
		} else {
			all = false;
			rc = kstrtoul(qpstr, 10, (long *)&qp_id);
			if (rc)
				return;

			if ((qp_id == 0) || (qp_id == 1))
#ifndef EXTERNAL_RELEASE
				/* QPs 0 and 1 are UDA only */
#endif
				qp_id = -1;
		}

		if (sscanf(cbuf + offset, "%8s", stype) == 1) {
			if (strncasecmp(stype, "active", 6) == 0)
				subtype = QP_ACTIVE;
			if (strncasecmp(stype, "info", 5) == 0)
				subtype = QP_INFO;
			else if (strncasecmp(stype, "q2", 2) == 0)
				subtype = QP_Q2;
#ifdef DEBUG_SUPPORT_INTERNAL
#if 0
			else if (strncasecmp(stype, "host_ctx", 8) == 0)
				subtype = QP_HOST_CTX;
#endif
#endif
			else if (strncasecmp(stype, "chip_ctx", 8) == 0)
				subtype = QP_CHIP_CTX;
			else if (strncasecmp(stype, "chip_raw", 8) == 0)
				subtype = QP_CHIP_RAW;
			else if (strncasecmp(stype, "chip_reg", 8) == 0)
				subtype = QP_CHIP_REG;
		}
	} else {
		if (!all)
			return;

		qp_id = find_next_qp_num(rf, qp_id);
	}

	if (qp_id >= 0)
		dump_qp(rf, qp_id, subtype);
	else
		cmddone = true;
}

/**
 * dump_cq_cmd -
 * @rf: RDMA PCI function
 * @cbuf: character buffer to read the cmd from
 */
static void dump_cq_cmd(struct irdma_pci_f *rf, const char *cbuf)
{
	static long cq_id;
	static bool all;
	char cqstr[11];
	int offset;
	int rc;

	if (cmdnew) {
		if (sscanf(cbuf, "%10s%n", cqstr, &offset) == 0)
			return;

		if (strncasecmp(cqstr, "all", 3) == 0) {
			all = true;
			cq_id = find_next_cq_num(rf, 1);
		} else {
			all = false;
			rc = kstrtoul(cqstr, 0, &cq_id);
			if (rc)
				return;

			if (cq_id < 3 ||
			    cq_id >= rf->cqp.sc_cqp.dev->hmc_info->hmc_obj[IRDMA_HMC_IW_CQ].max_cnt) {
				dbg_vsnprintf("CQ %ld is not valid\n", cq_id);
				return;
			}
		}
	} else {
		if (!all)
			return;

		cq_id = find_next_cq_num(rf, cq_id);
	}

	if (cq_id > 2) {
		struct irdma_cq *iwcq = rf->cq_table[cq_id];

		if (iwcq && !iwcq->user_mode) {
			unsigned long flags;

			spin_lock_irqsave(&iwcq->lock, flags);
			irdma_print_cqes(&iwcq->sc_cq.cq_uk);
			spin_unlock_irqrestore(&iwcq->lock, flags);
		}
	} else {
		cmddone = true;
	}
}

#ifdef DEBUG_SUPPORT_INTERNAL
/**
 * dump_cqp_cmd -
 * @rf: RDMA PCI function
 * @cbuf: character buffer to dump to
 */
static void dump_cqp_cmd(struct irdma_pci_f *rf, char *cbuf)
{
	struct irdma_sc_cqp *cqp = &rf->cqp.sc_cqp;

	if (!cqp) {
		dbg_vsnprintf("%s: cqp isn't available: cqp = NULL\n", __func__);
		cmddone = true;
		return;
	}

	dbg_vsnprintf("CQP (irdma_cqp)\n");
	dbg_vsnprintf("  size: 0x%x\n", cqp->size);
	dbg_vsnprintf("  sq_pa: 0x%llx\n", cqp->sq_pa);
	dbg_vsnprintf("  host_ctx_pa: 0x%llx\n", cqp->host_ctx_pa);
	dbg_vsnprintf("  back_cqp: 0x%p\n", cqp->back_cqp);
	dbg_vsnprintf("  dev: 0x%p\n", cqp->dev);
	dbg_vsnprintf("  sq_base: 0x%p\n", cqp->sq_base);
	dbg_vsnprintf("  host_ctx: 0x%p\n", cqp->host_ctx);
	dbg_vsnprintf("  scratch_array: 0x%p\n", cqp->scratch_array);
	dbg_vsnprintf("  cqp_id: 0x%x\n", cqp->cqp_id);
	dbg_vsnprintf("  sq_size: 0x%x\n", cqp->sq_size);
	dbg_vsnprintf("  hw_sq_size: 0x%x\n", cqp->hw_sq_size);
	dbg_vsnprintf("  struct_ver: 0x%x\n", cqp->struct_ver);
	dbg_vsnprintf("  polarity: 0x%x\n", cqp->polarity);
	dbg_vsnprintf("  en_datacenter_tcp: 0x%x\n", cqp->en_datacenter_tcp);
	dbg_vsnprintf("  sq_ring.head: 0x%x tail: 0x%x size: 0x%x\n",
		      cqp->sq_ring.head, cqp->sq_ring.tail, cqp->sq_ring.size);
	cmddone = true;
}

/**
 * dump_hmc_bp -
 * @bp: pointer to HMC backing page
 * @prefix: prefix string
 */
static void dump_hmc_bp(struct irdma_hmc_bp *bp, char *prefix)
{
	dbg_vsnprintf("%sentry_type: %d\n", prefix, bp->entry_type);

	if (bp->entry_type <= IRDMA_SD_TYPE_DIRECT) {
		/* Back up over \n and add to the line */
		irdma_dbg_dump_data_len--;
		dbg_vsnprintf(" %s\n", sd_type[bp->entry_type]);
	}

	dma_mem(&bp->addr, "addr", "    ");
	dbg_vsnprintf("%spd_index: 0x%x\n", prefix, bp->sd_pd_index);
	dbg_vsnprintf("%suse_cnt: %d\n", prefix, bp->use_cnt);
}

/**
 * dump_hmc_pd_entry -
 * @pd_entry: page descriptor entry
 */
static void dump_hmc_pd_entry(struct irdma_hmc_pd_entry *pd_entry)
{
	int i;

	dbg_vsnprintf("\npd_table 0x%p\n", pd_entry);

	/* Should we stop at the first not valid? */
	for (i = 0; i < 512; i++, pd_entry++) {
		if (!pd_entry->valid)
			continue;

		dump_hmc_bp(&pd_entry->bp, "  ");

		dbg_vsnprintf("%-3d\n", i);
		dbg_vsnprintf("  sd_index=%d\n", pd_entry->sd_index);
		dbg_vsnprintf("  valid: %d\n", pd_entry->valid);
		dbg_vsnprintf("  vmem: %d\n", pd_entry->rsrc_pg);
	}
}

/**
 * dump_hmc_pd_table -
 * @pd_table: page desc table
 * @prefix: prefix string
 */
static void dump_hmc_pd_table(struct irdma_hmc_pd_table *pd_table, char *prefix)
{
	dma_mem(&pd_table->pd_page_addr, "pd_page_addr", "    ");
	dbg_vsnprintf("%spd_entry: 0x%p - see below\n", prefix,
		      pd_table->pd_entry);
	virt_mem(&pd_table->pd_entry_virt_mem, "pd_entry_virt_mem", "  ");
	dump_hmc_pd_entry(pd_table->pd_entry);
}

/**
 * dump_hmc_sd -:
 * @info: HMC info pointer
*/
/* returns true when there is nothing more for this routine to do  */
static bool dump_hmc_sd(struct irdma_hmc_info *info)
{
	struct irdma_hmc_sd_entry *sd_entry;
	/* Static so we can continue if buf gets too full */
	static int sd_indx;
	static bool hmc_sd_new = true;

	if (hmc_sd_new) {
		sd_indx = 0;

		dbg_vsnprintf("\nsd_table\n");

		virt_mem(&info->sd_table.addr, "addr", "  ");
		dbg_vsnprintf("  sd_cnt=0x%x\n", info->sd_table.sd_cnt);
		dbg_vsnprintf("  use_cnt=%d\n", info->sd_table.use_cnt);
		hmc_sd_new = false;
	}

	/* Run through the entries in the sd_table */
	while ((sd_indx < info->sd_table.sd_cnt) &&
	       (irdma_dbg_dump_data_len < IRDMA_DUMP_BUF_HALF_FULL)) {
		sd_entry = &info->sd_table.sd_entry[sd_indx];
		if (sd_entry && sd_entry->valid) {
			dbg_vsnprintf("\nsd_entry 0x%x (0x%x)\n", sd_indx,
				      (sd_indx * 2 * 1024 * 1024));
			dbg_vsnprintf("  entry_type: %d\n",
				      sd_entry->entry_type);

			if (sd_entry->entry_type <= IRDMA_SD_TYPE_DIRECT) {
				/* Back up over \n and add to the line */
				irdma_dbg_dump_data_len--;
				dbg_vsnprintf(" %s\n",
					      sd_type[sd_entry->entry_type]);
			}

			dbg_vsnprintf("  valid: %d\n", sd_entry->valid);

			switch (sd_entry->entry_type) {
			case IRDMA_SD_TYPE_INVALID:
				break;
			case IRDMA_SD_TYPE_PAGED:
				dbg_vsnprintf("  u.pd_table\n");
				dump_hmc_pd_table(&sd_entry->u.pd_table,
						  "    ");
				break;
			case IRDMA_SD_TYPE_DIRECT:
				dbg_vsnprintf("  u.bp\n");
				dump_hmc_bp(&sd_entry->u.bp, "    ");
				break;
			}
		}

		sd_indx++;
	}

	if (sd_indx >= info->sd_table.sd_cnt) {
		hmc_sd_new = true;
		return true;
	}

	return false;
}

/* The routine expects to be called again to continue dumping. */
/* returns true when there is nothing more for this routine to do  */
/**
 * dump_hmc_info -:
 * @info: hmc info struct
*/
static bool dump_hmc_info(struct irdma_hmc_info *info)
{
	struct irdma_hmc_obj_info *obj_info;
	size_t lendumped;
	int i;
	static char *addr;
	static size_t len;
	static int offset;
	static int hmc_flags;

	/* Due to buffer limitation, it takes several passes through this routine. */
	/* Use the flags to track where we are */
	if (cmdnew)
		hmc_flags = DUMP_HMC_ALL;

	if (hmc_flags & DUMP_HMC_INFO) {
		dbg_vsnprintf("hmc_info\n");
		dbg_vsnprintf("  hmc_fn_id: 0x%x\n", info->hmc_fn_id);
		dbg_vsnprintf("  first_sd_index: 0x%x\n", info->first_sd_index);
		dbg_vsnprintf("  hmc_obj: 0x%p\n", info->hmc_obj);
		virt_mem(&info->hmc_obj_virt_mem, "hmc_obj_virt_mem", "  ");
		dbg_vsnprintf("  sd_table - see below\n");
		dbg_vsnprintf("\n  sd_indexes\n");

		/* Prepare to dump the sd_indexes */
		addr = (char *)info->sd_indexes;
		len = IRDMA_HMC_MAX_SD_COUNT * sizeof(u16);
		offset = 0;
		hmc_flags &= ~DUMP_HMC_INFO;
	}

	if (hmc_flags & DUMP_HMC_INDEXES) {
		/* Dump the sd_indexes */
		dump_to_buf(addr, len, offset, 16, 2, 0, &lendumped);

		/* A buffer full condition requires us to continue later */
		/* That is, if len is non-zero there is more to dump */
		addr += lendumped;
		offset = lendumped;
		len -= lendumped;
		if (len == 0)
			hmc_flags &= ~DUMP_HMC_INDEXES;
		/* empty the buffer before continuing */
		return false;
	}

	if (hmc_flags & DUMP_HMC_OBJ_INFO) {
		obj_info = info->hmc_obj;
		if (obj_info) {
			dbg_vsnprintf("\nhmc_obj_info\n");
			for (i = 0; i < IRDMA_HMC_IW_MAX; i++) {
				dbg_vsnprintf("\n  obj_info %d %s\n", i,
					      obj_info_type[i]);
				dbg_vsnprintf("    base: 0x%llx\n",
					      obj_info->base);
				dbg_vsnprintf("    max_cnt: 0x%x\n",
					      obj_info->max_cnt);
				dbg_vsnprintf("    cnt: 0x%x\n", obj_info->cnt);
				dbg_vsnprintf("    size: 0x%llx\n",
					      obj_info->size);
				obj_info++;
			}
		}

		hmc_flags &= ~DUMP_HMC_OBJ_INFO;
	}

	if (hmc_flags & DUMP_HMC_SD) {
		cmddone = dump_hmc_sd(info);
		if (cmddone)
			hmc_flags &= ~DUMP_HMC_SD;
	}

	if (hmc_flags == 0) {
		hmc_flags = DUMP_HMC_ALL;
		return true;
	}

	return false;
}

/**
 * dump_all_hmcs -
 * @rf: RDMA PCI function
 */
static void dump_all_hmcs(struct irdma_pci_f *rf)
{
#ifdef VCHNL_SUPPORT
	struct irdma_hmc_info *info;
	static u32 vf_hmc_idx;
#endif
	static bool pf_done;

	if (cmdnew) {
#ifdef VCHNL_SUPPORT
		vf_hmc_idx = 0;
#endif
		dbg_vsnprintf("hmc info PF\n");
		pf_done = false;
	}

	if (!pf_done) {
		pf_done = dump_hmc_info(rf->sc_dev.hmc_info);
		/* empty the buffer before continuing */
		return;
	}

	/* Dump the VF HMCs */
#ifdef VCHNL_SUPPORT
	while (vf_hmc_idx < IRDMA_MAX_PE_ENA_VF_COUNT) {
		if (rf->sc_dev.vf_dev[vf_hmc_idx]) {
			info = &rf->sc_dev.vf_dev[vf_hmc_idx]->hmc_info;
			dbg_vsnprintf("hmc info VF vf_idx 0x%x\n", vf_hmc_idx);
			if (dump_hmc_info(info))
				vf_hmc_idx++;
			return;
		}

		vf_hmc_idx++;
	}
#endif
}

#if 0
/**
 * dump_hmc_search_bp -
 * @bp: bp pointer
 * @hf: hmc
 */
static void dump_hmc_search_bp(struct irdma_hmc_bp *bp, struct hmc_find *hf)
{
	dma_addr_t bp_end = bp->addr.pa + bp->addr.size;

	if ((hf->paddr >= bp->addr.pa) && (hf->paddr < bp_end)) {
		hf->bp_pa = bp->addr.pa;
		hf->fnd = true;
	}
}

/**
 * dump_hmc_search_pd -
 * @pd_table: page desc. table
 * @hf: hf
 */
static void dump_hmc_search_pd(struct irdma_hmc_pd_table *pd_table,
			       struct hmc_find *hf)
{
	int i;

	if ((hf->paddr >= pd_table->pd_page_addr.pa) &&
	    (hf->paddr <
	     (pd_table->pd_page_addr.pa + pd_table->pd_page_addr.size))) {
		hf->in_pd_itself = true;
		hf->fnd = true;
		return;
	}

	for (i = 0; i < 512; i++) {
		if (pd_table->pd_entry[i].valid) {
			dump_hmc_search_bp(&pd_table->pd_entry[i].bp, hf);
			if (hf->fnd) {
				hf->paged = true;
				hf->pd_indx = i;
				break;
			}
		}
	}

}

/**
 * dump_hmc_search -
 * @info: hmc info
 * @hf: hf
 */
static void dump_hmc_search(struct irdma_hmc_info *info, struct hmc_find *hf)
{
	int sd_indx;
	struct irdma_hmc_sd_entry *sd_entry;

	for (sd_indx = 0; sd_indx < info->sd_table.sd_cnt; sd_indx++) {
		sd_entry = &info->sd_table.sd_entry[sd_indx];
		if (sd_entry->valid) {
			if (sd_entry->entry_type <= IRDMA_SD_TYPE_DIRECT)
				dump_hmc_search_bp(&sd_entry->u.bp, hf);
			else if (sd_entry->entry_type <= IRDMA_SD_TYPE_PAGED)
				dump_hmc_search_pd(&sd_entry->u.pd_table, hf);

			if (hf->fnd) {
				hf->sd_indx = sd_indx;
				break;
			}
		}
	}
}

/**
 * dump_hmc_search_obj_info -
 * @obj: obj
 * @hf: hf
 */
static void dump_hmc_search_obj_info(struct irdma_hmc_obj_info *obj,
				     struct hmc_find *hf)
{
	int i;
	u64 objend;

	hf->hmc_addr = hf->paddr - hf->bp_pa;
	hf->hmc_addr += hf->pd_indx * 4096;
	hf->hmc_addr += hf->sd_indx * 2 * 1024 * 1024;

	/* Find out what type of object it is */
	for (i = 0; i < IRDMA_HMC_IW_MAX; i++) {
		if (obj->cnt) {
			objend = obj->base + (obj->cnt * obj->size);
			if ((hf->hmc_addr >= obj->base) && (hf->hmc_addr < objend)) {
				hf->obj_type = i;
				hf->obj_indx =
				    (hf->hmc_addr - obj->base) / obj->size;
				hf->obj_info_fnd = true;
				break;
			}
		}
		obj++;
	}
}

/**
 * dump_hmc_find_paddr -
 * @rf: RDMA PCI function
 * @paddr: paddr
 */
static void dump_hmc_find_paddr(struct irdma_pci_f *rf, dma_addr_t paddr)
{
	struct hmc_find hf = { 0 };
	struct irdma_hmc_info *info;
	int i;

	if (paddr == 0ll) {
		dbg_vsnprintf("paddr of 0 is not valid\n");
		cmddone = true;
		return;
	}

	hf.paddr = paddr;

	/* Search PF hmc */
	dump_hmc_search(rf->sc_dev.hmc_info, &hf);

	if (!hf.fnd) {
		/* Search VFs */
		for (i = 0; i < IRDMA_MAX_PE_ENA_VF_COUNT; i++) {
			if (rf->sc_dev.vf_dev[i]) {
				info = &rf->sc_dev.vf_dev[i]->hmc_info;
				dump_hmc_search(info, &hf);
				if (hf.fnd) {
					hf.in_vf = true;
					hf.vf_id =
					    rf->sc_dev.vf_dev[i]->vf_id;
					break;
				}
			}
		}
	}

	if (hf.fnd) {
		dump_hmc_search_obj_info(rf->sc_dev.hmc_info->hmc_obj,
					 &hf);

		dbg_vsnprintf("Physical address 0x%lx was found in ", hf.paddr);
		if (hf.in_vf)
			dbg_vsnprintf("VF HMC tables 0x%x\n", hf.vf_id);
		else
			dbg_vsnprintf("Kernel HMC tables 0x%x\n", hf.vf_id);

		dbg_vsnprintf("HMC Address: 0x%lx SD Index: 0x%x\n",
			      hf.hmc_addr, hf.sd_indx);

		if (hf.paged) {
			if (hf.in_pd_itself)
				dbg_vsnprintf("In the PD Table itself\n");
			else
				dbg_vsnprintf("Paged table index 0x%x\n",
					      hf.pd_indx);
		} else {
			dbg_vsnprintf("Is a Direct entry\n");
		}

		if (hf.obj_info_fnd)
			dbg_vsnprintf("Obj Type: %s Obj Index: 0x%x\n",
				      obj_info_type[hf.obj_type], hf.obj_indx);
		else
			dbg_vsnprintf("Obj info not found\n");
	} else {
		dbg_vsnprintf
		    ("Physical address 0x%lx was not found in HMC tables\n");
	}

	cmddone = true;
}

#endif
/**
 * dump_hmc_cmd -
 * @rf: RDMA PCI function
 * @cbuf: char buffer
 */
static void dump_hmc_cmd(struct irdma_pci_f *rf, char *cbuf)
{
#if 0
	dma_addr_t paddr;
	int rc;

#endif
	if (!rf->sc_dev.hmc_info) {
		dbg_vsnprintf("hmc_info is not available\n");
		return;
	}

	if (strncasecmp(cbuf, "info", 4) == 0) {
		dump_all_hmcs(rf);
#if 0
	} else if (strncasecmp(cbuf, "paddr ", 6) == 0) {
		rc = kstrtoul(cbuf + 6, 16, (long *)&paddr);
		if (rc)
			return;
		dump_hmc_find_paddr(rf, paddr);
#endif
	} else {
		dump_help();
	}
}

#if 0
/* This routine is a start at adding a read register command - it is not complete or tested */
static void dump_rdreg_cmd(struct irdma_pci_f *rf, char *cbuf)
{
	struct irdma_sc_cqp *cqp = rf->sc_dev.cqp;
	int reg = -1;
	u32 reg_val;

	if (!cqp) {
		dbg_vsnprintf("%s: cqp isn't available: cqp = NULL\n", __func__);
		goto rdreg_cmd_exit;
	}

	if (kstrtoint(cbuf, 16, &reg) == 0)
		goto rdreg_cmd_exit;

	reg_val = rd32(cqp->dev->hw, reg);
	dbg_vsnprintf("rd32(0x%x)=0x%x\n", reg, reg_val);

rdreg_cmd_exit:
	cmddone = true;
}

#endif
/* Prints leaf or level1 pble */
/* The routine expects to be called again to continue dumping. */
/* returns true when there is nothing more for this routine to do  */
/**
 * dump_pble_info -:
 * @leaf: leaf
 * @fbo: fbo 
 * @pgsz: MR page size 
 * @type: type
 * @prefix: prefix
*/
static bool dump_pble_info(struct irdma_pble_info *leaf,
			   u64 fbo, u32 pgsz,
			   char *type,
			   char *prefix)
{
	u64 *paddrs = (u64 *)leaf->addr;
	static int pbleidx;
	static u64 data_offset;

	if (pbleidx == 0)
		dbg_vsnprintf("\n%s%s has 0x%x pbles\n", prefix, type,
			      leaf->cnt);

	if (cmdnew) {
		/* Prepare to dump list of data pages */
		dbg_vsnprintf("%s  pble-indx pble-phys-addr       data-phys-addr      data-offset       hmc-idx\n", prefix);

		/* Print first one here to add in the fbo */
		dbg_vsnprintf("%s  0x%-8x 0x%-16llx   0x%-16llx  0x%-16lx  0x%-8x\n", prefix,
			      0, virt_to_bus(paddrs), paddrs[0] + fbo, 0l, leaf->idx);

		data_offset = pgsz - fbo;
		pbleidx = 1;
	}

	while (pbleidx < leaf->cnt) {
		dbg_vsnprintf("%s  0x%-8x 0x%-16llx   0x%-16llx  0x%-16llx  0x%-8x\n", prefix,
			      pbleidx, virt_to_bus(&paddrs[pbleidx]), paddrs[pbleidx],
			      data_offset, leaf->idx + pbleidx);
		data_offset += pgsz;
		pbleidx++;

		if (irdma_dbg_dump_data_len > IRDMA_DUMP_BUF_HALF_FULL)
			return false;
	}

	pbleidx = 0;

	return true;
}

/* return true means that it is returning so the buffer can be emptied. */
/* The routine expects to be called again to continue dumping. */
/* returns true when there is nothing more for this routine to do  */

static bool dump_pbl_root(struct irdma_pble_level2 *lvl2, u32 fbo, u32 pgsz,
			  char *prefix)
{
	static int leafidx;
	static struct irdma_pble_info *leaf;
	char info[20];

	if (cmdnew) {
		leaf = lvl2->leaf;
		leafidx = 0;

		dbg_vsnprintf
		    ("%s root hmc-idx=0x%x paddr 0x%llx has 0x%x leaves\n",
		     prefix, lvl2->root.idx, virt_to_bus(&lvl2->root.addr),
		     lvl2->leaf_cnt);
	}

	while (leafidx < lvl2->leaf_cnt) {
		snprintf(info, sizeof(info), "leaf 0x%x", leafidx);
		if (dump_pble_info(leaf, fbo, pgsz, info, "   ") == false)
			return false;

		leaf++;
		leafidx++;
	}

	return true;
}

/* The routine expects to be called again to continue dumping. */
/* returns true when there is nothing more for this routine to do  */
static bool dump_mr(struct ib_mr *ibmr, bool kernel_mr, int vf_idx, bool show_pbles)
{
	u64 fbo;
	struct irdma_mr *iwmr = to_iwmr(ibmr);
	struct irdma_pd *iwpd = to_iwpd(ibmr->pd);
	int lvl;
	bool done = true;

	if (!iwmr->iwpbl.pbl_allocated)
		lvl = 0;
	else if (iwmr->iwpbl.pble_alloc.level == PBLE_LEVEL_1)
		lvl = 1;
	else
		lvl = 2;

	dbg_vsnprintf("pd=%d stag 0x%-8x len 0x%-16llx page_cnt 0x%-8x type %-4s level%d baseVA 0x%-16llx ",
		     iwpd->sc_pd.pd_id, iwmr->stag, iwmr->len, iwmr->page_cnt,
		     mr_types[iwmr->type], lvl, iwmr->iwpbl.user_base);

	if (kernel_mr)
		dbg_vsnprintf("kernel-mr ");
	else
		dbg_vsnprintf("user-mr ");

	if (vf_idx == -1)
		dbg_vsnprintf("PF-hmc\n");
	else
		dbg_vsnprintf("VF-hmc index %d\n", vf_idx);

	if (show_pbles) {
		if (cmdnew)
			fbo = iwmr->iwpbl.user_base & (iwmr->page_size - 1);
		else
			fbo = 0;

		if (lvl == 0)
			dbg_vsnprintf("  level0 pble pa 0x%llx\n",
				      iwmr->level0_pa + fbo);
		else if (lvl == 1)
			done = dump_pble_info(&iwmr->iwpbl.pble_alloc.level1,
					      fbo, iwmr->page_size, "level1", "  ");
		else
			done = dump_pbl_root(&iwmr->iwpbl.pble_alloc.level2,
					     fbo, iwmr->page_size,  "  ");
	}

	return done;
}

#if 0
/* Look for a paddr in leaf and level1 pages */
static void search_pbl_info(dma_addr_t paddr,
			    struct irdma_pble_info *leaf,
			    u64 mr_len,
			    u32 fbo,
			    u64 *area,
			    u64 *mr_offset,
			    enum dump_mr_type *type)
{
	bool candidate = false;
	u64 lbo;
	int pgidx;
	u64 *paddrs = (u64 *)leaf->addr;

	/* Search list of pages that hold the pbles */
	if (paddr >= virt_to_bus(paddrs) &&
	    (paddr < (virt_to_bus(&paddrs[leaf->cnt])))) {
		*type = DUMP_MR_LEVEL1;
		cmddone = true;
		return;
	}

	/* see if paddr is in any on the data pages */
	for (pgidx = 0; pgidx < leaf->cnt; pgidx++) {
		*area += 4096;
		if ((paddr >= paddrs[pgidx]) &&
		    (paddr < (paddrs[pgidx] + 4096))) {
			candidate = true;
			break;
		}
	}

	if (candidate) {
		/* found paddr - need to check first and last page */
		if ((*area - fbo) > mr_len) {
			/* last page of the mr - check that addr is in bounds */
			lbo = (fbo + mr_len - 1) & OFFSET_MASK_4K;
			if ((paddr & OFFSET_MASK_4K) > lbo)
				return;
		}

		if (*area == 4096) {
			/* found on first page, check the first byte offset */
			if ((paddr & OFFSET_MASK_4K) < fbo)
				return;

			*mr_offset = (paddr & OFFSET_MASK_4K) - fbo;
		} else {
			*mr_offset = 4096 - fbo;
			*mr_offset += *area - 8192;
			*mr_offset += paddr - paddrs[pgidx];
		}

		*type = DUMP_MR_DATA;
		cmddone = true;
	}
}

static void search_pbl_root(dma_addr_t paddr,
			    struct irdma_mr *iwmr,
			    u32 fbo,
			    u64 *area,
			    u64 *mr_offset,
			    enum dump_mr_type *type)
{
	int i;
	struct irdma_pble_info *leaf = iwmr->iwpbl.pble_alloc.level2.leaf;
	dma_addr_t rootaddr, rootend;

	/* Check for paddr in root info */
	rootaddr = virt_to_bus(&iwmr->iwpbl.pble_alloc.level2.root.addr);
	rootend = rootaddr +
	    (iwmr->iwpbl.pble_alloc.level2.leaf_cnt * sizeof(u64));
	if ((paddr >= rootaddr) && (paddr < rootend)) {
		*type = DUMP_MR_ROOT;
		cmddone = true;
		return;
	}

	/* See if paddr is in the leaf or data pages */
	for (i = 0; i < iwmr->iwpbl.pble_alloc.level2.leaf_cnt; i++) {
		search_pbl_info(paddr, leaf, iwmr->len, fbo, area, mr_offset,
				type);
		if (cmddone)
			break;
		leaf++;
	}

	/* Since there is a root page, it should be called a leaf not level1 */
	if (cmddone && (*type == DUMP_MR_LEVEL1))
		*type = DUMP_MR_LEAF;
}

static void search_mr_paddr(dma_addr_t paddr,
			    struct ib_mr *ibmr,
			    bool kernel_mr,
			    int vf_idx)
{
	u64 area = 0l;
	u64 mr_offset = 0l;
	struct irdma_mr *iwmr = to_iwmr(ibmr);
	u32 fbo = iwmr->iwpbl.user_base & OFFSET_MASK_4K;
	enum dump_mr_type type;

	if (!iwmr->iwpbl.pbl_allocated) {
		if ((paddr >= (iwmr->level0_pa + fbo)) &&
		    (paddr <= (iwmr->level0_pa + fbo + iwmr->len - 1))) {
			mr_offset = paddr - (iwmr->level0_pa + fbo);
			cmddone = true;
		}
	} else if (iwmr->iwpbl.pble_alloc.level == IRDMA_LEVEL_1) {
		search_pbl_info(paddr, &iwmr->iwpbl.pble_alloc.level1,
				iwmr->len, fbo, &area, &mr_offset, &type);
	} else {
		search_pbl_root(paddr, iwmr, fbo, &area, &mr_offset, &type);
	}

	if (cmddone) {
		/* ignore rc since short form of one stag will not fill buf */
		dump_mr(ibmr, false, vf_idx, false);
		if (type == DUMP_MR_DATA)
			dbg_vsnprintf("data offset 0x%lx\n", mr_offset);
		else if (type == DUMP_MR_LEVEL1)
			dbg_vsnprintf("level1 page\n");
		else if (type == DUMP_MR_ROOT)
			dbg_vsnprintf("root page\n");
		else
			dbg_vsnprintf("leaf page\n");
	}
}

#endif

/* This struct addition is quirk to use the Restrack DB since it is not exposed
 * in the public header to rdma drivers
 */
struct rdma_restrack_root {
      struct xarray xa;
      u32 next_id;
};

static void dump_mr_cmds(struct irdma_handler *hdl, enum mr_cmd_type cmd,
			 dma_addr_t paddr, u32 stag, int vf_idx)
{
	static struct ib_device *ibdev;
	static struct irdma_ucontext *ucontext;
	static unsigned long index;

	if (cmdnew) {
		if (list_empty(&hdl->ucontext_list))
			goto dump_mr_done;

		ucontext = list_first_entry(&hdl->ucontext_list,
					    struct irdma_ucontext, uctx_list);
		ibdev = &ucontext->iwdev->ibdev;
		index = 0;
	}

	while (1) {
		struct xarray *xa = &ibdev->res[RDMA_RESTRACK_MR].xa;
		struct rdma_restrack_entry *res;
		struct ib_mr *ibmr;

		if (irdma_dbg_dump_data_len > IRDMA_DUMP_BUF_HALF_FULL)
			return;

		res = xa_find(xa, &index, ULONG_MAX, XA_PRESENT);

		while (res) {
			if (!rdma_restrack_get(res))
				goto next;

			ibmr = container_of(res, struct ib_mr, res);

			switch (cmd) {
			case MR_CMD_LIST:
				if (!dump_mr(ibmr, rdma_is_kernel_res(res), vf_idx, true)) {
					rdma_restrack_put(res);
					dbg_vsnprintf("Dump buffer full, retrying MR\n");
					return;
				}

				break;
#if 0
			case MR_CMD_PADDR:
				search_mr_paddr(paddr, ibmr, false,
						vf_idx);
				if (cmddone)
					goto dump_mr_done;
				break;
#endif
			case MR_CMD_STAG:
				if (stag == ibmr->lkey) {
					if (!dump_mr(ibmr, rdma_is_kernel_res(res), vf_idx, true)) {
						rdma_restrack_put(res);
						dbg_vsnprintf("Dump buffer full, retrying MR\n");
						return;
					}

					rdma_restrack_put(res);
					goto dump_mr_done;
				}
			default:
				break;
			}

			rdma_restrack_put(res);
next:
			res = xa_find_after(xa, &index, ULONG_MAX, XA_PRESENT);

		}

		if (list_is_last(&ucontext->uctx_list, &hdl->ucontext_list))
			break;
		ucontext =
		    list_entry(ucontext->uctx_list.next, struct irdma_ucontext,
			       uctx_list);

		ibdev = &ucontext->iwdev->ibdev;
		index = 0;
	}

	if (cmd != MR_CMD_LIST)
		dbg_vsnprintf("not found\n");

dump_mr_done:
	cmddone = true;
}

/**
 * dump_mr_cmd -
 * @rf: RDMA PCI function
 * @cbuf: char buf
 */
static void dump_mr_cmd(struct irdma_pci_f *rf, char *cbuf)
{

	static dma_addr_t paddr;
	static u32 stag;
	int rc;
	static enum mr_cmd_type cmd;

	if (cmdnew) {
		if (strncasecmp(cbuf, "list", 4) == 0) {
			cmd = MR_CMD_LIST;
#if 0
		} else if (strncasecmp(cbuf, "paddr ", 6) == 0) {
			rc = kstrtoul(cbuf + 6, 10, (long *)&paddr);
			if (rc)
				return;
			cmd = MR_CMD_PADDR;

#endif
		} else if (strncasecmp(cbuf, "stag ", 5) == 0) {
			rc = kstrtoul(cbuf + 5, 16, (long *)&stag);
			if (rc)
				return;
			cmd = MR_CMD_STAG;
		} else {
			dump_help();
			return;
		}
	}

	dump_mr_cmds(rf->iwdev->hdl, cmd, paddr, stag, -1);
}


#if 0
//This function can be enabled to dump the CM nodes but can conflict with
// the cm_timer_tick.
static void irdma_debug_build_timer_list(struct list_head *timer_list,
					 struct list_head *hte)
{
	struct irdma_cm_node *cm_node;
	struct list_head *list_core_temp, *list_node;

	list_for_each_safe(list_node, list_core_temp, hte) {
		cm_node = container_of(list_node, struct irdma_cm_node, list);
		list_add(&cm_node->timer_entry, timer_list);
	}
}

static void irdma_dump_cm_nodes(struct irdma_device *iwdev)
{
	struct list_head timer_list;
	struct irdma_cm_core *cm_core = &iwdev->cm_core;
	struct list_head *list_core_temp;
	struct list_head *list_node;
	struct irdma_cm_node *cm_node;
	unsigned long flags;
	u32 cnt = 0;

	INIT_LIST_HEAD(&timer_list);

	spin_lock_irqsave(&cm_core->ht_lock, flags);
	irdma_debug_build_timer_list(&timer_list, &cm_core->non_accelerated_list);
	irdma_debug_build_timer_list(&timer_list, &cm_core->accelerated_list);

	dbg_vsnprintf("Allocated CM Nodes(First 20):\n");
	list_for_each_safe(list_node, list_core_temp, &timer_list) {
		cm_node = container_of(list_node, struct irdma_cm_node,
				       timer_entry);
		if (cnt < 20) {
			dbg_vsnprintf("cm_node=%px, iwqp=%px, accel=%d, close_entry=%px, send_entry=%px, refcnt=%d\n",
				      cm_node, cm_node->iwqp, cm_node->accelerated,
				      cm_node->close_entry, cm_node->send_entry,
				      refcount_read(&cm_node->refcnt));
		}
		++cnt;
	}
	spin_unlock_irqrestore(&cm_core->ht_lock, flags);
	dbg_vsnprintf("CM nodes count = %d\n", cnt);
}
#endif
#endif //DEBUG_SUPOORT_INTERNAL
/**
 * dump_stats_cmd - Dump statistics
 * @iwdev: device
 */
static void dump_stats_cmd(struct irdma_device *iwdev)
{
	struct irdma_sc_dev *dev = &iwdev->rf->sc_dev;
	struct irdma_cm_core *cm_core = &iwdev->cm_core;
	struct irdma_hmc_pble_rsrc *pble = iwdev->rf->pble_rsrc;
	struct irdma_puda_rsrc *ilq = iwdev->vsi.ilq;
	struct irdma_puda_rsrc *ieq = iwdev->vsi.ieq;

	dbg_vsnprintf("cm nodes created                 %lld\n",
		      cm_core->stats_nodes_created);
	dbg_vsnprintf("cm nodes destroyed               %lld\n",
		      cm_core->stats_nodes_destroyed);
	dbg_vsnprintf("cm listen called                 %lld\n",
		      cm_core->stats_listen_created);
	dbg_vsnprintf("cm listen removed                %lld\n",
		      cm_core->stats_listen_destroyed);
	dbg_vsnprintf("cm listen nodes created          %lld\n",
		      cm_core->stats_listen_nodes_created);
	dbg_vsnprintf("cm listen nodes destroyed        %lld\n",
		      cm_core->stats_listen_nodes_destroyed);
	dbg_vsnprintf("cm accepts                       %lld\n",
		      cm_core->stats_accepts);
	dbg_vsnprintf("cm rejects                       %lld\n",
		      cm_core->stats_rejects);
	dbg_vsnprintf("cm loopbacks                     %lld\n",
		      cm_core->stats_lpbs);
	dbg_vsnprintf("cm connect errors                %lld\n",
		      cm_core->stats_connect_errs);
	dbg_vsnprintf("cm passive errors                %lld\n",
		      cm_core->stats_passive_errs);
	dbg_vsnprintf("cm pkts retrans                  %lld\n",
		      cm_core->stats_pkt_retrans);
	dbg_vsnprintf("cm backlog drops                 %lld\n",
		      cm_core->stats_backlog_drops);

	dbg_vsnprintf("pble direct sds                  %d\n",
		      pble->stats_direct_sds);
	dbg_vsnprintf("pble paged sds                   %d\n",
		      pble->stats_paged_sds);
	dbg_vsnprintf("pble alloc ok                    %lld\n",
		      pble->stats_alloc_ok);
	dbg_vsnprintf("pble alloc fail                  %lld\n",
		      pble->stats_alloc_fail);
	dbg_vsnprintf("pble alloc freed                 %lld\n",
		      pble->stats_alloc_freed);
	dbg_vsnprintf("pble lvl1 alloc                  %lld\n",
		      pble->stats_lvl1);
	dbg_vsnprintf("pble lvl2 alloc                  %lld\n",
		      pble->stats_lvl2);
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 2, 0)
	dbg_vsnprintf("hugepages registered             %d\n",
		       iwdev->hugepgcnt);
#endif
	if (ilq) {
		dbg_vsnprintf("ilq packet sent                  %lld\n",
			      ilq->stats_pkt_sent);
		dbg_vsnprintf("ilq avail buffs                  %d\n",
			      ilq->avail_buf_count);
		dbg_vsnprintf("ilq alloc buffs                  %d\n",
			      ilq->alloc_buf_count);
		dbg_vsnprintf("ilq buf alloc fail               %lld\n",
			      ilq->stats_buf_alloc_fail);
		dbg_vsnprintf("ilq packet rcvd                  %lld\n",
			      ilq->stats_pkt_rcvd);
		dbg_vsnprintf("ilq packet rcv error             %lld\n",
			      ilq->stats_rcvd_pkt_err);
	}

	if (ieq) {
		dbg_vsnprintf("ieq fpdu_processed               %lld\n",
			      ieq->fpdu_processed);
		dbg_vsnprintf("ieq bad_seq_num                  %lld\n",
			      ieq->bad_seq_num);
		dbg_vsnprintf("ieq crc_err                      %lld\n",
			      ieq->crc_err);
		dbg_vsnprintf("ieq pmode_count                  %lld\n",
			      ieq->pmode_count);
		dbg_vsnprintf("ieq bad_qp_id(or RoCE pkts)      %lld\n\n",
			      ieq->stats_bad_qp_id);
	}

	dbg_vsnprintf("cqp requested ops                %llu\n",
		      dev->cqp->requested_ops);
	dbg_vsnprintf("cqp completed ops                %llu\n\n",
		      (u64)atomic64_read(&dev->cqp->completed_ops));

	/* sorted by cqp op type */
	dbg_vsnprintf("cqp OP_CEQ_DESTROY               %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_CEQ_DESTROY]);
	dbg_vsnprintf("cqp OP_AEQ_DESTROY               %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_AEQ_DESTROY]);
	dbg_vsnprintf("cqp OP_DELETE_ARP_CACHE_ENTRY    %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_DELETE_ARP_CACHE_ENTRY]);
	dbg_vsnprintf("cqp OP_MANAGE_APBVT_ENTRY        %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_MANAGE_APBVT_ENTRY]);
	dbg_vsnprintf("cqp OP_CEQ_CREATE                %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_CEQ_CREATE]);
	dbg_vsnprintf("cqp OP_AEQ_CREATE                %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_AEQ_CREATE]);
	dbg_vsnprintf("cqp OP_MANAGE_QHASH_TABLE_ENTRY  %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_MANAGE_QHASH_TABLE_ENTRY]);
	dbg_vsnprintf("cqp OP_QP_MODIFY                 %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_QP_MODIFY]);
	dbg_vsnprintf("cqp OP_QP_UPLOAD_CONTEXT         %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_QP_UPLOAD_CONTEXT]);
	dbg_vsnprintf("cqp OP_CQ_CREATE                 %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_CQ_CREATE]);
	dbg_vsnprintf("cqp OP_CQ_DESTROY                %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_CQ_DESTROY]);
	dbg_vsnprintf("cqp OP_QP_CREATE                 %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_QP_CREATE]);
	dbg_vsnprintf("cqp OP_QP_DESTROY                %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_QP_DESTROY]);
	dbg_vsnprintf("cqp OP_ALLOC_STAG                %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_ALLOC_STAG]);
	dbg_vsnprintf("cqp OP_MR_REG_NON_SHARED         %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_MR_REG_NON_SHARED]);
	dbg_vsnprintf("cqp OP_DEALLOC_STAG              %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_DEALLOC_STAG]);
	dbg_vsnprintf("cqp OP_MW_ALLOC                  %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_MW_ALLOC]);
	dbg_vsnprintf("cqp OP_QP_FLUSH_WQES             %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_QP_FLUSH_WQES]);
	dbg_vsnprintf("cqp OP_ADD_ARP_CACHE_ENTRY       %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_ADD_ARP_CACHE_ENTRY]);
	dbg_vsnprintf("cqp OP_MANAGE_PUSH_PAGE          %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_MANAGE_PUSH_PAGE]);
	dbg_vsnprintf("cqp OP_UPDATE_PE_SDS             %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_UPDATE_PE_SDS]);
	dbg_vsnprintf("cqp OP_MANAGE_HMC_PM_FUNC_TABLE  %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_MANAGE_HMC_PM_FUNC_TABLE]);
	dbg_vsnprintf("cqp OP_SUSPEND                   %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_SUSPEND]);
	dbg_vsnprintf("cqp OP_RESUME                    %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_RESUME]);
#ifdef CVL_VCHNL_UNPRIVILEGED
	dbg_vsnprintf("cqp OP_MANAGE_PBLE_BP		%lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_MANAGE_PBLE_BP]);
#endif
	dbg_vsnprintf("cqp OP_QUERY_FPM_VAL             %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_QUERY_FPM_VAL]);
	dbg_vsnprintf("cqp OP_COMMIT_FPM_VAL            %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_COMMIT_FPM_VAL]);
	dbg_vsnprintf("cqp OP_AH_CREATE                 %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_AH_CREATE]);
	dbg_vsnprintf("cqp OP_AH_MODIFY                 %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_AH_MODIFY]);
	dbg_vsnprintf("cqp OP_AH_DESTROY                %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_AH_DESTROY]);
	dbg_vsnprintf("cqp OP_MC_CREATE                 %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_MC_CREATE]);
	dbg_vsnprintf("cqp OP_MC_DESTROY                %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_MC_DESTROY]);
	dbg_vsnprintf("cqp OP_MC_MODIFY                 %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_MC_MODIFY]);
	dbg_vsnprintf("cqp OP_STATS_ALLOCATE            %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_STATS_ALLOCATE]);
	dbg_vsnprintf("cqp OP_STATS_FREE                %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_STATS_FREE]);
	dbg_vsnprintf("cqp OP_STATS_GATHER              %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_STATS_GATHER]);
	dbg_vsnprintf("cqp OP_WS_ADD_NODE               %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_WS_ADD_NODE]);
	dbg_vsnprintf("cqp OP_WS_MODIFY_NODE            %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_WS_MODIFY_NODE]);
	dbg_vsnprintf("cqp OP_WS_DELETE_NODE            %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_WS_DELETE_NODE]);
	dbg_vsnprintf("cqp OP_WS_FAILOVER_START         %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_WS_FAILOVER_START]);
	dbg_vsnprintf("cqp OP_WS_FAILOVER_COMPLETE      %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_WS_FAILOVER_COMPLETE]);
	dbg_vsnprintf("cqp OP_SET_UP_MAP                %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_SET_UP_MAP]);
	dbg_vsnprintf("cqp OP_GEN_AE                    %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_GEN_AE]);
	dbg_vsnprintf("cqp OP_QUERY_RDMA_FEATURES       %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_QUERY_RDMA_FEATURES]);
	dbg_vsnprintf("cqp OP_ALLOC_LOCAL_MAC_ENTRY     %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_ALLOC_LOCAL_MAC_ENTRY]);
	dbg_vsnprintf("cqp OP_ADD_LOCAL_MAC_ENTRY       %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_ADD_LOCAL_MAC_ENTRY]);
	dbg_vsnprintf("cqp OP_DELETE_LOCAL_MAC_ENTRY    %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_DELETE_LOCAL_MAC_ENTRY]);
	dbg_vsnprintf("cqp OP_CQ_MODIFY                 %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_CQ_MODIFY]);
#ifdef MEV_SUPPORT
	dbg_vsnprintf("cqp OP_SRQ_CREATE                %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_SRQ_CREATE]);
	dbg_vsnprintf("cqp OP_SRQ_MODIFY                %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_SRQ_MODIFY]);
	dbg_vsnprintf("cqp OP_SRQ_DESTROY               %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_SRQ_DESTROY]);
#ifdef MEV_RCA
	dbg_vsnprintf("cqp OP_RCA_EXEC_FWD_OP           %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_RCA_EXEC_FWD_OP]);
	dbg_vsnprintf("cqp OP_RET_CQP_CMPL              %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_RET_CQP_CMPL]);
	dbg_vsnprintf("cqp OP_COPY_DATA                 %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_COPY_DATA]);
#endif /* MEV_RCA */
#if defined(MEV_RCA) || defined(SV_DRIVER) || defined(DEBUG_SUPPORT_INTERNAL)
	dbg_vsnprintf("cqp OP_NOP                       %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_NOP]);
#endif /* defined(MEV_RCA) || defined(SV_DRIVER) || defined(DEBUG_SUPPORT_INTERNAL) */
#if defined(CPF_LIB) || defined(SV_DRIVER)
	dbg_vsnprintf("cqp OP_SET_INTERRUPT             %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_SET_INTERRUPT]);
	dbg_vsnprintf("cqp OP_MANAGE_RESOURCE_LIMITS    %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_MANAGE_RESOURCE_LIMITS]);
#endif /* defined(CPF_LIB) || defined(SV_DRIVER) */
#ifdef CPF_LIB
	dbg_vsnprintf("cqp OP_FWD_CQP_OPS               %lld\n",
		      dev->cqp_cmd_stats[IRDMA_OP_FWD_CQP_OPS]);
#endif /* CPF_LIB */
#endif /* MEV_SUPPORT */

#ifdef AH_REUSE
	dbg_vsnprintf("AH Reused Count                  %lld\n",
		      iwdev->ah_reused);
	dbg_vsnprintf("AH Current List Count            %d\n",
		      iwdev->ah_list_cnt);
	dbg_vsnprintf("AH list cnt HWM                  %d\n",
		      iwdev->ah_list_hwm);
#endif

#ifdef CONFIGFS_SUPPORT
#if IS_ENABLED(CONFIG_CONFIGFS_FS)
	if  (iwdev->roce_mode) {
		dbg_vsnprintf("roce rtomin  = %d\n", iwdev->roce_rtomin);
		dbg_vsnprintf("roce cwnd    = %d\n", iwdev->roce_cwnd);
	} else {
		dbg_vsnprintf("iwarp rtomin = %d\n", iwdev->iwarp_rtomin);
		dbg_vsnprintf("iwarp rcvwnd = %d\n", iwdev->rcv_wnd);
	}
#endif /* CONFIG_CONFIGFS_FS */
#endif /* CONFIGFS_SUPPORT */
#if 0
	irdma_dump_cm_nodes(iwdev);
#endif
	cmddone = true;
}

#ifdef DEBUG_SUPPORT_INTERNAL
void read_cache_stats(struct irdma_sc_dev *dev)
{
	u64 high_bit, reg_val;
	u32 low_bit, val;
	int i;

#define MAX_OBJECT_CNT		32
	dbg_vsnprintf("----------------------\n");
	for (i = 0; i < MAX_OBJECT_CNT; i++) {
#define GLPEOC0_STAT_CTL        0x00514004
#define GLPEOC0_STAT_READ_HIT_LO        0x0051400C
#define GLPEOC0_STAT_READ_HIT_HI        0x00514010
#define GLPEOC0_STAT_READ_MISS_LO	0x00514014
#define GLPEOC0_STAT_READ_MISS_HI	0x00514018
#define GLPEOC0_STAT_WR_HIT_LO		0x0051401C
#define GLPEOC0_STAT_WR_HIT_HI		0x00514020
#define GLPEOC0_STAT_WR_MISS_LO		0x00514024
#define GLPEOC0_STAT_WR_MISS_HI		0x00514028
		val = rd32(dev->hw, GLPEOC0_STAT_CTL) & ~0x1f;
		val |= i;
		wr32(dev->hw, GLPEOC0_STAT_CTL, val);
		irdma_wmb();
		low_bit = rd32(dev->hw, GLPEOC0_STAT_READ_HIT_LO);
		high_bit = rd32(dev->hw, GLPEOC0_STAT_READ_HIT_HI) & 0xffffffULL;
		reg_val = (high_bit << 32) | low_bit;
		dbg_vsnprintf("GLPEOC0_READ_HIT = %llu\n", reg_val);
		low_bit = rd32(dev->hw, GLPEOC0_STAT_READ_MISS_LO);
		high_bit = rd32(dev->hw, GLPEOC0_STAT_READ_MISS_HI) & 0xffffffULL;
		reg_val = (high_bit << 32) | low_bit;
		dbg_vsnprintf("GLPEOC0_READ_MISS = %llu\n", reg_val);
		low_bit = rd32(dev->hw, GLPEOC0_STAT_WR_HIT_LO);
		high_bit = rd32(dev->hw, GLPEOC0_STAT_WR_HIT_HI) & 0xffffffULL;
		reg_val = (high_bit << 32) | low_bit;
		dbg_vsnprintf("GLPEOC0_WR_HIT = %llu\n", reg_val);
		low_bit = rd32(dev->hw, GLPEOC0_STAT_WR_MISS_LO);
		high_bit = rd32(dev->hw, GLPEOC0_STAT_WR_MISS_HI) & 0xffffffULL;
		reg_val = (high_bit << 32) | low_bit;
		dbg_vsnprintf("GLPEOC0_WR_MISS = %llu\n", reg_val);

#define GLPDOC_STAT_CTL                 0x00530004
#define GLPDOC_STAT_RD_HIT_LO           0x0053000C
#define GLPDOC_STAT_RD_HIT_HI           0x00530010
#define GLPDOC_STAT_RD_MISS_LO		0x00530014
#define GLPDOC_STAT_RD_MISS_HI		0x00530018
		val = rd32(dev->hw, GLPDOC_STAT_CTL) & ~0x1f;
		val |= i;
		wr32(dev->hw, GLPDOC_STAT_CTL, val);
		irdma_wmb();
		low_bit = rd32(dev->hw, GLPDOC_STAT_RD_HIT_LO);
		high_bit = rd32(dev->hw, GLPDOC_STAT_RD_HIT_HI) & 0xffffffULL;
		reg_val = (high_bit << 32) | low_bit;
		dbg_vsnprintf("GLPDOC_READ_HIT = %llu\n", reg_val);
		low_bit = rd32(dev->hw, GLPDOC_STAT_RD_MISS_LO);
		high_bit = rd32(dev->hw, GLPDOC_STAT_RD_MISS_HI) & 0xffffffULL;
		reg_val = (high_bit << 32) | low_bit;
		dbg_vsnprintf("GLPDOC_READ_MISS = %llu\n", reg_val);

#define GLPBLOC0_STAT_CTL               0x00518004
#define GLPBLOC0_STAT_RD_HIT_LO         0x0051800C
#define GLPBLOC0_STAT_RD_HIT_HI         0x00518010
#define GLPBLOC0_STAT_RD_MISS_LO	0x00518014
#define GLPBLOC0_STAT_RD_MISS_HI	0x00518018
#define GLPBLOC0_STAT_WR_HIT_LO		0x0051801C
#define GLPBLOC0_STAT_WR_HIT_HI		0x00518020
#define GLPBLOC0_STAT_WR_MISS_LO	0x00518024
#define GLPBLOC0_STAT_WR_MISS_HI	0x00518028
		val = rd32(dev->hw, GLPBLOC0_STAT_CTL) & ~0x1f;
		val |= i;
		wr32(dev->hw, GLPBLOC0_STAT_CTL, val);
		irdma_wmb();
		low_bit = rd32(dev->hw, GLPBLOC0_STAT_RD_HIT_LO);
		high_bit = rd32(dev->hw, GLPBLOC0_STAT_RD_HIT_HI) & 0xffffffULL;
		reg_val = (high_bit << 32) | low_bit;
		dbg_vsnprintf("GLPBLOC0_RD_HIT = %llu\n", reg_val);
		low_bit = rd32(dev->hw, GLPBLOC0_STAT_RD_MISS_LO);
		high_bit = rd32(dev->hw, GLPBLOC0_STAT_RD_MISS_HI) & 0xffffffULL;
		reg_val = (high_bit << 32) | low_bit;
		dbg_vsnprintf("GLPBLOC0_RD_MISS = %llu\n", reg_val);
		low_bit = rd32(dev->hw, GLPBLOC0_STAT_WR_HIT_LO);
		high_bit = rd32(dev->hw, GLPBLOC0_STAT_WR_HIT_HI) & 0xffffffULL;
		reg_val = (high_bit << 32) | low_bit;
		dbg_vsnprintf("GLPBLOC0_WR_HIT = %llu\n", reg_val);
		low_bit = rd32(dev->hw, GLPBLOC0_STAT_WR_MISS_LO);
		high_bit = rd32(dev->hw, GLPBLOC0_STAT_WR_MISS_HI) & 0xffffffULL;
		reg_val = (high_bit << 32) | low_bit;
		dbg_vsnprintf("GLPBLOC0_WR_MISS = %llu\n", reg_val);

#define GLMDOC_STAT_CTL                 0x0051C004
#define GLMDOC_STAT_RD_HIT_LO           0x0051C00C
#define GLMDOC_STAT_RD_HIT_HI           0x0051C010
#define GLMDOC_STAT_RD_MISS_LO		0x0051C014
#define GLMDOC_STAT_RD_MISS_HI		0x0051C018
#define GLMDOC_STAT_WR_HIT_LO		0x0051C01C
#define GLMDOC_STAT_WR_HIT_HI		0x0051C020
#define GLMDOC_STAT_WR_MISS_LO		0x0051C024
#define GLMDOC_STAT_WR_MISS_HI		0x0051C028
		val = rd32(dev->hw, GLMDOC_STAT_CTL) & ~0x1f;
		val |= i;
		wr32(dev->hw, GLMDOC_STAT_CTL, val);
		irdma_wmb();
		low_bit = rd32(dev->hw, GLMDOC_STAT_RD_HIT_LO);
		high_bit = rd32(dev->hw, GLMDOC_STAT_RD_HIT_HI) & 0xffffffULL;
		reg_val = (high_bit << 32) | low_bit;
		dbg_vsnprintf("GLMDOC_RD_HIT = %llu\n", reg_val);
		low_bit = rd32(dev->hw, GLMDOC_STAT_RD_MISS_LO);
		high_bit = rd32(dev->hw, GLMDOC_STAT_RD_MISS_HI) & 0xffffffULL;
		reg_val = (high_bit << 32) | low_bit;
		dbg_vsnprintf("GLMDOC_RD_MISS = %llu\n", reg_val);
		low_bit = rd32(dev->hw, GLMDOC_STAT_WR_HIT_LO);
		high_bit = rd32(dev->hw, GLMDOC_STAT_WR_HIT_HI) & 0xffffffULL;
		reg_val = (high_bit << 32) | low_bit;
		dbg_vsnprintf("GLMDOC_WR_HIT = %llu\n", reg_val);
		low_bit = rd32(dev->hw, GLMDOC_STAT_WR_MISS_LO);
		high_bit = rd32(dev->hw, GLMDOC_STAT_WR_MISS_HI) & 0xffffffULL;
		reg_val = (high_bit << 32) | low_bit;
		dbg_vsnprintf("GLMDOC_WR_MISS = %llu\n", reg_val);
	}
}

#endif
/**
 * dump_hw_stats_cmd -
 * @iwdev: iwarp device
 */
static void dump_hw_stats_cmd(struct irdma_device *iwdev)
{
	struct irdma_sc_dev *dev = &iwdev->rf->sc_dev;
	struct irdma_vsi_pestat *devstat = iwdev->vsi.pestat;
	struct irdma_dev_hw_stats *hw_stats = &devstat->hw_stats;

#ifndef EXTERNAL_RELEASE
	/* FIXME: Use for HW debug if needed, remove in final source dumpCSR(dev);
	dumpCSR(dev);

	return;
	 */
#endif
	if (iwdev->rf->rdma_ver >= IRDMA_GEN_2)
		irdma_cqp_gather_stats_cmd(dev, devstat, true);
#ifdef FPK_SUPPORT
	else
		irdma_cqp_gather_stats_gen1(dev, devstat);
#endif

	dbg_vsnprintf("IPV4 octs recvd                      %llu\n",
		      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_IP4RXOCTS]);
	dbg_vsnprintf("IPV6 octs recvd                      %llu\n",
		      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_IP6RXOCTS]);
	dbg_vsnprintf("IPV4 pkts recvd                      %llu\n",
		      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_IP4RXPKTS]);
	dbg_vsnprintf("IPV6 pkts recvd                      %llu\n",
		      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_IP6RXPKTS]);
	dbg_vsnprintf("IPV4 frag recvd                      %llu\n",
		      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_IP4RXFRAGS]);
	dbg_vsnprintf("IPV6 frag recvd                      %llu\n",
		      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_IP6RXFRAGS]);

	if (iwdev->rf->rdma_ver >= IRDMA_GEN_2) {
		dbg_vsnprintf("IPV4 multicast pkts recvd            %llu\n",
			      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_IP4RXMCPKTS]);
		dbg_vsnprintf("IPV6 multicast pkts recvd            %llu\n",
			      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_IP6RXMCPKTS]);
		dbg_vsnprintf("IPV4 multicast octs recvd            %llu\n",
			      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_IP4RXMCOCTS]);
		dbg_vsnprintf("IPV6 multicast octs recvd            %llu\n",
			      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_IP6RXMCOCTS]);
	}

	dbg_vsnprintf("IPV4 octs trans                      %llu\n",
		      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_IP4TXOCTS]);
	dbg_vsnprintf("IPV6 octs trans                      %llu\n",
		      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_IP6TXOCTS]);
	dbg_vsnprintf("IPV4 pkts trans                      %llu\n",
		      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_IP4TXPKTS]);
	dbg_vsnprintf("IPV6 pkts trans                      %llu\n",
		      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_IP6TXPKTS]);
	dbg_vsnprintf("IPV4 frag trans                      %llu\n",
		      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_IP4TXFRAGS]);
	dbg_vsnprintf("IPV6 frag trans                      %llu\n",
		      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_IP6TXFRAGS]);
	dbg_vsnprintf("IPV4 multicast pkts trans            %llu\n",
		      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_IP4TXMCPKTS]);
	dbg_vsnprintf("IPV6 multicast pkts trans            %llu\n",
		      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_IP6TXMCPKTS]);
	dbg_vsnprintf("IPV4 multicast octs trans            %llu\n",
		      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_IP4TXMCOCTS]);
	dbg_vsnprintf("IPV6 multicast octs trans            %llu\n",
	       hw_stats->stats_val[IRDMA_HW_STAT_INDEX_IP6TXMCOCTS]);
	dbg_vsnprintf("IPV4 pkts recvd & discarded          %llu\n",
		      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_IP4RXDISCARD]);
	dbg_vsnprintf("IPV6 pkts recvd & discarded          %llu\n",
		      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_IP6RXDISCARD]);
	dbg_vsnprintf("IPV4 pkts recvd & truncated          %llu\n",
		      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_IP4RXTRUNC]);
	dbg_vsnprintf("IPV6 pkts recvd & truncated          %llu\n",
		      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_IP6RXTRUNC]);
	dbg_vsnprintf("IPV4 dtgms discarded (no ARP hit)    %llu\n",
		      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_IP4TXNOROUTE]);
	dbg_vsnprintf("IPV6 dtgms discarded (no ARP hit)    %llu\n",
		      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_IP6TXNOROUTE]);
	dbg_vsnprintf("Rx VLAN err                          %llu\n",
		      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_RXVLANERR]);

	dbg_vsnprintf("TCP sgmnts recvd                     %llu\n",
		      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_TCPRXSEGS]);
	dbg_vsnprintf("TCP sgmnts trans                     %llu\n",
		      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_TCPTXSEG]);
	dbg_vsnprintf("TCP sgmnts retrans                   %llu\n",
		      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_TCPRTXSEG]);
	dbg_vsnprintf("TCP sgmnts recvd w/ err              %llu\n",
		      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_TCPRXOPTERR]);
	dbg_vsnprintf("TCP sgmnts recvd w/ protocol-err     %llu\n",
		      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_TCPRXPROTOERR]);

	if (iwdev->rf->rdma_ver >= IRDMA_GEN_2) {
		dbg_vsnprintf("Rx CNP HANDLED                       %llu\n",
			      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_RXRPCNPHANDLED]);
		dbg_vsnprintf("Rx CNP IGNORED                       %llu\n",
			      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_RXRPCNPIGNORED]);
		dbg_vsnprintf("Tx CNP SENT                          %llu\n",
			      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_TXNPCNPSENT]);
	}

	dbg_vsnprintf("RDMA read-req msgs recvd             %llu\n",
		      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_RDMARXRDS]);
	dbg_vsnprintf("RDMA send-type msgs recvd            %llu\n",
		      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_RDMARXSNDS]);
	dbg_vsnprintf("RDMA write msgs recvd                %llu\n",
		      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_RDMARXWRS]);
	dbg_vsnprintf("RDMA read-req msgs sent              %llu\n",
		      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_RDMATXRDS]);
	dbg_vsnprintf("RDMA send-type msgs sent             %llu\n",
		      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_RDMATXSNDS]);
	dbg_vsnprintf("RDMA write msgs sent                 %llu\n",
		      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_RDMATXWRS]);
	dbg_vsnprintf("RDMA verb-bind ops                   %llu\n",
		      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_RDMAVBND]);
	dbg_vsnprintf("RDMA verb-inv ops                    %llu\n",
		      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_RDMAVINV]);

	if (iwdev->rf->rdma_ver >= IRDMA_GEN_2) {
		dbg_vsnprintf("UDP recvd                            %llu\n",
			      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_UDPRXPKTS]);
		dbg_vsnprintf("UDP trans                            %llu\n",
			      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_UDPTXPKTS]);
	}

	if (iwdev->rf->rdma_ver >= IRDMA_GEN_2 && iwdev->roce_mode) {
		dbg_vsnprintf("RX ECN MARKED PKTS           %llu\n",
			      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_RXNPECNMARKEDPKTS]);
	}
#ifdef MEV_SUPPORT

	if (iwdev->rf->rdma_ver >= IRDMA_GEN_3) {
		dbg_vsnprintf("RNR sent                     %llu\n",
			      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_RNR_SENT]);
		dbg_vsnprintf("RNR received                 %llu\n",
			      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_RNR_RCVD]);
		dbg_vsnprintf("ord limit count              %llu\n",
			      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_RDMAORDLMTCNT]);
		dbg_vsnprintf("ird limit count              %llu\n",
			      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_RDMAIRDLMTCNT]);
		dbg_vsnprintf("Rx ATS                       %llu\n",
			      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_RDMARXATS]);
		dbg_vsnprintf("Tx ATS                       %llu\n",
			      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_RDMATXATS]);
		dbg_vsnprintf("Nak Sequence Error           %llu\n",
			      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_NAKSEQERR]);
		dbg_vsnprintf("Nak Sequence Error Implied   %llu\n",
			      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_NAKSEQERR_IMPLIED]);
		dbg_vsnprintf("RTO                          %llu\n",
			      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_RTO]);
		dbg_vsnprintf("Rcvd Out of order packets    %llu\n",
			      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_RXOOOPKTS]);
		dbg_vsnprintf("CRC errors                   %llu\n",
			      hw_stats->stats_val[IRDMA_HW_STAT_INDEX_ICRCERR]);
	}
#endif

#ifdef DEBUG_SUPPORT_INTERNAL
	if (iwdev->rf->rdma_ver == IRDMA_GEN_2) {
#define GLPES_RDMARXOOODDPL	0x0055E010
#define GLPES_TCPRXONEHOLE	0x0055E020
#define GLPES_TCPRXTWOHOLE	0x0055E028
#define GLPES_TCPRXTHREEHOLE	0x0055E030
#define GLPES_TCPRXFOURHOLE	0x0055E038

		dbg_vsnprintf("GLPES_RDMARXOOODDPL              = 0x%x\n", rd32(dev->hw, GLPES_RDMARXOOODDPL));
		dbg_vsnprintf("GLPES_GLPES_TCPRXONEHOLE         = 0x%x\n", rd32(dev->hw, GLPES_TCPRXONEHOLE));
		dbg_vsnprintf("GLPES_GLPES_TCPRXTWOHOLE         = 0x%x\n", rd32(dev->hw, GLPES_TCPRXTWOHOLE));
		dbg_vsnprintf("GLPES_GLPES_TCPRXTHREEHOLE       = 0x%x\n", rd32(dev->hw, GLPES_TCPRXTHREEHOLE));
		dbg_vsnprintf("GLPES_GLPES_TCPRXFOURHOLE        = 0x%x\n", rd32(dev->hw, GLPES_TCPRXFOURHOLE));
		read_cache_stats(dev);
	}

	/* The following debug registers are useful for debugging issues like DVM issues */
	if (iwdev->rf->rdma_ver == IRDMA_GEN_2) {
		u32 value;

#define GGLQF_DFD_PE_CNT_SEL 0x471290
		value = rd32(dev->hw, GGLQF_DFD_PE_CNT_SEL);
		dbg_vsnprintf("GGLQF_DFD_PE_CNT_SEL (0x471290) = 0x%x. Expecting 0xec41148 for following values to be trusted.\n", value);
		value = rd32(dev->hw, 0x4712c4);
		dbg_vsnprintf("QH hit (0x4712c4) = 0x%x\n", value);
		value = rd32(dev->hw, 0x4712c8);
		dbg_vsnprintf("QH miss(0x4712c8) = 0x%x\n", value);

		value = rd32(dev->hw, 0x00455500);
		dbg_vsnprintf("APBVT CNT(0x00455500) = 0x%x\n", value);

		{
			u32 regs[10] = {0x50AC2C, 0x50AC30, 0x50AC34,
					0x50AC38, 0x50AC3C, 0x50AC40,
					0x50AC44, 0x50AC48, 0x50AC4C, 0x50AC50};
			int i;

			for (i = 0; i < 10; ++i) {
				value = rd32(dev->hw, regs[i]);
				dbg_vsnprintf("reg(0x%x) = 0x%x\n", regs[i], value);
			}
		}
	}
#endif /* DEBUG_INTERNAL */
	cmddone = true;
}

#if 0
/**
 * dump_mem_cmd -
 * @rf: RDMA PCI function
 * @cbuf: character buffer
 */
static void dump_mem_cmd(struct irdma_pci_f *rf, char *cbuf)
{
	static char *addr;
	static int offset;
	static int groupsize;
	static size_t dumpsize;
	static size_t lendumped;
	static int ascii;
	int delta = 0;
	int ret;

	if (cmdnew) {
		if (sscanf(cbuf, "%llx%n", (u64 *)&addr, &delta) == 0)
			return;

		cbuf += delta;
		if (sscanf(cbuf, "%zu%n", &dumpsize, &delta) == 0)
			return;

		cbuf += delta;
		ret = kstrtoint(cbuf, 10, &groupsize);
		if (ret < 0)
			groupsize = 4;

		if (dumpsize == 0 || (dumpsize & 0x80000000))
			dumpsize = 32;

		switch (groupsize) {
		case 1:
			ascii = true;
			break;
		case 8:
			break;
		default:
			groupsize = 4;
		}
		offset = 0;
	}

	dump_to_buf(addr, dumpsize, offset, 16, groupsize, ascii,
		       &lendumped);

	addr += lendumped;
	offset += lendumped;
	dumpsize -= lendumped;
	if (dumpsize == 0)
		cmddone = true;
}

/**
 * dump_pmem_cmd -
 * @rf: RDMA PCI function
 * @cbuf: char buffer
 */
static void dump_pmem_cmd(struct irdma_pci_f *rf, char *cbuf)
{
	static ulong paddr;
	static char *addr;
	static int offset;
	static int groupsize;
	static size_t dumpsize;
	static size_t lendumped;
	static int ascii;
	int delta = 0;
	int ret;

	if (cmdnew) {
		if (sscanf(cbuf, "%llx%n", (u64 *)&paddr, &delta) == 0)
			return;

		cbuf += delta;
		if (sscanf(cbuf, "%zu%n", &dumpsize, &delta) == 0)
			return;

		cbuf += delta;
		ret = kstrtoint(cbuf, 10, &groupsize);
		if (ret < 0)
			groupsize = 4;

		if (dumpsize == 0 || (dumpsize & 0x80000000))
			dumpsize = 32;

		switch (grpsize) {
		case 1:
			ascii = true;
			break;
		case 8:
			break;
		default:
			groupsize = 4;
		}
		offset = 0;
	}

	addr = ioremap(paddr, dumpsize);
	if (addr) {
		dump_to_buf(addr, dumpsize, offset, 16, groupsize, ascii,
			       &lendumped);
		iounmap(addr);
	}

	paddr += lendumped;
	offset += lendumped;
	dumpsize -= lendumped;
	if (dumpsize == 0)
		cmddone = true;
}

#endif
/**
 * irdma_dbg_dump_read - read the dump data
 * @filp: the opened file
 * @buf: where to write the data for the user to read
 * @count: the size of the user's buffer
 * @ppos: file position offset
 *
 * When a read happens, the file system will keep calling this routine
 * until it returns with no data.  This is because the buffers may be too
 * small to contain all the data at once.  Some of the cmds count on being
 * able to return the results in pieces.
 */
static ssize_t irdma_dbg_dump_read(struct file *filp,
				   char __user *buf,
				   size_t count,
				   loff_t *ppos)
{
	struct irdma_handler *hdl = filp->private_data;
	struct irdma_device *iwdev = hdl->iwdev;
	struct irdma_pci_f *rf;
	int bytes_not_copied;
	int len;

	rf = iwdev->rf;
	if (cmddone) {
		cmdnew = true;
		cmddone = false;
	} else {
		if (cmdnew)
			dbg_vsnprintf("cmd: %s", cmd_buf);

		if (strncasecmp(cmd_buf, "qp ", 3) == 0)
			dump_qp_cmd(rf, &cmd_buf[3]);
		else if (strncasecmp(cmd_buf, "cq ", 3) == 0)
			dump_cq_cmd(rf, &cmd_buf[3]);
		else if (strncasecmp(cmd_buf, "sw-stats", 8) == 0)
			dump_stats_cmd(iwdev);
		else if (strncasecmp(cmd_buf, "hw-stats", 8) == 0)
			dump_hw_stats_cmd(iwdev);
#if defined(DEBUG_SUPPORT_INTERNAL) || defined(MEV_RCA)
		else if (strncasecmp(cmd_buf, "cqp-nop", 7) == 0)
			irdma_cqp_nop(&rf->sc_dev);
#endif /* defined(DEBUG_SUPPORT_INTERNAL) || defined(MEV_RCA) */
#ifdef MEV_RCA
		else if (strncasecmp(cmd_buf, "cqp-rq", 6) == 0)
			irdma_cqp_rca_post_rqes(&rf->sc_dev);
#endif /* MEV_RCA */
#ifdef DEBUG_SUPPORT_INTERNAL
		else if (strncasecmp(cmd_buf, "cqp", 3) == 0)
			dump_cqp_cmd(rf, &cmd_buf[3]);
		else if (strncasecmp(cmd_buf, "hmc", 3) == 0)
			dump_hmc_cmd(rf, &cmd_buf[4]);
		else if (strncasecmp(cmd_buf, "mr ", 3) == 0)
			dump_mr_cmd(rf, &cmd_buf[3]);
		else if (strncasecmp(cmd_buf, "hw-regs", 7) == 0)
			dumpCSR(&rf->sc_dev);
#if 0
		else if (strncasecmp(cmd_buf, "mem ", 4) == 0)
			dump_mem_cmd(rf, &cmd_buf[4]);

		else if (strncasecmp(cmd_buf, "pmem ", 5) == 0)
			dump_pmem_cmd(rf, &cmd_buf[5]);

		else if (strncasecmp(cmd_buf, "rdreg ", 6) == 0)
			dump_rdreg_cmd(rf, &cmd_buf[6]);
#endif
		else if (strncasecmp(cmd_buf, "reset ", 5) == 0) {
			if (rf->sc_dev.hw_attrs.uk_attrs.hw_rev == IRDMA_GEN_2 && !rf->reset) {
				struct iidc_core_dev_info *cdev_info = rf->cdev;

				irdma_pr_err("Reset request is requested\n");
				rf->reset = true;
				cdev_info->ops->request_reset(cdev_info, IIDC_PFR);
			}
		} else if (strncasecmp(cmd_buf, "en filter", 9) == 0) {
			struct iidc_core_dev_info *cdev_info = rf->cdev;

			cdev_info->ops->update_vport_filter(cdev_info, iwdev->vsi_num, true);
			irdma_pr_err("enabling filter, vsi_num=%d\n", iwdev->vsi_num);
		} else if (strncasecmp(cmd_buf, "dis filter", 10) == 0) {
			struct iidc_core_dev_info *cdev_info = rf->cdev;

			cdev_info->ops->update_vport_filter(cdev_info, iwdev->vsi_num, false);
			irdma_pr_err("disabling filter vsi_num=%d\n", iwdev->vsi_num);
		} else if (strncasecmp(cmd_buf, "suspend", 7) == 0) {
			irdma_sc_suspend_resume_qps(&iwdev->vsi, IRDMA_OP_SUSPEND);
		} else if (strncasecmp(cmd_buf, "resume", 6) == 0) {
			irdma_sc_suspend_resume_qps(&iwdev->vsi, IRDMA_OP_RESUME);
		} else {
			dump_help();
		}
#else
		else
			dump_help();
#endif
		cmdnew = false;
	}

#ifndef EXTERNAL_RELEASE
	/* be sure to not write beyond the end of available data */
#endif
	len = min_t(int, count, (irdma_dbg_dump_data_len - *ppos));
	if (len == 0) {
		*ppos = 0;
		irdma_dbg_dump_data_len = 0;
		return 0;
	}

	bytes_not_copied =
	    copy_to_user(buf, &irdma_dbg_dump_buf[*ppos], len);
	if (bytes_not_copied) {
		dev_warn(iwdev->ibdev.dma_device,
			 "copy_to_user returned 0x%x\n", bytes_not_copied);
	}

	*ppos += len;
	if (*ppos >= irdma_dbg_dump_data_len) {
		*ppos = 0;
		irdma_dbg_dump_data_len = 0;
	}

	return len;
}

/**
 * irdma_dbg_dump_write - trigger a datadump snapshot
 * @filp: the opened file
 * @buf: where to find the user's data
 * @count: the length of the user's data
 * @ppos: file position offset
 */
static ssize_t irdma_dbg_dump_write(struct file *filp,
				    const char __user *buf,
				    size_t count,
				    loff_t *ppos)
{
	int bytes_not_copied;

	/* don't allow partial writes */
	if (*ppos != 0)
		return 0;
	if (count >= sizeof(cmd_buf))
		return -ENOSPC;

	bytes_not_copied = copy_from_user(cmd_buf, buf, count);
	if (bytes_not_copied < 0)
		return bytes_not_copied;
	if (bytes_not_copied > 0)
		count -= bytes_not_copied;

	cmd_buf[count] = '\0';
	cmdnew = true;
	cmddone = false;
	*ppos = 0;
	irdma_dbg_dump_data_len = 0;
	return count;
}

/**
 * irdma_dbg_prep_dump_buf
 * @hdl: the iWARP handler we're working with
 * @buflen: the desired buffer length
 *
 * Return positive if success, 0 if failed
 */
static int irdma_dbg_prep_dump_buf(struct irdma_handler *hdl, int buflen)
{
#ifndef EXTERNAL_RELEASE
	/* if not already big enough, prep for re alloc */
#endif
	if (irdma_dbg_dump_buf_len && irdma_dbg_dump_buf_len < buflen) {
		kfree(irdma_dbg_dump_buf);
		irdma_dbg_dump_buf_len = 0;
		irdma_dbg_dump_buf = NULL;
	}

#ifndef EXTERNAL_RELEASE
	/* get a new buffer if needed */
#endif
	if (!irdma_dbg_dump_buf) {
		irdma_dbg_dump_buf = kzalloc(buflen, GFP_KERNEL);
		if (!irdma_dbg_dump_buf) {
			irdma_dbg_dump_buf_len = 0;
			pr_err("%s: memory alloc for snapshot failed\n",
			       __func__);
		} else {
			irdma_dbg_dump_buf_len = buflen;
			pr_err("%s: irdma_dbg_dump_buf_len = %d\n",
			       __func__, (int)irdma_dbg_dump_buf_len);
		}
	}

	return irdma_dbg_dump_buf_len;
}

static const struct file_operations irdma_dbg_dump_fops = {
	.owner = THIS_MODULE,
	.open = simple_open,
	.read = irdma_dbg_dump_read,
	.write = irdma_dbg_dump_write,
};

/**
 * irdma_dbg_pf_init - setup the debugfs directory for the pf
 * @hdl: the iWARP handler that is starting up
 */
void irdma_dbg_pf_init(struct irdma_handler *hdl)
{
	const char *name = pci_name(hdl->iwdev->rf->pcidev);
	struct dentry *pfile __attribute__ ((unused));

	spin_lock_init(&hdl->uctx_list_lock);
	INIT_LIST_HEAD(&hdl->ucontext_list);
	if (irdma_dbg_prep_dump_buf(hdl, IRDMA_DUMP_BUF_SIZE) == 0) {
		pr_err("irdma_dbg_pf_init: unable to allocate debugfs dump buffer\n");
		return;
	}

	hdl->irdma_dbg_dentry = debugfs_create_dir(name, irdma_dbg_root);
	if (hdl->irdma_dbg_dentry)
		pfile =
		    debugfs_create_file("dump", 0600, hdl->irdma_dbg_dentry,
					hdl, &irdma_dbg_dump_fops);
	else
		pr_err("%s: debugfs entry for %s failed\n", __func__, name);
}

/**
 * irdma_dbg_pf_exit - clear out the pf's debugfs entries
 * @hdl: the iWARP handler that is stopping
 */
void irdma_dbg_pf_exit(struct irdma_handler *hdl)
{
	if (hdl) {
		pr_err("%s: removing debugfs entries\n", __func__);
		debugfs_remove_recursive(hdl->irdma_dbg_dentry);
		hdl->irdma_dbg_dentry = NULL;
	}
}

/**
 * irdma_dbg_init - start up debugfs for the driver
 */
void irdma_dbg_init(void)
{
	irdma_dbg_root = debugfs_create_dir(irdma_driver_name, NULL);
	if (!irdma_dbg_root)
		pr_err("%s: init of debugfs failed\n", __func__);
}

/**
 * irdma_dbg_exit - clean out the driver's debugfs entries
 */
void irdma_dbg_exit(void)
{
	kfree(irdma_dbg_dump_buf);
	irdma_dbg_dump_buf_len = 0;
	irdma_dbg_dump_buf = NULL;
	debugfs_remove_recursive(irdma_dbg_root);
}

#endif /* CONFIG_DEBUG_FS */
#endif /* DEBUG_SUPPORT */
