/* Copyright (c) 2015 - 2023 Intel Corporation */
#ifndef LICENSE_REPLACED
/* For licensing information, see the file 'LICENSE' in the root folder */
#endif /* LICENSE_REPLACED */
#include "main.h"
#ifdef IIDC_UPSTREAM
#include "../../../net/ethernet/intel/ice/ice.h"
#endif
#ifdef IRDMA_FLOW_CONTROL_WARNING
/* TODO: Adding this here is not ideal. Can we remove this warning now? */
#include "icrdma_hw.h"
#endif
#ifndef UPSTREAM_RELEASE
#define DRV_VER_MAJOR @VERSION_MAJOR@
#define DRV_VER_MINOR @VERSION_MINOR@
#define DRV_VER_BUILD @VERSION_BUILD@
#ifdef MEV_HVL
#define DRV_VER	__stringify(DRV_VER_MAJOR) "."		\
	__stringify(DRV_VER_MINOR) "." __stringify(DRV_VER_BUILD) "-hvl"
#else
#define DRV_VER	__stringify(DRV_VER_MAJOR) "."		\
	__stringify(DRV_VER_MINOR) "." __stringify(DRV_VER_BUILD)
#endif /* MEV_HVL */
#endif

#ifndef UPSTREAM_RELEASE
static u8 resource_profile;
module_param(resource_profile, byte, 0444);
MODULE_PARM_DESC(resource_profile, "Resource Profile: 0=PF only(default), 1=Weighted VF, 2=Even Distribution");

#ifdef VCHNL_PRIVILEGED
static unsigned short max_rdma_vfs = IRDMA_MAX_PE_ENA_VF_COUNT;
module_param(max_rdma_vfs, ushort, 0444);
MODULE_PARM_DESC(max_rdma_vfs, "Maximum VF count: 0-32, default=32");

#endif
bool irdma_upload_context;
module_param(irdma_upload_context, bool, 0644);
MODULE_PARM_DESC(irdma_upload_context, "Upload QP context, default=false");

static unsigned int limits_sel = 3;
module_param(limits_sel, uint, 0444);
MODULE_PARM_DESC(limits_sel, "Resource limits selector, Range: 0-7, default=3");

static unsigned int gen1_limits_sel = 1;
module_param(gen1_limits_sel, uint, 0444);
MODULE_PARM_DESC(gen1_limits_sel, "x722 resource limits selector, Range: 0-5, default=1");

static unsigned int roce_ena;
module_param(roce_ena, uint, 0444);
MODULE_PARM_DESC(roce_ena, "RoCE enable: 1=enable RoCEv2 on all ports (not supported on x722), 0=iWARP(default)");

static ulong roce_port_cfg;
module_param(roce_port_cfg, ulong, 0444);
MODULE_PARM_DESC(roce_port_cfg, "RoCEv2 per port enable: 1=port0 RoCEv2 all others iWARP, 2=port1 RoCEv2 etc. not supported on X722");

#ifdef RDPU_ASSIST_WA
static u8 rdpu_assist = 0x0F;
module_param(rdpu_assist, byte, 0444);
MODULE_PARM_DESC(rdpu_assist, "RDPU rate of assist, Range: 0-0xF");

#endif
#ifndef REM_ENDPOINT_TRACKING_UNSUPPORTED

static bool en_rem_endpoint_trk;
module_param(en_rem_endpoint_trk, bool, 0444);
MODULE_PARM_DESC(en_rem_endpoint_trk, "Remote Endpoint Tracking: 1=enabled (not supported on x722), 0=disabled(default)");
#endif

static u8 fragment_count_limit = 6;
module_param(fragment_count_limit, byte, 0444);
MODULE_PARM_DESC(fragment_count_limit, "adjust maximum values for queue depth and inline data size, default=6, Range: 2-13");

#ifdef MEV_HVL
static u8 rrf_m = 8;
module_param(rrf_m, byte, 0444);
MODULE_PARM_DESC(rrf_m, "rrf_multiplier for HMC resource default to 8");

static u8 xf_m = 16;
module_param(xf_m, byte, 0444);
MODULE_PARM_DESC(xf_m, "xf_multiplier for HMC resource default to 16");

static u8 min_ird = 8;
module_param(min_ird, byte, 0444);
MODULE_PARM_DESC(min_ird, "minimum ird for Q1 resources");

static bool host_mem_mrte;
module_param(host_mem_mrte, bool, 0444);
MODULE_PARM_DESC(host_mem_mrte, "true if mrte host memory false local memory default = false");

#ifdef MEV_RCA
bool irdma_rca_ena;
module_param(irdma_rca_ena, bool, 0444);
MODULE_PARM_DESC(irdma_rca_ena, "driver enable rca default=false");

bool irdma_rca_rq_post = true;
module_param(irdma_rca_rq_post, bool, 0444);
MODULE_PARM_DESC(irdma_rca_rq_post, "Post RQEs at RCA's RQ initialization, default=true");

bool irdma_rca_rq_polarity = true;
module_param(irdma_rca_rq_polarity, bool, 0444);
MODULE_PARM_DESC(irdma_rca_rq_polarity, "Use polarity as Frag_Valid in RCA's RQEs, default=true");

unsigned int irdma_rca_rq_size = IRDMA_CQP_SW_RQSIZE_2048;
module_param(irdma_rca_rq_size, uint, 0444);
MODULE_PARM_DESC(irdma_rca_rq_size, "RCA CQP RQ size, default=2048");

#endif /* MEV_RCA */
unsigned int wa_mem_pages;
module_param(wa_mem_pages, uint, 0444);
MODULE_PARM_DESC(wa_mem_pages, "to override memory pages in local memory default = 0");

unsigned int hw_type_wa = 0x45;
module_param(hw_type_wa, uint, 0444);
MODULE_PARM_DESC(hw_type_wa, "to setup hw work around for specific release default = Veloce SWR39, if set to 0xFFFF, take hw_wa_bitmask value.");

static ulong hw_wa_bitmask;
module_param(hw_wa_bitmask, ulong, 0444);
MODULE_PARM_DESC(hw_wa_bitmask, "Used to manually set HW WAs; default=0");

/*
 * crt_ena is to allow HVL to indicate CRT protocol family to LIBCPF
 * (imc cp) only for setting hmc function cqp operation.
 */
static bool crt_ena;
module_param(crt_ena, bool, 0444);
MODULE_PARM_DESC(crt_ena, "CRT enable, 1=enable CRT protocol on all ports, 0=RoCEv2 on all ports");

/*
 * rdma_key - allow rdma load if key is success
 * default is rdma load fail if proper key is not entered
 */
static unsigned int rdma_key;
module_param(rdma_key, uint, 0444);
MODULE_PARM_DESC(rdma_key, "Driver probe on GEN3 B0 will fail by default if proper key is not used");

#endif /* MEV_HVL */
#ifdef DCQCN_CONFIG_SUPPORT
/******************Advanced RoCEv2 congestion knobs***********************************************/
static bool dcqcn_enable;
module_param(dcqcn_enable, bool, 0444);
MODULE_PARM_DESC(dcqcn_enable, "enables DCQCN algorithm for RoCEv2 on all ports, default=false ");

static bool dcqcn_cc_cfg_valid;
module_param(dcqcn_cc_cfg_valid, bool, 0444);
MODULE_PARM_DESC(dcqcn_cc_cfg_valid, "set DCQCN parameters to be valid, default=false");

static u8 dcqcn_min_dec_factor = 1;
module_param(dcqcn_min_dec_factor, byte, 0444);
MODULE_PARM_DESC(dcqcn_min_dec_factor, "set minimum percentage factor by which tx rate can be changed for CNP, Range: 1-100, default=1");

static u8 dcqcn_min_rate_MBps;
module_param(dcqcn_min_rate_MBps, byte, 0444);
MODULE_PARM_DESC(dcqcn_min_rate_MBps, "set minimum rate limit value, in MBits per second, default=0");

static u8 dcqcn_F = 5;
module_param(dcqcn_F, byte, 0444);
MODULE_PARM_DESC(dcqcn_F, "set number of times to stay in each stage of bandwidth recovery, default=5");

static unsigned short dcqcn_T = 0x37;
module_param(dcqcn_T, ushort, 0444);
MODULE_PARM_DESC(dcqcn_T, "set number of usecs that should elapse before increasing the CWND in DCQCN mode, default=0x37");

static unsigned int dcqcn_B = 0x249f0;
module_param(dcqcn_B, uint, 0444);
MODULE_PARM_DESC(dcqcn_B, "The number of bytes to transmit before updating CWND in DCQCN mode. default=0x249f0");

static unsigned short dcqcn_rai_factor = 1;
module_param(dcqcn_rai_factor, ushort, 0444);
MODULE_PARM_DESC(dcqcn_rai_factor, "set number of MSS to add to the congestion window in additive increase mode, default=1");

static unsigned short dcqcn_hai_factor = 5;
module_param(dcqcn_hai_factor, ushort, 0444);
MODULE_PARM_DESC(dcqcn_hai_factor, "set number of MSS to add to the congestion window in hyperactive increase mode, default=5");

static unsigned int dcqcn_rreduce_mperiod = 50;
module_param(dcqcn_rreduce_mperiod, uint, 0444);
MODULE_PARM_DESC(dcqcn_rreduce_mperiod, "set minimum time between 2 consecutive rate reductions for a single flow, default=50");

/**************************************************************************************************/
#endif /* DCQCN_CONFIG_SUPPORT */
#endif /* UPSTREAM_RELEASE */

MODULE_ALIAS("i40iw");
MODULE_AUTHOR("Intel Corporation, <e1000-rdma@lists.sourceforge.net>");
MODULE_DESCRIPTION("Intel(R) Ethernet Protocol Driver for RDMA");
MODULE_LICENSE("Dual BSD/GPL");
#ifndef UPSTREAM_RELEASE
MODULE_VERSION(DRV_VER);
#endif

#ifndef UPSTREAM_RELEASE
/**
 * set_protocol_used - set protocol_used against HW generation and roce_ena flag
 * @rf: RDMA PCI function
 * @roce_ena: RoCE enabled bit flag
 */
static inline void set_protocol_used(struct irdma_pci_f *rf, uint roce_ena)
{
	switch (rf->rdma_ver) {
#ifdef MEV_SUPPORT
	case IRDMA_GEN_3:
#ifdef MMG_SUPPORT
	case IRDMA_GEN_4:
#endif
		rf->protocol_used = IRDMA_ROCE_PROTOCOL_ONLY;
#ifdef MEV_HVL
		if (crt_ena)
			rf->protocol_used = IRDMA_CRT_PROTOCOL_ONLY;
#endif
		break;
#endif
	case IRDMA_GEN_2:
#ifdef DEVLINK_PORT_CFG_WA
		rf->protocol_used = roce_ena & BIT(PCI_FUNC(rf->pcidev->devfn)) ?
			IRDMA_ROCE_PROTOCOL_ONLY : IRDMA_IWARP_PROTOCOL_ONLY;
#else
		rf->protocol_used = roce_ena ? IRDMA_ROCE_PROTOCOL_ONLY : IRDMA_IWARP_PROTOCOL_ONLY;
#endif

		break;
	case IRDMA_GEN_1:
		rf->protocol_used = IRDMA_IWARP_PROTOCOL_ONLY;
		break;
	}
}

/**
 * irdma_set_rf_user_cfg_params - Setup RF configurations from module parameters
 * @rf: RDMA PCI function
 */
void irdma_set_rf_user_cfg_params(struct irdma_pci_f *rf)
{
#ifndef EXTERNAL_RELEASE
	/*TODO: Fixup range checks on all integer module params */
#endif
	if (limits_sel > 7)
		limits_sel = 7;

	if (gen1_limits_sel > 5)
		gen1_limits_sel = 5;

	rf->limits_sel = (rf->rdma_ver == IRDMA_GEN_1) ? gen1_limits_sel :
							 limits_sel;
#ifndef EXTERNAL_RELEASE
/* Once devlink options are supported in ICE, the roce_port_cfg module
 * parameter should be made obolete and ignored.
 */
#endif
	if (roce_ena)
		pr_warn_once("irdma: Because roce_ena is ENABLED, roce_port_cfg will be ignored.");
	set_protocol_used(rf, roce_ena ? 0xFFFFFFFF : roce_port_cfg);
	rf->rsrc_profile = (resource_profile < IRDMA_HMC_PROFILE_EQUAL) ?
			    (u8)resource_profile + IRDMA_HMC_PROFILE_DEFAULT :
			    IRDMA_HMC_PROFILE_DEFAULT;
#ifdef VCHNL_PRIVILEGED
	if (max_rdma_vfs > IRDMA_MAX_PE_ENA_VF_COUNT) {
		pr_warn_once("irdma: Requested VF count [%d] is above max supported. Setting to %d.",
			     max_rdma_vfs, IRDMA_MAX_PE_ENA_VF_COUNT);
		max_rdma_vfs = IRDMA_MAX_PE_ENA_VF_COUNT;
	}
	rf->max_rdma_vfs = (rf->rsrc_profile != IRDMA_HMC_PROFILE_DEFAULT) ?
				max_rdma_vfs : 0;
#endif /* RDMA_VF_DRIVER */
#ifdef RDPU_ASSIST_WA
	if (rdpu_assist > 0xF)
		rdpu_assist = 0xF;
	rf->rdpu_assist = rdpu_assist;
#endif
#ifndef REM_ENDPOINT_TRACKING_UNSUPPORTED
	rf->en_rem_endpoint_trk = en_rem_endpoint_trk;
#endif
#ifdef CONFIG_DEV_LIMITS
	rf->fragcnt_limit = fragment_count_limit;
	if (rf->fragcnt_limit > 13 || rf->fragcnt_limit < 2) {
		rf->fragcnt_limit = 6;
		pr_warn_once("irdma: Requested [%d] fragment count limit out of range (2-13), setting to default=6.",
			     fragment_count_limit);
	}
#endif
#ifdef DCQCN_CONFIG_SUPPORT
	rf->dcqcn_ena = dcqcn_enable;

	/* Skip over all checking if no dcqcn */
	if (!dcqcn_enable)
		return;

	rf->dcqcn_params.cc_cfg_valid = dcqcn_cc_cfg_valid;
	rf->dcqcn_params.dcqcn_b = dcqcn_B;

#define DCQCN_B_MAX GENMASK(25, 0)
	if (rf->dcqcn_params.dcqcn_b > DCQCN_B_MAX) {
		rf->dcqcn_params.dcqcn_b = DCQCN_B_MAX;
		pr_warn_once("irdma: Requested [%d] dcqcn_b value too high, setting to %d.",
			     dcqcn_B, rf->dcqcn_params.dcqcn_b);
	}

#define DCQCN_F_MAX 8
	rf->dcqcn_params.dcqcn_f = dcqcn_F;
	if (dcqcn_F > DCQCN_F_MAX) {
		rf->dcqcn_params.dcqcn_f = DCQCN_F_MAX;
		pr_warn_once("irdma: Requested [%d] dcqcn_f value too high, setting to %d.",
			     dcqcn_F, DCQCN_F_MAX);
	}

	rf->dcqcn_params.dcqcn_t = dcqcn_T;
	rf->dcqcn_params.hai_factor = dcqcn_hai_factor;
	rf->dcqcn_params.min_dec_factor = dcqcn_min_dec_factor;
	if (dcqcn_min_dec_factor < 1 || dcqcn_min_dec_factor > 100) {
		rf->dcqcn_params.dcqcn_b = 1;
		pr_warn_once("irdma: Requested [%d] dcqcn_min_dec_factor out of range (1-100) , setting to default=1",
			     dcqcn_min_dec_factor);
	}

	rf->dcqcn_params.min_rate = dcqcn_min_rate_MBps;
	rf->dcqcn_params.rai_factor = dcqcn_rai_factor;
	rf->dcqcn_params.rreduce_mperiod = dcqcn_rreduce_mperiod;
#endif
}

static int irdma_init_dbg_and_configfs(void)
{
#ifdef CONFIGFS_SUPPORT
#if IS_ENABLED(CONFIG_CONFIGFS_FS)
	int ret;

#endif /* CONFIG_CONFIGFS_FS */
#endif /* CONFIGFS_SUPPORT */
#ifdef DEBUG_SUPPORT
#ifdef CONFIG_DEBUG_FS
	irdma_dbg_init();
#endif
#endif
#ifdef CONFIGFS_SUPPORT
#if IS_ENABLED(CONFIG_CONFIGFS_FS)
	ret = irdma_configfs_init();
	if (ret) {
		irdma_pr_err("Failed to register irdma to configfs subsystem\n");
#ifdef DEBUG_SUPPORT
#ifdef CONFIG_DEBUG_FS
		irdma_dbg_exit();
#endif
#endif
		return ret;
	}
#endif /* CONFIG_CONFIGFS_FS */
#endif /* CONFIGFS_SUPPORT */
	return 0;
}

static inline void irdma_deinit_dbg_and_configfs(void)
{
#ifdef CONFIGFS_SUPPORT
#if IS_ENABLED(CONFIG_CONFIGFS_FS)
	irdma_configfs_exit();
#endif
#endif
#ifdef DEBUG_SUPPORT
#ifdef CONFIG_DEBUG_FS
	irdma_dbg_exit();
#endif
#endif
}

#endif /* UPSTREAM_RELEASE */
#ifdef VCHNL_SUPPORT
static int irdma_vchnl_receive(struct iidc_core_dev_info *cdev_info, u32 vf_id,
			       u8 *msg, u16 len)
{
	struct irdma_device *iwdev = dev_get_drvdata(&cdev_info->adev->dev);
	struct irdma_sc_dev *dev = &iwdev->rf->sc_dev;

	if (WARN_ON(!len || !msg))
		return -EINVAL;

#ifndef EXTERNAL_RELEASE
	/* TODO: Remove (u16) typecast of vf_id when IDC changes vf_id to 16-bits */
#endif
#if defined(VCHNL_PRIVILEGED) && defined(VCHNL_UNPRIVILEGED)
	return dev->vchnl_if->vchnl_recv(dev, (u16)vf_id, msg, len);
#elif defined(VCHNL_PRIVILEGED)
	return irdma_vchnl_recv_pf(dev, (u16)vf_id, msg, len);
#elif defined(VCHNL_UNPRIVILEGED)
	return irdma_vchnl_req_recv(dev, (u16)vf_id, msg, len);
#else
	#error "Either VCHNL_PRIVILEGED or VCHNL_UNPRIVILEGED must be defined"
#endif /* defined(VCHNL_PRIVILEGED) && defined(VCHNL_UNPRIVILEGED) */
}

#ifdef VCHNL_PRIVILEGED
int irdma_vchnl_send_pf(struct irdma_sc_dev *dev, u16 vf_id, u8 *msg, u16 len)
{
	struct iidc_core_dev_info *cdev_info = dev_to_rf(dev)->cdev;

	cdev_info->ops->vc_send(cdev_info, vf_id, msg, len);

	return 0;
}

#endif /* VCHNL_PRIVILEGED */
#ifdef VCHNL_UNPRIVILEGED
int irdma_vchnl_send_sync(struct irdma_sc_dev *dev, u8 *msg, u16 len,
			  u8 *recv_msg, u16 *recv_len)
{
	struct iidc_core_dev_info *cdev_info = dev_to_rf(dev)->cdev;
	int ret;

	ret = cdev_info->ops->vc_send_sync(cdev_info, msg, len, recv_msg,
					   recv_len);
	if (ret == -ETIMEDOUT) {
		ibdev_err(&(dev_to_rf(dev)->iwdev->ibdev),
			  "Virtual channel Req <-> Resp completion timeout = 0x%x\n", ret);
		dev->vchnl_up = false;
	}

	return ret;
}

#endif /* VCHNL_UNPRIVILEGED */
#if defined(VCHNL_PRIVILEGED) && defined(VCHNL_UNPRIVILEGED)
static struct irdma_vchnl_if irdma_vchnl_if_pf = {
	.vchnl_recv = irdma_vchnl_recv_pf,
};

static struct irdma_vchnl_if irdma_vchnl_if_req = {
	.vchnl_recv = irdma_vchnl_req_recv,
};

#endif /* defined(VCHNL_PRIVILEGED) && defined(VCHNL_UNPRIVILEGED) */
#endif /* VCHNL_SUPPORT */
static void irdma_prep_tc_change(struct irdma_device *iwdev)
{
	iwdev->vsi.tc_change_pending = true;
	irdma_sc_suspend_resume_qps(&iwdev->vsi, IRDMA_OP_SUSPEND);

	/* Wait for all qp's to suspend */
	wait_event_timeout(iwdev->suspend_wq,
			   !atomic_read(&iwdev->vsi.qp_suspend_reqs),
			   msecs_to_jiffies(IRDMA_EVENT_TIMEOUT_MS));

#ifndef EXTERNAL_RELEASE
	/* TBD: CVL does irdma_ws_reset(). What about MEV? */
#endif
	if (iwdev->rf->rdma_ver == IRDMA_GEN_2)
		irdma_ws_reset(&iwdev->vsi);
}

static void irdma_log_invalid_mtu(u16 mtu, struct irdma_sc_dev *dev)
{
	if (mtu < IRDMA_MIN_MTU_IPV4)
		ibdev_warn(to_ibdev(dev),
			   "MTU setting [%d] too low for RDMA traffic. Minimum MTU is 576 for IPv4\n",
			   mtu);
	else if (mtu < IRDMA_MIN_MTU_IPV6)
		ibdev_warn(to_ibdev(dev),
			   "MTU setting [%d] too low for RDMA traffic. Minimum MTU is 1280 for IPv6\\n",
			   mtu);
}

static void irdma_fill_qos_info(struct irdma_l2params *l2params,
				struct iidc_qos_params *qos_info)
{
	int i;

	l2params->num_tc = qos_info->num_tc;
	l2params->vsi_prio_type = qos_info->vport_priority_type;
	l2params->vsi_rel_bw = qos_info->vport_relative_bw;
	for (i = 0; i < l2params->num_tc; i++) {
		l2params->tc_info[i].egress_virt_up =
			qos_info->tc_info[i].egress_virt_up;
		l2params->tc_info[i].ingress_virt_up =
			qos_info->tc_info[i].ingress_virt_up;
		l2params->tc_info[i].prio_type = qos_info->tc_info[i].prio_type;
		l2params->tc_info[i].rel_bw = qos_info->tc_info[i].rel_bw;
		l2params->tc_info[i].tc_ctx = qos_info->tc_info[i].tc_ctx;
	}
	for (i = 0; i < IIDC_MAX_USER_PRIORITY; i++)
		l2params->up2tc[i] = qos_info->up2tc[i];

#ifndef EXTERNAL_RELEASE
	/* TBD: This is not implemented in MEV driver */
#endif
	if (qos_info->pfc_mode == IIDC_DSCP_PFC_MODE) {
		l2params->dscp_mode = true;
		memcpy(l2params->dscp_map, qos_info->dscp_map,
		       sizeof(l2params->dscp_map));
	}
}

#ifdef VCHNL_PRIVILEGED
static void irdma_free_one_vf(struct irdma_vchnl_dev *vc_dev)
{
	struct irdma_sc_dev *dev = vc_dev->pf_dev;

	irdma_ws_reset(vc_dev->vf_vsi);
	irdma_del_hmc_objects(dev, &vc_dev->hmc_info, true, false,
			      dev->hw_attrs.uk_attrs.hw_rev);
	irdma_pf_put_vf_hmc_fcn(dev, vc_dev);
	irdma_put_vfdev(dev, vc_dev);
}

static void irdma_free_all_vf_rsrc(struct irdma_sc_dev *dev)
{
	u16 vf_idx;

	for (vf_idx = 0; vf_idx < dev->num_vfs; vf_idx++) {
		if (dev->vc_dev[vf_idx])
			irdma_free_one_vf(dev->vc_dev[vf_idx]);
	}
}

#endif /* VCHNL_PRIVILEGED */
#ifdef LAG_SUPPORT
/**
 * irdma_failover_start - Handle failover start
 * @iwdev: rdma device
 * @failing_port: Port number failing
 **/
static void irdma_failover_start(struct irdma_device *iwdev, u8 failing_port)
{
#ifdef DEBUG_SUPPORT_INTERNAL
	if (iwdev->vsi.failover_pending) {
		ibdev_warn(&iwdev->ibdev, "Failover already pending\n");
		return;
	}
#endif
	iwdev->vsi.failover_pending = true;
	irdma_ws_failover_cmd(&iwdev->vsi, IRDMA_OP_WS_FAILOVER_START, failing_port, 0);
}

/**
 * irdma_failover_complete - Handle fail over complete
 * @iwdev: rdma device structure
 * @failing_port: Port number failing
 * @failover_port: fail_over_port number
 **/
static void irdma_failover_complete(struct irdma_device *iwdev, u8 failing_port, u8 failover_port)
{
#ifdef DEBUG_SUPPORT_INTERNAL
	if (!iwdev->vsi.failover_pending) {
		ibdev_warn(&iwdev->ibdev, "Failover complete with no failover pending!\n");
		return;
	}
#endif
	irdma_ws_failover_cmd(&iwdev->vsi, IRDMA_OP_WS_FAILOVER_COMPLETE, failing_port, failover_port);
	iwdev->vsi.failover_pending = false;
}

#endif /* LAG_SUPPORT */

#ifdef IIDC_UPSTREAM
static void irdma_iidc_event_handler(struct ice_pf *pf, struct iidc_event *event)
{
	struct irdma_device *iwdev = dev_get_drvdata(&pf->adev->dev);
#else
static void irdma_iidc_event_handler(struct iidc_core_dev_info *cdev_info, struct iidc_event *event)
{
	struct irdma_device *iwdev = dev_get_drvdata(&cdev_info->adev->dev);
#endif /* IIDC_UPSTREAM */
	struct irdma_l2params l2params = {};

	if (iwdev->rf->reset)
		return;

	if (*event->type & BIT(IIDC_EVENT_AFTER_MTU_CHANGE)) {
		ibdev_dbg(&iwdev->ibdev, "CLNT: new MTU = %d\n", iwdev->netdev->mtu);
		if (iwdev->vsi.mtu != iwdev->netdev->mtu) {
			l2params.mtu = iwdev->netdev->mtu;
			l2params.mtu_changed = true;
			irdma_log_invalid_mtu(l2params.mtu, &iwdev->rf->sc_dev);
			if (iwdev->vsi.tc_change_pending) {
				iwdev->vsi.mtu_change_pending = true;
				iwdev->vsi.mtu = iwdev->netdev->mtu;
				return;
			}
			irdma_change_l2params(&iwdev->vsi, &l2params);
		}
#ifdef VCHNL_PRIVILEGED
	} else if (*event->type & BIT(IIDC_EVENT_VF_RESET)) {
		struct irdma_sc_dev *dev = &iwdev->rf->sc_dev;
		struct irdma_vchnl_dev *vc_dev =
			irdma_find_vc_dev(dev, event->info.vf_id);

		if (vc_dev)
			irdma_free_one_vf(vc_dev);
#endif
	} else if (*event->type & BIT(IIDC_EVENT_BEFORE_TC_CHANGE)) {
		if (iwdev->vsi.tc_change_pending)
			return;

		irdma_prep_tc_change(iwdev);
	} else if (*event->type & BIT(IIDC_EVENT_AFTER_TC_CHANGE)) {
#ifdef IIDC_UPSTREAM
		struct iidc_qos_params qos_info = {};
#endif

		if (!iwdev->vsi.tc_change_pending)
			return;

		if (iwdev->vsi.mtu_change_pending) {
			iwdev->vsi.mtu_change_pending = false;
			l2params.mtu = iwdev->vsi.mtu;
			l2params.mtu_changed = true;
		}

		l2params.tc_changed = true;
		ibdev_dbg(&iwdev->ibdev, "CLNT: TC Change\n");

#ifdef IIDC_UPSTREAM
		ice_get_qos_params(pf, &qos_info);
		irdma_fill_qos_info(&l2params, &qos_info);
#else
		irdma_fill_qos_info(&l2params, &cdev_info->qos_info);
#endif
#ifndef DISABLE_IWARP_VLAN_0_WA
		iwdev->dcb_vlan_mode = l2params.num_tc > 1 && !l2params.dscp_mode;

#else
		if (iwdev->rf->protocol_used != IRDMA_IWARP_PROTOCOL_ONLY)
			iwdev->dcb_vlan_mode = l2params.num_tc > 1 && !l2params.dscp_mode;
#endif
#ifdef IRDMA_FLOW_CONTROL_WARNING
#ifdef VCHNL_SUPPORT
		if (iwdev->rf->sc_dev.privileged)
			irdma_check_fc_for_tc_update(&iwdev->vsi, &l2params);
#else
		irdma_check_fc_for_tc_update(&iwdev->vsi, &l2params);
#endif /* VCHNL_SUPPORT */
#endif /* IRDMA_FLOW_CONTROL_WARNING */
		irdma_change_l2params(&iwdev->vsi, &l2params);
#ifdef LAG_SUPPORT
	} else if (*event->type & BIT(IIDC_EVENT_FAILOVER_START)) {
		irdma_failover_start(iwdev, cdev_info->rdma_active_port);
	} else if (*event->type & BIT(IIDC_EVENT_FAILOVER_FINISH)) {
		irdma_failover_complete(iwdev, 0, cdev_info->rdma_active_port);
#endif
	} else if (*event->type & BIT(IIDC_EVENT_CRIT_ERR)) {
#ifdef IIDC_UPSTREAM
		ibdev_warn(&iwdev->ibdev, "ICE OICR event notification: oicr = 0x%08x\n",
			   event->reg);
		if (event->reg & IRDMAPFINT_OICR_PE_CRITERR_M) {
#else
		ibdev_warn(&iwdev->ibdev, "ICE OICR event notification: oicr = 0x%08x\n",
			   event->info.reg);
		if (event->info.reg & IRDMAPFINT_OICR_PE_CRITERR_M) {
#endif /* IIDC_UPSTREAM */
			u32 pe_criterr;

			pe_criterr = rd32v(iwdev->rf->sc_dev.hw_regs[IRDMA_GLPE_CRITERR]);
#define IRDMA_Q1_RESOURCE_ERR 0x0001024d
			if (pe_criterr != IRDMA_Q1_RESOURCE_ERR) {
				ibdev_err(&iwdev->ibdev, "critical PE Error, GLPE_CRITERR=0x%08x\n",
					  pe_criterr);
				iwdev->rf->reset = true;
			} else {
				ibdev_warn(&iwdev->ibdev, "Q1 Resource Check\n");
			}
		}
#ifdef IIDC_UPSTREAM
		if (event->reg & IRDMAPFINT_OICR_HMC_ERR_M) {
#else
		if (event->info.reg & IRDMAPFINT_OICR_HMC_ERR_M) {
#endif
			ibdev_err(&iwdev->ibdev, "HMC Error\n");
			iwdev->rf->reset = true;
		}
#ifdef IIDC_UPSTREAM
		if (event->reg & IRDMAPFINT_OICR_PE_PUSH_M) {
#else
		if (event->info.reg & IRDMAPFINT_OICR_PE_PUSH_M) {
#endif
			ibdev_err(&iwdev->ibdev, "PE Push Error\n");
			iwdev->rf->reset = true;
		}
		if (iwdev->rf->reset)
			iwdev->rf->gen_ops.request_reset(iwdev->rf);
#ifndef UPSTREAM_RELEASE
	} else if (*event->type & BIT(IIDC_EVENT_WARN_RESET)) {
		iwdev->rf->reset = true;
	}
#else
	}
#endif
}

/**
 * irdma_request_reset - Request a reset
 * @rf: RDMA PCI function
 */
static void irdma_request_reset(struct irdma_pci_f *rf)
{
#ifdef IIDC_UPSTREAM
	struct ice_pf *pf = rf->cdev;

	ibdev_warn(&rf->iwdev->ibdev, "Requesting a reset\n");
	ice_rdma_request_reset(pf, IIDC_PFR);
#else
	struct iidc_core_dev_info *cdev_info = rf->cdev;

	ibdev_warn(&rf->iwdev->ibdev, "Requesting a reset\n");
#ifdef VCHNL_SUPPORT
	rf->sc_dev.vchnl_up = false;
#endif
	cdev_info->ops->request_reset(rf->cdev, IIDC_PFR);
#endif /* IIDC_UPSTREAM */
}

#ifdef CVL_VCHNL_UNPRIVILEGED
/*
 * irdma_vchnl_req_aeq_vec_map_gen2 - Virt channel AEQ configuration
 * @dev: device
 * @idx: function relative MSI-X vector
 *
 * Call the IDC to send a AEQ configuration request.
 * Return 0 if successful, otherwise return error
 */
int irdma_vchnl_req_aeq_vec_map_gen2(struct irdma_sc_dev *dev, u32 idx)
{
	struct iidc_core_dev_info *cdev_info = dev_to_rf(dev)->cdev;
	struct iidc_qvlist_info qvl_info = { 0 };
	struct iidc_qv_info *qvinfo = &qvl_info.qv_info[0];

#ifndef EXTERNAL_RELEASE
	/* TBD:
	 * Unmapping individual IRQs is a bit problematic, as the
	 * VIRTCHNL_OP_RELEASE_IWARP_IRQ_MAP is intended to nuke all
	 * the mapped IRQs in one shot and does not include any data.
	 * To adopt it to our purposes, we either need to pass
	 * CEQ/AEQ index (requires modification of virtchnl API and
	 * breaks compatibility) or to keep track on all the mapped
	 * IRQs on the CP side.
	 * For now, we can simply skip this step.  It seems to be safe,
	 * as it is called right before the entire HMC function is
	 * to be destroyed.
	 */
#endif
	qvl_info.num_vectors = 1;
	qvinfo->ceq_idx = IRDMA_Q_INVALID_IDX;
	qvinfo->v_idx = idx;
	qvinfo->itr_idx = IRDMA_IDX_ITR0;

	return cdev_info->ops->vc_queue_vec_map_unmap(cdev_info, &qvl_info,
						      true);
}

/*
 * irdma_vchnl_req_ceq_vec_map_gen2 - Virt channel CEQ configuration
 * @dev: shared code device
 * @ceq_id: function relative CEQ id
 * @idx: function relative MSI-X vector
 *
 * Call the IDC to send a CEQ configuration request.
 * Return 0 if successful, otherwise return error
 */
int irdma_vchnl_req_ceq_vec_map_gen2(struct irdma_sc_dev *dev, u16 ceq_id, u32 idx)
{
	struct iidc_core_dev_info *cdev_info = dev_to_rf(dev)->cdev;
	struct iidc_qvlist_info qvl_info = { 0 };
	struct iidc_qv_info *qvinfo = &qvl_info.qv_info[0];

#ifndef EXTERNAL_RELEASE
	/* TBD:
	 * Unmapping individual IRQs is a bit problematic, as the
	 * VIRTCHNL_OP_RELEASE_IWARP_IRQ_MAP is intended to nuke all
	 * the mapped IRQs in one shot and does not include any data.
	 * To adopt it to our purposes, we either need to pass
	 * CEQ/AEQ index (requires modification of virtchnl API and
	 * breaks compatibility) or to keep track on all the mapped
	 * IRQs on the CP side.
	 * For now, we can simply skip this step.  It seems to be safe,
	 * as it is called right before the entire HMC function is
	 * to be destroyed.
	 */
#endif
	qvl_info.num_vectors = 1;
	qvinfo->aeq_idx = IRDMA_Q_INVALID_IDX;
	qvinfo->ceq_idx = ceq_id;
	qvinfo->v_idx = idx;
	qvinfo->itr_idx = IRDMA_IDX_ITR0;

	return cdev_info->ops->vc_queue_vec_map_unmap(cdev_info, &qvl_info,
						      true);
}

#endif /* CVL_VCHNL_UNPRIVILEGED */
#ifdef LAG_AA_SUPPORT
/*
 * irdma_lan_register_qset - Register qsets with LAN driver
 * @vsi: vsi structure
 * @tc_node1: List of Traffic class node ponters
 * @tc_node2: List of Traffic class node ponters
 */
static int irdma_lan_register_qset(struct irdma_sc_vsi *vsi,
				   struct irdma_ws_node *tc_node1,
				   struct irdma_ws_node *tc_node2)
{
	struct irdma_device *iwdev = vsi->back_vsi;
#ifdef IIDC_UPSTREAM
	struct ice_pf *pf = iwdev->rf->cdev;
#else
	struct iidc_core_dev_info *cdev_info = iwdev->rf->cdev;
#endif
	int ret;

	if (tc_node2) {
		struct iidc_rdma_multi_qset_params mqset = { };

		mqset.qs_handle[0] = tc_node1->qs_handle;
		mqset.qs_handle[1] = tc_node2->qs_handle;
		mqset.tc = tc_node2->traffic_class;
		mqset.vport_id = vsi->vsi_idx;
		mqset.num = 2;
#ifdef IIDC_UPSTREAM
		ret = ice_add_rdma_qset(pf, &mqset);
#else
		ret = cdev_info->ops->alloc_multi_res(cdev_info, &mqset);
#endif
		if (ret) {
			ibdev_warn(&iwdev->ibdev, "WS: LAN alloc_multi_res for rdma qset failed.\n");
			/* return ret; Ignore return code since RDMA still works when it fails, until ICE is debugged */
		}
		tc_node1->l2_sched_node_id = mqset.teid[0];
		tc_node2->l2_sched_node_id = mqset.teid[1];
		vsi->qos[tc_node1->user_pri].l2_sched_node_id[0] = mqset.teid[0];
		vsi->qos[tc_node1->user_pri].l2_sched_node_id[1] = mqset.teid[1];
	} else {
		struct iidc_rdma_qset_params qset = { };

		qset.qs_handle = tc_node1->qs_handle;
		qset.tc = tc_node1->traffic_class;
		qset.vport_id = vsi->vsi_idx;
#ifdef IIDC_UPSTREAM
		ret = ice_add_rdma_qset(pf, &qset);
#else
		ret = cdev_info->ops->alloc_res(cdev_info, &qset);
#endif
		if (ret) {
			ibdev_warn(&iwdev->ibdev, "WS: LAN alloc_res for rdma qset failed.\n");
			//return ret;
		}

		tc_node1->l2_sched_node_id = qset.teid;
		vsi->qos[tc_node1->user_pri].l2_sched_node_id[0] = qset.teid;
	}
	return 0;
}

/**
 * irdma_lan_unregister_qset - Unregister qsets with LAN driver
 * @vsi: vsi structure
 * @tc_node1: Traffic class node
 * @tc_node2: Traffic class node
 */
static void irdma_lan_unregister_qset(struct irdma_sc_vsi *vsi,
				      struct irdma_ws_node *tc_node1,
				      struct irdma_ws_node *tc_node2)
{
	struct irdma_device *iwdev = vsi->back_vsi;
#ifdef IIDC_UPSTREAM
	struct ice_pf *pf = iwdev->rf->cdev;
#else
	struct iidc_core_dev_info *cdev_info = iwdev->rf->cdev;
#endif

	if (vsi->lag_aa) {
		struct iidc_rdma_multi_qset_params mqset = { };

		mqset.qs_handle[0] = tc_node1->qs_handle;
		mqset.qs_handle[1] = tc_node2->qs_handle;
		mqset.tc = tc_node1->traffic_class; /* TC is the same for both nodes */
		mqset.vport_id = vsi->vsi_idx;
		mqset.num = 2;
		mqset.teid[0] = tc_node1->l2_sched_node_id;
		mqset.teid[1] = tc_node2->l2_sched_node_id;
#ifdef IIDC_UPSTREAM
		if (ice_del_rdma_qsets(pf, &mqset))
#else
		if (cdev_info->ops->free_multi_res(cdev_info, &mqset))
#endif
			ibdev_warn(&iwdev->ibdev, "WS: LAN free_res for rdma qset failed.\n");

	} else {
		struct iidc_rdma_qset_params qset = { };
		struct irdma_ws_node *tc_node = tc_node1 ? tc_node1 : tc_node2;

		qset.qs_handle = tc_node->qs_handle;
		qset.tc = tc_node->traffic_class;
		qset.vport_id = vsi->vsi_idx;
		qset.teid = tc_node->l2_sched_node_id;

#ifdef IIDC_UPSTREAM
		if (ice_del_rdma_qset(pf, &qset))
#else
		if (cdev_info->ops->free_res(cdev_info, &qset))
#endif
			ibdev_warn(&iwdev->ibdev, "WS: LAN free_res for rdma qset failed.\n");
	}
}

#else  /* LAG_AA_SUPPORT */
/*
 * irdma_lan_register_qset - Register qset with LAN driver
 * @vsi: vsi structure
 * @tc_node: Traffic class node
 */
static int irdma_lan_register_qset(struct irdma_sc_vsi *vsi,
				   struct irdma_ws_node *tc_node)
{
	struct irdma_device *iwdev = vsi->back_vsi;
#ifdef IIDC_UPSTREAM
	struct ice_pf *pf = iwdev->rf->cdev;
#else
	struct iidc_core_dev_info *cdev_info = iwdev->rf->cdev;
#endif
	struct iidc_rdma_qset_params qset = {};
	int ret;

	qset.qs_handle = tc_node->qs_handle;
	qset.tc = tc_node->traffic_class;
	qset.vport_id = vsi->vsi_idx;
#ifdef IIDC_UPSTREAM
	ret = ice_add_rdma_qset(pf, &qset);
#else
	ret = cdev_info->ops->alloc_res(cdev_info, &qset);
#endif
	if (ret) {
		ibdev_dbg(&iwdev->ibdev, "WS: LAN alloc_res for rdma qset failed.\n");
		return ret;
	}

	tc_node->l2_sched_node_id = qset.teid;
	vsi->qos[tc_node->user_pri].l2_sched_node_id = qset.teid;

	return 0;
}

/**
 * irdma_lan_unregister_qset - Unregister qset with LAN driver
 * @vsi: vsi structure
 * @tc_node: Traffic class node
 */
static void irdma_lan_unregister_qset(struct irdma_sc_vsi *vsi,
				      struct irdma_ws_node *tc_node)
{
	struct irdma_device *iwdev = vsi->back_vsi;
#ifdef IIDC_UPSTREAM
	struct ice_pf *pf = iwdev->rf->cdev;
#else
	struct iidc_core_dev_info *cdev_info = iwdev->rf->cdev;
#endif
	struct iidc_rdma_qset_params qset = {};

	qset.qs_handle = tc_node->qs_handle;
	qset.tc = tc_node->traffic_class;
	qset.vport_id = vsi->vsi_idx;
	qset.teid = tc_node->l2_sched_node_id;

#ifdef IIDC_UPSTREAM
	if (ice_del_rdma_qset(pf, &qset))
#else
	if (cdev_info->ops->free_res(cdev_info, &qset))
#endif
		ibdev_dbg(&iwdev->ibdev, "WS: LAN free_res for rdma qset failed.\n");
}
#endif /* LAG_AA_SUPPORT */

#ifdef CPK_A0_ERRATA
static void irdma_enable_lb(struct irdma_device *iwdev)
{
	u32 val;

	val = rd32(iwdev->rf->sc_dev.hw, VSI_RXSWCTRL(iwdev->vsi.vsi_idx));
	val &= ~VSI_RXSWCTRL_SRCPRUNEENABLE_M;
	wr32(iwdev->rf->sc_dev.hw, VSI_RXSWCTRL(iwdev->vsi.vsi_idx), val);
}

#endif
#ifdef CPK_DESTROY_QP_FAIL_WA
void irdma_cleanup_dead_qps(struct irdma_sc_vsi *vsi)
{
	struct irdma_sc_qp *qp = NULL;
	struct irdma_qp *iwqp;
	struct irdma_pci_f *rf;
	u8 i;

	for (i = 0; i < IRDMA_MAX_USER_PRIORITY; i++) {
		qp = irdma_get_qp_from_list(&vsi->qos[i].qplist, qp);
		while (qp) {
			if (qp->qp_uk.qp_type == IRDMA_QP_TYPE_UDA) {
				qp = irdma_get_qp_from_list(&vsi->qos[i].qplist, qp);
				continue;
			}
			iwqp = qp->qp_uk.back_qp;
			rf = iwqp->iwdev->rf;
			dma_free_coherent(rf->hw.device,
					  iwqp->q2_ctx_mem.size,
					  iwqp->q2_ctx_mem.va,
					  iwqp->q2_ctx_mem.pa);
			dma_free_coherent(rf->hw.device,
					  iwqp->kqp.dma_mem.size,
					  iwqp->kqp.dma_mem.va,
					  iwqp->kqp.dma_mem.pa);
			kfree(iwqp->kqp.sq_wrid_mem);
			kfree(iwqp->kqp.rq_wrid_mem);
			qp = irdma_get_qp_from_list(&vsi->qos[i].qplist, qp);
			kfree(iwqp);
		}
	}
}

#endif
#ifdef MEV_HVL
static void irdma_poll_cq3(struct irdma_pci_f *rf)
{
	struct irdma_cq *cq = rf->cq_id_3;
	struct irdma_cq_uk *ukcq  = &cq->sc_cq.cq_uk;
	u64 qword3;
	__le64 *cqe;
	u8 polarity;

	cqe = IRDMA_GET_CURRENT_CQ_ELEM(ukcq);
	get_64bit_val(cqe, 24, &qword3);
	polarity = (u8)FIELD_GET(IRDMA_CQ_VALID, qword3);

	if (polarity == ukcq->polarity && cq->ibcq.comp_handler)
		cq->ibcq.comp_handler(&cq->ibcq, cq->ibcq.cq_context);
}

static int poll_thread(void *context)
{
	struct irdma_pci_f *rf = context;

	msleep(200);
	do {
		msleep(1);
		if (rf->sc_dev.hw_wa & AEQ_POLL) {
			irdma_process_aeq(rf);
			continue;
		}
		if (rf->sc_dev.hw_wa & CCQ_CQ3_POLL) {
			struct irdma_sc_cq *ccq = &rf->ccq.sc_cq;

			if (ccq)
				irdma_cqp_ce_handler(rf, ccq);
			if (rf->cq_id_3)
				irdma_poll_cq3(rf);
			continue;
		}
		if (rf->sc_dev.hw_wa & CEQ_POLL) {
			if (rf->ceqlist)
				irdma_process_ceq(rf, rf->ceqlist);
			irdma_process_aeq(rf);
		}
	} while (!kthread_should_stop());

	return 0;
}

#endif /* MEV_HVL */
#ifndef UPSTREAM_RELEASE
#ifdef HAVE_AUXILIARY_DRIVER_INT_REMOVE
static int irdma_remove(struct auxiliary_device *aux_dev)
#else /* HAVE_AUXILIARY_DRIVER_INT_REMOVE */
static void irdma_remove(struct auxiliary_device *aux_dev)
#endif /* HAVE_AUXILIARY_DRIVER_INT_REMOVE */
#else  /* UPSTREAM_RELEASE */
static void irdma_remove(struct auxiliary_device *aux_dev)
#endif /* UPSTREAM_RELEASE */
{
	struct iidc_auxiliary_dev *iidc_adev = container_of(aux_dev,
							    struct iidc_auxiliary_dev,
							    adev);
#ifdef IIDC_UPSTREAM
	struct ice_pf *pf = iidc_adev->pf;
#else
	struct iidc_core_dev_info *cdev_info = iidc_adev->cdev_info;
#endif
	struct irdma_device *iwdev = auxiliary_get_drvdata(aux_dev);
	u8 rdma_ver = iwdev->rf->rdma_ver;

#ifndef EXTERNAL_RELEASE
	/* TBD: what about MEV? */
#endif
	if (rdma_ver == IRDMA_GEN_2 && !iwdev->rf->reset) {
#ifdef VCHNL_PRIVILEGED
		if (iwdev->rf->sc_dev.privileged)
			irdma_free_all_vf_rsrc(&iwdev->rf->sc_dev);
#endif
#ifdef CONFIGFS_SUPPORT
#if IS_ENABLED(CONFIG_CONFIGFS_FS)
#ifdef CPK_UP_MAP_WA
		if (rdma_ver <= IRDMA_GEN_2 && iwdev->up_map_en) {
			struct irdma_up_info up_map_info = {};

			*((u64 *)up_map_info.map) = IRDMA_DEFAULT_UP_UP_MAP;
			up_map_info.use_cnp_up_override = false;
			up_map_info.cnp_up_override = 0;
			up_map_info.hmc_fcn_idx = iwdev->rf->sc_dev.hmc_fn_id;
			irdma_cqp_up_map_cmd(&iwdev->rf->sc_dev,
					     IRDMA_OP_SET_UP_MAP,
					     &up_map_info);
		}
#endif /* CPK_UP_MAP_WA */
#endif /* CONFIG_CONFIGFS_FS */
#endif /* CONFIGFS_SUPPORT */
#ifdef CPK_CQP_HANG_ON_SUSPENDED_QPS_WA
		if (iwdev->vsi.tc_change_pending) {
			iwdev->vsi.tc_change_pending = false;
			irdma_sc_suspend_resume_qps(&iwdev->vsi,
						    IRDMA_OP_RESUME);
		}

#endif
	}

	if (rdma_ver == IRDMA_GEN_2) {
#ifdef IIDC_UPSTREAM
		ice_rdma_update_vsi_filter(pf, iwdev->vsi_num, false);
#else
#ifdef CVL_VCHNL_UNPRIVILEGED
		if (iwdev->rf->sc_dev.privileged)
			cdev_info->ops->update_vport_filter(
				cdev_info, iwdev->vsi_num, false);
#else
		cdev_info->ops->update_vport_filter(cdev_info,
						    iwdev->vsi_num, false);
#endif
#endif
	}

	irdma_ib_unregister_device(iwdev);
#ifdef LAG_SUPPORT
#ifndef EXTERNAL_RELEASE
	/* Release bonding netdev associated with ibdev before calling
	 * irdma_unregister_notifiers. This avoids a deadlock with the netdev
	 * notifier unregister waiting on the rtnl lock held up by an unloading
	 * bonding driver waiting on bonding netdev refcnt to be released.
	 */
#endif
#endif
	irdma_unregister_notifiers(iwdev);
	irdma_deinit_device(iwdev);
	ib_dealloc_device(&iwdev->ibdev);
#ifdef IIDC_UPSTREAM
	pr_debug("INIT: Gen2 func[%d] device remove success\n", PCI_FUNC(pf->pdev->devfn));
#else
	pr_debug("INIT: Gen[%d] func[%d] device remove success\n",
		 rdma_ver, PCI_FUNC(cdev_info->pdev->devfn));
#endif /* IIDC_UPSTREAM */
#ifndef UPSTREAM_RELEASE
#ifdef HAVE_AUXILIARY_DRIVER_INT_REMOVE
	return 0;
#endif /* HAVE_AUXILIARY_DRIVER_INT_REMOVE */
#endif /* UPSTREAM_RELEASE */
}

#ifdef VCHNL_SUPPORT
static int irdma_vchnl_init(struct irdma_device *iwdev,
			    struct iidc_core_dev_info *cdev_info, u8 *rdma_ver)
{
	struct irdma_vchnl_init_info virt_info;
	struct irdma_pci_f *rf = iwdev->rf;
	u8 gen = cdev_info->rdma_caps.gen;
	int ret;

	rf->vchnl_wq = alloc_ordered_workqueue("irdma-virtchnl-wq", 0);
	if (!rf->vchnl_wq)
		return -ENOMEM;

	irdma_mutex_init(&rf->sc_dev.vchnl_mutex);

#ifndef EXTERNAL_RELEASE
	/* Initialize rdma_caps.gen to IRDMA_GEN_2 to be compatible with older
	 * ice that does not set rdma_caps.gen.
	 */
#endif
	virt_info.hw_rev = !gen ? IRDMA_GEN_2 : gen;
#ifdef MEV_SUPPORT
	virt_info.is_pf = !cdev_info->ftype;
#endif
#ifndef EXTERNAL_RELEASE
/**
 * The cdev_info->rdma_caps.gen is not populated in the idpf and on ice
 * is populated starting with ice-1.10.6 which has iidc version 10.2.
 * Unfortunately ice-1.9.12 still used iidc version 10.2, so for that
 * release it will have an issue if new irdma driver is used.
 */
#endif /* !EXTERNAL_RELEASE */

#ifdef UPSTREAM_RELEASE
	virt_info.privileged = !cdev_info->ftype && gen == IRDMA_GEN_2;
#else /* UPSTREAM_RELEASE */
	if (cdev_info->ftype) {
		virt_info.privileged = false;
	} else {
		if (cdev_info->ver.major >= 10 && cdev_info->ver.minor >= 2)
			virt_info.privileged = cdev_info->rdma_caps.gen == IRDMA_GEN_2;
		else
			virt_info.privileged = true;
	}
#if defined(VCHNL_PRIVILEGED) && defined(VCHNL_UNPRIVILEGED)
	virt_info.vchnl_if = virt_info.privileged ? &irdma_vchnl_if_pf :
						    &irdma_vchnl_if_req;
#endif /* defined(VCHNL_PRIVILEGED) && defined(VCHNL_UNPRIVILEGED) */
#endif /* UPSTREAM_RELEASE */
	virt_info.vchnl_wq = rf->vchnl_wq;
	ret = irdma_sc_vchnl_init(&rf->sc_dev, &virt_info);
	if (ret) {
		destroy_workqueue(rf->vchnl_wq);
		return ret;
	}

	*rdma_ver = rf->sc_dev.hw_attrs.uk_attrs.hw_rev;
	return 0;
}
#endif /* VCHNL_SUPPORT */

#ifdef IIDC_UPSTREAM
static void irdma_fill_device_info(struct irdma_device *iwdev, struct ice_pf *pf)
#else
static int irdma_fill_device_info(struct irdma_device *iwdev, struct iidc_core_dev_info *cdev_info)
#endif /* IIDC_UPSTREAM */
{
	struct irdma_pci_f *rf = iwdev->rf;
#ifdef IIDC_UPSTREAM
	struct ice_vsi *vsi = ice_get_main_vsi(pf);
#else
	int err;
#endif

	rf->sc_dev.hw = &rf->hw;
	rf->iwdev = iwdev;
#ifdef IIDC_UPSTREAM
	rf->rdma_ver = IRDMA_GEN_2;
	rf->sc_dev.hw_attrs.uk_attrs.hw_rev = IRDMA_GEN_2;
	rf->cdev = pf;
	rf->gen_ops.register_qset = irdma_lan_register_qset;
	rf->gen_ops.unregister_qset = irdma_lan_unregister_qset;
	rf->hw.hw_addr = pf->hw.hw_addr;
	rf->pcidev = pf->pdev;
	rf->hw.device = &rf->pcidev->dev;
	rf->msix_count =  pf->num_rdma_msix;
	rf->msix_entries = &pf->msix_entries[pf->rdma_base_vector];
	rf->default_vsi.vsi_idx = vsi->vsi_num;
	rf->protocol_used = IRDMA_ROCE_PROTOCOL_ONLY;
	rf->pf_id = pf->hw.pf_id;
#else
	rf->cdev = cdev_info;
	rf->hw.hw_addr = cdev_info->hw_addr;
	rf->pcidev = cdev_info->pdev;
	rf->hw.device = &rf->pcidev->dev;
	rf->ftype = cdev_info->ftype;
	rf->msix_count = cdev_info->msix_count;
	rf->msix_entries = cdev_info->msix_entries;

#ifdef VCHNL_SUPPORT
	err = irdma_vchnl_init(iwdev, cdev_info, &rf->rdma_ver);
	if (err)
		return err;

#endif /* VCHNL_SUPPORT */
	if (!cdev_info->ftype && cdev_info->ver.major == 10 &&
	    cdev_info->ver.minor == 0 && rf->rdma_ver == IRDMA_GEN_2) {
		u32 val;
#define PF_FUNC_RID 0x0009E880
#define PF_FUNC_RID_FUNCTION_NUMBER GENMASK(2, 0)
		rf->hw.hw_addr = cdev_info->hw_addr;
		val = rd32(&rf->hw, PF_FUNC_RID);
		rf->pf_id = (u8)FIELD_GET(PF_FUNC_RID_FUNCTION_NUMBER, val);
#ifdef MEV_IIDC_WA
	} else if (!cdev_info->ftype && rf->rdma_ver >= IRDMA_GEN_3) {
#ifndef EXTERNAL_RELEASE
	/* hardcode here as ingen3 LAN driver, pf_id is not reflected in iidc.h */
#endif
		rf->pf_id = PCI_FUNC(cdev_info->pdev->devfn);
#endif /* MEV_IIDC_WA */
	} else {
		rf->pf_id = cdev_info->pf_id;
	}

	if (!cdev_info->ftype && rf->rdma_ver == IRDMA_GEN_2) {
		rf->gen_ops.register_qset = irdma_lan_register_qset;
		rf->gen_ops.unregister_qset = irdma_lan_unregister_qset;
	}

#ifdef MEV_HVL
	if (rf->rdma_ver >= IRDMA_GEN_3) {
		if (cdev_info->pdev->revision < MEV_PCI_VER_C0) {
#define IRDMA_MEV_B0_RDMA_KEY	0xb
			if (rdma_key != IRDMA_MEV_B0_RDMA_KEY) {
				dev_err(rf->hw.device,
					"IRDMA: Invalid RDMA key used for B0\n");
				return -EINVAL;
			}
		}

		mev_enable_hw_wa(&rf->sc_dev, hw_type_wa, wa_mem_pages,
				 hw_wa_bitmask, host_mem_mrte);
		rf->sc_dev.rrf_multiplier = rrf_m;
		rf->sc_dev.xf_multiplier = xf_m;
		rf->sc_dev.min_ird = min_ird;
	}

	if (rf->sc_dev.hw_wa & CEQ_POLL)
		rf->msix_count = 1;

#endif
	rf->default_vsi.vsi_idx = cdev_info->vport_id;
	rf->protocol_used = cdev_info->rdma_protocol == IIDC_RDMA_PROTOCOL_ROCEV2 ?
			    IRDMA_ROCE_PROTOCOL_ONLY : IRDMA_IWARP_PROTOCOL_ONLY;
#ifdef MEV_IIDC_WA
	if (rf->rdma_ver >= IRDMA_GEN_3)
#ifndef EXTERNAL_RELEASE
	/* In gen3 LAN driver, cdev_info->rdma_protocol should be set to RoCEv2 */
#endif
		rf->protocol_used = IRDMA_ROCE_PROTOCOL_ONLY;
#endif
#endif /* IIDC_UPSTREAM */
	rf->rsrc_profile = IRDMA_HMC_PROFILE_DEFAULT;
#ifdef IRDMA_FLOW_CONTROL_WARNING
	if (rf->rdma_ver == IRDMA_GEN_2)
		rf->check_fc = irdma_check_fc_for_qp;
#endif
	rf->gen_ops.request_reset = irdma_request_reset;
#ifndef UPSTREAM_RELEASE
	/* Can override limits_sel, protocol_used */
	irdma_set_rf_user_cfg_params(rf);
#else
	rf->limits_sel = 7;
#endif

#ifdef AH_REUSE
	irdma_mutex_init(&iwdev->ah_tbl_lock);
#endif

#ifdef LAG_SUPPORT
#ifdef IIDC_UPSTREAM
	rcu_read_lock();
	iwdev->netdev =  netdev_master_upper_dev_get_rcu(vsi->netdev);
	rcu_read_unlock();
	if (!iwdev->netdev || cdev_info->rdma_active_port == 0xFF)
		iwdev->netdev = vsi->netdev;
	else
		iwdev->lag_active = true;
	iwdev->vsi_num = vsi->vsi_num;
#else
	rcu_read_lock();
	iwdev->netdev =  netdev_master_upper_dev_get_rcu(cdev_info->netdev);
	rcu_read_unlock();
	if (!iwdev->netdev || cdev_info->rdma_active_port == 0xFF) {
		iwdev->netdev = cdev_info->netdev;
	} else {
#ifndef EXTERNAL_RELEASE
		/**
		 *  When a PF is added to a bond, only RoCEv2 mode is support,
		 *  so if the irdma driver was loaded in iWARP mode we need to
		 *  disable it to avoid undefined behavior. This is not an issue
		 *  for in-tree since the ICE driver will know the mode and
		 *  avoid putting the rdma device on the bus and there are no
		 *  module parameters to override the mode.
		 */
#endif
#ifndef UPSTREAM_RELEASE
		if (rf->protocol_used != IRDMA_ROCE_PROTOCOL_ONLY) {
                        dev_err(rf->hw.device, "LAG is only supported in RoCEv2 mode.\n");
                        return -EINVAL;
                }
#endif
		iwdev->lag_active = true;
	}
	iwdev->vsi_num = cdev_info->vport_id;
#endif /* IIDC_UPSTREAM */
#else /* LAG_SUPPORT */

#ifdef IIDC_UPSTREAM
	iwdev->netdev = vsi->netdev;
	iwdev->vsi_num = vsi->vsi_num;
#else
	iwdev->netdev = cdev_info->netdev;
	iwdev->vsi_num = cdev_info->vport_id;
#endif /* IIDC_UPSTREAM */
#endif /* LAG_SUPPORT */
	iwdev->init_state = INITIAL_STATE;
	iwdev->roce_cwnd = IRDMA_ROCE_CWND_DEFAULT;
	iwdev->roce_ackcreds = IRDMA_ROCE_ACKCREDS_DEFAULT;
	iwdev->rcv_wnd = IRDMA_CM_DEFAULT_RCV_WND_SCALED;
	iwdev->rcv_wscale = IRDMA_CM_DEFAULT_RCV_WND_SCALE;
	iwdev->push_mode = false;
#ifdef CONFIGFS_SUPPORT
#if IS_ENABLED(CONFIG_CONFIGFS_FS)
	iwdev->iwarp_ecn_en = iwdev->rf->rdma_ver == IRDMA_GEN_2 ? true : false;
	iwdev->iwarp_rtomin = 5;
	iwdev->up_up_map = IRDMA_DEFAULT_UP_UP_MAP;
#endif
#endif
	if (iwdev->rf->protocol_used != IRDMA_IWARP_PROTOCOL_ONLY) {
#if defined(IRDMA_FLOW_CONTROL_WARNING) || defined(CONFIGFS_SUPPORT)
		iwdev->roce_rtomin = 5;
#endif
#ifdef DCQCN_CONFIG_SUPPORT
		iwdev->roce_dcqcn_en = iwdev->rf->dcqcn_ena;
#endif
		iwdev->roce_mode = true;
	}
#ifndef IIDC_UPSTREAM

	return 0;
#endif
}

static int irdma_probe(struct auxiliary_device *aux_dev, const struct auxiliary_device_id *id)
{
	struct iidc_auxiliary_dev *iidc_adev = container_of(aux_dev,
							    struct iidc_auxiliary_dev,
							    adev);
#ifdef IIDC_UPSTREAM
	struct ice_pf *pf = iidc_adev->pf;
	struct iidc_qos_params qos_info = {};
#else
	struct iidc_core_dev_info *cdev_info = iidc_adev->cdev_info;
#endif /* IIDC_UPSTREAM */
	struct irdma_device *iwdev;
	struct irdma_pci_f *rf;
	struct irdma_l2params l2params = { 0 };
	int err;
#ifdef GLOBAL_HDL_LIST
	struct irdma_handler *hdl;
#endif

#ifndef IIDC_UPSTREAM
	if (cdev_info->ver.major != IIDC_MAJOR_VER) {
		pr_err("version mismatch:\n");
		pr_err("expected major ver %d, caller specified major ver %d\n",
		       IIDC_MAJOR_VER, cdev_info->ver.major);
		pr_err("expected minor ver %d, caller specified minor ver %d\n",
		       IIDC_MINOR_VER, cdev_info->ver.minor);
		return -EINVAL;
	}
	if (cdev_info->ver.minor != IIDC_MINOR_VER)
		irdma_pr_info("probe: minor version mismatch: expected %0d.%0d caller specified %0d.%0d\n",
			      IIDC_MAJOR_VER, IIDC_MINOR_VER,
			      cdev_info->ver.major, cdev_info->ver.minor);
#ifndef LAG_SUPPORT
	irdma_pr_info("probe: cdev_info=%p, cdev_info->dev.aux_dev.bus->number=%d, netdev=%s\n",
		      cdev_info, cdev_info->pdev->bus->number, netdev_name(cdev_info->netdev));
#else
	irdma_pr_info("probe: cdev_info=%p, cdev_info->dev.aux_dev.bus->number=%d, cdev_info->rdma_active_port=0x%x netdev=%s\n",
		      cdev_info, cdev_info->pdev->bus->number, cdev_info->rdma_active_port,
		      netdev_name(cdev_info->netdev));
#endif
#endif /* IIDC_UPSTREAM */
	iwdev = ib_alloc_device(irdma_device, ibdev);
	if (!iwdev)
		return -ENOMEM;
	iwdev->rf = kzalloc(sizeof(*rf), GFP_KERNEL);
	if (!iwdev->rf) {
		ib_dealloc_device(&iwdev->ibdev);
		return -ENOMEM;
	}

#ifdef IIDC_UPSTREAM
	irdma_fill_device_info(iwdev, pf);
#else
	err = irdma_fill_device_info(iwdev, cdev_info);
	if (err)
		goto err_fill_devinfo;
#endif /* IIDC_UPSTREAM */
	rf = iwdev->rf;
	iwdev->aux_dev = aux_dev;

#ifdef GLOBAL_HDL_LIST
	hdl = kzalloc(sizeof(*hdl), GFP_KERNEL);
	if (!hdl)
		goto err_hdl;

	hdl->iwdev = iwdev;
	iwdev->hdl = hdl;
#endif /* GLOBAL_HDL_LIST */
	err = irdma_ctrl_init_hw(rf);
	if (err)
		goto err_ctrl_init;

#ifdef MEV_HVL
	if (rf->rdma_ver >= IRDMA_GEN_3 &&
	    rf->sc_dev.hw_wa & TIMER_NEEDED)
		rf->poll_thread =
			kthread_run(poll_thread, rf, "dpc polling thread");

#endif
#ifndef EXTERNAL_RELEASE
	/* TODO: Update the fragcnt limits for MEV before enabling GEN3 */
#endif
#ifdef CONFIG_DEV_LIMITS
	if (rf->rdma_ver == IRDMA_GEN_2) {
		if (irdma_set_attr_from_fragcnt(&rf->sc_dev, rf->fragcnt_limit))
			dev_warn(rf->hw.device,
				 "device limit update failed for fragment count %d\n",
				 rf->fragcnt_limit);
	}
#endif
	l2params.mtu = iwdev->netdev->mtu;
#ifdef IIDC_UPSTREAM
	ice_get_qos_params(pf, &qos_info);
	irdma_fill_qos_info(&l2params, &qos_info);
#else
	irdma_fill_qos_info(&l2params, &cdev_info->qos_info);
#endif
#ifndef DISABLE_IWARP_VLAN_0_WA
	iwdev->dcb_vlan_mode = l2params.num_tc > 1 && !l2params.dscp_mode;

#else
	if (rf->protocol_used != IRDMA_IWARP_PROTOCOL_ONLY)
		iwdev->dcb_vlan_mode = l2params.num_tc > 1 && !l2params.dscp_mode;
#endif
	err = irdma_rt_init_hw(iwdev, &l2params);
	if (err)
		goto err_rt_init;

#ifdef CPK_A0_ERRATA
	irdma_enable_lb(iwdev);
#endif
#ifndef EXTERNAL_RELEASE
	/**
	 * We need to register notfifiers first in order to avoid a hang 
	 * when the bonding driver is removed and that causes an 
	 * irdma_probe. The bonding driver takes the rtnl lock and waits 
	 * for the netdev refcount to go to zero which it cannot if the 
	 * register notifiers is waiting on the rtnl lock. 
	 */
#endif
	err = irdma_register_notifiers(iwdev);
	if (err)
		goto err_notifier_reg;
#ifdef GLOBAL_HDL_LIST
	irdma_add_handler(hdl);
#ifdef DEBUG_SUPPORT
#ifdef CONFIG_DEBUG_FS
	irdma_dbg_pf_init(hdl);
#endif
#endif
#endif /* GLOBAL_HDL_LIST */

	err = irdma_ib_register_device(iwdev);
	if (err)
		goto err_ibreg;

#ifndef EXTERNAL_RELEASE
	/* TBD: MEV doesn't implement this yet */
#endif
	if (rf->rdma_ver == IRDMA_GEN_2) {
#ifdef IIDC_UPSTREAM
		ice_rdma_update_vsi_filter(pf, iwdev->vsi_num, true);
#else
#ifdef CVL_VCHNL_UNPRIVILEGED
		if (rf->sc_dev.privileged)
			cdev_info->ops->update_vport_filter(
				cdev_info, iwdev->vsi_num, true);
#else
		cdev_info->ops->update_vport_filter(cdev_info,
						    iwdev->vsi_num, true);
#endif
#endif
	}

	ibdev_dbg(&iwdev->ibdev, "INIT: Gen[%d] PF[%d] device probe success\n",
		  rf->rdma_ver, PCI_FUNC(rf->pcidev->devfn));

	auxiliary_set_drvdata(aux_dev, iwdev);

	return 0;

err_ibreg:
#ifdef GLOBAL_HDL_LIST
#ifdef DEBUG_SUPPORT
#ifdef CONFIG_DEBUG_FS
	irdma_dbg_pf_exit(iwdev->hdl);
#endif
#endif
	irdma_del_handler(iwdev->hdl);
#endif
	irdma_unregister_notifiers(iwdev);
err_notifier_reg:
	irdma_rt_deinit_hw(iwdev);
err_rt_init:
	irdma_ctrl_deinit_hw(rf);
err_ctrl_init:
#ifdef GLOBAL_HDL_LIST
	kfree(hdl);
err_hdl:
#endif
#ifndef IIDC_UPSTREAM
	destroy_workqueue(rf->vchnl_wq);
err_fill_devinfo:
#endif
	kfree(iwdev->rf);
	ib_dealloc_device(&iwdev->ibdev);

	return err;
}

static const struct auxiliary_device_id irdma_auxiliary_id_table[] = {
	{.name = "ice.iwarp", },
	{.name = "ice.roce", },
#ifdef VCHNL_SUPPORT
#ifdef MEV_IIDC_WA
#ifndef EXTERNAL_RELEASE
	/* TODO: Remove this once fixed in idpf */ 
#endif
	{.name = "idpf.iwarp", },
#endif /* MEV_IIDC_WA */
	{.name = "idpf.roce", },
#ifdef CVL_VCHNL_UNPRIVILEGED
	{.name = "iavf.iwarp", },
	{.name = "iavf.roce", },
#endif /* CVL_VCHNL_UNPRIVILEGED */
#endif /* VCHNL_SUPPORT */
	{},
};

MODULE_DEVICE_TABLE(auxiliary, irdma_auxiliary_id_table);

static struct iidc_auxiliary_drv irdma_auxiliary_drv = {
	.adrv = {
	    .id_table = irdma_auxiliary_id_table,
	    .probe = irdma_probe,
	    .remove = irdma_remove,
	},
	.event_handler = irdma_iidc_event_handler,
#ifdef VCHNL_SUPPORT
	.vc_receive = irdma_vchnl_receive,
#endif
};

static int __init irdma_init_module(void)
{
	int ret;

#ifndef UPSTREAM_RELEASE
	pr_info("irdma driver version: %d.%d.%d\n", DRV_VER_MAJOR,
		DRV_VER_MINOR, DRV_VER_BUILD);
	ret = irdma_init_dbg_and_configfs();
	if (ret)
		return ret;

#endif /* UPSTREAM_RELEASE */
	ret = auxiliary_driver_register(&i40iw_auxiliary_drv);
	if (ret) {
		irdma_pr_err("Failed i40iw(gen_1) auxiliary_driver_register() ret=%d\n", ret);
#ifndef UPSTREAM_RELEASE
		irdma_deinit_dbg_and_configfs();
#endif
		return ret;
	}

	ret = auxiliary_driver_register(&irdma_auxiliary_drv.adrv);
	if (ret) {
		auxiliary_driver_unregister(&i40iw_auxiliary_drv);
		irdma_pr_err("Failed irdma auxiliary_driver_register() ret=%d\n", ret);
#ifndef UPSTREAM_RELEASE
		irdma_deinit_dbg_and_configfs();
#endif
		return ret;
	}

	return 0;
}

static void __exit irdma_exit_module(void)
{
	auxiliary_driver_unregister(&irdma_auxiliary_drv.adrv);
	auxiliary_driver_unregister(&i40iw_auxiliary_drv);
#ifndef UPSTREAM_RELEASE
	irdma_deinit_dbg_and_configfs();
#endif
}

module_init(irdma_init_module);
module_exit(irdma_exit_module);
