/* Copyright (c) 2016 - 2021 Intel Corporation */
#ifndef LICENSE_REPLACED
/* For licensing information, see the file 'LICENSE' in the root folder */
#endif /* LICENSE_REPLACED */
#ifndef IRDMA_MSG_H
#define IRDMA_MSG_H

#include "irdma_hmc.h"

/* must be not zero to aviod confusion after zeroing memory */
#define OPAGUE_CHANNEL_VERSION  0x1234
#define IRDMA_MSG_REQUEST  0xA
#define IRDMA_MSG_OK  0x1
#define IRDMA_MSG_ERROR  0xF

#define ALLOC_ALL_HMC_OBJ_MSG_LEN (I40E_HMC_IW_MAX * ALLOC_ALL_HMC_OBJ_LINES + HEADER_LEN)
#define FREE_ALL_HMC_OBJ_MSG_LEN (I40E_HMC_IW_MAX * FREE_ALL_HMC_OBJ_LINES + HEADER_LEN)
#define ALLOC_HMC_FNID_MSG_LEN (HEADER_LEN + ALLOC_HMC_FNID_LINES)
#define CHANNEL_VER_MSG_LEN (HEADER_LEN + CHANNEL_VER_LINES)
#define MBOX_ID GENMASK(27, 16)	/* [WF] should be only 12 bits !!!!! */
#define MSG_ID GENMASK(15, 0)
#define SEQ_NUM GENMASK(15, 0)
#define MSG_LEN GENMASK(27, 16)
#define RESP_CODE GENMASK(31, 28)

/* HMC objects allocation and free*/
#define HEADER_LEN 2
#define ALLOC_ALL_HMC_OBJ_LINES 5
#define FREE_ALL_HMC_OBJ_LINES 3
#define ALLOC_HMC_FNID_LINES 1
#define CHANNEL_VER_LINES 1
#define OBJECT_TYPE GENMASK(15, 0)
#define OBJECT_TYPE_LINE 0
#define HMC_ENTRY_TYPE GENMASK(19, 16)
#define HMC_ENTRY_TYPE_LINE 0
#define FUNCTION_INDEX GENMASK(31, 24)
#define FUNCTION_INDEX_LINE 1
#define OBJECT_INDEX GENMASK(23, 0)
#define OBJECT_INDEX_LINE 1
#define OBJECT_CNT GENMASK(31, 0)
#define OBJECT_CNT_LINE 2
#define HIGH_ADDR_OFF GENMASK_ULL(63, 32)
#define HIGH_ADDR_OFF_LINE 3
#define LOW_ADDR_OFF GENMASK_ULL(31, 0)
#define LOW_ADDR_OFF_LINE 4

/* alloc HMC fn id*/
#define HMC_FNID_SHIFT 0
#define HMC_FNID_MASK  (0x3F << LOW_ADDR_OFF_SHIFT)
#define HMC_FNID_LINE 0
#define CHANNEL_VER GENMASK(15, 0)
#define CHANNEL_VER_LINE 0

#define GET_ALLOC_MSG(buf, index, x) (buf[(x ## _LINE + HEADER_LEN) + index * ALLOC_ALL_HMC_OBJ_LINES])
#define GET_FREE_MSG(buf, index, x) (buf[(x ## _LINE + HEADER_LEN) + index * FREE_ALL_HMC_OBJ_LINES])

#define SET_ALLOC_MSG(buf, index, x, field_val) iw_vf_set_field((u32 *)buf + (x ## _LINE + HEADER_LEN) + index * ALLOC_ALL_HMC_OBJ_LINES, field_val, x ## _SHIFT, x ## _MASK)
#define SET_FREE_MSG(buf, index, x, field_val) iw_vf_set_field((u32 *)buf + (x ## _LINE + HEADER_LEN) + index * FREE_ALL_HMC_OBJ_LINES, field_val, x ## _SHIFT, x ## _MASK)

#define SET_FIELD(buf, x, field_val) iw_vf_set_field((u32 *)buf + (x ## _LINE + HEADER_LEN), (u64)field_val, x ## _SHIFT, x ## _MASK)
#define GET_FIELD(buf, x) (buf[(x ## _LINE + HEADER_LEN)])

enum {
	IRDMA_VF_ALLOC_ALL_HMC_OBJ = 0x01,	/* VF requests all HMC object allocation */
	IRDMA_VF_FREE_ALL_HMC_OBJ,	/* VF requests the HMC object free */
	IRDMA_VF_ALLOC_HMC_FNID,	/* VF requests the alloc HMC FN id */
	IRDMA_VF_CHANNEL_VER,	/* VF requests PF channel version for compatibility checking */
	IRDMA_VF_END_MSG,
};

struct channel_version_resp {
	u16 version;
};

struct hmc_object_alloc_resp {
	u8 rsrc_type;
	u32 count;
	u32 start_idx;		/* only 24 bits */
	u64 addr_offset;	/* HMC address offset for object type */
	u8 entry_type;
};

struct hmc_object_free_resp {
	u8 rsrc_type;
	u32 count;
	u32 start_idx;		/* only 24 bits */
};

struct alloc_hmc_fnid_resp {
	u8 vf_id;
	u8 hmc_fn_id;
	bool free_pm_fcn;
};

struct iwarp_msg {
	u16 mbox_id;
	u16 msg_id;
	u16 msg_len;		/* message length in words it include iwarp header length */
	u16 responce_code;
	u16 seq_num;		/* increment */
	union {
		struct hmc_object_alloc_resp alloc_obj[I40E_HMC_IW_MAX];
		struct hmc_object_free_resp free_obj[I40E_HMC_IW_MAX];
		struct alloc_hmc_fnid_resp hmc_fnid_obj;
		struct channel_version_resp channel_ver_rsp;
		/* TODO for new function */
	} mtype;
	struct mutex msg_lock;

	struct list_head list;
	struct semaphore msg_sem;
	struct timer_list msg_timer;
};

int parse_iw_msg(struct iwarp_msg *iw_msg_out, u32 *msg_in, u16 msg_len);
int pack_iw_msg(struct iwarp_msg *iw_msg_in, u32 *msg_out, u16 *msg_len);

#endif				/* _IRDMA_MSG_H */
