#ifndef INTERFACE_H
#define INTERFACE_H


#include "kernel_abstract.h"
#include "defs.h"
#include "list.h"
#include "status.h"
#include "irdma.h"

//THEM

enum ib_send_flags {
	IB_SEND_FENCE           = 1,
	IB_SEND_SIGNALED        = (1<<1),
	IB_SEND_SOLICITED       = (1<<2),
	IB_SEND_INLINE          = (1<<3),
	IB_SEND_IP_CSUM         = (1<<4),

	/* reserve bits 26-31 for low level drivers' internal use */
	IB_SEND_RESERVED_START  = (1 << 26),
	IB_SEND_RESERVED_END    = (1 << 31),
};

enum ib_wr_opcode {
	IB_WR_RDMA_WRITE,
	IB_WR_RDMA_WRITE_WITH_IMM,
	IB_WR_SEND,
	IB_WR_SEND_WITH_IMM,
	IB_WR_RDMA_READ,
	IB_WR_ATOMIC_CMP_AND_SWP,
	IB_WR_ATOMIC_FETCH_AND_ADD,
	IB_WR_LSO,
	IB_WR_SEND_WITH_INV,
	IB_WR_RDMA_READ_WITH_INV,
	IB_WR_LOCAL_INV,
	IB_WR_REG_MR,
	IB_WR_MASKED_ATOMIC_CMP_AND_SWP,
	IB_WR_MASKED_ATOMIC_FETCH_AND_ADD,
	IB_WR_REG_SIG_MR,
	/* reserve values for low level drivers' internal use.
	 * These values will not be used at all in the ib core layer.
	 */
	IB_WR_RESERVED1 = 0xf0,
	IB_WR_RESERVED2,
	IB_WR_RESERVED3,
	IB_WR_RESERVED4,
	IB_WR_RESERVED5,
	IB_WR_RESERVED6,
	IB_WR_RESERVED7,
	IB_WR_RESERVED8,
	IB_WR_RESERVED9,
	IB_WR_RESERVED10,
};

enum ib_qp_type {
	/*
	 * IB_QPT_SMI and IB_QPT_GSI have to be the first two entries
	 * here (and in that order) since the MAD layer uses them as
	 * indices into a 2-entry table.
	 */
	IB_QPT_SMI,
	IB_QPT_GSI, //qp for UD connection management, similar to ILQ for UD?

	IB_QPT_RC, //can be either RoCE or iWARP
	IB_QPT_UC,
	IB_QPT_UD, //RoCE
	IB_QPT_RAW_IPV6,
	IB_QPT_RAW_ETHERTYPE,
	IB_QPT_RAW_PACKET = 8,
	IB_QPT_XRC_INI = 9,
	IB_QPT_XRC_TGT,
	IB_QPT_MAX,
	/* Reserve a range for qp types internal to the low level driver.
	 * These qp types will not be visible at the IB core layer, so the
	 * IB_QPT_MAX usages should not be affected in the core layer
	 */
	IB_QPT_RESERVED1 = 0x1000,
	IB_QPT_RESERVED2,
	IB_QPT_RESERVED3,
	IB_QPT_RESERVED4,
	IB_QPT_RESERVED5,
	IB_QPT_RESERVED6,
	IB_QPT_RESERVED7,
	IB_QPT_RESERVED8,
	IB_QPT_RESERVED9,
	IB_QPT_RESERVED10,
};

enum rdma_ah_attr_type {
        RDMA_AH_ATTR_TYPE_IB,
        RDMA_AH_ATTR_TYPE_ROCE,
        RDMA_AH_ATTR_TYPE_OPA,
};

enum ib_mw_type {
        IB_MW_TYPE_1 = 1,
        IB_MW_TYPE_2 = 2
};

struct ib_qp {
	enum ib_qp_type         qp_type;
};

struct ib_sge {
	u64	addr;
	u32	length;
	u32	lkey;
};

struct ib_send_wr {
	struct ib_send_wr	*next;
	union {
		u64		wr_id;
	};
	struct ib_sge		*sg_list;
	int			num_sge;
	enum ib_wr_opcode	opcode;
	int			send_flags;
	union {
		__be32		imm_data;
		u32		invalidate_rkey;
	} ex;

};

struct ib_device {
	u64 placeholder;
};

struct ib_pd {
	u64 placeholder;
};

struct ib_ah {
        struct ib_device        *device;
        struct ib_pd            *pd;
        struct ib_uobject       *uobject;
        enum rdma_ah_attr_type  type;
};

struct ib_uobject {
	u64 placeholder;
};

struct ib_atomic_wr {
	struct ib_send_wr	wr;
	u64			remote_addr;
	u64			compare_add;
	u64			swap;
	u64			compare_add_mask;
	u64			swap_mask;
	u32			rkey;
};

static inline struct ib_atomic_wr *atomic_wr(struct ib_send_wr *wr)
{
	return container_of(wr, struct ib_atomic_wr, wr);
}

struct ib_ud_wr {
        struct ib_send_wr       wr;
        struct ib_ah            *ah;
        void                    *header;
        int                     hlen;
        int                     mss;
        u32                     remote_qpn;
        u32                     remote_qkey;
        u16                     pkey_index; /* valid for GSI only */
        u8                      port_num;   /* valid for DR SMPs on switch only */
};

struct ib_rdma_wr {
        struct ib_send_wr       wr;
        u64                     remote_addr;
        u32                     rkey;
};

struct ib_mr {
        struct ib_device  *device;
        struct ib_pd      *pd;
        u32                lkey;
        u32                rkey;
        u64                iova;
        u64                length;
        unsigned int       page_size;
        bool               need_inval;
        union {
                struct ib_uobject       *uobject;       /* user */
                struct list_head        qp_entry;       /* FR */
        };
};

struct ib_mw {
        struct ib_device        *device;
        struct ib_pd            *pd;
        struct ib_uobject       *uobject;
        u32                     rkey;
        enum ib_mw_type         type;
};

struct ib_fmr {
        struct ib_device        *device;
        struct ib_pd            *pd;
        struct list_head        list;
        u32                     lkey;
        u32                     rkey;
};
struct ib_reg_wr {
        struct ib_send_wr       wr;
        struct ib_mr            *mr;
        u32                     key;
        int                     access;
};

//FIXME: Eventually this should be replaced by our .h files
//US

enum irdma_qp_caps {
	IRDMA_WRITE_WITH_IMM = 1,
	IRDMA_SEND_WITH_IMM = 2,
};

enum irdma_device_caps_const {
	IRDMA_WQE_SIZE =                        4,
	IRDMA_CQP_WQE_SIZE =                    8,
	IRDMA_CQE_SIZE =                        4,
	IRDMA_EXTENDED_CQE_SIZE =               8,
	IRDMA_AEQE_SIZE =                       2,
	IRDMA_CEQE_SIZE =                       1,
	IRDMA_CQP_CTX_SIZE =                    8,
	IRDMA_SHADOW_AREA_SIZE =                8,
	IRDMA_QUERY_FPM_BUF_SIZE =              176,
	IRDMA_COMMIT_FPM_BUF_SIZE =             176,
	IRDMA_GATHER_STATS_BUF_SIZE =           304,
	IRDMA_MIN_IW_QP_ID =                    0,
	IRDMA_MAX_IW_QP_ID =                    262143,
	IRDMA_MIN_CEQID =                       0,
	IRDMA_MAX_CEQID =                       1023,
	IRDMA_CEQ_MAX_COUNT =                   IRDMA_MAX_CEQID + 1,
	IRDMA_MIN_CQID =                        0,
	IRDMA_MAX_CQID =                        524287,
	IRDMA_MIN_AEQ_ENTRIES =                 1,
	IRDMA_MAX_AEQ_ENTRIES =                 524287,
	IRDMA_MIN_CEQ_ENTRIES =                 1,
	IRDMA_MAX_CEQ_ENTRIES =                 524288,
	IRDMA_MIN_CQ_SIZE =                     1,
	IRDMA_MAX_CQ_SIZE =                     1048575,
	IRDMA_DB_ID_ZERO =                      0,
	IRDMA_MAX_WQ_FRAGMENT_COUNT =           14,
	IRDMA_MAX_SGE_RD =                      14,
	IRDMA_MAX_OUTBOUND_MSG_SIZE =           2147483647,
	IRDMA_MAX_INBOUND_MSG_SIZE =            2147483647,
	IRDMA_MAX_PUSH_PAGE_COUNT =             4096,
	IRDMA_MAX_PE_ENA_VF_COUNT =             32,
	IRDMA_MAX_VF_FPM_ID =                   47,
	IRDMA_MAX_SQ_PAYLOAD_SIZE =             2145386496,
	IRDMA_MAX_INLINE_DATA_SIZE =            224,
	IRDMA_MAX_PUSHMODE_INLINE_DATA_SIZE =   224,
	IRDMA_MAX_IRD_SIZE =                    127,
	IRDMA_MAX_ORD_SIZE =                    255,
	IRDMA_MAX_WQ_ENTRIES =                  16384,
	IRDMA_Q2_BUF_SIZE =                     256,
	IRDMA_QP_CTX_SIZE =                     256,
	IRDMA_MAX_PDS =                         262144,
};

enum irdma_page_size {
	IRDMA_PAGE_SIZE_4K = 0,
	IRDMA_PAGE_SIZE_2M,
	IRDMA_PAGE_SIZE_1G,
};

enum irdma_addressing_type {
        IRDMA_ADDR_TYPE_ZERO_BASED = 0,
        IRDMA_ADDR_TYPE_VA_BASED = 1,
};

enum ib_access_flags {
        IB_ACCESS_LOCAL_WRITE   = 1,
        IB_ACCESS_REMOTE_WRITE  = (1<<1),
        IB_ACCESS_REMOTE_READ   = (1<<2),
        IB_ACCESS_REMOTE_ATOMIC = (1<<3),
        IB_ACCESS_MW_BIND       = (1<<4),
        IB_ZERO_BASED           = (1<<5),
        IB_ACCESS_ON_DEMAND     = (1<<6),
        IB_ACCESS_HUGETLB       = (1<<7),
};

#define IRDMA_OP_TYPE_RDMA_WRITE                0x00
#define IRDMA_OP_TYPE_RDMA_READ                 0x01
#define IRDMA_OP_TYPE_SEND                      0x03
#define IRDMA_OP_TYPE_SEND_INV                  0x04
#define IRDMA_OP_TYPE_SEND_SOL                  0x05
#define IRDMA_OP_TYPE_SEND_SOL_INV              0x06
#define IRDMA_OP_TYPE_RDMA_WRITE_SOL            0x0d
#define IRDMA_OP_TYPE_BIND_MW                   0x08
#define IRDMA_OP_TYPE_FAST_REG_NSMR             0x09
#define IRDMA_OP_TYPE_INV_STAG                  0x0a
#define IRDMA_OP_TYPE_RDMA_READ_INV_STAG        0x0b
#define IRDMA_OP_TYPE_NOP                       0x0c
#define IRDMA_OP_TYPE_REC       0x3e
#define IRDMA_OP_TYPE_REC_IMM   0x3f

#define IRDMA_MIN_PAGES_PER_FMR		1

#define irdma_tagged_offset u64
#define irdma_stag_index u32
#define irdma_stag u32
#define irdma_stag_key u8
//#define irdma_sgl \
//	struct irdma_sge *

typedef struct irdma_sge* irdma_sgl;

#define IRDMA_MAX_SAVED_PHY_PGADDR 4
#define IRDMA_ACCESS_FLAGS_LOCALREAD 0x01
#define IRDMA_ACCESS_FLAGS_LOCALWRITE           0x02
#define IRDMA_ACCESS_FLAGS_REMOTEREAD_ONLY      0x04
#define IRDMA_ACCESS_FLAGS_REMOTEREAD           0x05
#define IRDMA_ACCESS_FLAGS_REMOTEWRITE_ONLY     0x08
#define IRDMA_ACCESS_FLAGS_REMOTEWRITE          0x0a
#define IRDMA_ACCESS_FLAGS_BIND_WINDOW          0x10
#define IRDMA_ACCESS_FLAGS_ALL                  0x1f
#define IRDMA_DEBUG_WQE "IRDMA_WQE"
#define kc_typeq_ib_wr const

struct irdma_sge {
	irdma_tagged_offset tag_off;
	u32 len;
	irdma_stag stag;
};

struct irdma_ring {
	u32 head;
	u32 tail;
	u32 size;
};

struct irdma_cqe {
	__le64 buf[IRDMA_CQE_SIZE];
};

struct irdma_sc_dev;
struct irdma_device;
struct irdma_hw_attrs;

struct irdma_sq_uk_wr_trk_info {
        u64 wrid;
        u32 wr_len;
        u16 quanta;
        u8 reserved[2];
};

struct irdma_qp_quanta {
        __le64 elem[IRDMA_WQE_SIZE];
};

struct irdma_priv_qp_ops {
	enum irdma_status_code (*qp_init)(struct irdma_sc_qp *qp,
					  struct irdma_qp_init_info *info);
	enum irdma_status_code (*qp_create)(struct irdma_sc_qp *qp,
					    struct irdma_create_qp_info *info,
					    u64 scratch, bool post_sq);
	enum irdma_status_code (*qp_modify)(struct irdma_sc_qp *qp,
					    struct irdma_modify_qp_info *info,
					    u64 scratch, bool post_sq);
	enum irdma_status_code (*qp_destroy)(struct irdma_sc_qp *qp, u64 scratch,
					     bool remove_hash_idx,
					     bool ignore_mw_bnd, bool post_sq);
	enum irdma_status_code (*qp_flush_wqes)(struct irdma_sc_qp *qp,
						struct irdma_qp_flush_info *info,
						u64 scratch, bool post_sq);
	enum irdma_status_code (*qp_upload_context)(struct irdma_sc_dev *dev,
						    struct irdma_upload_context_info *info,
						    u64 scratch, bool post_sq);
	void (*qp_setctx)(struct irdma_sc_qp *qp, __le64 *qp_ctx,
			  struct irdma_qp_host_ctx_info *info);
	void (*qp_setctx_roce)(struct irdma_sc_qp *qp, __le64 *qp_ctx,
			       struct irdma_qp_host_ctx_info *info);
	void (*qp_send_lsmm)(struct irdma_sc_qp *qp, void *lsmm_buf, u32 size, irdma_stag stag);
	void (*qp_send_lsmm_nostag)(struct irdma_sc_qp *qp, void *lsmm_buf, u32 size);
	void (*qp_send_rtt)(struct irdma_sc_qp *qp, bool read);
	enum irdma_status_code (*iw_mr_fast_register)(struct irdma_sc_qp *qp,
						      struct irdma_fast_reg_stag_info *info,
						      bool post_sq);
};

struct irdma_qp_uk_ops {
	void (*iw_qp_post_wr)(struct irdma_qp_uk *qp);
	void (*iw_qp_ring_push_db)(struct irdma_qp_uk *qp, u32 wqe_index);
	enum irdma_status_code (*iw_rdma_write)(struct irdma_qp_uk *qp,
						struct irdma_post_sq_info *info,
						bool post_sq);
	enum irdma_status_code (*iw_rdma_read)(struct irdma_qp_uk *qp,
					       struct irdma_post_sq_info *info,
					       bool inv_stag, bool post_sq);
	enum irdma_status_code (*iw_send)(struct irdma_qp_uk *qp,
					  struct irdma_post_sq_info *info,
					  bool post_sq);
	enum irdma_status_code (*iw_inline_rdma_write)(struct irdma_qp_uk *qp,
						       struct irdma_post_sq_info *info,
						       bool post_sq);
	enum irdma_status_code (*iw_inline_send)(struct irdma_qp_uk *qp,
						 struct irdma_post_sq_info *info,
						 bool post_sq);
	enum irdma_status_code (*iw_stag_local_invalidate)(struct irdma_qp_uk *qp,
							   struct irdma_post_sq_info *info,
							   bool post_sq);
	enum irdma_status_code (*iw_mw_bind)(struct irdma_qp_uk *qp,
					     struct irdma_post_sq_info *info,
					     bool post_sq);
	enum irdma_status_code (*iw_post_receive)(struct irdma_qp_uk *qp,
						  struct irdma_post_rq_info *info);
	enum irdma_status_code (*iw_post_nop)(struct irdma_qp_uk *qp,
					      u64 wr_id,
					      bool signaled,
					      bool post_sq);
	void (*iw_set_fragment)(__le64 *wqe,
				u32 offset,
				struct irdma_sge *sge,
				u8 valid);
};

struct irdma_qp_uk {
        struct irdma_qp_quanta *sq_base;
        struct irdma_qp_quanta *rq_base;
        struct irdma_hw_attrs *hw_attrs;
	u32 __iomem *wqe_alloc_db;
	struct irdma_sq_uk_wr_trk_info *sq_wrtrk_array;
	__le64 *shadow_area;
	u32 *push_db;
        __le64 *push_wqe;
	struct irdma_ring sq_ring;
	struct irdma_ring rq_ring;
	struct irdma_ring initial_ring;
	u32 qp_id;
	u32 qp_caps;
	u32 max_sq_frag_cnt;
	u32 max_rq_frag_cnt;
	struct irdma_qp_uk_ops ops;
	u8 swqe_polarity;
	bool push_mode;
};

struct irdma_sc_qp {
	struct irdma_qp_uk qp_uk;
	struct irdma_sc_dev *dev;
	u8 qp_type;
};

struct irdma_qp {
	struct ib_qp ibqp;
	struct irdma_sc_qp sc_qp;
	struct irdma_device *iwdev;
	spinlock_t lock;
	u8 flush_issued:1;
	u8 sig_all:1;
};

struct irdma_virt_mem {
        void *va;
        u32 size;
}; //__packed

enum irdma_pble_level {
        PBLE_LEVEL_0 = 0,
        PBLE_LEVEL_1 = 1,
        PBLE_LEVEL_2 = 2,
};

enum irdma_alloc_type {
        PBLE_NO_ALLOC     = 0,
        PBLE_SD_CONTIGOUS = 1,
        PBLE_SD_PAGED     = 2,
};
struct irdma_chunk;

struct irdma_pble_chunkinfo {
        struct irdma_chunk      *pchunk;
        u64                     bit_idx;
        u64                     bits_used;
};

struct irdma_pble_info {
        u64                             addr;
        u32                             idx;
        u32                             cnt;
        struct irdma_pble_chunkinfo     chunkinfo;
};

struct irdma_pble_level2 {
        struct irdma_pble_info  root;
        struct irdma_pble_info *leaf;
        struct irdma_virt_mem   leafmem;
        u32                     leaf_cnt;
};

struct irdma_pble_alloc {
        u32                     total_cnt;
        enum irdma_pble_level   level;
        union {
                struct irdma_pble_info          level1;
                struct irdma_pble_level2        level2;
        };
};

struct irdma_hmc_pble {
        union {
                u32 idx;
                dma_addr_t addr;
        };
};

struct irdma_cq_mr {
        struct irdma_hmc_pble cq_pbl;
        dma_addr_t shadow;
};

struct irdma_qp_mr {
        struct irdma_hmc_pble sq_pbl;
        struct irdma_hmc_pble rq_pbl;
        dma_addr_t shadow;
        struct page *sq_page;
};

struct irdma_pbl {
	struct list_head list;
	union {
		struct irdma_qp_mr qp_mr;
		struct irdma_cq_mr cq_mr;
	};

	bool pbl_allocated;
	bool on_list;
	u64 user_base;
	struct irdma_pble_alloc pble_alloc;
	struct irdma_mr *iwmr;
};

struct irdma_dma_info {
        dma_addr_t *dmaaddrs;
};

struct irdma_chunk {
        struct list_head        list;
        struct irdma_dma_info   dmainfo;
        void                    *bitmapbuf;

        u32                     sizeofbitmap;
        u64                     size;
        u64                     vaddr;
        u64                     fpm_addr;
        u32                     pg_cnt;
        enum irdma_alloc_type   type;
        struct irdma_sc_dev     *dev;
        struct irdma_virt_mem   bitmapmem;
        struct irdma_virt_mem   chunkmem;
};

struct irdma_mr {
        union {
                struct ib_mr ibmr;
                struct ib_mw ibmw;
                struct ib_fmr ibfmr;
        };
        struct ib_umem *region;
        u16 type;
        u32 page_cnt;
        u32 page_size;
        u64 page_msk;
        u32 npages;
        u32 stag;
        u64 len;
        u64 pgaddrmem[IRDMA_MAX_SAVED_PHY_PGADDR];
        u64 level0_pa;
        struct irdma_pbl iwpbl;
};

//struct irdma_hw_attrs {
//	u32 max_hw_read_sges;
//};

struct irdma_sc_dev {
	struct irdma_hw_attrs hw_attrs;
	struct irdma_priv_qp_ops *iw_priv_qp_ops;
};

struct irdma_post_send {
        irdma_sgl sg_list;
        u32 num_sges;
        u32 qkey;
        u32 dest_qp;
        u32 ah_id;
};

struct irdma_rdma_read {
        irdma_sgl lo_sg_list;
        u32 num_lo_sges;
        struct irdma_sge rem_addr;
};

struct irdma_inv_local_stag {
        irdma_stag target_stag;
};

struct irdma_fast_reg_stag_info {
        u64 wr_id;
        u64 reg_addr_pa;
        u64 fbo;
        void *va;
        u64 total_len;
        u32 page_size;
        u32 chunk_size;
        u32 first_pm_pbl_index;
        enum irdma_addressing_type addr_type;
        irdma_stag_index stag_idx;
        u16 access_rights;
        u32 pd_id;
        irdma_stag_key stag_key;
        bool local_fence;
        bool read_fence;
        bool signaled;
        bool push_wqe;
        bool use_hmc_fcn_index;
        u8 hmc_fcn_index;
        bool use_pf_rid;
        bool defer_flag;
};

struct irdma_post_inline_send {
        void *data;
        u32 len;
        u32 qkey;
        u32 dest_qp;
        u32 ah_id;
};

struct irdma_rdma_write {
        irdma_sgl lo_sg_list;
        u32 num_lo_sges;
        struct irdma_sge rem_addr;
};

struct irdma_inline_rdma_write {
        void *data;
        u32 len;
        struct irdma_sge rem_addr;
};

struct irdma_sc_ah;

struct irdma_ah_info {
        struct irdma_sc_ah *ah;
//        struct irdma_sc_vsi *vsi;
        u32 pd_idx;
        u32 dst_arpindex;
        u32 dest_ip_addr[4];
        u32 src_ip_addr[4];
        u32 flow_label;
        u32 ah_idx;
        bool ipv4_valid;
        bool do_lpbk;
        u16 vlan_tag;
        u8 insert_vlan_tag;
        u8 tc_tos;
        u8 hop_ttl;
        u8 mac_addr[ETH_ALEN];
        bool ah_valid;
};

struct irdma_sc_ah {
        struct irdma_sc_dev *dev;
        struct irdma_ah_info ah_info;
};

struct irdma_ah {
	struct ib_ah ibah;
	struct irdma_sc_ah sc_ah;
};

struct irdma_post_sq_info {
	u64 wr_id;
	u8 op_type;
	u8 l4len;
	bool signaled;
	bool read_fence;
	bool local_fence;
	bool inline_data;
	bool imm_data_valid;
	bool push_wqe;
	bool report_rtt;
	bool udp_hdr;
	u32 imm_data;
	u32 stag_to_inv;
	bool defer_flag;
	union {
		struct irdma_post_send send;
		struct irdma_rdma_write rdma_write;
		struct irdma_rdma_read rdma_read;
		struct irdma_inv_local_stag inv_local_stag;
		struct irdma_inline_rdma_write inline_rdma_write;
		struct irdma_post_inline_send inline_send;
	} op;
};

struct irdma_pci_f {
        struct irdma_sc_dev sc_dev;
};

struct irdma_device {
        struct irdma_pci_f *rf;
};

// FUNCTIONS

// THEM

typedef unsigned long uintptr_t;

inline void writel(u32 b, volatile void *addr)
{
	*((u32 *)addr) = b;
	mb();
}


static inline struct ib_ud_wr *ud_wr(struct ib_send_wr *wr)
{
        return container_of(wr, struct ib_ud_wr, wr);
}

static inline struct ib_rdma_wr *rdma_wr(struct ib_send_wr *wr)
{
        return container_of(wr, struct ib_rdma_wr, wr);
}

static inline struct ib_reg_wr *reg_wr(struct ib_send_wr *wr)
{
        return container_of(wr, struct ib_reg_wr, wr);
}

//FIXME: Eventually this should be replaced by our .h files
// US
#define irdma_debug(a, b, c, d...)
#define irdma_debug_buf(a, b, c, d...)

static inline struct irdma_ah *to_iwah(struct ib_ah *ibah)
{
        return container_of(ibah, struct irdma_ah, ibah);
}

static inline struct irdma_mr *to_iwmr(struct ib_mr *ibmr)
{
        return container_of(ibmr, struct irdma_mr, ibmr);
}

//from protos.h
static inline void irdma_insert_wqe_hdr(__le64 *wqe, u64 hdr)
{
        wmb();   /* make sure WQE is populated before polarity is set */
        set_64bit_val(wqe, 24, hdr);
}

//in defs.h
//set_64bit_val

#endif /* INTERFACE_H */
