#ifndef __KERNEL_ABSTRACT
#define __KERNEL_ABSTRACT

#include <stdint.h>
#include <algorithm>
#include <linux/types.h>

#define BIT_ULL(nr)             (1ULL << (nr))

#define ntohl(a) a
#define cpu_to_le64(a) (a)
#define le64_to_cpu(a) (a)
#define __init
#define __iomem
#define __force
#define likely(a) (a)
#define unlikely(a) (a)
#define __always_unused __attribute__((unused))

static inline void prefetch(void *addr __always_unused)
{
}

#define	EPERM		 1	/* Operation not permitted */
#define	ENOENT		 2	/* No such file or directory */
#define	ESRCH		 3	/* No such process */
#define	EINTR		 4	/* Interrupted system call */
#define	EIO		 5	/* I/O error */
#define	ENXIO		 6	/* No such device or address */
#define	E2BIG		 7	/* Argument list too long */
#define	ENOEXEC		 8	/* Exec format error */
#define	EBADF		 9	/* Bad file number */
#define	ECHILD		10	/* No child processes */
#define	EAGAIN		11	/* Try again */
#define	ENOMEM		12	/* Out of memory */
#define	EACCES		13	/* Permission denied */
#define	EFAULT		14	/* Bad address */
#define	ENOTBLK		15	/* Block device required */
#define	EBUSY		16	/* Device or resource busy */
#define	EEXIST		17	/* File exists */
#define	EXDEV		18	/* Cross-device link */
#define	ENODEV		19	/* No such device */
#define	ENOTDIR		20	/* Not a directory */
#define	EISDIR		21	/* Is a directory */
#define	EINVAL		22	/* Invalid argument */
#define	ENFILE		23	/* File table overflow */
#define	EMFILE		24	/* Too many open files */
#define	ENOTTY		25	/* Not a typewriter */
#define	ETXTBSY		26	/* Text file busy */
#define	EFBIG		27	/* File too large */
#define	ENOSPC		28	/* No space left on device */
#define	ESPIPE		29	/* Illegal seek */
#define	EROFS		30	/* Read-only file system */
#define	EMLINK		31	/* Too many links */
#define	EPIPE		32	/* Broken pipe */
#define	EDOM		33	/* Math argument out of domain of func */
#define	ERANGE		34	/* Math result not representable */
#define	EADDRNOTAVAIL	99	/* Cannot assign requested address */
#define	ENOTSUPP	524	/* Operation is not supported */

#define L1_CACHE_BYTES 64
#define ____cacheline_aligned_in_smp
#define ____cacheline_internodealigned_in_smp
#define __aligned(n)
#define	IFNAMSIZ	16
#define smp_mb()
#define wmb()
#define mb()
#define in_interrupt() 0
#define WARN_ON(a) do { if (a) fprintf(stderr, "WARN file: %s, line: %d\n", __FILE__, __LINE__); } while (0)
#define WARN(a, b) do { if (a) fprintf(stderr, "WARN_ON file: %s, line: %d %s\n", __FILE__, __LINE__, (b)); } while (0)
#define __ALIGN_KERNEL_MASK(x, mask)	(((x) + (mask)) & ~(mask))
#define __ALIGN_KERNEL(x, a)		__ALIGN_KERNEL_MASK(x, (typeof(x))(a) - 1)
#define ALIGN(x, a) __ALIGN_KERNEL((x), (a))
#define GFP_ATOMIC 0
#define __GFP_NOWARN 0
#define MAKEMASK(m, s) ((m) << (s))
//#define min(a, b)    (std::min((a),(b)))
//#define min_t(t,a,b) min((t)(a), (t)(b))
#define ARRAY_SIZE(x) (sizeof(x) / sizeof((x)[0]))
#define ETH_ALEN 6
#define ETH_MIN_MTU 68

typedef uint8_t u8;
typedef uint16_t u16;
typedef uint32_t u32;
typedef uint64_t u64;
typedef int16_t s16;
typedef int32_t s32;
typedef int64_t s64;
typedef u16 __sum16;
typedef u32 __wsum;

enum lock_state {
	LOCK_NEW,
	LOCK_INITIALIZED,
	LOCK_LOCKED,
	LOCK_UNLOCKED,
};

struct mutex {
	char *mem;
	enum lock_state state;
};

static inline void mutex_init(struct mutex *lock)
{
	lock->mem = new char();
	if (lock->state != LOCK_NEW)
		fprintf(stderr, "something is wrong with mutex\n");
	lock->state = LOCK_INITIALIZED;
}

static inline void mutex_lock(struct mutex *lock)
{
	if (lock->state == LOCK_INITIALIZED ||
	    lock->state == LOCK_UNLOCKED)
		lock->state = LOCK_LOCKED;
	else
		fprintf(stderr, "mutex: locking order bad\n");
}

static inline void mutex_unlock(struct mutex *lock)
{
	if (lock->state == LOCK_LOCKED)
		lock->state = LOCK_UNLOCKED;
	else
		fprintf(stderr, "mutex: unlocking order bad\n");
}

static inline void mutex_destroy(struct mutex *lock)
{
	if (lock->state == LOCK_INITIALIZED ||
	    lock->state == LOCK_UNLOCKED)
		delete lock->mem;
	else
		fprintf(stderr, "mutex: destroy order bad\n");
}

//struct callback_head {
//	struct callback_head *next;
//	void (*func)(struct callback_head *head);
//} __attribute__((aligned(sizeof(void *))));
//#define rcu_head callback_head
//#define rcu_read_lock()
//#define rcu_read_unlock()

//struct work_struct {
//	unsigned int handle;
//};

#define smp_rmb()

#define READ_ONCE(x) (x)

/*
 * These can be used to print at the various log levels.
 * All of these will print unconditionally, although note that pr_debug()
 * and other debug macros are compiled out unless either DEBUG is defined
 * or CONFIG_DYNAMIC_DEBUG is set.
 */
#define pr_emerg(fmt, ...) fprintf(stderr, (fmt), ##__VA_ARGS__)
#define pr_alert(fmt, ...) fprintf(stderr, (fmt), ##__VA_ARGS__)
#define pr_crit(fmt, ...)  fprintf(stderr, (fmt), ##__VA_ARGS__)
#define pr_err(fmt, ...)   fprintf(stderr, (fmt), ##__VA_ARGS__)
#define pr_warning(fmt, ...) fprintf(stderr, (fmt), ##__VA_ARGS__)
#define pr_warn pr_warning
#define pr_notice(fmt, ...) fprintf(stderr, (fmt), ##__VA_ARGS__)
#define pr_info(fmt, ...)  fprintf(stderr, (fmt), ##__VA_ARGS__)

#define dev_err(dev, fmt, ...) fprintf(stderr, (fmt), ##__VA_ARGS__)
#define dev_warn(dev, fmt, ...) fprintf(stderr, (fmt), ##__VA_ARGS__)
#define dev_info(dev, fmt, ...) fprintf(stderr, (fmt), ##__VA_ARGS__)
#define dev_dbg(dev, fmt, ...) fprintf(stderr, (fmt), ##__VA_ARGS__)

/*
 * ktime_t:
 *
 * A single 64-bit variable is used to store the hrtimers
 * internal representation of time values in scalar nanoseconds. The
 * design plays out best on 64-bit CPUs, where most conversions are
 * NOPs and most arithmetic ktime_t operations are plain arithmetic
 * operations.
 *
 */

union ktime {
	s64	tv64;
};

typedef struct {
	int counter;
} atomic_t;

typedef struct {
	long long counter;
} atomic_long_t;

static inline void atomic_inc(atomic_t *atom)
{
	atom->counter++;
}

static inline void atomic_dec(atomic_t *atom)
{
	atom->counter--;
}

typedef union ktime ktime_t;		/* Kill this */

#define NR_CPUS 4096

typedef struct refcount_struct {
	atomic_t refs;
} refcount_t;

#define __rcu

//typedef struct cpumask { DECLARE_BITMAP(bits, NR_CPUS); } cpumask_t;


struct device {
	struct device		*parent;
	struct bus_type	*bus;		/* type of bus device is on */
	struct device_driver *driver;	/* which driver has allocated this */
	void *driver_data;
	void (*release)(struct device *dev);
};

#define STATIC
#define INLINE
#define GFP_KERNEL 0
#define __GFP_ZERO 0

typedef unsigned gfp_t;

#define FIELD_SIZEOF(t, f) (sizeof(((t*)0)->f))

typedef u64 resource_size_t;

/* for dma_addr_t */
#include "dma-mapping.h"

extern void *dmam_alloc_coherent(struct device *dev, size_t size,
                                 dma_addr_t *dma_handle, gfp_t gfp);
extern void dmam_free_coherent(struct device *dev, size_t size, void *vaddr,
                               dma_addr_t dma_handle);

#define page_to_nid(p) 0
#define page_is_pfmemalloc(x) 0
#define numa_mem_id() 0

typedef struct qspinlock {
	atomic_t	val;
} arch_spinlock_t;

typedef struct raw_spinlock {
	arch_spinlock_t raw_lock;
} raw_spinlock_t;

typedef struct spinlock {
	union {
		struct raw_spinlock rlock;

#ifdef CONFIG_DEBUG_LOCK_ALLOC
# define LOCK_PADSIZE (offsetof(struct raw_spinlock, dep_map))
		struct {
			u8 __padding[LOCK_PADSIZE];
			struct lockdep_map dep_map;
		};
#endif
	};
} spinlock_t;


static inline void spin_lock_bh(spinlock_t *lock)
{
	if (lock->rlock.raw_lock.val.counter)
		fprintf(stderr, "lock already acquired!\n");
	else
		lock->rlock.raw_lock.val.counter = 1;
}

static inline void spin_unlock_bh(spinlock_t *lock)
{
	if (!lock->rlock.raw_lock.val.counter)
		fprintf(stderr, "lock already released\n");
	else
		lock->rlock.raw_lock.val.counter = 0;
}

static inline void spin_lock_irqsave(spinlock_t *lock, unsigned long flags)
{
	flags = flags;
	if (lock->rlock.raw_lock.val.counter)
		fprintf(stderr, "lock already acquired!\n");
	else
		lock->rlock.raw_lock.val.counter = 1;
}

static inline void spin_unlock_irqrestore(spinlock_t *lock, unsigned long flags)
{
	flags = flags;
	if (!lock->rlock.raw_lock.val.counter)
		fprintf(stderr, "lock already released\n");
	else
		lock->rlock.raw_lock.val.counter = 0;
}

static inline void spin_lock_init(spinlock_t *lock)
{
	lock->rlock.raw_lock.val.counter = 0;
}

static inline void synchronize_irq(unsigned int irq)
{
}

enum net_device_flags {
/* for compatibility with glibc net/if.h */
	IFF_UP				= 1<<0,  /* sysfs */
	IFF_BROADCAST			= 1<<1,  /* volatile */
	IFF_DEBUG			= 1<<2,  /* sysfs */
	IFF_LOOPBACK			= 1<<3,  /* volatile */
	IFF_POINTOPOINT			= 1<<4,  /* volatile */
	IFF_NOTRAILERS			= 1<<5,  /* sysfs */
	IFF_RUNNING			= 1<<6,  /* volatile */
	IFF_NOARP			= 1<<7,  /* sysfs */
	IFF_PROMISC			= 1<<8,  /* sysfs */
	IFF_ALLMULTI			= 1<<9,  /* sysfs */
	IFF_MASTER			= 1<<10, /* volatile */
	IFF_SLAVE			= 1<<11, /* volatile */
	IFF_MULTICAST			= 1<<12, /* sysfs */
	IFF_PORTSEL			= 1<<13, /* sysfs */
	IFF_AUTOMEDIA			= 1<<14, /* sysfs */
	IFF_DYNAMIC			= 1<<15, /* sysfs */
	IFF_LOWER_UP			= 1<<16, /* volatile */
	IFF_DORMANT			= 1<<17, /* volatile */
	IFF_ECHO			= 1<<18, /* volatile */
};

/* for compatibility with glibc net/if.h */
#define IFF_UP				IFF_UP
#define IFF_BROADCAST			IFF_BROADCAST
#define IFF_DEBUG			IFF_DEBUG
#define IFF_LOOPBACK			IFF_LOOPBACK
#define IFF_POINTOPOINT			IFF_POINTOPOINT
#define IFF_NOTRAILERS			IFF_NOTRAILERS
#define IFF_RUNNING			IFF_RUNNING
#define IFF_NOARP			IFF_NOARP
#define IFF_PROMISC			IFF_PROMISC
#define IFF_ALLMULTI			IFF_ALLMULTI
#define IFF_MASTER			IFF_MASTER
#define IFF_SLAVE			IFF_SLAVE
#define IFF_MULTICAST			IFF_MULTICAST
#define IFF_PORTSEL			IFF_PORTSEL
#define IFF_AUTOMEDIA			IFF_AUTOMEDIA
#define IFF_DYNAMIC			IFF_DYNAMIC

#define jiffies 0x12345678ULL

#define time_before(x,y) 0
/* FIXME: maybe this needs to be a mock instead so that we can force taking
 * different branches?
 */
#define time_after(a,b) \
	((long)((b) - (a)) < 0)

#define smp_mb__before_atomic() do {} while(0)

#define kfree_rcu(ptr, rcu_head) free(ptr)

#define kmalloc(size, flags) calloc(size, sizeof(char))
#define kfree(obj) free(obj)

#define HZ 1000

#define htonl(x) cpu_to_be32(x)
#define cpu_to_be16(x)

//#define _K_SS_MAXSIZE 128
//#define _K_SS_ALIGNSIZE sizeof(struct sockaddr *))

typedef unsigned short __kernel_sa_family_t;

//struct __kernel_sockaddr_storage {
//	__kernel_sa_family_t    ss_family;
//	char                    __data[_K_SS_MAXSIZE - sizeof(unsigned short)];
//};

//#define sockaddr_storage __kernel_sockaddr_storage

static inline void cpumask_set_cpu(unsigned int cpu, struct cpumask *dstp) {}

struct tcphdr {
	__be16  source;
	__be16  dest;
	__be32  seq;
	__be32  ack_seq;
	__u16	res1:4,
		doff:4,
		fin:1,
		syn:1,
		rst:1,
		psh:1,
		ack:1,
		urg:1,
		ece:1,
		cwr:1;
	__be16  window;
	__sum16 check;
	__be16  urg_ptr;
};

//struct udphdr {
//	__be16  source;
//	__be16  dest;
//	__be16  len;
//	__sum16 check;
//};

//#define IPPROTO_TCP	6
//#define IPPROTO_UDP	17
//#define IPPROTO_SCTP	132

/* maybe someone should include linux/stringify.h */
#define __stringify_1(x...)	#x
#define __stringify(x...)	__stringify_1(x)

static inline void * ERR_PTR(long error)
{
	return (void *) error;
}

#endif /* __KERNEL_ABSTRACT */
