#ifndef TB_WQE_H
#define TB_WQE_H


#include "tb_base.h"
#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <string>
#include "kernel_abstract.h"
#include "cqp_fields.h"

enum permissions {
	RSVD,
	READWRITE,
	READONLY,
	WRITEONLY
};

class Field : public TB_OBJ
{
	public:
	std::string name;
	u64 data;
	u64 bit_mask;
	u8 bit_shift;
	u8 bit_length;
	u16 byte_offset;
	enum permissions perm;

	virtual void init(std::string i_name, u64 i_data, u64 i_bit_mask,
			  u8 i_bit_shift, u8 i_bit_length,
			  enum permissions i_perm, u16 i_byte_offset);

	virtual u64 get_masked_data_64(void);
	virtual u64 get_masked_shifted_data_64(void);
	virtual void print_debug_high(void);
};

class INLINE_FIELD : public Field
{
	public:
	u16 quanta; //which quanta this field is in
	u16 inline_len; //total inline data length

	virtual void init(std::string i_name, u64 i_data, u64 i_bit_mask,
			  u8 i_bit_shift, u8 i_bit_length,
			  enum permissions i_perm, u16 i_byte_offset);
	void set_inline_len(u16);
	void set_mask(void);
};

class WQE : public TB_OBJ
{
	public:
	Field *fields; //child class will need to create and destroy this
	int num_fields;
	u64 mem; //used to check actual mem against
	u16 avail_bytes;
	u64 sq_tail_addr[4];
	bool wqe_will_not_fit = false;
	u32 num_quanta;

	virtual void gen_field_vs_rules(int field_idx, u64 *field_data) = 0;
	virtual u64 gen_byte_offset(u16 byte);
	virtual u16 min_wqe_words(void);
	virtual bool compare_64bits(u64 mem_val, u16 byte_off);
	virtual bool compare_wqe_words(u64 *sq_wqe, u16 head, u16 size, u8 num_words);
	virtual void write_wr(u64 *sq_wqe, u16 head, u16 size, u8 num_words);
	virtual Field *find_field_by_name(const std::string& find);
	virtual u64 get_valid(void);
	virtual void randomize(void) = 0;
	virtual int set_field_by_name(const std::string& find, u64 new_data);
	virtual Field *find_field_by_int(int idx);
	virtual void set_field_by_int(int idx, u64 new_val);
};

class CQP_SQ_WR : public WQE
{
	public:

	virtual u16 min_wqe_words(void);
};

class QP_SQ_WR : public WQE
{
	public:

	virtual u16 min_wqe_words(void);
};

#endif /* TB_WQE_H */
