#include "wr.h"

void QP_SEND::set_debug(enum debug_level level)
{
	QP_SQ_WR::set_debug(level);
	if (fields)
		for (int i=0; i < QPSQ_SEND_FIELD_NUM_FIELDS; i++)
			fields[i].set_debug(level);
	else
		if (level >= DEBUG_HIGH)
			printf("WARNING: %s:fields undefined, debug won't work!\n",
			       __func__);
}

void QP_SEND::set_setup_config(enum qp_send_setup var)
{
	setup_config |= var;
}

QP_SEND::QP_SEND(void)
{
	fields = new Field[QPSQ_SEND_FIELD_NUM_FIELDS];
	num_fields = QPSQ_SEND_FIELD_NUM_FIELDS;
	init();
}

QP_SEND::~QP_SEND(void)
{
	delete [] fields;
}

void QP_SEND::update_frag_perm(void)
{
	enum qp_sq_send send_enum;
}

void QP_SEND::init(void)
{
	int offset = 0;
	fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_0].init(
		"tagged_offset_0", //i_name
		0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		READWRITE, //i_perm
		0); //i_byte_offset
	fields[QPSQ_SEND_FIELD_UD_IMM_DATA].init(
		"ud_imm_data", //i_name
		0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		READWRITE, //i_perm
		0); //i_byte_offset
	fields[QPSQ_SEND_FIELD_FRAG_0_VALID].init(
		"frag_0_valid", //i_name
		0, //i_data
		0x1, //i_bit_mask
		63, //i_bit_shift
		1, //i_bit_length
		READWRITE, //i_perm
		8); //i_byte_offset
	fields[QPSQ_SEND_FIELD_FRAG_0_LENGTH].init(
		"frag_0_length", //i_name
		0, //i_data
		0x7fffffff, //i_bit_mask
		32, //i_bit_shift
		31, //i_bit_length
		READWRITE, //i_perm
		8); //i_byte_offset
	fields[QPSQ_SEND_FIELD_STAG_0].init(
		"stag_0", //i_name
		0, //i_data
		0xffffffff, //i_bit_mask
		0, //i_bit_shift
		32, //i_bit_length
		READWRITE, //i_perm
		8); //i_byte_offset
	fields[QPSQ_SEND_FIELD_IW_RSVD0].init(
		"iw_rsvd0", //i_name
		0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		16); //i_byte_offset
	fields[QPSQ_SEND_FIELD_UD_RSVD0].init(
		"ud_rsvd0", //i_name
		0, //i_data
		0xff, //i_bit_mask
		56, //i_bit_shift
		8, //i_bit_length
		RSVD, //i_perm
		16); //i_byte_offset
	fields[QPSQ_SEND_FIELD_UD_DEST_QPN].init(
		"ud_dest_qpn", //i_name
		0, //i_data
		0xffffff, //i_bit_mask
		32, //i_bit_shift
		24, //i_bit_length
		READWRITE, //i_perm
		16); //i_byte_offset
	fields[QPSQ_SEND_FIELD_UD_DEST_QKEY].init(
		"ud_dest_qkey", //i_name
		0, //i_data
		0xffffffff, //i_bit_mask
		0, //i_bit_shift
		32, //i_bit_length
		READWRITE, //i_perm
		16); //i_byte_offset
	fields[QPSQ_SEND_FIELD_WQE_VALID].init(
		"wqe_valid", //i_name
		0, //i_data
		0x1, //i_bit_mask
		63, //i_bit_shift
		1, //i_bit_length
		READWRITE, //i_perm
		24); //i_byte_offset
	fields[QPSQ_SEND_FIELD_SIGNALED_CMPL].init(
		"signaled_cmpl", //i_name
		0, //i_data
		0x1, //i_bit_mask
		62, //i_bit_shift
		1, //i_bit_length
		READWRITE, //i_perm
		24); //i_byte_offset
	fields[QPSQ_SEND_FIELD_IW_LOCAL_FENCE].init(
		"iw_local_fence", //i_name
		0, //i_data
		0x1, //i_bit_mask
		61, //i_bit_shift
		1, //i_bit_length
		READWRITE, //i_perm
		24); //i_byte_offset
	fields[QPSQ_SEND_FIELD_IW_READ_FENCE].init(
		"iw_read_fence", //i_name
		0, //i_data
		0x1, //i_bit_mask
		60, //i_bit_shift
		1, //i_bit_length
		READWRITE, //i_perm
		24); //i_byte_offset
	fields[QPSQ_SEND_FIELD_IW_WAIT_FOR_RCVFPDU].init(
		"iw_wait_for_rcvfpdu", //i_name
		0, //i_data
		0x1, //i_bit_mask
		59, //i_bit_shift
		1, //i_bit_length
		READWRITE, //i_perm
		24); //i_byte_offset
	fields[QPSQ_SEND_FIELD_IW_STREAMING_MODE].init(
		"iw_streaming_mode", //i_name
		0, //i_data
		0x1, //i_bit_mask
		59, //i_bit_shift
		1, //i_bit_length
		READWRITE, //i_perm
		24); //i_byte_offset
	fields[QPSQ_SEND_FIELD_UDP_HEADER].init(
		"ud_header", //i_name
		0, //i_data
		0x1, //i_bit_mask
		61, //i_bit_shift
		1, //i_bit_length
		READWRITE, //i_perm
		24); //i_byte_offset
	fields[QPSQ_SEND_FIELD_UD_RSVD1].init(
		"ud_rsvd1", //i_name
		0, //i_data
		0x1, //i_bit_mask
		60, //i_bit_shift
		1, //i_bit_length
		RSVD, //i_perm
		24); //i_byte_offset
	fields[QPSQ_SEND_FIELD_UD_RSVD2].init(
		"ud_rsvd2", //i_name
		0, //i_data
		0x1, //i_bit_mask
		59, //i_bit_shift
		1, //i_bit_length
		RSVD, //i_perm
		24); //i_byte_offset
	fields[QPSQ_SEND_FIELD_UD_RSVD3].init(
		"ud_rsvd3", //i_name
		0, //i_data
		0x1, //i_bit_mask
		58, //i_bit_shift
		1, //i_bit_length
		RSVD, //i_perm
		24); //i_byte_offset
	fields[QPSQ_SEND_FIELD_INLINE_DATA_FLAG].init(
		"inline_data_flag", //i_name
		0, //i_data
		0x1, //i_bit_mask
		57, //i_bit_shift
		1, //i_bit_length
		READWRITE, //i_perm
		24); //i_byte_offset
	fields[QPSQ_SEND_FIELD_PUSH_WQE].init(
		"push_wqe", //i_name
		0, //i_data
		0x1, //i_bit_mask
		56, //i_bit_shift
		1, //i_bit_length
		READWRITE, //i_perm
		24); //i_byte_offset
	fields[QPSQ_SEND_FIELD_UD_RSVD4].init(
		"ud_rsvd4", //i_name
		0, //i_data
		0xff, //i_bit_mask
		48, //i_bit_shift
		8, //i_bit_length
		RSVD, //i_perm
		24); //i_byte_offset
	fields[QPSQ_SEND_FIELD_IW_RSVD1].init(
		"iw_rsvd1", //i_name
		0, //i_data
		0xff, //i_bit_mask
		48, //i_bit_shift
		8, //i_bit_length
		RSVD, //i_perm
		24); //i_byte_offset
	fields[QPSQ_SEND_FIELD_INLINE_DATA_LENGTH].init(
		"inline_data_length", //i_name
		0, //i_data
		0xff, //i_bit_mask
		48, //i_bit_shift
		8, //i_bit_length
		RSVD, //i_perm
		24); //i_byte_offset
	fields[QPSQ_SEND_FIELD_IMM_DATA_FLAG].init(
		"imm_data_flag", //i_name
		0, //i_data
		0x1, //i_bit_mask
		47, //i_bit_shift
		1, //i_bit_length
		READWRITE, //i_perm
		24); //i_byte_offset
	fields[QPSQ_SEND_FIELD_REPORT_RTT].init(
		"report_rtt", //i_name
		0, //i_data
		0x1, //i_bit_mask
		46, //i_bit_shift
		1, //i_bit_length
		READWRITE, //i_perm
		24); //i_byte_offset
	fields[QPSQ_SEND_FIELD_UD_L4LEN].init(
		"l4len", //i_name
		0, //i_data
		0xf, //i_bit_mask
		42, //i_bit_shift
		4, //i_bit_length
		READWRITE, //i_perm
		24); //i_byte_offset
	fields[QPSQ_SEND_FIELD_ADDITIONAL_FRAG_COUNT].init(
		"additional_frag_count", //i_name
		0, //i_data
		0xf, //i_bit_mask
		38, //i_bit_shift
		4, //i_bit_length
		READWRITE, //i_perm
		24); //i_byte_offset
	fields[QPSQ_SEND_FIELD_OP].init(
		"op", //i_name
		0, //i_data
		0x3f, //i_bit_mask
		32, //i_bit_shift
		6, //i_bit_length
		READWRITE, //i_perm
		24); //i_byte_offset
	fields[QPSQ_SEND_FIELD_UD_RSVD5].init(
		"ud_rsvd5", //i_name
		0, //i_data
		0x7fff, //i_bit_mask
		32, //i_bit_shift
		15, //i_bit_length
		RSVD, //i_perm
		24); //i_byte_offset
	fields[QPSQ_SEND_FIELD_UD_AH_ID].init(
		"ud_ah_id", //i_name
		0, //i_data
		0x1ffff, //i_bit_mask
		0, //i_bit_shift
		17, //i_bit_length
		READWRITE, //i_perm
		24); //i_byte_offset
	fields[QPSQ_SEND_FIELD_IW_REMOTE_INV_STAG].init(
		"iw_remote_inv_stag", //i_name
		0, //i_data
		0xffffffff, //i_bit_mask
		0, //i_bit_shift
		32, //i_bit_length
		READWRITE, //i_perm
		24); //i_byte_offset
	fields[QPSQ_SEND_QP_TYPE].init(
		"qp_type", //i_name
		0, //i_data
		0x7, //i_bit_mask
		0, //i_bit_shift
		3, //i_bit_length
		RSVD, //i_perm
		500); //i_byte_offset
	fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_1].init(
		"tagged_offset_1", //i_name
		0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		32); //i_byte_offset
	fields[QPSQ_SEND_FIELD_FRAG_1_VALID].init(
		"frag_1_valid", //i_name
		0, //i_data
		0x1, //i_bit_mask
		63, //i_bit_shift
		1, //i_bit_length
		RSVD, //i_perm
		40); //i_byte_offset
	fields[QPSQ_SEND_FIELD_FRAG_1_LENGTH].init(
		"frag_1_length", //i_name
		0, //i_data
		0x7fffffff, //i_bit_mask
		32, //i_bit_shift
		31, //i_bit_length
		RSVD, //i_perm
		40); //i_byte_offset
	fields[QPSQ_SEND_FIELD_STAG_1].init(
		"stag_1", //i_name
		0, //i_data
		0xffffffff, //i_bit_mask
		0, //i_bit_shift
		32, //i_bit_length
		RSVD, //i_perm
		40); //i_byte_offset
	fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_2].init(
		"tagged_offset_2", //i_name
		0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		48); //i_byte_offset

	fields[QPSQ_SEND_FIELD_FRAG_2_VALID].init(
		"frag_2_valid", //i_name
		0, //i_data
		0x1, //i_bit_mask
		63, //i_bit_shift
		1, //i_bit_length
		RSVD, //i_perm
		56); //i_byte_offset
	fields[QPSQ_SEND_FIELD_FRAG_2_LENGTH].init(
		"frag_2_length", //i_name
		0, //i_data
		0x7fffffff, //i_bit_mask
		32, //i_bit_shift
		31, //i_bit_length
		RSVD, //i_perm
		56); //i_byte_offset
	fields[QPSQ_SEND_FIELD_STAG_2].init(
		"stag_2", //i_name
		0, //i_data
		0xffffffff, //i_bit_mask
		0, //i_bit_shift
		32, //i_bit_length
		RSVD, //i_perm
		56); //i_byte_offset
	fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_3].init(
		"tagged_offset_3", //i_name
		0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		64); //i_byte_offset

	fields[QPSQ_SEND_FIELD_FRAG_3_VALID].init(
		"frag_3_valid", //i_name
		0, //i_data
		0x1, //i_bit_mask
		63, //i_bit_shift
		1, //i_bit_length
		RSVD, //i_perm
		72); //i_byte_offset
	fields[QPSQ_SEND_FIELD_FRAG_3_LENGTH].init(
		"frag_3_length", //i_name
		0, //i_data
		0x7fffffff, //i_bit_mask
		32, //i_bit_shift
		31, //i_bit_length
		RSVD, //i_perm
		72); //i_byte_offset
	fields[QPSQ_SEND_FIELD_STAG_3].init(
		"stag_3", //i_name
		0, //i_data
		0xffffffff, //i_bit_mask
		0, //i_bit_shift
		32, //i_bit_length
		RSVD, //i_perm
		72); //i_byte_offset
	fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_4].init(
		"tagged_offset_4", //i_name
		0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		80); //i_byte_offset

	fields[QPSQ_SEND_FIELD_FRAG_4_VALID].init(
		"frag_4_valid", //i_name
		0, //i_data
		0x1, //i_bit_mask
		63, //i_bit_shift
		1, //i_bit_length
		RSVD, //i_perm
		88); //i_byte_offset
	fields[QPSQ_SEND_FIELD_FRAG_4_LENGTH].init(
		"frag_4_length", //i_name
		0, //i_data
		0x7fffffff, //i_bit_mask
		32, //i_bit_shift
		31, //i_bit_length
		RSVD, //i_perm
		88); //i_byte_offset
	fields[QPSQ_SEND_FIELD_STAG_4].init(
		"stag_4", //i_name
		0, //i_data
		0xffffffff, //i_bit_mask
		0, //i_bit_shift
		32, //i_bit_length
		RSVD, //i_perm
		88); //i_byte_offset
	fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_5].init(
		"tagged_offset_5", //i_name
		0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		96); //i_byte_offset

	fields[QPSQ_SEND_FIELD_FRAG_5_VALID].init(
		"frag_5_valid", //i_name
		0, //i_data
		0x1, //i_bit_mask
		63, //i_bit_shift
		1, //i_bit_length
		RSVD, //i_perm
		104); //i_byte_offset
	fields[QPSQ_SEND_FIELD_FRAG_5_LENGTH].init(
		"frag_5_length", //i_name
		0, //i_data
		0x7fffffff, //i_bit_mask
		32, //i_bit_shift
		31, //i_bit_length
		RSVD, //i_perm
		104); //i_byte_offset
	fields[QPSQ_SEND_FIELD_STAG_5].init(
		"stag_5", //i_name
		0, //i_data
		0xffffffff, //i_bit_mask
		0, //i_bit_shift
		32, //i_bit_length
		RSVD, //i_perm
		104); //i_byte_offset
	fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_6].init(
		"tagged_offset_6", //i_name
		0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		112); //i_byte_offset

	fields[QPSQ_SEND_FIELD_FRAG_6_VALID].init(
		"frag_6_valid", //i_name
		0, //i_data
		0x1, //i_bit_mask
		63, //i_bit_shift
		1, //i_bit_length
		RSVD, //i_perm
		120); //i_byte_offset
	fields[QPSQ_SEND_FIELD_FRAG_6_LENGTH].init(
		"frag_6_length", //i_name
		0, //i_data
		0x7fffffff, //i_bit_mask
		32, //i_bit_shift
		31, //i_bit_length
		RSVD, //i_perm
		120); //i_byte_offset
	fields[QPSQ_SEND_FIELD_STAG_6].init(
		"stag_6", //i_name
		0, //i_data
		0xffffffff, //i_bit_mask
		0, //i_bit_shift
		32, //i_bit_length
		RSVD, //i_perm
		120); //i_byte_offset
	fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_7].init(
		"tagged_offset_7", //i_name
		0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		128); //i_byte_offset

	fields[QPSQ_SEND_FIELD_FRAG_7_VALID].init(
		"frag_7_valid", //i_name
		0, //i_data
		0x1, //i_bit_mask
		63, //i_bit_shift
		1, //i_bit_length
		RSVD, //i_perm
		136); //i_byte_offset
	fields[QPSQ_SEND_FIELD_FRAG_7_LENGTH].init(
		"frag_7_length", //i_name
		0, //i_data
		0x7fffffff, //i_bit_mask
		32, //i_bit_shift
		31, //i_bit_length
		RSVD, //i_perm
		136); //i_byte_offset
	fields[QPSQ_SEND_FIELD_STAG_7].init(
		"stag_7", //i_name
		0, //i_data
		0xffffffff, //i_bit_mask
		0, //i_bit_shift
		32, //i_bit_length
		RSVD, //i_perm
		136); //i_byte_offset
	fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_8].init(
		"tagged_offset_8", //i_name
		0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		144); //i_byte_offset

	fields[QPSQ_SEND_FIELD_FRAG_8_VALID].init(
		"frag_8_valid", //i_name
		0, //i_data
		0x1, //i_bit_mask
		63, //i_bit_shift
		1, //i_bit_length
		RSVD, //i_perm
		152); //i_byte_offset
	fields[QPSQ_SEND_FIELD_FRAG_8_LENGTH].init(
		"frag_8_length", //i_name
		0, //i_data
		0x7fffffff, //i_bit_mask
		32, //i_bit_shift
		31, //i_bit_length
		RSVD, //i_perm
		152); //i_byte_offset
	fields[QPSQ_SEND_FIELD_STAG_8].init(
		"stag_8", //i_name
		0, //i_data
		0xffffffff, //i_bit_mask
		0, //i_bit_shift
		32, //i_bit_length
		RSVD, //i_perm
		152); //i_byte_offset
	fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_9].init(
		"tagged_offset_9", //i_name
		0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		160); //i_byte_offset
	fields[QPSQ_SEND_FIELD_FRAG_9_VALID].init(
		"frag_9_valid", //i_name
		0, //i_data
		0x1, //i_bit_mask
		63, //i_bit_shift
		1, //i_bit_length
		RSVD, //i_perm
		168); //i_byte_offset
	fields[QPSQ_SEND_FIELD_FRAG_9_LENGTH].init(
		"frag_9_length", //i_name
		0, //i_data
		0x7fffffff, //i_bit_mask
		32, //i_bit_shift
		31, //i_bit_length
		RSVD, //i_perm
		168); //i_byte_offset
	fields[QPSQ_SEND_FIELD_STAG_9].init(
		"stag_9", //i_name
		0, //i_data
		0xffffffff, //i_bit_mask
		0, //i_bit_shift
		32, //i_bit_length
		RSVD, //i_perm
		168); //i_byte_offset

	fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_10].init(
		"tagged_offset_10", //i_name
		0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		176); //i_byte_offset

	fields[QPSQ_SEND_FIELD_FRAG_10_VALID].init(
		"frag_10_valid", //i_name
		0, //i_data
		0x1, //i_bit_mask
		63, //i_bit_shift
		1, //i_bit_length
		RSVD, //i_perm
		184); //i_byte_offset
	fields[QPSQ_SEND_FIELD_FRAG_10_LENGTH].init(
		"frag_10_length", //i_name
		0, //i_data
		0x7fffffff, //i_bit_mask
		32, //i_bit_shift
		31, //i_bit_length
		RSVD, //i_perm
		184); //i_byte_offset
	fields[QPSQ_SEND_FIELD_STAG_10].init(
		"stag_10", //i_name
		0, //i_data
		0xffffffff, //i_bit_mask
		0, //i_bit_shift
		32, //i_bit_length
		RSVD, //i_perm
		184); //i_byte_offset
	fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_11].init(
		"tagged_offset_11", //i_name
		0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		192); //i_byte_offset

	fields[QPSQ_SEND_FIELD_FRAG_11_VALID].init(
		"frag_11_valid", //i_name
		0, //i_data
		0x1, //i_bit_mask
		63, //i_bit_shift
		1, //i_bit_length
		RSVD, //i_perm
		200); //i_byte_offset
	fields[QPSQ_SEND_FIELD_FRAG_11_LENGTH].init(
		"frag_11_length", //i_name
		0, //i_data
		0x7fffffff, //i_bit_mask
		32, //i_bit_shift
		31, //i_bit_length
		RSVD, //i_perm
		200); //i_byte_offset
	fields[QPSQ_SEND_FIELD_STAG_11].init(
		"stag_11", //i_name
		0, //i_data
		0xffffffff, //i_bit_mask
		0, //i_bit_shift
		32, //i_bit_length
		RSVD, //i_perm
		200); //i_byte_offset
	fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_12].init(
		"tagged_offset_12", //i_name
		0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		208); //i_byte_offset

	fields[QPSQ_SEND_FIELD_FRAG_12_VALID].init(
		"frag_12_valid", //i_name
		0, //i_data
		0x1, //i_bit_mask
		63, //i_bit_shift
		1, //i_bit_length
		RSVD, //i_perm
		216); //i_byte_offset
	fields[QPSQ_SEND_FIELD_FRAG_12_LENGTH].init(
		"frag_12_length", //i_name
		0, //i_data
		0x7fffffff, //i_bit_mask
		32, //i_bit_shift
		31, //i_bit_length
		RSVD, //i_perm
		216); //i_byte_offset
	fields[QPSQ_SEND_FIELD_STAG_12].init(
		"stag_12", //i_name
		0, //i_data
		0xffffffff, //i_bit_mask
		0, //i_bit_shift
		32, //i_bit_length
		RSVD, //i_perm
		216); //i_byte_offset
	fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_13].init(
		"tagged_offset_13", //i_name
		0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		224); //i_byte_offset

	fields[QPSQ_SEND_FIELD_FRAG_13_VALID].init(
		"frag_13_valid", //i_name
		0, //i_data
		0x1, //i_bit_mask
		63, //i_bit_shift
		1, //i_bit_length
		RSVD, //i_perm
		232); //i_byte_offset
	fields[QPSQ_SEND_FIELD_FRAG_13_LENGTH].init(
		"frag_13_length", //i_name
		0, //i_data
		0x7fffffff, //i_bit_mask
		32, //i_bit_shift
		31, //i_bit_length
		RSVD, //i_perm
		232); //i_byte_offset
	fields[QPSQ_SEND_FIELD_STAG_13].init(
		"stag_13", //i_name
		0, //i_data
		0xffffffff, //i_bit_mask
		0, //i_bit_shift
		32, //i_bit_length
		RSVD, //i_perm
		232); //i_byte_offset
	fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_14].init(
		"tagged_offset_14", //i_name
		0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		240); //i_byte_offset

	fields[QPSQ_SEND_FIELD_FRAG_14_VALID].init(
		"frag_14_valid", //i_name
		0, //i_data
		0x1, //i_bit_mask
		63, //i_bit_shift
		1, //i_bit_length
		RSVD, //i_perm
		248); //i_byte_offset
	fields[QPSQ_SEND_FIELD_FRAG_14_LENGTH].init(
		"frag_14_length", //i_name
		0, //i_data
		0x7fffffff, //i_bit_mask
		32, //i_bit_shift
		31, //i_bit_length
		RSVD, //i_perm
		248); //i_byte_offset
	fields[QPSQ_SEND_FIELD_STAG_14].init(
		"stag_14", //i_name
		0, //i_data
		0xffffffff, //i_bit_mask
		0, //i_bit_shift
		32, //i_bit_length
		RSVD, //i_perm
		248); //i_byte_offset

	fields[QPSQ_SEND_FIELD_NOP_0_0].init(
		"nop_0_0", //i_name
		0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		0); //i_byte_offset
	fields[QPSQ_SEND_FIELD_NOP_0_1].init(
		"nop_0_1", //i_name
		0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		8); //i_byte_offset
	fields[QPSQ_SEND_FIELD_NOP_0_2].init(
		"nop_0_2", //i_name
		0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		16); //i_byte_offset
	fields[QPSQ_SEND_FIELD_NOP_0_3].init(
		"nop_0_3", //i_name
		0xc00000000, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		24); //i_byte_offset

	fields[QPSQ_SEND_FIELD_NOP_1_0].init(
		"nop_1_0", //i_name
		0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		32); //i_byte_offset
	fields[QPSQ_SEND_FIELD_NOP_1_1].init(
		"nop_1_1", //i_name
		0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		40); //i_byte_offset
	fields[QPSQ_SEND_FIELD_NOP_1_2].init(
		"nop_1_2", //i_name
		0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		48); //i_byte_offset
	fields[QPSQ_SEND_FIELD_NOP_1_3].init(
		"nop_1_3", //i_name
		0xc00000000, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		56); //i_byte_offset

	fields[QPSQ_SEND_FIELD_NOP_2_0].init(
		"nop_2_0", //i_name
		0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		64); //i_byte_offset
	fields[QPSQ_SEND_FIELD_NOP_2_1].init(
		"nop_2_1", //i_name
		0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		72); //i_byte_offset
	fields[QPSQ_SEND_FIELD_NOP_2_2].init(
		"nop_2_2", //i_name
		0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		80); //i_byte_offset
	fields[QPSQ_SEND_FIELD_NOP_2_3].init(
		"nop_2_3", //i_name
		0xc00000000, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		88); //i_byte_offset

	fields[QPSQ_SEND_FIELD_NOP_3_0].init(
		"nop_3_0", //i_name
		0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		96); //i_byte_offset
	fields[QPSQ_SEND_FIELD_NOP_3_1].init(
		"nop_3_1", //i_name
		0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		104); //i_byte_offset
	fields[QPSQ_SEND_FIELD_NOP_3_2].init(
		"nop_3_2", //i_name
		0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		112); //i_byte_offset
	fields[QPSQ_SEND_FIELD_NOP_3_3].init(
		"nop_3_3", //i_name
		0xc00000000, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		120); //i_byte_offset

	fields[QPSQ_SEND_FIELD_NOP_4_0].init(
		"nop_4_0", //i_name
		0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		128); //i_byte_offset
	fields[QPSQ_SEND_FIELD_NOP_4_1].init(
		"nop_4_1", //i_name
		0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
	64, //i_bit_length
		RSVD, //i_perm
		136); //i_byte_offset
	fields[QPSQ_SEND_FIELD_NOP_4_2].init(
		"nop_4_2", //i_name
		0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		144); //i_byte_offset
	fields[QPSQ_SEND_FIELD_NOP_4_3].init(
		"nop_4_3", //i_name
		0xc00000000, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		152); //i_byte_offset

	fields[QPSQ_SEND_FIELD_NOP_5_0].init(
		"nop_5_0", //i_name
		0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		160); //i_byte_offset
	fields[QPSQ_SEND_FIELD_NOP_5_1].init(
		"nop_5_1", //i_name
		0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		168); //i_byte_offset
	fields[QPSQ_SEND_FIELD_NOP_5_2].init(
		"nop_5_2", //i_name
		0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		176); //i_byte_offset
	fields[QPSQ_SEND_FIELD_NOP_5_3].init(
		"nop_5_3", //i_name
		0xc00000000, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		184); //i_byte_offset

	fields[QPSQ_SEND_FIELD_NOP_6_0].init(
		"nop_6_0", //i_name
		0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		192); //i_byte_offset
	fields[QPSQ_SEND_FIELD_NOP_6_1].init(
		"nop_6_1", //i_name
		0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		200); //i_byte_offset
	fields[QPSQ_SEND_FIELD_NOP_6_2].init(
		"nop_6_2", //i_name
		0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		208); //i_byte_offset
	fields[QPSQ_SEND_FIELD_NOP_6_3].init(
		"nop_6_3", //i_name
		0xc00000000, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		216); //i_byte_offset

	fields[QPSQ_SEND_FIELD_NOP_7_0].init(
		"nop_7_0", //i_name
		0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		224); //i_byte_offset
	fields[QPSQ_SEND_FIELD_NOP_7_1].init(
		"nop_7_1", //i_name
		0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		232); //i_byte_offset
	fields[QPSQ_SEND_FIELD_NOP_7_2].init(
		"nop_7_2", //i_name
		0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		240); //i_byte_offset
	fields[QPSQ_SEND_FIELD_NOP_7_3].init(
		"nop_7_3", //i_name
		0xc00000000, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		248); //i_byte_offset

	fields[QPSQ_SEND_FIELD_INLINE_DATA_0].init(
		"inline_0", //i_name
		0x0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		8); //i_byte_offset
	fields[QPSQ_SEND_FIELD_INLINE_DATA_1].init(
		"inline_1", //i_name
		0x0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		32); //i_byte_offset
	fields[QPSQ_SEND_FIELD_INLINE_DATA_2].init(
		"inline_2", //i_name
		0x0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		40); //i_byte_offset
	fields[QPSQ_SEND_FIELD_INLINE_DATA_3].init(
		"inline_3", //i_name
		0x0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		48); //i_byte_offset
	fields[QPSQ_SEND_FIELD_INLINE_DATA_4].init(
		"inline_4", //i_name
		0x0, //i_data
		0xffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		56, //i_bit_length
		RSVD, //i_perm
		56); //i_byte_offset
	fields[QPSQ_SEND_FIELD_INLINE_DATA_VALID_4].init(
		"inline_valid_4", //i_name
		0x0, //i_data
		0x1, //i_bit_mask
		63, //i_bit_shift
		1, //i_bit_length
		RSVD, //i_perm
		56);
	fields[QPSQ_SEND_FIELD_INLINE_DATA_RSVD_4].init(
		"inline_rsvd_4", //i_name
		0x0, //i_data
		0x7f, //i_bit_mask
		56, //i_bit_shift
		7, //i_bit_length
		RSVD, //i_perm
		56);
	fields[QPSQ_SEND_FIELD_INLINE_DATA_5].init(
		"inline_5", //i_name
		0x0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		64); //i_byte_offset
	fields[QPSQ_SEND_FIELD_INLINE_DATA_6].init(
		"inline_6", //i_name
		0x0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		72); //i_byte_offset
	fields[QPSQ_SEND_FIELD_INLINE_DATA_7].init(
		"inline_7", //i_name
		0x0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		80); //i_byte_offset
	fields[QPSQ_SEND_FIELD_INLINE_DATA_8].init(
		"inline_8", //i_name
		0x0, //i_data
		0xffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		56, //i_bit_length
		RSVD, //i_perm
		88); //i_byte_offset
	fields[QPSQ_SEND_FIELD_INLINE_DATA_VALID_8].init(
		"inline_valid_8", //i_name
		0x0, //i_data
		0x1, //i_bit_mask
		63, //i_bit_shift
		1, //i_bit_length
		RSVD, //i_perm
		88);
	fields[QPSQ_SEND_FIELD_INLINE_DATA_RSVD_8].init(
		"inline_rsvd_8", //i_name
		0x0, //i_data
		0x7f, //i_bit_mask
		56, //i_bit_shift
		7, //i_bit_length
		RSVD, //i_perm
		88);
	fields[QPSQ_SEND_FIELD_INLINE_DATA_9].init(
		"inline_9", //i_name
		0x0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		96); //i_byte_offset
	fields[QPSQ_SEND_FIELD_INLINE_DATA_10].init(
		"inline_10", //i_name
		0x0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		104); //i_byte_offset
	fields[QPSQ_SEND_FIELD_INLINE_DATA_11].init(
		"inline_11", //i_name
		0x0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		112); //i_byte_offset
	fields[QPSQ_SEND_FIELD_INLINE_DATA_12].init(
		"inline_12", //i_name
		0x0, //i_data
		0xffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		56, //i_bit_length
		RSVD, //i_perm
		120); //i_byte_offset
	fields[QPSQ_SEND_FIELD_INLINE_DATA_VALID_12].init(
		"inline_valid_12", //i_name
		0x0, //i_data
		0x1, //i_bit_mask
		63, //i_bit_shift
		1, //i_bit_length
		RSVD, //i_perm
		120);
	fields[QPSQ_SEND_FIELD_INLINE_DATA_RSVD_12].init(
		"inline_rsvd_12", //i_name
		0x0, //i_data
		0x7f, //i_bit_mask
		56, //i_bit_shift
		7, //i_bit_length
		RSVD, //i_perm
		120);
	fields[QPSQ_SEND_FIELD_INLINE_DATA_13].init(
		"inline_13", //i_name
		0x0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		128); //i_byte_offset
	fields[QPSQ_SEND_FIELD_INLINE_DATA_14].init(
		"inline_14", //i_name
		0x0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		136); //i_byte_offset
	fields[QPSQ_SEND_FIELD_INLINE_DATA_15].init(
		"inline_15", //i_name
		0x0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		144); //i_byte_offset
	fields[QPSQ_SEND_FIELD_INLINE_DATA_16].init(
		"inline_16", //i_name
		0x0, //i_data
		0xffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		56, //i_bit_length
		RSVD, //i_perm
		152); //i_byte_offset
	fields[QPSQ_SEND_FIELD_INLINE_DATA_VALID_16].init(
		"inline_valid_16", //i_name
		0x0, //i_data
		0x1, //i_bit_mask
		63, //i_bit_shift
		1, //i_bit_length
		RSVD, //i_perm
		152);
	fields[QPSQ_SEND_FIELD_INLINE_DATA_RSVD_16].init(
		"inline_rsvd_16", //i_name
		0x0, //i_data
		0x7f, //i_bit_mask
		56, //i_bit_shift
		7, //i_bit_length
		RSVD, //i_perm
		152);
	fields[QPSQ_SEND_FIELD_INLINE_DATA_17].init(
		"inline_17", //i_name
		0x0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		160); //i_byte_offset
	fields[QPSQ_SEND_FIELD_INLINE_DATA_18].init(
		"inline_18", //i_name
		0x0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		168); //i_byte_offset
	fields[QPSQ_SEND_FIELD_INLINE_DATA_19].init(
		"inline_19", //i_name
		0x0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		176); //i_byte_offset
	fields[QPSQ_SEND_FIELD_INLINE_DATA_20].init(
		"inline_20", //i_name
		0x0, //i_data
		0xffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		56, //i_bit_length
		RSVD, //i_perm
		184); //i_byte_offset
	fields[QPSQ_SEND_FIELD_INLINE_DATA_VALID_20].init(
		"inline_valid_20", //i_name
		0x0, //i_data
		0x1, //i_bit_mask
		63, //i_bit_shift
		1, //i_bit_length
		RSVD, //i_perm
		184);
	fields[QPSQ_SEND_FIELD_INLINE_DATA_RSVD_20].init(
		"inline_rsvd_20", //i_name
		0x0, //i_data
		0x7f, //i_bit_mask
		56, //i_bit_shift
		7, //i_bit_length
		RSVD, //i_perm
		184);
	fields[QPSQ_SEND_FIELD_INLINE_DATA_21].init(
		"inline_21", //i_name
		0x0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		192); //i_byte_offset
	fields[QPSQ_SEND_FIELD_INLINE_DATA_22].init(
		"inline_22", //i_name
		0x0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		200); //i_byte_offset
	fields[QPSQ_SEND_FIELD_INLINE_DATA_23].init(
		"inline_23", //i_name
		0x0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		208); //i_byte_offset
	fields[QPSQ_SEND_FIELD_INLINE_DATA_24].init(
		"inline_24", //i_name
		0x0, //i_data
		0xffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		56, //i_bit_length
		RSVD, //i_perm
		216); //i_byte_offset
	fields[QPSQ_SEND_FIELD_INLINE_DATA_VALID_24].init(
		"inline_valid_24", //i_name
		0x0, //i_data
		0x1, //i_bit_mask
		63, //i_bit_shift
		1, //i_bit_length
		RSVD, //i_perm
		216);
	fields[QPSQ_SEND_FIELD_INLINE_DATA_RSVD_24].init(
		"inline_rsvd_24", //i_name
		0x0, //i_data
		0x7f, //i_bit_mask
		56, //i_bit_shift
		7, //i_bit_length
		RSVD, //i_perm
		216);
	fields[QPSQ_SEND_FIELD_INLINE_DATA_25].init(
		"inline_25", //i_name
		0x0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		224); //i_byte_offset
	fields[QPSQ_SEND_FIELD_INLINE_DATA_26].init(
		"inline_26", //i_name
		0x0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		232); //i_byte_offset
	fields[QPSQ_SEND_FIELD_INLINE_DATA_27].init(
		"inline_27", //i_name
		0x0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		240); //i_byte_offset
	fields[QPSQ_SEND_FIELD_INLINE_DATA_28].init(
		"inline_28", //i_name
		0x0, //i_data
		0xffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		56, //i_bit_length
		RSVD, //i_perm
		248); //i_byte_offset
	fields[QPSQ_SEND_FIELD_INLINE_DATA_VALID_28].init(
		"inline_valid_28", //i_name
		0x0, //i_data
		0x1, //i_bit_mask
		63, //i_bit_shift
		1, //i_bit_length
		RSVD, //i_perm
		248);
	fields[QPSQ_SEND_FIELD_INLINE_DATA_RSVD_28].init(
		"inline_rsvd_28", //i_name
		0x0, //i_data
		0x7f, //i_bit_mask
		56, //i_bit_shift
		7, //i_bit_length
		RSVD, //i_perm
		248);
	fields[QPSQ_WRITE_FIELD_RC_REMOTE_TAG_OFFSET].init(
		"rc_remote_tag_offset", //i_name
		0, //i_data
		0xffffffffffffffff, //i_bit_mask
		0, //i_bit_shift
		64, //i_bit_length
		RSVD, //i_perm
		16); //i_byte_offset
	fields[QPSQ_WRITE_FIELD_RC_REMOTE_STAG].init(
		"rc_remote_stag", //i_name
		0, //i_data
		0xffffffff, //i_bit_mask
		0, //i_bit_shift
		32, //i_bit_length
		RSVD, //i_perm
		24); //i_byte_offset
}

void QP_SEND::randomize(void)
{
	int start_idx = QPSQ_SEND_FIELD_INLINE_DATA_0;
	u64 base_array[32]; //will hold randomized data with byte gaps
	u8 *base_ptr = (u8 *) &base_array[0];

	fields[QPSQ_SEND_FIELD_WQE_VALID].data = rand() % 2;
	fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_0].data = (u64) rand() << 32 | rand();
	fields[QPSQ_SEND_FIELD_FRAG_0_VALID].data = fields[QPSQ_SEND_FIELD_WQE_VALID].data; //rand();
	fields[QPSQ_SEND_FIELD_FRAG_0_LENGTH].data = rand();
	fields[QPSQ_SEND_FIELD_STAG_0].data = rand();
	fields[QPSQ_SEND_FIELD_IW_RSVD0].data = rand();
	fields[QPSQ_SEND_FIELD_UD_DEST_QKEY].data = rand();
	fields[QPSQ_SEND_FIELD_SIGNALED_CMPL].data = rand() % 2;
	fields[QPSQ_SEND_FIELD_IW_LOCAL_FENCE].data = rand() % 2;
	fields[QPSQ_SEND_FIELD_IW_READ_FENCE].data = rand() % 2;
	fields[QPSQ_SEND_FIELD_IW_WAIT_FOR_RCVFPDU].data = rand() % 2;
	fields[QPSQ_SEND_FIELD_IW_STREAMING_MODE].data = rand() % 2;
	fields[QPSQ_SEND_FIELD_UDP_HEADER].data = rand() % 2;
	fields[QPSQ_SEND_FIELD_UD_RSVD0].data = rand();
	fields[QPSQ_SEND_FIELD_UD_RSVD1].data = rand();
	fields[QPSQ_SEND_FIELD_UD_RSVD2].data = rand();
	if (setup_config & NO_INLINE)
		fields[QPSQ_SEND_FIELD_INLINE_DATA_FLAG].data = 0;//rand() % 2;
	else
		fields[QPSQ_SEND_FIELD_INLINE_DATA_FLAG].data = rand() % 2;
	fields[QPSQ_SEND_FIELD_PUSH_WQE].data = rand() % 2;
	fields[QPSQ_SEND_FIELD_UD_RSVD3].data = rand();
	fields[QPSQ_SEND_FIELD_INLINE_DATA_LENGTH].data =  (u64) rand() << 32 | rand();
	if (setup_config & NO_IMM_DATA) {
		fields[QPSQ_SEND_FIELD_IMM_DATA_FLAG].data = 0;//rand() % 2;
		fields[QPSQ_SEND_FIELD_UD_IMM_DATA].data = 0;
	} else {
		fields[QPSQ_SEND_FIELD_IMM_DATA_FLAG].data = rand() % 2;
		fields[QPSQ_SEND_FIELD_UD_IMM_DATA].data = rand() << 32 | rand();
	}
	fields[QPSQ_SEND_FIELD_REPORT_RTT].data = rand() % 2;
	fields[QPSQ_SEND_FIELD_UD_L4LEN].data = rand();
	fields[QPSQ_SEND_FIELD_ADDITIONAL_FRAG_COUNT].data = rand() % 17;
	if (setup_config & WRITE_WR) {
		switch (rand() % 2) {
			case 0:
				fields[QPSQ_SEND_FIELD_OP].data = TB_IRDMA_OP_TYPE_WRITE;
				break;
			case 1:
				fields[QPSQ_SEND_FIELD_OP].data = TB_IRDMA_OP_TYPE_WRITE_SE;
				break;
		}
	} else {
		switch (rand() % 4) {
			case 0:
				fields[QPSQ_SEND_FIELD_OP].data = TB_IRDMA_OP_TYPE_SEND;
				break;
			case 1:
				fields[QPSQ_SEND_FIELD_OP].data = TB_IRDMA_OP_TYPE_SEND_INV;
				break;
			case 2:
				fields[QPSQ_SEND_FIELD_OP].data = TB_IRDMA_OP_TYPE_SEND_SE;
				break;
			case 3:
				fields[QPSQ_SEND_FIELD_OP].data = TB_IRDMA_OP_TYPE_SEND_SE_INV;
				break;
		}
	}
	fields[QPSQ_SEND_FIELD_UD_RSVD4].data = rand();
	fields[QPSQ_SEND_FIELD_UD_AH_ID].data = rand();
	fields[QPSQ_SEND_FIELD_IW_REMOTE_INV_STAG].data = rand();
	if (setup_config & RELIABLE_CONN) {
		switch (rand() % 2) {
			case 0:
				fields[QPSQ_SEND_QP_TYPE].data = QP_TYPE_IWARP;
				break;
			case 1:
				fields[QPSQ_SEND_QP_TYPE].data = QP_TYPE_ROCEV2_RC;
				break;
		}
	} else
		fields[QPSQ_SEND_QP_TYPE].data = rand() % 4 + 1; //Will not be CQP
		if (debug >= DEBUG_MEDIUM)
			printf("TB: rand: qp_type=0x%llx, op_type=0x%llx\n", fields[QPSQ_SEND_QP_TYPE].data, fields[QPSQ_SEND_FIELD_OP].data);
	fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_1].data = (u64) rand() << 32 | rand();
	fields[QPSQ_SEND_FIELD_FRAG_1_VALID].data = fields[QPSQ_SEND_FIELD_WQE_VALID].data; //rand();
	fields[QPSQ_SEND_FIELD_FRAG_1_LENGTH].data = rand();
	fields[QPSQ_SEND_FIELD_STAG_1].data = rand();
	fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_2].data = (u64) rand() << 32 | rand();
	fields[QPSQ_SEND_FIELD_FRAG_2_VALID].data = fields[QPSQ_SEND_FIELD_WQE_VALID].data; //rand();
	fields[QPSQ_SEND_FIELD_FRAG_2_LENGTH].data = rand();
	fields[QPSQ_SEND_FIELD_STAG_2].data = rand();
	fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_3].data = (u64) rand() << 32 | rand();
	fields[QPSQ_SEND_FIELD_FRAG_3_VALID].data = fields[QPSQ_SEND_FIELD_WQE_VALID].data; //rand();
	fields[QPSQ_SEND_FIELD_FRAG_3_LENGTH].data = rand();
	fields[QPSQ_SEND_FIELD_STAG_3].data = rand();
	fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_4].data = (u64) rand() << 32 | rand();
	fields[QPSQ_SEND_FIELD_FRAG_4_VALID].data = fields[QPSQ_SEND_FIELD_WQE_VALID].data; //rand();
	fields[QPSQ_SEND_FIELD_FRAG_4_LENGTH].data = rand();
	fields[QPSQ_SEND_FIELD_STAG_4].data = rand();
	fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_5].data = (u64) rand() << 32 | rand();
	fields[QPSQ_SEND_FIELD_FRAG_5_VALID].data = fields[QPSQ_SEND_FIELD_WQE_VALID].data; //rand();
	fields[QPSQ_SEND_FIELD_FRAG_5_LENGTH].data = rand();
	fields[QPSQ_SEND_FIELD_STAG_5].data = rand();
	fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_6].data = (u64) rand() << 32 | rand();
	fields[QPSQ_SEND_FIELD_FRAG_6_VALID].data = fields[QPSQ_SEND_FIELD_WQE_VALID].data; //rand();
	fields[QPSQ_SEND_FIELD_FRAG_6_LENGTH].data = rand();
	fields[QPSQ_SEND_FIELD_STAG_6].data = rand();
	fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_7].data = (u64) rand() << 32 | rand();
	fields[QPSQ_SEND_FIELD_FRAG_7_VALID].data = fields[QPSQ_SEND_FIELD_WQE_VALID].data; //rand();
	fields[QPSQ_SEND_FIELD_FRAG_7_LENGTH].data = rand();
	fields[QPSQ_SEND_FIELD_STAG_7].data = rand();
	fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_8].data = (u64) rand() << 32 | rand();
	fields[QPSQ_SEND_FIELD_FRAG_8_VALID].data = fields[QPSQ_SEND_FIELD_WQE_VALID].data; //rand();
	fields[QPSQ_SEND_FIELD_FRAG_8_LENGTH].data = rand();
	fields[QPSQ_SEND_FIELD_STAG_8].data = rand();
	fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_9].data = (u64) rand() << 32 | rand();
	fields[QPSQ_SEND_FIELD_FRAG_9_VALID].data = fields[QPSQ_SEND_FIELD_WQE_VALID].data; //rand();
	fields[QPSQ_SEND_FIELD_FRAG_9_LENGTH].data = rand();
	fields[QPSQ_SEND_FIELD_STAG_9].data = rand();
	fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_10].data = (u64) rand() << 32 | rand();
	fields[QPSQ_SEND_FIELD_FRAG_10_VALID].data = fields[QPSQ_SEND_FIELD_WQE_VALID].data; //rand();
	fields[QPSQ_SEND_FIELD_FRAG_10_LENGTH].data = rand();
	fields[QPSQ_SEND_FIELD_STAG_10].data = rand();
	fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_11].data = (u64) rand() << 32 | rand();
	fields[QPSQ_SEND_FIELD_FRAG_11_VALID].data = fields[QPSQ_SEND_FIELD_WQE_VALID].data; //rand();
	fields[QPSQ_SEND_FIELD_FRAG_11_LENGTH].data = rand();
	fields[QPSQ_SEND_FIELD_STAG_11].data = rand();
	fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_12].data = (u64) rand() << 32 | rand();
	fields[QPSQ_SEND_FIELD_FRAG_12_VALID].data = fields[QPSQ_SEND_FIELD_WQE_VALID].data; //rand();
	fields[QPSQ_SEND_FIELD_FRAG_12_LENGTH].data = rand();
	fields[QPSQ_SEND_FIELD_STAG_12].data = rand();
	fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_13].data = (u64) rand() << 32 | rand();
	fields[QPSQ_SEND_FIELD_FRAG_13_VALID].data = fields[QPSQ_SEND_FIELD_WQE_VALID].data; //rand();
	fields[QPSQ_SEND_FIELD_FRAG_13_LENGTH].data = rand();
	fields[QPSQ_SEND_FIELD_STAG_13].data = rand();
	fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_14].data = (u64) rand() << 32 | rand();
	fields[QPSQ_SEND_FIELD_FRAG_14_VALID].data = fields[QPSQ_SEND_FIELD_WQE_VALID].data; //rand();
	fields[QPSQ_SEND_FIELD_FRAG_14_LENGTH].data = rand();
	fields[QPSQ_SEND_FIELD_STAG_14].data = rand();
	fields[QPSQ_SEND_FIELD_INLINE_DATA_0].data = (u64) rand() << 32 | rand();
	fields[QPSQ_SEND_FIELD_INLINE_DATA_1].data = (u64) rand() << 32 | rand();
	fields[QPSQ_SEND_FIELD_INLINE_DATA_2].data = (u64) rand() << 32 | rand();
	fields[QPSQ_SEND_FIELD_INLINE_DATA_3].data = (u64) rand() << 32 | rand();
	fields[QPSQ_SEND_FIELD_INLINE_DATA_4].data = (u64) rand() << 32 | rand();
	fields[QPSQ_SEND_FIELD_INLINE_DATA_5].data = (u64) rand() << 32 | rand();
	fields[QPSQ_SEND_FIELD_INLINE_DATA_6].data = (u64) rand() << 32 | rand();
	fields[QPSQ_SEND_FIELD_INLINE_DATA_7].data = (u64) rand() << 32 | rand();
	fields[QPSQ_SEND_FIELD_INLINE_DATA_8].data = (u64) rand() << 32 | rand();
	fields[QPSQ_SEND_FIELD_INLINE_DATA_9].data = (u64) rand() << 32 | rand();
	fields[QPSQ_SEND_FIELD_INLINE_DATA_10].data = (u64) rand() << 32 | rand();
	fields[QPSQ_SEND_FIELD_INLINE_DATA_11].data = (u64) rand() << 32 | rand();
	fields[QPSQ_SEND_FIELD_INLINE_DATA_12].data = (u64) rand() << 32 | rand();
	fields[QPSQ_SEND_FIELD_INLINE_DATA_13].data = (u64) rand() << 32 | rand();
	fields[QPSQ_SEND_FIELD_INLINE_DATA_14].data = (u64) rand() << 32 | rand();
	fields[QPSQ_SEND_FIELD_INLINE_DATA_15].data = (u64) rand() << 32 | rand();
	fields[QPSQ_SEND_FIELD_INLINE_DATA_16].data = (u64) rand() << 32 | rand();
	fields[QPSQ_SEND_FIELD_INLINE_DATA_17].data = (u64) rand() << 32 | rand();
	fields[QPSQ_SEND_FIELD_INLINE_DATA_18].data = (u64) rand() << 32 | rand();
	fields[QPSQ_SEND_FIELD_INLINE_DATA_19].data = (u64) rand() << 32 | rand();
	fields[QPSQ_SEND_FIELD_INLINE_DATA_20].data = (u64) rand() << 32 | rand();
	fields[QPSQ_SEND_FIELD_INLINE_DATA_21].data = (u64) rand() << 32 | rand();
	fields[QPSQ_SEND_FIELD_INLINE_DATA_22].data = (u64) rand() << 32 | rand();
	fields[QPSQ_SEND_FIELD_INLINE_DATA_23].data = (u64) rand() << 32 | rand();
	fields[QPSQ_SEND_FIELD_INLINE_DATA_24].data = (u64) rand() << 32 | rand();
	fields[QPSQ_SEND_FIELD_INLINE_DATA_25].data = (u64) rand() << 32 | rand();
	fields[QPSQ_SEND_FIELD_INLINE_DATA_26].data = (u64) rand() << 32 | rand();
	fields[QPSQ_SEND_FIELD_INLINE_DATA_27].data = (u64) rand() << 32 | rand();
	fields[QPSQ_SEND_FIELD_INLINE_DATA_28].data = (u64) rand() << 32 | rand();
	//this only works if SQ doesn't roll over!!!!
	if (debug >= DEBUG_NONE)
		printf("TB: doesn't work if SQ rolls over!!!\n");
	fields[QPSQ_SEND_FIELD_INLINE_DATA_VALID_4].data = fields[QPSQ_SEND_FIELD_WQE_VALID].data;
	fields[QPSQ_SEND_FIELD_INLINE_DATA_VALID_8].data = fields[QPSQ_SEND_FIELD_WQE_VALID].data;
	fields[QPSQ_SEND_FIELD_INLINE_DATA_VALID_12].data = fields[QPSQ_SEND_FIELD_WQE_VALID].data;
	fields[QPSQ_SEND_FIELD_INLINE_DATA_VALID_16].data = fields[QPSQ_SEND_FIELD_WQE_VALID].data;
	fields[QPSQ_SEND_FIELD_INLINE_DATA_VALID_20].data = fields[QPSQ_SEND_FIELD_WQE_VALID].data;
	fields[QPSQ_SEND_FIELD_INLINE_DATA_VALID_24].data = fields[QPSQ_SEND_FIELD_WQE_VALID].data;
	fields[QPSQ_SEND_FIELD_INLINE_DATA_VALID_28].data = fields[QPSQ_SEND_FIELD_WQE_VALID].data;

	fields[QPSQ_WRITE_FIELD_RC_REMOTE_TAG_OFFSET].data = (u64) rand() << 32 | rand();
	fields[QPSQ_WRITE_FIELD_RC_REMOTE_STAG].data = rand();

	if (debug >= DEBUG_HIGH) {
		printf("TB: inline_data_0(0x%llx)\n", fields[QPSQ_SEND_FIELD_INLINE_DATA_0].get_masked_data_64());
		printf("TB: inline_data_1(0x%llx)\n", fields[QPSQ_SEND_FIELD_INLINE_DATA_1].get_masked_data_64());
		printf("TB: inline_data_2(0x%llx)\n", fields[QPSQ_SEND_FIELD_INLINE_DATA_2].get_masked_data_64());
		printf("TB: inline_data_3(0x%llx)\n", fields[QPSQ_SEND_FIELD_INLINE_DATA_3].get_masked_data_64());
		printf("TB: inline_data_4(0x%llx)\n", fields[QPSQ_SEND_FIELD_INLINE_DATA_4].get_masked_data_64());
		printf("TB: inline_data_5(0x%llx)\n", fields[QPSQ_SEND_FIELD_INLINE_DATA_5].get_masked_data_64());
		printf("TB: inline_data_6(0x%llx)\n", fields[QPSQ_SEND_FIELD_INLINE_DATA_6].get_masked_data_64());
		printf("TB: inline_data_7(0x%llx)\n", fields[QPSQ_SEND_FIELD_INLINE_DATA_7].get_masked_data_64());
		printf("TB: inline_data_8(0x%llx)\n", fields[QPSQ_SEND_FIELD_INLINE_DATA_8].get_masked_data_64());
		printf("TB: inline_data_9(0x%llx)\n", fields[QPSQ_SEND_FIELD_INLINE_DATA_9].get_masked_data_64());
		printf("TB: inline_data_10(0x%llx)\n", fields[QPSQ_SEND_FIELD_INLINE_DATA_10].get_masked_data_64());
		printf("TB: inline_data_11(0x%llx)\n", fields[QPSQ_SEND_FIELD_INLINE_DATA_11].get_masked_data_64());
		printf("TB: inline_data_12(0x%llx)\n", fields[QPSQ_SEND_FIELD_INLINE_DATA_12].get_masked_data_64());
		printf("TB: inline_data_13(0x%llx)\n", fields[QPSQ_SEND_FIELD_INLINE_DATA_13].get_masked_data_64());
		printf("TB: inline_data_14(0x%llx)\n", fields[QPSQ_SEND_FIELD_INLINE_DATA_14].get_masked_data_64());
		printf("TB: inline_data_15(0x%llx)\n", fields[QPSQ_SEND_FIELD_INLINE_DATA_15].get_masked_data_64());
		printf("TB: inline_data_16(0x%llx)\n", fields[QPSQ_SEND_FIELD_INLINE_DATA_16].get_masked_data_64());
		printf("TB: inline_data_17(0x%llx)\n", fields[QPSQ_SEND_FIELD_INLINE_DATA_17].get_masked_data_64());
	}

	if (fields[QPSQ_SEND_FIELD_INLINE_DATA_LENGTH].get_masked_data_64() <= 224) {

		//Fill base_array with INLINE_DATA_* (will have "GAP" Bytes)
		for (int qp_enum=QPSQ_SEND_FIELD_INLINE_DATA_0, base_idx=0; base_idx<29; base_idx++, qp_enum++) {
			if (base_idx>=5 && (base_idx-5)%4==0) {
				qp_enum += 2; //skip RSVD & Valid Fields
			}
			base_array[base_idx] = fields[qp_enum].get_masked_data_64();
		}

		//Fill inline_data with base_array data, inline_data will not have "Gap" Bytes
		for (int n=0, g=0; n<225; n++, g++) {
			if (debug >= DEBUG_HIGH)
				printf("TB: n(%d) g(%d), base_ptr[g](0x%x)\n", n, g, base_ptr[g]);
			if (((g-39)%32)==0 && g>=39)
				g++;
			if (n >= fields[QPSQ_SEND_FIELD_INLINE_DATA_LENGTH].get_masked_data_64())
				base_ptr[g] = 0;
			if (debug >= DEBUG_HIGH)
				printf("TB: n(%d) g(%d), base_ptr[g](0x%x)\n", n, g, base_ptr[g]);
			inline_data[n] = base_ptr[g];
		}

		//re-init inline data values to be based on length
		for (int qp_enum=QPSQ_SEND_FIELD_INLINE_DATA_0, base_idx=0; base_idx<29; base_idx++, qp_enum++) {
			if (base_idx>=5 && (base_idx-5)%4==0) {
				qp_enum += 2; //skip RSVD & Valid Fields
			}
			fields[qp_enum].data = base_array[base_idx];
		}

		if (debug >= DEBUG_HIGH)
			for (int k=0; k<200; k+=8) {
				u64 *tmp = (u64 *) &inline_data[k];
				printf("TB: inline_data[%d](0x%llx) @ %p\n", k, *tmp, &inline_data[k]);
			}
	}
}

u64 QP_SEND::concat_sge_fields(Field *sge_field_array)
{
	u64 ret = sge_field_array[0].get_masked_data_64() << sge_field_array[0].bit_shift;
	ret |= sge_field_array[1].get_masked_data_64() << sge_field_array[1].bit_shift;
	ret |= sge_field_array[2].get_masked_data_64() << sge_field_array[2].bit_shift;
	if (debug >= DEBUG_HIGH)
		printf("HGO: setting sge to (0x%llx)\n",ret);
	return ret;
}

void QP_SEND::fill_rdma_send_intf_mixed(struct irdma_post_sq_info *info)
{
	if (debug >= DEBUG_MEDIUM) {
		printf("HGO: inside fill_...intf mixed\n");
		printf("TB: qp type: 0x%llx\n", fields[QPSQ_SEND_QP_TYPE].data);
	}
	if (!(setup_config & NO_IMM_DATA)) {
		info->imm_data       = fields[QPSQ_SEND_FIELD_UD_IMM_DATA].data;
		info->imm_data_valid = fields[QPSQ_SEND_FIELD_IMM_DATA_FLAG].data;
	}
	info->l4len          = fields[QPSQ_SEND_FIELD_UD_L4LEN].data;
	info->read_fence     = fields[QPSQ_SEND_FIELD_IW_READ_FENCE].data;
	info->local_fence    = fields[QPSQ_SEND_FIELD_IW_LOCAL_FENCE].data;
	info->report_rtt     = fields[QPSQ_SEND_FIELD_REPORT_RTT].data;
	info->udp_hdr        = fields[QPSQ_SEND_FIELD_UDP_HEADER].data;
	if (fields[QPSQ_SEND_FIELD_OP].data == TB_IRDMA_OP_TYPE_SEND_INV ||
	    fields[QPSQ_SEND_FIELD_OP].data == TB_IRDMA_OP_TYPE_SEND_SE_INV)
		info->stag_to_inv    = fields[QPSQ_SEND_FIELD_IW_REMOTE_INV_STAG].data;

	info->op.send.qkey      = fields[QPSQ_SEND_FIELD_UD_DEST_QKEY].data;
	info->op.send.dest_qp   = fields[QPSQ_SEND_FIELD_UD_DEST_QPN].data;
	info->op.send.ah_id     = fields[QPSQ_SEND_FIELD_UD_AH_ID].data;
}

void QP_SEND::fill_rdma_send_intf_mixed_inline(struct irdma_post_sq_info *info)
{
	if (debug >= DEBUG_MEDIUM) {
		printf("HGO: inside fill_...intf inline\n");
		printf("TB: qp type: 0x%llx\n", fields[QPSQ_SEND_QP_TYPE].data);
	}
	if (!(setup_config & NO_IMM_DATA)) {
		info->imm_data       = fields[QPSQ_SEND_FIELD_UD_IMM_DATA].data;
		info->imm_data_valid = fields[QPSQ_SEND_FIELD_IMM_DATA_FLAG].data;
	}
	info->l4len          = fields[QPSQ_SEND_FIELD_UD_L4LEN].data;
	info->read_fence     = fields[QPSQ_SEND_FIELD_IW_READ_FENCE].data;
	info->local_fence    = fields[QPSQ_SEND_FIELD_IW_LOCAL_FENCE].data;
	info->report_rtt     = fields[QPSQ_SEND_FIELD_REPORT_RTT].data;
	info->udp_hdr        = fields[QPSQ_SEND_FIELD_UDP_HEADER].data;
	if (fields[QPSQ_SEND_FIELD_OP].data == TB_IRDMA_OP_TYPE_SEND_INV ||
	    fields[QPSQ_SEND_FIELD_OP].data == TB_IRDMA_OP_TYPE_SEND_SE_INV)
		info->stag_to_inv    = fields[QPSQ_SEND_FIELD_IW_REMOTE_INV_STAG].data;

	info->op.inline_send.qkey      = fields[QPSQ_SEND_FIELD_UD_DEST_QKEY].data;
	info->op.inline_send.dest_qp   = fields[QPSQ_SEND_FIELD_UD_DEST_QPN].data;
	info->op.inline_send.ah_id     = fields[QPSQ_SEND_FIELD_UD_AH_ID].data;
}

void QP_SEND::fill_rdma_send_intf_mutex(struct irdma_post_sq_info *info)
{
	if (debug >= DEBUG_MEDIUM)
		printf("TB: qp type: 0x%llx\n", fields[QPSQ_SEND_QP_TYPE].data);
	if (fields[QPSQ_SEND_QP_TYPE].data == QP_TYPE_ROCEV2_UD ||
	    fields[QPSQ_SEND_QP_TYPE].data == QP_TYPE_UDA) { //ROCE
		if (!(setup_config & NO_IMM_DATA)) {
			info->imm_data       = fields[QPSQ_SEND_FIELD_UD_IMM_DATA].data;
			info->imm_data_valid = fields[QPSQ_SEND_FIELD_IMM_DATA_FLAG].data;
		}
		info->l4len             = fields[QPSQ_SEND_FIELD_UD_L4LEN].data;
		info->op.send.qkey      = fields[QPSQ_SEND_FIELD_UD_DEST_QKEY].data;
		info->op.send.dest_qp   = fields[QPSQ_SEND_FIELD_UD_DEST_QPN].data;
		info->op.send.ah_id     = fields[QPSQ_SEND_FIELD_UD_AH_ID].data;
		if (fields[QPSQ_SEND_QP_TYPE].data == QP_TYPE_UDA)
			info->udp_hdr        = fields[QPSQ_SEND_FIELD_UDP_HEADER].data;

	} else {
		info->report_rtt     = fields[QPSQ_SEND_FIELD_REPORT_RTT].data;
		info->read_fence     = fields[QPSQ_SEND_FIELD_IW_READ_FENCE].data;
		info->local_fence    = fields[QPSQ_SEND_FIELD_IW_LOCAL_FENCE].data;
		if (fields[QPSQ_SEND_FIELD_OP].data == TB_IRDMA_OP_TYPE_SEND_INV ||
		    fields[QPSQ_SEND_FIELD_OP].data == TB_IRDMA_OP_TYPE_SEND_SE_INV)
			info->stag_to_inv    = fields[QPSQ_SEND_FIELD_IW_REMOTE_INV_STAG].data;

	}

}

void QP_SEND::fill_rdma_send_intf_mutex_inline(struct irdma_post_sq_info *info)
{
	if (debug >= DEBUG_MEDIUM)
		printf("TB: qp type: 0x%llx\n", fields[QPSQ_SEND_QP_TYPE].data);
	if (fields[QPSQ_SEND_QP_TYPE].data == QP_TYPE_ROCEV2_UD ||
	    fields[QPSQ_SEND_QP_TYPE].data == QP_TYPE_UDA) { //ROCE
		if (!(setup_config & NO_IMM_DATA)) {
			info->imm_data       = fields[QPSQ_SEND_FIELD_UD_IMM_DATA].data;
			info->imm_data_valid = fields[QPSQ_SEND_FIELD_IMM_DATA_FLAG].data;
		}
		info->l4len          = fields[QPSQ_SEND_FIELD_UD_L4LEN].data;
		info->op.inline_send.qkey      = fields[QPSQ_SEND_FIELD_UD_DEST_QKEY].data;
		info->op.inline_send.dest_qp   = fields[QPSQ_SEND_FIELD_UD_DEST_QPN].data;
		info->op.inline_send.ah_id     = fields[QPSQ_SEND_FIELD_UD_AH_ID].data;
		if (fields[QPSQ_SEND_QP_TYPE].data == QP_TYPE_UDA)
			info->udp_hdr        = fields[QPSQ_SEND_FIELD_UDP_HEADER].data;

	} else { // if (fields[QPSQ_SEND_QP_TYPE].data == ) { //other
		info->report_rtt     = fields[QPSQ_SEND_FIELD_REPORT_RTT].data;
		info->read_fence     = fields[QPSQ_SEND_FIELD_IW_READ_FENCE].data;
		info->local_fence    = fields[QPSQ_SEND_FIELD_IW_LOCAL_FENCE].data;
		if (fields[QPSQ_SEND_FIELD_OP].data == TB_IRDMA_OP_TYPE_SEND_INV ||
		    fields[QPSQ_SEND_FIELD_OP].data == TB_IRDMA_OP_TYPE_SEND_SE_INV)
			info->stag_to_inv    = fields[QPSQ_SEND_FIELD_IW_REMOTE_INV_STAG].data;
	}

}

void QP_SEND::fill_rdma_write_intf_mixed(struct irdma_post_sq_info *info)
{
	if (fields[QPSQ_SEND_QP_TYPE].data == QP_TYPE_ROCEV2_RC ||
	    fields[QPSQ_SEND_QP_TYPE].data == QP_TYPE_IWARP) {
		if (debug >= DEBUG_HIGH)
			printf("HGO: inside fill_...intf\n");
		if (!(setup_config & NO_IMM_DATA)) {
			info->imm_data       = fields[QPSQ_SEND_FIELD_UD_IMM_DATA].data;
			info->imm_data_valid = fields[QPSQ_SEND_FIELD_IMM_DATA_FLAG].data;
		}
		info->l4len          = fields[QPSQ_SEND_FIELD_UD_L4LEN].data;
		info->read_fence     = fields[QPSQ_SEND_FIELD_IW_READ_FENCE].data;
		info->local_fence    = fields[QPSQ_SEND_FIELD_IW_LOCAL_FENCE].data;
		info->report_rtt     = fields[QPSQ_SEND_FIELD_REPORT_RTT].data;
	}
}

void QP_SEND::fill_rdma_write_intf_mixed_inline(struct irdma_post_sq_info *info)
{
	if (fields[QPSQ_SEND_QP_TYPE].data == QP_TYPE_ROCEV2_RC ||
	    fields[QPSQ_SEND_QP_TYPE].data == QP_TYPE_IWARP) {
		if (debug >= DEBUG_HIGH)
			printf("HGO: inside fill_...intf\n");
		if (!(setup_config & NO_IMM_DATA)) {
			info->imm_data       = fields[QPSQ_SEND_FIELD_UD_IMM_DATA].data;
			info->imm_data_valid = fields[QPSQ_SEND_FIELD_IMM_DATA_FLAG].data;
		}
		info->l4len          = fields[QPSQ_SEND_FIELD_UD_L4LEN].data;
		info->read_fence     = fields[QPSQ_SEND_FIELD_IW_READ_FENCE].data;
		info->local_fence    = fields[QPSQ_SEND_FIELD_IW_LOCAL_FENCE].data;
		info->report_rtt     = fields[QPSQ_SEND_FIELD_REPORT_RTT].data;
	}
}

void QP_SEND::fill_rdma_write_intf_mutex(struct irdma_post_sq_info *info)
{
	//should only be for roce_rc and iwarp
	if (fields[QPSQ_SEND_QP_TYPE].data == QP_TYPE_ROCEV2_RC ||
	    fields[QPSQ_SEND_QP_TYPE].data == QP_TYPE_IWARP) {
		info->report_rtt     = fields[QPSQ_SEND_FIELD_REPORT_RTT].data;
		info->read_fence     = fields[QPSQ_SEND_FIELD_IW_READ_FENCE].data;
		info->local_fence    = fields[QPSQ_SEND_FIELD_IW_LOCAL_FENCE].data;
	}
}

void QP_SEND::fill_rdma_write_intf_mutex_inline(struct irdma_post_sq_info *info)
{
	//should only be for roce_rc and iwarp
	if (fields[QPSQ_SEND_QP_TYPE].data == QP_TYPE_ROCEV2_RC ||
	    fields[QPSQ_SEND_QP_TYPE].data == QP_TYPE_IWARP) {
		info->report_rtt     = fields[QPSQ_SEND_FIELD_REPORT_RTT].data;
		info->read_fence     = fields[QPSQ_SEND_FIELD_IW_READ_FENCE].data;
		info->local_fence    = fields[QPSQ_SEND_FIELD_IW_LOCAL_FENCE].data;
	}
}

void QP_SEND::fill_nops(struct irdma_sc_qp *sc_qp, u16 wqe_size_bytes, u8 polarity)
{
	u16 bytes_left_in_block = 0;
	u16 num_nops=0;
	enum qp_sq_send qp_send_enum = QPSQ_SEND_FIELD_NOP_0_0;

	bytes_left_in_block = TB_MAX_WQE_SIZE - (u16) (((u64)&sc_qp->qp_uk.sq_base[sc_qp->qp_uk.sq_ring.head]) & 0xffULL);
//	avail_bytes = (u16)sc_qp->qp_uk.sq_ring.size - ( (&sc_qp->qp_uk.sq_base[sc_qp->qp_uk.sq_ring.head]) + sc_qp->qp_uk.sq_ring.size - (&sc_qp->qp_uk.sq_base[sc_qp->qp_uk.sq_ring.tail])) % sc_qp->qp_uk.sq_ring.size;

	avail_bytes = (u16)sc_qp->qp_uk.sq_ring.size + sc_qp->qp_uk.sq_ring.head - sc_qp->qp_uk.sq_ring.tail % sc_qp->qp_uk.sq_ring.size;

	avail_bytes *= (8*min_wqe_words());
	avail_bytes -= 8*min_wqe_words(); // honor space between head and tail
	if (debug >= DEBUG_HIGH)
		printf("TB: %s: avail_bytes(%d)\n", __func__, avail_bytes);
	if (wqe_size_bytes > bytes_left_in_block)
		num_nops = bytes_left_in_block / (8 * min_wqe_words());

	if (avail_bytes < (num_nops*8*min_wqe_words() + wqe_size_bytes)) {
		wqe_will_not_fit = true;
	} else if (num_nops > 0) {

		//adjust initial offset to account for NOPs
		for (int i=0; i < qp_send_enum; i++) {
			fields[i].byte_offset += num_nops*8*min_wqe_words();
		}
		for (int i=QPSQ_SEND_FIELD_INLINE_DATA_0; i<QPSQ_SEND_FIELD_INLINE_DATA_RSVD_28;
		     i++)
			fields[i].byte_offset += num_nops*8*min_wqe_words();

		//enable NOPs to be seen in SQ
		for (int j=0; j < num_nops*4; j++) {
			fields[qp_send_enum + j].perm = READWRITE;
			if (j % 4 == 3)
				fields[qp_send_enum + j].data |= fields[QPSQ_SEND_FIELD_WQE_VALID].get_masked_shifted_data_64();
			if (debug >= DEBUG_HIGH) {
				std::cout << "TB: enabling " << fields[j].name;
				printf(", idx(%d) to READWRITE\n", j);
			}
		}
	}

}

void QP_SEND::fill_nops(struct irdma_ring *sqr, struct irdma_sc_qp *sc_qp, u16 wqe_size_bytes, u8 polarity)
{
	u16 bytes_left_in_block = 0;
	u16 num_nops=0;
	enum qp_sq_send qp_send_enum = QPSQ_SEND_FIELD_NOP_0_0;
	u32 q_nops = 0;
	u32 avail_q = 0;
	u32 wqe_q = 0;

	tb_debug(DEBUG_HIGH, "overloaded function: has irdma_ring as arg\n");

	//avail_bytes still seems to be needed by something else
	avail_bytes = (u16)sqr->size + sqr->head - sqr->tail % sqr->size;

	avail_bytes *= (8*min_wqe_words());
	avail_bytes -= 8*min_wqe_words(); // honor space between head and tail
	if (debug >= DEBUG_HIGH)
		printf("TB: %s: avail_bytes(%d)\n", __func__, avail_bytes);
	if (wqe_size_bytes > bytes_left_in_block)
		num_nops = bytes_left_in_block / (8 * min_wqe_words());

	q_nops = TB_MAX_HW_SQ_CHUNK_CPK - sqr->head % TB_MAX_HW_SQ_CHUNK_CPK;
	tb_debug(DEBUG_HIGH, "space left in chunk(%d)\n", q_nops);
	wqe_q = wqe_size_bytes / (8*min_wqe_words());
	if (wqe_q <= q_nops)
		q_nops = 0;
	tb_debug(DEBUG_HIGH, "sqr.size(%d), sqr.head(%d), sqr.tail(%d)\n",
		 sqr->size, sqr->head, sqr->tail);
	avail_q = (sqr->size + sqr->tail - sqr->head - 1) % sqr->size;
	if (debug >= DEBUG_NONE)
		printf("TB: %s: q_nops(%d), wqe_q(%d), wqe_size(%d), avail_q(%d)\n",
		       __func__, q_nops, wqe_q, wqe_size_bytes, avail_q);

	if (avail_q < (q_nops + wqe_q))
		wqe_will_not_fit = true;
	if (q_nops > 0) {
		//adjust initial offset to account for NOPs
		for (int i=0; i < qp_send_enum; i++) {
			fields[i].byte_offset += q_nops*8*min_wqe_words();
		}
		for (int i=QPSQ_SEND_FIELD_INLINE_DATA_0; i<QPSQ_SEND_FIELD_INLINE_DATA_RSVD_28;
		     i++)
			fields[i].byte_offset += q_nops*8*min_wqe_words();

		//enable NOPs to be seen in SQ
		for (int j=0; j < q_nops*4; j++) {
			fields[qp_send_enum + j].perm = READWRITE;
			if (j % 4 == 3)
				fields[qp_send_enum + j].data |= fields[QPSQ_SEND_FIELD_WQE_VALID].get_masked_shifted_data_64();
			if (debug >= DEBUG_HIGH) {
				std::cout << "TB: enabling " << fields[j].name;
				printf(", idx(%d) to READWRITE\n", j);
			}
		}
	}
}

void QP_SEND::enable_inline_fields(u16 inline_wqes)
{
	fields[QPSQ_SEND_FIELD_INLINE_DATA_0].perm = READWRITE;
	if (debug >= DEBUG_HIGH)
		printf("TB: INLINE_DATA_0 enum val (%d), inline_wqes(%d)\n", QPSQ_SEND_FIELD_INLINE_DATA_0, inline_wqes);
	for (int i=QPSQ_SEND_FIELD_INLINE_DATA_1; i<(QPSQ_SEND_FIELD_INLINE_DATA_1+inline_wqes*6); i++) {
		fields[i].perm = READWRITE;
		if (debug >= DEBUG_HIGH)
			printf("TB: enabling idx(%d) to READWRITE\n", i);
	}
	fields[QPSQ_SEND_FIELD_INLINE_DATA_LENGTH].perm = READWRITE;
}

void QP_SEND::fill_rdma_send_intf(struct irdma_sc_qp *sc_qp, struct irdma_post_sq_info *info,
			    struct irdma_sge *sges)
{
	enum qp_sq_send qp_send_enum = QPSQ_SEND_FIELD_FRAG_1_VALID;
	struct irdma_sge *sge_ptr = (struct irdma_sge *) sges;
	u16 wqe_size_bytes = 32;
	int tmp = 0;

	if (debug >= DEBUG_LOW) {
		if (!sc_qp)
			tmp |= 1;
		if (!sc_qp->qp_uk.sq_base)
			tmp |= 2;
	}

	sq_tail_addr[0] = sc_qp->qp_uk.sq_base[sc_qp->qp_uk.sq_ring.tail].elem[0];
	sq_tail_addr[1] = sc_qp->qp_uk.sq_base[sc_qp->qp_uk.sq_ring.tail].elem[1];
	sq_tail_addr[2] = sc_qp->qp_uk.sq_base[sc_qp->qp_uk.sq_ring.tail].elem[2];
	sq_tail_addr[3] = sc_qp->qp_uk.sq_base[sc_qp->qp_uk.sq_ring.tail].elem[3];

	if (sc_qp->qp_uk.sq_ring.head == 0)
		sc_qp->qp_uk.swqe_polarity = !fields[QPSQ_SEND_FIELD_WQE_VALID].data;
	else
		sc_qp->qp_uk.swqe_polarity = fields[QPSQ_SEND_FIELD_WQE_VALID].data;
	sc_qp->qp_type = fields[QPSQ_SEND_QP_TYPE].data;
	info->op_type        = fields[QPSQ_SEND_FIELD_OP].data;
	info->signaled       = fields[QPSQ_SEND_FIELD_SIGNALED_CMPL].data;
	info->inline_data    = fields[QPSQ_SEND_FIELD_INLINE_DATA_FLAG].data;
	// this one is slightly different. The code checks to see if there is an
	// address associated with qp_uk.push_db, if there is
	// it always sets the push_wqe field in the WQE
	// otherwise, it leaves that field clear
	if (sc_qp->qp_uk.push_db)
		fields[QPSQ_SEND_FIELD_PUSH_WQE].data = 1ULL;
	else
		fields[QPSQ_SEND_FIELD_PUSH_WQE].data = 0ULL;

	if(fields[QPSQ_SEND_FIELD_INLINE_DATA_FLAG].data) {
		u16 wqes_for_inline_data = 0;
		u16 bytes_for_inline_data = 0;
		//set inline data pointer
		info->op.inline_send.data = &inline_data[0];
		info->op.inline_send.len = fields[QPSQ_SEND_FIELD_INLINE_DATA_LENGTH].get_masked_data_64();

		if (fields[QPSQ_SEND_FIELD_INLINE_DATA_LENGTH].get_masked_data_64() <= 224) {
			int valid_bytes = 0;
			if (config == MIXED_QP_TYPES) {
				fill_rdma_send_intf_mixed_inline(info);
			} else if (config == MUTEX_QP_TYPES) {
				fill_rdma_send_intf_mutex_inline(info);
			}

			for(int vb=info->op.inline_send.len-8; vb > 0; vb-=32, valid_bytes++)
				if (vb-32 >= 0)
					vb+=1;

			wqes_for_inline_data = (info->op.inline_send.len - 8 + valid_bytes)/(8*min_wqe_words());

			if ((info->op.inline_send.len - 8 + valid_bytes) % (8*min_wqe_words()))
				wqes_for_inline_data++;

			bytes_for_inline_data = 32*wqes_for_inline_data;
			wqe_size_bytes += bytes_for_inline_data;

			enable_inline_fields(wqes_for_inline_data);
			fill_nops(sc_qp, wqe_size_bytes, sc_qp->qp_uk.swqe_polarity);
		}

	} else {
		//bool set_next_valid = false; //needed if last frag only takes up half of min_wqe_size
		int valid_idx = 0;

		if (config == MIXED_QP_TYPES) {
			fill_rdma_send_intf_mixed(info);
		} else if (config == MUTEX_QP_TYPES) {
			fill_rdma_send_intf_mutex(info);
		}

		info->op.send.num_sges  = fields[QPSQ_SEND_FIELD_ADDITIONAL_FRAG_COUNT].get_masked_data_64() + 1;

		if (info->op.send.num_sges < 0)
			info->op.send.num_sges = 0;

		wqe_size_bytes += (info->op.send.num_sges-1)*16;
		if ((info->op.send.num_sges-1) % 2) //ensure on 32 byte boundary
			wqe_size_bytes += 16;

		fill_nops(sc_qp, wqe_size_bytes, sc_qp->qp_uk.swqe_polarity);

		for (int i=0, j=0; j < info->op.send.num_sges && info->op.send.num_sges <= 14; i+=2, j++) {

			sge_ptr[j].tag_off = (irdma_tagged_offset) fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_0+j*4].data;

			sge_ptr[j].len = (u32) fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_0+j*4+2].get_masked_data_64();
			sge_ptr[j].stag = (irdma_stag) fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_0+j*4+3].get_masked_data_64();
			if ((((fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_0+j*4].byte_offset) % 32)==0) && j) {
				set_next_valid = true;
				valid_idx = QPSQ_SEND_FIELD_TAGGED_OFFSET_0+(j+1)*4+1;
			} else
				set_next_valid = false;
		}

		if (set_next_valid) {
			if (debug >= DEBUG_HIGH) {
				printf("TB: set_next_valid true idx(%d)\n", valid_idx);
				std::cout << "\t" << fields[valid_idx].name << ", " <<
					  fields[valid_idx].data << "\n";
			}
			fields[valid_idx-1].perm = RSVD;
			fields[valid_idx].perm = READWRITE;
			fields[valid_idx+1].perm = RSVD;
			fields[valid_idx+2].perm = RSVD;
		} else {
			if (debug >= DEBUG_HIGH)
				printf("TB: set_next_valid false\n");
		}

		info->op.send.sg_list   = (struct irdma_sge *) &sges[0];
	}
	u16 add_sges = (u16) fields[QPSQ_SEND_FIELD_ADDITIONAL_FRAG_COUNT].get_masked_data_64();
	enable_frags(add_sges);

}

void QP_SEND::enable_write_fields(void)
{
	fields[QPSQ_WRITE_FIELD_RC_REMOTE_TAG_OFFSET].perm = READWRITE;
	fields[QPSQ_WRITE_FIELD_RC_REMOTE_STAG].perm = READWRITE;
}

void QP_SEND::fill_rdma_write_intf(struct irdma_sc_qp *sc_qp, struct irdma_post_sq_info *info,
			    u64 *sges)
{
	enum qp_sq_send qp_send_enum = QPSQ_SEND_FIELD_FRAG_1_VALID;
	struct irdma_sge *sge_ptr = (struct irdma_sge *) sges;
	u16 wqe_size_bytes = 32;

	sq_tail_addr[0] = sc_qp->qp_uk.sq_base[sc_qp->qp_uk.sq_ring.tail].elem[0];
	sq_tail_addr[1] = sc_qp->qp_uk.sq_base[sc_qp->qp_uk.sq_ring.tail].elem[1];
	sq_tail_addr[2] = sc_qp->qp_uk.sq_base[sc_qp->qp_uk.sq_ring.tail].elem[2];
	sq_tail_addr[3] = sc_qp->qp_uk.sq_base[sc_qp->qp_uk.sq_ring.tail].elem[3];

	if (sc_qp->qp_uk.sq_ring.head == 0)
		sc_qp->qp_uk.swqe_polarity = !fields[QPSQ_SEND_FIELD_WQE_VALID].data;
	else
		sc_qp->qp_uk.swqe_polarity = fields[QPSQ_SEND_FIELD_WQE_VALID].data;
	sc_qp->qp_type = fields[QPSQ_SEND_QP_TYPE].data;
	info->op_type        = fields[QPSQ_SEND_FIELD_OP].data;
	info->signaled       = fields[QPSQ_SEND_FIELD_SIGNALED_CMPL].data;
	info->inline_data    = fields[QPSQ_SEND_FIELD_INLINE_DATA_FLAG].data;
	// this one is slightly different. The code checks to see if there is an
	// address associated with qp_uk.push_db, if there is
	// it always sets the push_wqe field in the WQE
	// otherwise, it leaves that field clear
	if (sc_qp->qp_uk.push_db)
		fields[QPSQ_SEND_FIELD_PUSH_WQE].data = 1ULL;
	else
		fields[QPSQ_SEND_FIELD_PUSH_WQE].data = 0ULL;

	enable_write_fields();

	if(fields[QPSQ_SEND_FIELD_INLINE_DATA_FLAG].data) {
		u16 wqes_for_inline_data = 0;
		u16 bytes_for_inline_data = 0;
		//set inline data pointer
		info->op.inline_rdma_write.data = &inline_data[0];
		info->op.inline_rdma_write.len = fields[QPSQ_SEND_FIELD_INLINE_DATA_LENGTH].get_masked_data_64();
		info->op.inline_rdma_write.rem_addr.len = 0;
		info->op.inline_rdma_write.rem_addr.tag_off = (irdma_tagged_offset) fields[QPSQ_WRITE_FIELD_RC_REMOTE_TAG_OFFSET].data;
		info->op.inline_rdma_write.rem_addr.stag = (irdma_stag) fields[QPSQ_WRITE_FIELD_RC_REMOTE_STAG].data;
		if (debug >= DEBUG_HIGH)
			printf("TB: rem_addr.stag(0x%x)\n", info->op.inline_rdma_write.rem_addr.stag);

		if (fields[QPSQ_SEND_FIELD_INLINE_DATA_LENGTH].get_masked_data_64() <= 224) {
			int valid_bytes = 0;
			if (config == MIXED_QP_TYPES) {
				fill_rdma_write_intf_mixed_inline(info);
			} else if (config == MUTEX_QP_TYPES) {
				fill_rdma_write_intf_mutex_inline(info);
			}

			for(int vb=info->op.inline_rdma_write.len-8; vb > 0; vb-=32, valid_bytes++)
				if (vb-32 >= 0)
					vb+=1;

			wqes_for_inline_data = (info->op.inline_rdma_write.len - 8 + valid_bytes)/(8*min_wqe_words());

			if ((info->op.inline_rdma_write.len - 8 + valid_bytes) % (8*min_wqe_words()))
				wqes_for_inline_data++;

			bytes_for_inline_data = 32*wqes_for_inline_data;
			wqe_size_bytes += bytes_for_inline_data;

			enable_inline_fields(wqes_for_inline_data);
			fill_nops(sc_qp, wqe_size_bytes, sc_qp->qp_uk.swqe_polarity);
		}

	} else {
		int valid_idx = 0;

		if (config == MIXED_QP_TYPES) {
			fill_rdma_write_intf_mixed(info);
		} else if (config == MUTEX_QP_TYPES) {
			fill_rdma_write_intf_mutex(info);
		}

		info->op.rdma_write.num_lo_sges  = fields[QPSQ_SEND_FIELD_ADDITIONAL_FRAG_COUNT].get_masked_data_64() + 1;
		info->op.rdma_write.rem_addr.len = 0;
		info->op.rdma_write.rem_addr.tag_off = (irdma_tagged_offset) fields[QPSQ_WRITE_FIELD_RC_REMOTE_TAG_OFFSET].data;
		info->op.rdma_write.rem_addr.stag = (irdma_stag) fields[QPSQ_WRITE_FIELD_RC_REMOTE_STAG].data;
		if (debug >= DEBUG_HIGH)
			printf("TB: rem_addr.stag(0x%x)\n", info->op.rdma_write.rem_addr.stag);

		if (info->op.rdma_write.num_lo_sges < 0)
			info->op.rdma_write.num_lo_sges = 0;

		wqe_size_bytes += (info->op.rdma_write.num_lo_sges-1)*16;
		if ((info->op.rdma_write.num_lo_sges-1) % 2) //ensure on 32 byte boundary
			wqe_size_bytes += 16;

		fill_nops(sc_qp, wqe_size_bytes, sc_qp->qp_uk.swqe_polarity);

		if (debug >= DEBUG_HIGH)
			printf("TB: filling sges\n");
		for (int i=0, j=0; j < info->op.rdma_write.num_lo_sges && info->op.rdma_write.num_lo_sges <= 14; i+=2, j++) {
			if (debug >= DEBUG_HIGH)
				printf("TB: filling idx %d @ %p\n", j, &sge_ptr[j]);
			sge_ptr[j].tag_off = (irdma_tagged_offset) fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_0+j*4].data;

			if (setup_config & CONSTRAIN_SGE_LEN) {
				fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_0+j*4+2].bit_mask = 0xffff;
				sge_ptr[j].len = (u32) fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_0+j*4+2].get_masked_data_64();
			} else
				sge_ptr[j].len = (u32) fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_0+j*4+2].get_masked_data_64();
			sge_ptr[j].stag = (irdma_stag) fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_0+j*4+3].get_masked_data_64();
			if ((((fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_0+j*4].byte_offset) % 32) ==0) && j) {
				set_next_valid = true;
				valid_idx = QPSQ_SEND_FIELD_TAGGED_OFFSET_0+(j+1)*4+1;
			} else
				set_next_valid = false;
		}
		if (debug >= DEBUG_HIGH)
			printf("TB: &sge(%p), &sge_ptr[0](%p)\n", &sges[0], &sge_ptr[0]);
		info->op.rdma_write.lo_sg_list   = (struct irdma_sge *) &sges[0];
		if (set_next_valid) {
			if (debug >= DEBUG_HIGH)
				printf("TB: set_next_valid true\n");
			fields[valid_idx].perm = READWRITE;
			fields[valid_idx-1].perm = RSVD;
			fields[valid_idx+1].perm = RSVD;
			fields[valid_idx+2].perm = RSVD;
		} else {
			if (debug >= DEBUG_HIGH)
				printf("TB: set_next_valid false\n");
		}
	}
}

void QP_SEND::rule_not_used_for_iw(u64 *field_data)
{
	if (fields[QPSQ_SEND_QP_TYPE].data == QP_TYPE_IWARP)
		*field_data = 0;
}

void QP_SEND::rule_not_used_for_cqp(u64 *field_data)
{
	if (fields[QPSQ_SEND_QP_TYPE].data == QP_TYPE_CQP)
		*field_data = 0;
}

void QP_SEND::rule_not_used_for_uda(u64 *field_data)
{
	if (fields[QPSQ_SEND_QP_TYPE].data == QP_TYPE_UDA)
		*field_data = 0;
}

void QP_SEND::rule_not_used_for_rocev2_rc(u64 *field_data)
{
	if (fields[QPSQ_SEND_QP_TYPE].data == QP_TYPE_ROCEV2_RC)
		*field_data = 0;
}

void QP_SEND::rule_not_used_for_rocev2_ud(u64 *field_data)
{
	if (fields[QPSQ_SEND_QP_TYPE].data == QP_TYPE_ROCEV2_UD)
		*field_data = 0;
}

void QP_SEND::rule_only_used_for_rc_qps(u64 *field_data)
{
	rule_not_used_for_cqp(field_data);
	rule_not_used_for_uda(field_data);
	rule_not_used_for_rocev2_ud(field_data);
}

void QP_SEND::rule_only_used_for_ud_qps(u64 *field_data)
{
	rule_not_used_for_cqp(field_data);
	rule_not_used_for_iw(field_data);
	rule_not_used_for_rocev2_rc(field_data);
}

void QP_SEND::rule_only_used_for_iw(u64 *field_data)
{
	rule_not_used_for_cqp(field_data);
	rule_not_used_for_uda(field_data);
	rule_not_used_for_rocev2_rc(field_data);
	rule_not_used_for_rocev2_ud(field_data);
}

void QP_SEND::rule_should_always_be(u64 *field_data, u64 x)
{
	*field_data = x;
}

void QP_SEND::rule_not_valid_if_flag_set(u64 *field_data, enum qp_sq_send send_enum)
{
	if (fields[send_enum].data)
		*field_data = 0;
}

void QP_SEND::rule_not_valid_if_flag_clear(u64 *field_data, enum qp_sq_send send_enum)
{
	if (!fields[send_enum].data)
		*field_data = 0;
}

void QP_SEND::rule_based_on_num_add_frags(u64 *field_data, enum qp_sq_send send_enum, u64 imm_data)
{
	//3 fields per Additional frag
	int tmp = fields[QPSQ_SEND_FIELD_ADDITIONAL_FRAG_COUNT].get_masked_data_64()*4;
	//FIXME: might need to adjust how imm_data works in this calculation
	if (imm_data)
		tmp+=4; //was tmp++;
	if (set_next_valid)
		tmp+=4;
	if (imm_data)
		*field_data = fields[send_enum - 4].data;
	if ((send_enum - QPSQ_SEND_FIELD_TAGGED_OFFSET_1) >= tmp)
		*field_data = 0;
		
}

void QP_SEND::rule_has_valid_number_of_frags(u64 *field_data)
{
	if (fields[QPSQ_SEND_FIELD_ADDITIONAL_FRAG_COUNT].get_masked_data_64() >= 14 &&
	    !fields[QPSQ_SEND_FIELD_INLINE_DATA_FLAG].get_masked_data_64())
		*field_data = 0;
}

void QP_SEND::rule_only_used_for_connection_setup(u64 *field_data)
{
	*field_data = 0;
}

void QP_SEND::rule_based_on_avail_wqes(int field_idx, u64 *field_data)
{
	if (fields[field_idx].byte_offset > avail_bytes) {
		*field_data = 0;
		if (debug >= DEBUG_HIGH)
			printf("TB: avail_wqes clearing idx(%d), avail(%d)\n",
			       field_idx,
			       avail_bytes);
	}
}

void QP_SEND::rule_honor_head_tail_gap(int field_idx, u64 *field_data)
{
	if (fields[field_idx].byte_offset >= avail_bytes - 32) {
		*field_data = 0;
		if (debug >= DEBUG_HIGH)
			printf("TB: avail_wqes clearing idx(%d), avail(%d)\n",
			       field_idx,
			       avail_bytes - 32);
	}
}

// Used to overwrite "TAIL" pointer with preset TAIL data
// This will change the expected data to match what is in the sq_tail
void QP_SEND::rule_valid_sq_data(int field_idx, u64 *field_data)
{
	if (fields[field_idx].byte_offset >= avail_bytes) {
		int bytes_per_wqe = min_wqe_words() * 8;
		if (fields[field_idx].byte_offset == avail_bytes + bytes_per_wqe) {
			*field_data = sq_tail_addr[0];
			fields[field_idx].bit_shift = 0;
			fields[field_idx].bit_mask = 0xffffffffffffffff;
			fields[field_idx].perm = READWRITE;
			if (debug >= DEBUG_HIGH)
				printf("byte_offset: sq_tail[0]=0x%llx\n", *field_data);
		} else if (fields[field_idx].byte_offset == avail_bytes + bytes_per_wqe + 8) {
			*field_data = sq_tail_addr[1];
			fields[field_idx].bit_shift = 0;
			fields[field_idx].bit_mask = 0xffffffffffffffff;
			fields[field_idx].perm = READWRITE;
			if (debug >= DEBUG_HIGH)
				printf("byte_offset: sq_tail[1]=0x%llx\n", *field_data);
		} else if (fields[field_idx].byte_offset == avail_bytes + bytes_per_wqe + 16) {
			*field_data = sq_tail_addr[2];
			fields[field_idx].bit_shift = 0;
			fields[field_idx].bit_mask = 0xffffffffffffffff;
			fields[field_idx].perm = READWRITE;
			if (debug >= DEBUG_HIGH)
				printf("byte_offset: sq_tail[2]=0x%llx\n", *field_data);
		} else if (fields[field_idx].byte_offset == avail_bytes + bytes_per_wqe + 24) {
			*field_data = sq_tail_addr[3];
			fields[field_idx].bit_shift = 0;
			fields[field_idx].bit_mask = 0xffffffffffffffff;
			fields[field_idx].perm = READWRITE;
			if (debug >= DEBUG_HIGH)
				printf("byte_offset: sq_tail[3]=0x%llx\n", *field_data);
		}
		if (debug >= DEBUG_HIGH)
			printf("TB: valid_sq_data clearing idx(%d)\n", field_idx);
	}
		
}

void QP_SEND::rule_wqe_fits_in_avail_space(int field_idx, u64 *field_data)
{
	if ((field_idx < QPSQ_SEND_FIELD_NOP_0_0 ||
	    field_idx > QPSQ_SEND_FIELD_NOP_7_3) &&
	    field_idx != QPSQ_SEND_QP_TYPE &&
	    wqe_will_not_fit) {
		*field_data = 0ULL;
		if (debug >= DEBUG_HIGH)
			printf("TB: avail_wqes clearing idx(%d)\n", field_idx);
	}
}

void QP_SEND::rule_not_used_for_inline(u64 *field_data)
{
	if(fields[QPSQ_SEND_FIELD_INLINE_DATA_FLAG].get_masked_data_64())
		*field_data = 0ULL;
}

void QP_SEND::rule_only_for_uda(u64 *field_data)
{
	if (fields[QPSQ_SEND_QP_TYPE].data != QP_TYPE_UDA)
		*field_data = 0;
}

void QP_SEND::rule_only_used_for_inv_ops(u64 *field_data)
{
	if (!(fields[QPSQ_SEND_FIELD_OP].data == TB_IRDMA_OP_TYPE_SEND_INV ||
	    fields[QPSQ_SEND_FIELD_OP].data == TB_IRDMA_OP_TYPE_SEND_SE_INV))
		*field_data = 0;
}

void QP_SEND::rule_not_used_for_write(u64 *field_data)
{
	if (fields[QPSQ_SEND_FIELD_OP].data == TB_IRDMA_OP_TYPE_WRITE ||
	    fields[QPSQ_SEND_FIELD_OP].data == TB_IRDMA_OP_TYPE_WRITE_SE)
		*field_data = 0;
}

/*
 * QP_SEND::gen_field_vs_rules
 * Upper level will create a copy of the field data to pass into this function.
 * This way we can manipulate the data written to the SQ without modifying the
 * original value of the field.
 *
 * @field_idx: index of the field
 * @field_data: a pointer to a copy of the data in the field
 */
void QP_SEND::gen_field_vs_rules(int field_idx, u64 *field_data)
{
	//rule_not_used_for_cqp(field_data); //CQP should not come up
	rule_has_valid_number_of_frags(field_data);
	rule_based_on_avail_wqes(field_idx, field_data);
	rule_honor_head_tail_gap(field_idx, field_data);
	switch ((enum qp_sq_send) field_idx) {
		case QPSQ_SEND_FIELD_IMM_DATA_FLAG:
//HGO			if (fields[QPSQ_SEND_QP_TYPE].data == QP_TYPE_IWARP || fields[QPSQ_SEND_QP_TYPE].data == QP_TYPE_ROCEV2_RC)
			if (fields[QPSQ_SEND_QP_TYPE].data == QP_TYPE_IWARP)
				rule_should_always_be(field_data, 0);
			break;
		case QPSQ_SEND_FIELD_UDP_HEADER:
			rule_only_for_uda(field_data);
			break;
		case QPSQ_SEND_FIELD_UD_IMM_DATA:
			rule_not_valid_if_flag_clear(field_data, QPSQ_SEND_FIELD_IMM_DATA_FLAG);
			break;
		case QPSQ_SEND_FIELD_UD_DEST_QPN:
		case QPSQ_SEND_FIELD_UD_DEST_QKEY:
		case QPSQ_SEND_FIELD_UD_L4LEN:
		case QPSQ_SEND_FIELD_UD_AH_ID:
			rule_only_used_for_ud_qps(field_data);
			break;
		case QPSQ_SEND_FIELD_REPORT_RTT:
		case QPSQ_SEND_FIELD_IW_LOCAL_FENCE:
		case QPSQ_SEND_FIELD_IW_READ_FENCE:
			rule_only_used_for_rc_qps(field_data);
			break;
		case QPSQ_SEND_FIELD_IW_WAIT_FOR_RCVFPDU:
		case QPSQ_SEND_FIELD_IW_STREAMING_MODE:
			rule_only_used_for_connection_setup(field_data);
			break;
		case QPSQ_SEND_FIELD_FRAG_0_VALID:
		case QPSQ_SEND_FIELD_FRAG_0_LENGTH:
		case QPSQ_SEND_FIELD_STAG_0:
			if (fields[QPSQ_SEND_QP_TYPE].data != QP_TYPE_IWARP)
				rule_not_valid_if_flag_set(field_data, QPSQ_SEND_FIELD_IMM_DATA_FLAG);
			rule_not_used_for_inline(field_data);
			break;
		case QPSQ_SEND_FIELD_TAGGED_OFFSET_0:
			if(fields[QPSQ_SEND_QP_TYPE].data != QP_TYPE_IWARP &&
			   fields[QPSQ_SEND_FIELD_IMM_DATA_FLAG].data)
				rule_should_always_be(field_data, 0);
			rule_not_used_for_inline(field_data);
			break;
		case QPSQ_SEND_FIELD_ADDITIONAL_FRAG_COUNT:
			rule_not_used_for_inline(field_data);
			break;
		case QPSQ_SEND_FIELD_IW_REMOTE_INV_STAG:
			rule_only_used_for_inv_ops(field_data);
			rule_only_used_for_rc_qps(field_data);
			rule_not_used_for_write(field_data);
			break;
		case QPSQ_SEND_FIELD_TAGGED_OFFSET_1:
		case QPSQ_SEND_FIELD_TAGGED_OFFSET_2:
		case QPSQ_SEND_FIELD_TAGGED_OFFSET_3:
		case QPSQ_SEND_FIELD_TAGGED_OFFSET_4:
		case QPSQ_SEND_FIELD_TAGGED_OFFSET_5:
		case QPSQ_SEND_FIELD_TAGGED_OFFSET_6:
		case QPSQ_SEND_FIELD_TAGGED_OFFSET_7:
		case QPSQ_SEND_FIELD_TAGGED_OFFSET_8:
		case QPSQ_SEND_FIELD_TAGGED_OFFSET_9:
		case QPSQ_SEND_FIELD_TAGGED_OFFSET_10:
		case QPSQ_SEND_FIELD_TAGGED_OFFSET_11:
		case QPSQ_SEND_FIELD_TAGGED_OFFSET_12:
		case QPSQ_SEND_FIELD_TAGGED_OFFSET_13:
		case QPSQ_SEND_FIELD_TAGGED_OFFSET_14:
		case QPSQ_SEND_FIELD_FRAG_1_VALID:
		case QPSQ_SEND_FIELD_FRAG_1_LENGTH:
		case QPSQ_SEND_FIELD_STAG_1:
		case QPSQ_SEND_FIELD_FRAG_2_VALID:
		case QPSQ_SEND_FIELD_FRAG_2_LENGTH:
		case QPSQ_SEND_FIELD_STAG_2:
		case QPSQ_SEND_FIELD_FRAG_3_VALID:
		case QPSQ_SEND_FIELD_FRAG_3_LENGTH:
		case QPSQ_SEND_FIELD_STAG_3:
		case QPSQ_SEND_FIELD_FRAG_4_VALID:
		case QPSQ_SEND_FIELD_FRAG_4_LENGTH:
		case QPSQ_SEND_FIELD_STAG_4:
		case QPSQ_SEND_FIELD_FRAG_5_VALID:
		case QPSQ_SEND_FIELD_FRAG_5_LENGTH:
		case QPSQ_SEND_FIELD_STAG_5:
		case QPSQ_SEND_FIELD_FRAG_6_VALID:
		case QPSQ_SEND_FIELD_FRAG_6_LENGTH:
		case QPSQ_SEND_FIELD_STAG_6:
		case QPSQ_SEND_FIELD_FRAG_7_VALID:
		case QPSQ_SEND_FIELD_FRAG_7_LENGTH:
		case QPSQ_SEND_FIELD_STAG_7:
		case QPSQ_SEND_FIELD_FRAG_8_VALID:
		case QPSQ_SEND_FIELD_FRAG_8_LENGTH:
		case QPSQ_SEND_FIELD_STAG_8:
		case QPSQ_SEND_FIELD_FRAG_9_VALID:
		case QPSQ_SEND_FIELD_FRAG_9_LENGTH:
		case QPSQ_SEND_FIELD_STAG_9:
		case QPSQ_SEND_FIELD_FRAG_10_VALID:
		case QPSQ_SEND_FIELD_FRAG_10_LENGTH:
		case QPSQ_SEND_FIELD_STAG_10:
		case QPSQ_SEND_FIELD_FRAG_11_VALID:
		case QPSQ_SEND_FIELD_FRAG_11_LENGTH:
		case QPSQ_SEND_FIELD_STAG_11:
		case QPSQ_SEND_FIELD_FRAG_12_VALID:
		case QPSQ_SEND_FIELD_FRAG_12_LENGTH:
		case QPSQ_SEND_FIELD_STAG_12:
		case QPSQ_SEND_FIELD_FRAG_13_VALID:
		case QPSQ_SEND_FIELD_FRAG_13_LENGTH:
		case QPSQ_SEND_FIELD_STAG_13:
		case QPSQ_SEND_FIELD_FRAG_14_VALID:
		case QPSQ_SEND_FIELD_FRAG_14_LENGTH:
		case QPSQ_SEND_FIELD_STAG_14:
			rule_based_on_num_add_frags(field_data, field_idx, fields[QPSQ_SEND_FIELD_IMM_DATA_FLAG].data);
			rule_not_used_for_inline(field_data);
			break;
		/* INLINE FIELD CASES RSVD by default (won't be put in "expected" SQ) */
		default:
			break;
	}
	rule_wqe_fits_in_avail_space(field_idx, field_data);
	rule_valid_sq_data(field_idx, field_data);
}

/*
 * QP_SEND::enable_frags
 * This is a helper function to properly enable fragment data fields in the TB
 *
 * Notes on fields indecies for inline data
 *   The 1st index of inline data is at QPSQ_SEND_FIELD_TAGGED_OFFSET_0
 *   The indecies follow a repeating pattern per half quantum
 *   of the WR.
 *
 *   Example for quantum 2:
 *     QPSQ_SEND_FIELD_TAGGED_OFFSET_1
 *     QPSQ_SEND_FIELD_FRAG_1_VALID
 *     QPSQ_SEND_FIELD_FRAG_1_LENGTH
 *     QPSQ_SEND_FIELD_STAG_1
 *     QPSQ_SEND_FIELD_TAGGED_OFFSET_2
 *     QPSQ_SEND_FIELD_FRAG_2_VALID
 *     QPSQ_SEND_FIELD_FRAG_2_LENGTH
 *     QPSQ_SEND_FIELD_STAG_2
 *
 *   Also note that the numeral at the end corresponds to the fragment.
 *   The quantum bit valid is in the valid for the 2nd fragment of the
 *   quantum and will need to be set if a fragment is valid in the 1st
 *   fragment.
 *
 * This helper function depends on the order of the enum delcaration (found in
 * cqp_fields.h). All enums related to frag data fields are grouped together.
 *
 * @add_sges: number of additional sges for this WR.
 *
 */
void QP_SEND::enable_frags(u16 add_sges)
{
	u16 idx = QPSQ_SEND_FIELD_TAGGED_OFFSET_1;

	if (!add_sges)
		return;

	for (int i=0; i<add_sges; i++) {
		if (debug >= DEBUG_HIGH)
			std::cout << "TB: " << __func__ << ": add sges(" <<
			add_sges << "), enabling:\n" <<
			"\t" << fields[idx].name << "\n" <<
			"\t" << fields[idx+1].name << "\n" <<
			"\t" << fields[idx+2].name << "\n" <<
			"\t" << fields[idx+3].name << std::endl;
		fields[idx++].perm = READWRITE; //TAGGED_OFFSET
		fields[idx++].perm = READWRITE; // VALID
		fields[idx++].perm = READWRITE; // LENGTH
		fields[idx++].perm = READWRITE; // STAG
	}

	if (set_next_valid) // don't need to check against last stag as set_next_valid should
		fields[idx+1].perm = READWRITE; //sge valid
}

/*
 * QP_SEND::set_inline_data
 * This is a helper function to properly set the inline data fields in the TB
 *
 * Notes on fields indecies for inline data
 *   The 1st index of inline data is at QPSQ_SEND_FIELD_INLINE_DATA_0
 *   After the 1st index, the indecies follow a repeating pattern per quantum
 *   of the WR.
 *
 *   Example for quantum 2:
 *     QPSQ_SEND_FIELD_INLINE_DATA_1
 *     QPSQ_SEND_FIELD_INLINE_DATA_2
 *     QPSQ_SEND_FIELD_INLINE_DATA_3
 *     QPSQ_SEND_FIELD_INLINE_DATA_4
 *     QPSQ_SEND_FIELD_INLINE_DATA_VALID_4
 *     QPSQ_SEND_FIELD_INLINE_DATA_RSVD_4
 *
 *   Also note that the numeral at the end corresponds to the 64B word line.
 *   The first index for inline data in quantum 3 would be:
 *     QPSQ_SEND_FIELD_INLINE_DATA_5
 *
 * This helper function depends on the order of the enum delcaration (found in
 * cqp_fields.h). All enums related to inline data fields are grouped together.
 *
 * @enum_idx: pointer to where the fields array index will be stored
 * @data_idx: pointer to where inline data byte index will be stored
 * @ib_wr: pointer to ib_send_wr being sent to CODE
 *
 */
void QP_SEND::set_inline_data(u16 *enum_idx, u16 *data_idx, struct ib_send_wr *ib_wr)
{
	u64 *data_ptr;
	u16 in_len = ib_wr->sg_list[0].length;
	u64 mask = 0ULL;
	u64 tmp_mask = 0ULL;
	u16 e_idx = QPSQ_SEND_FIELD_INLINE_DATA_0;
	u16 bytes_in_word = 0;

	if (debug >= DEBUG_NONE) {
		printf("TB: %s: enum(%d), e(%d), idx(%d)\n", __func__, *enum_idx, e_idx, *data_idx);
		printf("TB: %s: ib_wr(%p), addr(0x%llx)\n", __func__, ib_wr, ib_wr->sg_list[0].addr);
	}

	if (*enum_idx == 0) {
		tmp_mask = 0xffffffffffffffff;
		bytes_in_word = 8;
	} else if (*enum_idx % 6 == 4){
		tmp_mask = 0xffffffffffffff;
		bytes_in_word = 7;
	} else {
		tmp_mask = 0xffffffffffffffff;
		bytes_in_word = 8;
	}

	data_ptr = (u64*) (ib_wr->sg_list[0].addr + *data_idx);
	if (*data_idx + bytes_in_word < in_len) {
		mask = tmp_mask;
	} else if (*data_idx < in_len) {
		mask = 0xff;
		for (int i=1; i < (in_len-*data_idx); i++)
			mask = mask << 8 | 0xff;
	}
	if (debug >= DEBUG_NONE)
		printf("TB: %s: e(%d)\n", __func__, e_idx);
	fields[e_idx + *enum_idx].data = *data_ptr & mask;
	*data_idx += bytes_in_word;

	if (*enum_idx % 6 == 4) {
		*enum_idx += 1;
		fields[e_idx+*enum_idx].data = fields[QPSQ_SEND_FIELD_WQE_VALID].data;
		*enum_idx += 1; //skip reserved
	}
	*enum_idx += 1;

}

void QP_SEND::fill_fields(struct irdma_ring *sqr,
			  struct irdma_qp *iwqp,
			  struct ib_send_wr *ib_wr,
			  u64 polarity)
			  
{
	u32 byte_offset;
	u32 orig_quanta;

	switch (ib_wr->opcode) {
		case IB_WR_RDMA_WRITE:
		case IB_WR_RDMA_WRITE_WITH_IMM:
		case IB_WR_SEND:
		case IB_WR_SEND_WITH_IMM:
		case IB_WR_SEND_WITH_INV:
			fill_fields_send(sqr, iwqp, ib_wr, polarity);
			break;
		default:
			if (debug >= DEBUG_NONE)
				printf("TB: ERR: %s: invalid IB SQ WR opcode(0x%llx)\n",
				       __func__, ib_wr->opcode);
			return;
	}

	byte_offset = 0;
	for (int i=0; i<QPSQ_SEND_FIELD_NUM_FIELDS; i++)
		if (fields[i].perm != RSVD)
			byte_offset = byte_offset < fields[i].byte_offset ?
				      fields[i].byte_offset : byte_offset;

	byte_offset += 8; // add that line of 8 bytes
	orig_quanta = byte_offset / (8 * min_wqe_words()); //should turn this into quanta

	//Add any NOPs needed
	fill_nops(sqr, &iwqp->sc_qp, orig_quanta*8*min_wqe_words(), iwqp->sc_qp.qp_uk.swqe_polarity);

	// Determine size of this WR with NOPs
	byte_offset = 0;
	for (int i=0; i<QPSQ_SEND_FIELD_NUM_FIELDS; i++)
		if (fields[i].perm != RSVD)
			byte_offset = byte_offset < fields[i].byte_offset ?
				      fields[i].byte_offset : byte_offset;

	byte_offset += 8; // add that line of 8 bytes
	num_quanta = byte_offset / (8 * min_wqe_words()); //should turn this into quanta
}

/*
 * QP_SEND::fill_fields_send
 * This function needs to be used if the testbench is utilizing
 * the software structures (ib_send_wr, irdma_qp, etc) for randomization
 * rather than utilizing the fields for randomization.
 * It fills in the fields based on the randomized/set structures.
 * It also sets the total number of quanta used by this WR in data member
 * WQE::num_quanta. This can be used by the testbench for SQ management.
 *
 * @sqr: tb sq ring
 * @qp: pointer to irdma_qp used by CODE
 * @ib_wr: pointer to ib_send_wr used by CODE
 * polarity: This is determined by the caller.
 */
void QP_SEND::fill_fields_send(struct irdma_ring *sqr,
			       struct irdma_qp *iwqp,
			       struct ib_send_wr *ib_wr,
			       u64 polarity)
{
	struct ib_qp *ibqp = &iwqp->ibqp;
	set_next_valid = false;

	if (iwqp->ibqp.qp_type == IB_QPT_UD ||
	    iwqp->ibqp.qp_type == IB_QPT_GSI)
		fields[QPSQ_SEND_QP_TYPE].data = QP_TYPE_ROCEV2_UD;
	else if (iwqp->ibqp.qp_type == IB_QPT_RC)
		if (iwqp->sc_qp.qp_uk.qp_caps == (IRDMA_SEND_WITH_IMM | IRDMA_WRITE_WITH_IMM))
			fields[QPSQ_SEND_QP_TYPE].data = QP_TYPE_ROCEV2_RC;
		else if (iwqp->sc_qp.qp_uk.qp_caps == IRDMA_WRITE_WITH_IMM)
			fields[QPSQ_SEND_QP_TYPE].data = QP_TYPE_IWARP;

	if (debug >= DEBUG_HIGH)
		printf("TB: %s: internal QP_TYPE(%d) external(%d)\n", __func__,
		       fields[QPSQ_SEND_QP_TYPE].data, iwqp->ibqp.qp_type);

	if (iwqp->ibqp.qp_type == IB_QPT_UD ||
	    iwqp->ibqp.qp_type == IB_QPT_GSI) {
		fields[QPSQ_SEND_FIELD_UD_DEST_QPN].data = ud_wr(ib_wr)->remote_qpn;
		fields[QPSQ_SEND_FIELD_UD_DEST_QKEY].data = ud_wr(ib_wr)->remote_qkey;
		fields[QPSQ_SEND_FIELD_UD_DEST_QPN].perm = READWRITE;
		fields[QPSQ_SEND_FIELD_UD_DEST_QKEY].perm = READWRITE;
	}

	fields[QPSQ_SEND_FIELD_WQE_VALID].data = polarity;
	if (ib_wr->send_flags & IB_SEND_SIGNALED || iwqp->sig_all)
		fields[QPSQ_SEND_FIELD_SIGNALED_CMPL].data = 1;

	if (ib_wr->send_flags & IB_SEND_FENCE) {
		if (ib_wr->opcode == IB_WR_SEND || 
		    ib_wr->opcode == IB_WR_SEND_WITH_IMM ||
		    ib_wr->opcode == IB_WR_SEND_WITH_INV ||
		    ib_wr->opcode == IB_WR_RDMA_WRITE ||
		    ib_wr->opcode == IB_WR_RDMA_WRITE_WITH_IMM)
			fields[QPSQ_SEND_FIELD_IW_READ_FENCE].data = 1;
		if (ib_wr->opcode == IB_WR_LOCAL_INV) //Mem INV
			fields[QPSQ_SEND_FIELD_IW_LOCAL_FENCE].data = 1;
	}

	// Only used for connection setup, not used in normal QP operation
	fields[QPSQ_SEND_FIELD_IW_WAIT_FOR_RCVFPDU].data = rand() % 2;
	// Only used for connection setup, not used in normal QP operation
	fields[QPSQ_SEND_FIELD_IW_STREAMING_MODE].data = rand() % 2;

// This is only used for UDA
//	fields[QPSQ_SEND_FIELD_UDP_HEADER].data = (u64) *ud_wr(ib_wr)->header;
	if (iwqp->sc_qp.qp_uk.push_db)
		fields[QPSQ_SEND_FIELD_PUSH_WQE].data = 1;

	if (ib_wr->opcode == IB_WR_RDMA_WRITE_WITH_IMM ||
	    ib_wr->opcode == IB_WR_SEND_WITH_IMM)
		fields[QPSQ_SEND_FIELD_IMM_DATA_FLAG].data = 1;

	//not used for production ... only for debug
	//fields[QPSQ_SEND_FIELD_REPORT_RTT].data = 0;
	//only for UDA QPs which is not supported
	//fields[QPSQ_SEND_FIELD_UD_L4LEN].data = 0;

	if (ib_wr->opcode == IB_WR_SEND ||
	    ib_wr->opcode == IB_WR_SEND_WITH_IMM)
		if (ib_wr->send_flags & IB_SEND_SOLICITED)
			fields[QPSQ_SEND_FIELD_OP].data = TB_IRDMA_OP_TYPE_SEND_SE;
		else
			fields[QPSQ_SEND_FIELD_OP].data = TB_IRDMA_OP_TYPE_SEND;

	if (ib_wr->opcode == IB_WR_SEND_WITH_INV)	
		if (ib_wr->send_flags & IB_SEND_SOLICITED)
			fields[QPSQ_SEND_FIELD_OP].data = TB_IRDMA_OP_TYPE_SEND_SE_INV;
		else
			fields[QPSQ_SEND_FIELD_OP].data = TB_IRDMA_OP_TYPE_SEND_INV;

	if (ib_wr->opcode == IB_WR_RDMA_WRITE ||
	    ib_wr->opcode == IB_WR_RDMA_WRITE_WITH_IMM)
		if (ib_wr->send_flags & IB_SEND_SOLICITED)
			fields[QPSQ_SEND_FIELD_OP].data = TB_IRDMA_OP_TYPE_WRITE_SE;
		else
			fields[QPSQ_SEND_FIELD_OP].data = TB_IRDMA_OP_TYPE_WRITE;

	if (iwqp->ibqp.qp_type == IB_QPT_UD ||
	    iwqp->ibqp.qp_type == IB_QPT_GSI) {
		fields[QPSQ_SEND_FIELD_UD_AH_ID].data = to_iwah(ud_wr(ib_wr)->ah)->sc_ah.ah_info.ah_idx;
		fields[QPSQ_SEND_FIELD_UD_DEST_QKEY].data = ud_wr(ib_wr)->remote_qkey;
		fields[QPSQ_SEND_FIELD_UD_DEST_QPN].data  = ud_wr(ib_wr)->remote_qpn;
		fields[QPSQ_SEND_FIELD_UD_AH_ID].perm     = READWRITE;
		fields[QPSQ_SEND_FIELD_UD_DEST_QKEY].perm = READWRITE;
		fields[QPSQ_SEND_FIELD_UD_DEST_QPN].perm  = READWRITE;
	} else {
		fields[QPSQ_SEND_FIELD_UD_AH_ID].data     = rand() | rand() << 32;
		fields[QPSQ_SEND_FIELD_UD_DEST_QKEY].data = rand() | rand() << 32;
		fields[QPSQ_SEND_FIELD_UD_DEST_QPN].data  = rand() | rand() << 32;

	}

	fields[QPSQ_SEND_FIELD_IW_REMOTE_INV_STAG].data = ib_wr->ex.invalidate_rkey;
	fields[QPSQ_SEND_FIELD_UD_IMM_DATA].data = ib_wr->ex.imm_data;

	if (iwqp->ibqp.qp_type == IB_QPT_RC &&
	    (ib_wr->opcode == IB_WR_RDMA_WRITE ||
	     ib_wr->opcode == IB_WR_RDMA_WRITE_WITH_IMM)) {
		fields[QPSQ_WRITE_FIELD_RC_REMOTE_TAG_OFFSET].data = rdma_wr(ib_wr)->remote_addr;
		fields[QPSQ_WRITE_FIELD_RC_REMOTE_STAG].data       = rdma_wr(ib_wr)->rkey;
		fields[QPSQ_WRITE_FIELD_RC_REMOTE_TAG_OFFSET].perm = READWRITE;
		fields[QPSQ_WRITE_FIELD_RC_REMOTE_STAG].perm       = READWRITE;
	} else {
		fields[QPSQ_WRITE_FIELD_RC_REMOTE_TAG_OFFSET].data = rand() | rand() << 32;
		fields[QPSQ_WRITE_FIELD_RC_REMOTE_STAG].data = rand() | rand() << 32;
	}

	if (ib_wr->opcode == IB_WR_SEND_WITH_IMM ||
	    ib_wr->opcode == IB_WR_RDMA_WRITE_WITH_IMM)
		fields[QPSQ_SEND_FIELD_ADDITIONAL_FRAG_COUNT].data = ib_wr->num_sge;
	else
		if (ib_wr->num_sge) //check if not 0 first
			fields[QPSQ_SEND_FIELD_ADDITIONAL_FRAG_COUNT].data = ib_wr->num_sge - 1;
		else
			fields[QPSQ_SEND_FIELD_ADDITIONAL_FRAG_COUNT].data = ib_wr->num_sge;

	if (fields[QPSQ_SEND_FIELD_ADDITIONAL_FRAG_COUNT].data % 2)
		set_next_valid = true;

	u16 add_frags = (u16) fields[QPSQ_SEND_FIELD_ADDITIONAL_FRAG_COUNT].get_masked_data_64();

	enable_frags(add_frags);

	int sge_idx = 0;
	fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_0].data = ib_wr->sg_list[sge_idx].addr;
	fields[QPSQ_SEND_FIELD_FRAG_0_VALID].data    = polarity;
	fields[QPSQ_SEND_FIELD_FRAG_0_LENGTH].data   = ib_wr->sg_list[sge_idx].length;
	fields[QPSQ_SEND_FIELD_STAG_0].data          = ib_wr->sg_list[sge_idx].lkey;
	sge_idx++;
	fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_1].data = ib_wr->sg_list[sge_idx].addr;
	fields[QPSQ_SEND_FIELD_FRAG_1_VALID].data    = polarity;
	fields[QPSQ_SEND_FIELD_FRAG_1_LENGTH].data   = ib_wr->sg_list[sge_idx].length;
	fields[QPSQ_SEND_FIELD_STAG_1].data          = ib_wr->sg_list[sge_idx].lkey;
	sge_idx++;
	fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_2].data = ib_wr->sg_list[sge_idx].addr;
	fields[QPSQ_SEND_FIELD_FRAG_2_VALID].data    = polarity;
	fields[QPSQ_SEND_FIELD_FRAG_2_LENGTH].data   = ib_wr->sg_list[sge_idx].length;
	fields[QPSQ_SEND_FIELD_STAG_2].data          = ib_wr->sg_list[sge_idx].lkey;
	sge_idx++;
	fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_3].data = ib_wr->sg_list[sge_idx].addr;
	fields[QPSQ_SEND_FIELD_FRAG_3_VALID].data    = polarity;
	fields[QPSQ_SEND_FIELD_FRAG_3_LENGTH].data   = ib_wr->sg_list[sge_idx].length;
	fields[QPSQ_SEND_FIELD_STAG_3].data          = ib_wr->sg_list[sge_idx].lkey;
	sge_idx++;
	fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_4].data = ib_wr->sg_list[sge_idx].addr;
	fields[QPSQ_SEND_FIELD_FRAG_4_VALID].data    = polarity;
	fields[QPSQ_SEND_FIELD_FRAG_4_LENGTH].data   = ib_wr->sg_list[sge_idx].length;
	fields[QPSQ_SEND_FIELD_STAG_4].data          = ib_wr->sg_list[sge_idx].lkey;
	sge_idx++;
	fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_5].data = ib_wr->sg_list[sge_idx].addr;
	fields[QPSQ_SEND_FIELD_FRAG_5_VALID].data    = polarity;
	fields[QPSQ_SEND_FIELD_FRAG_5_LENGTH].data   = ib_wr->sg_list[sge_idx].length;
	fields[QPSQ_SEND_FIELD_STAG_5].data          = ib_wr->sg_list[sge_idx].lkey;
	sge_idx++;
	fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_6].data = ib_wr->sg_list[sge_idx].addr;
	fields[QPSQ_SEND_FIELD_FRAG_6_VALID].data    = polarity;
	fields[QPSQ_SEND_FIELD_FRAG_6_LENGTH].data   = ib_wr->sg_list[sge_idx].length;
	fields[QPSQ_SEND_FIELD_STAG_6].data          = ib_wr->sg_list[sge_idx].lkey;
	sge_idx++;
	fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_7].data = ib_wr->sg_list[sge_idx].addr;
	fields[QPSQ_SEND_FIELD_FRAG_7_VALID].data    = polarity;
	fields[QPSQ_SEND_FIELD_FRAG_7_LENGTH].data   = ib_wr->sg_list[sge_idx].length;
	fields[QPSQ_SEND_FIELD_STAG_7].data          = ib_wr->sg_list[sge_idx].lkey;
	sge_idx++;
	fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_8].data = ib_wr->sg_list[sge_idx].addr;
	fields[QPSQ_SEND_FIELD_FRAG_8_VALID].data    = polarity;
	fields[QPSQ_SEND_FIELD_FRAG_8_LENGTH].data   = ib_wr->sg_list[sge_idx].length;
	fields[QPSQ_SEND_FIELD_STAG_8].data          = ib_wr->sg_list[sge_idx].lkey;
	sge_idx++;
	fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_9].data = ib_wr->sg_list[sge_idx].addr;
	fields[QPSQ_SEND_FIELD_FRAG_9_VALID].data    = polarity;
	fields[QPSQ_SEND_FIELD_FRAG_9_LENGTH].data   = ib_wr->sg_list[sge_idx].length;
	fields[QPSQ_SEND_FIELD_STAG_9].data          = ib_wr->sg_list[sge_idx].lkey;
	sge_idx++;
	fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_10].data = ib_wr->sg_list[sge_idx].addr;
	fields[QPSQ_SEND_FIELD_FRAG_10_VALID].data    = polarity;
	fields[QPSQ_SEND_FIELD_FRAG_10_LENGTH].data   = ib_wr->sg_list[sge_idx].length;
	fields[QPSQ_SEND_FIELD_STAG_10].data          = ib_wr->sg_list[sge_idx].lkey;
	sge_idx++;
	fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_11].data = ib_wr->sg_list[sge_idx].addr;
	fields[QPSQ_SEND_FIELD_FRAG_11_VALID].data    = polarity;
	fields[QPSQ_SEND_FIELD_FRAG_11_LENGTH].data   = ib_wr->sg_list[sge_idx].length;
	fields[QPSQ_SEND_FIELD_STAG_11].data          = ib_wr->sg_list[sge_idx].lkey;
	sge_idx++;
	fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_12].data = ib_wr->sg_list[sge_idx].addr;
	fields[QPSQ_SEND_FIELD_FRAG_12_VALID].data    = polarity;
	fields[QPSQ_SEND_FIELD_FRAG_12_LENGTH].data   = ib_wr->sg_list[sge_idx].length;
	fields[QPSQ_SEND_FIELD_STAG_12].data          = ib_wr->sg_list[sge_idx].lkey;
	sge_idx++;
	fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_13].data = ib_wr->sg_list[sge_idx].addr;
	fields[QPSQ_SEND_FIELD_FRAG_13_VALID].data    = polarity;
	fields[QPSQ_SEND_FIELD_FRAG_13_LENGTH].data   = ib_wr->sg_list[sge_idx].length;
	fields[QPSQ_SEND_FIELD_STAG_13].data          = ib_wr->sg_list[sge_idx].lkey;
	sge_idx++;
	fields[QPSQ_SEND_FIELD_TAGGED_OFFSET_14].data = ib_wr->sg_list[sge_idx].addr;
	fields[QPSQ_SEND_FIELD_FRAG_14_VALID].data    = polarity;
	fields[QPSQ_SEND_FIELD_FRAG_14_LENGTH].data   = ib_wr->sg_list[sge_idx].length;
	fields[QPSQ_SEND_FIELD_STAG_14].data          = ib_wr->sg_list[sge_idx].lkey;

	if (ib_wr->send_flags & IB_SEND_INLINE) {
		fields[QPSQ_SEND_FIELD_INLINE_DATA_FLAG].data = 1;
		fields[QPSQ_SEND_FIELD_INLINE_DATA_LENGTH].data = ib_wr->sg_list[0].length;
	} else {
		fields[QPSQ_SEND_FIELD_INLINE_DATA_FLAG].data = 0;
		fields[QPSQ_SEND_FIELD_INLINE_DATA_LENGTH].data = 0;
	}

	u16 inline_quanta = 0;
	u16 in_len = ib_wr->sg_list[0].length;
	u64 *data_ptr;
	u16 inline_data_idx;
	u16 enum_idx;
	u64 tmp = 0;
	u64 mask = 0;

	if (!(ib_wr->send_flags & IB_SEND_INLINE))
		return;

	if (in_len > 8)
		in_len -= 8;
	else
		in_len = 0;

	if (in_len % 31)
		inline_quanta++;

	inline_quanta += in_len/31;

	enable_inline_fields(inline_quanta);

	if (debug >= DEBUG_HIGH)
		printf("TB: inline addr(0x%llx)\n", ib_wr->sg_list[0].addr);
	data_ptr = (u64*) ib_wr->sg_list[0].addr;
	in_len = ib_wr->sg_list[0].length;//+8;
	inline_data_idx = 0;
	enum_idx = 0;
	set_inline_data(&enum_idx, &inline_data_idx, ib_wr);

	for (int i=0; i<7; i++) {//only 7 extra quanta in chunk

		//call set inline 4 times as 4 words per quanta
		for (int j=0; j<min_wqe_words(); j++)
			set_inline_data(&enum_idx, &inline_data_idx, ib_wr);
	}

}

void QP_SEND::set_polarity(u8 *polarity, bool flip)
{
	u64 mask_poll = 0x7fffffffffffffff;

	fields[QPSQ_SEND_FIELD_NOP_0_3].data &= mask_poll;
	fields[QPSQ_SEND_FIELD_NOP_0_3].data |= (u64) *polarity << 63;
	fields[QPSQ_SEND_FIELD_NOP_1_3].data &= mask_poll;
	fields[QPSQ_SEND_FIELD_NOP_1_3].data |= (u64) *polarity << 63;
	fields[QPSQ_SEND_FIELD_NOP_2_3].data &= mask_poll;
	fields[QPSQ_SEND_FIELD_NOP_2_3].data |= (u64) *polarity << 63;
	fields[QPSQ_SEND_FIELD_NOP_3_3].data &= mask_poll;
	fields[QPSQ_SEND_FIELD_NOP_3_3].data |= (u64) *polarity << 63;
	fields[QPSQ_SEND_FIELD_NOP_4_3].data &= mask_poll;
	fields[QPSQ_SEND_FIELD_NOP_4_3].data |= (u64) *polarity << 63;
	fields[QPSQ_SEND_FIELD_NOP_5_3].data &= mask_poll;
	fields[QPSQ_SEND_FIELD_NOP_5_3].data |= (u64) *polarity << 63;
	fields[QPSQ_SEND_FIELD_NOP_6_3].data &= mask_poll;
	fields[QPSQ_SEND_FIELD_NOP_6_3].data |= (u64) *polarity << 63;
	fields[QPSQ_SEND_FIELD_NOP_7_3].data &= mask_poll;
	fields[QPSQ_SEND_FIELD_NOP_7_3].data |= (u64) *polarity << 63;

	if (flip)
		if (*polarity)
			*polarity = 0;
		else
			*polarity = 1;
	fields[QPSQ_SEND_FIELD_WQE_VALID].data       = *polarity;
	fields[QPSQ_SEND_FIELD_FRAG_0_VALID].data    = *polarity;
	fields[QPSQ_SEND_FIELD_FRAG_1_VALID].data    = *polarity;
	fields[QPSQ_SEND_FIELD_FRAG_2_VALID].data    = *polarity;
	fields[QPSQ_SEND_FIELD_FRAG_3_VALID].data    = *polarity;
	fields[QPSQ_SEND_FIELD_FRAG_4_VALID].data    = *polarity;
	fields[QPSQ_SEND_FIELD_FRAG_5_VALID].data    = *polarity;
	fields[QPSQ_SEND_FIELD_FRAG_6_VALID].data    = *polarity;
	fields[QPSQ_SEND_FIELD_FRAG_7_VALID].data    = *polarity;
	fields[QPSQ_SEND_FIELD_FRAG_8_VALID].data    = *polarity;
	fields[QPSQ_SEND_FIELD_FRAG_9_VALID].data    = *polarity;
	fields[QPSQ_SEND_FIELD_FRAG_10_VALID].data    = *polarity;
	fields[QPSQ_SEND_FIELD_FRAG_11_VALID].data    = *polarity;
	fields[QPSQ_SEND_FIELD_FRAG_12_VALID].data    = *polarity;
	fields[QPSQ_SEND_FIELD_FRAG_13_VALID].data    = *polarity;
	fields[QPSQ_SEND_FIELD_FRAG_14_VALID].data    = *polarity;
	fields[QPSQ_SEND_FIELD_INLINE_DATA_VALID_4].data  = *polarity;
	fields[QPSQ_SEND_FIELD_INLINE_DATA_VALID_8].data  = *polarity;
	fields[QPSQ_SEND_FIELD_INLINE_DATA_VALID_12].data = *polarity;
	fields[QPSQ_SEND_FIELD_INLINE_DATA_VALID_16].data = *polarity;
	fields[QPSQ_SEND_FIELD_INLINE_DATA_VALID_20].data = *polarity;
	fields[QPSQ_SEND_FIELD_INLINE_DATA_VALID_24].data = *polarity;
	fields[QPSQ_SEND_FIELD_INLINE_DATA_VALID_28].data = *polarity;
}
