/*******************************************************************************
*
* Copyright (c) 2015-2016 Intel Corporation.  All rights reserved.
*
* This software is available to you under a choice of one of two
* licenses.  You may choose to be licensed under the terms of the GNU
* General Public License (GPL) Version 2, available from the file
* COPYING in the main directory of this source tree, or the
* OpenFabrics.org BSD license below:
*
*   Redistribution and use in source and binary forms, with or
*   without modification, are permitted provided that the following
*   conditions are met:
*
*    - Redistributions of source code must retain the above
*	copyright notice, this list of conditions and the following
*	disclaimer.
*
*    - Redistributions in binary form must reproduce the above
*	copyright notice, this list of conditions and the following
*	disclaimer in the documentation and/or other materials
*	provided with the distribution.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
* NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
* BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
* ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
* CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
*******************************************************************************/

#ifndef IRDMA_OSDEP_H
#define IRDMA_OSDEP_H

#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/bitops.h>
#include <linux/pci.h>
#include <net/tcp.h>
#include <crypto/hash.h>
/* get readq/writeq support for 32 bit kernels, use the low-first version */
#include <linux/io-64-nonatomic-lo-hi.h>
typedef unsigned long uintptr_t;

#define MAKEMASK(m, s) ((m) << (s))

#define ICRDDMA_MAX_STATS_32	0xFFFFFFFF
#define ICRDDMA_MAX_STATS_48	0xFFFFFFFFFFFF

#define IRDMA_STATS_DELTA(a, b, c) ((a) >= (b) ? (a) - (b) : (a) + (c) - (b))

#define STATS_TIMER_DELAY 10000

struct irdma_dma_mem {
	void *va;
	dma_addr_t pa;
	u32 size;
} __packed;

struct irdma_virt_mem {
	void *va;
	u32 size;
} __packed;

struct irdma_dma_info {
        dma_addr_t *dmaaddrs;
};

#define to_device(ptr)						\
	(&((struct pci_dev *)((ptr)->hw->dev_context))->dev)

#ifdef CONFIG_DYNAMIC_DEBUG
#define irdma_debug(dev, prefix, ...)
#define irdma_debug_buf(dev, prefix, desc, buf, size)
#else
#define irdma_debug(dev, mask, fmt, ...)			\
do {                                                            \
} while (0)
#define irdma_debug_buf(dev, mask, desc, buf, size)		\
do {                                                            \
} while (0)
#endif

#define irdma_flush(a)          readl((a)->hw_addr + I40E_GLGEN_STAT)
#ifdef IRDMA_FPGA_SUPPORT
#else
#endif

struct irdma_sc_dev;
struct irdma_sc_qp;
struct irdma_puda_buf;
struct irdma_puda_completion_info;
struct irdma_update_sds_info;
struct irdma_hmc_fcn_info;
struct irdma_manage_vf_pble_info;
struct irdma_device;
struct irdma_hmc_info;
struct irdma_hw;
#ifdef VCHNL_SUPPORT
struct irdma_virtchnl_req;
#endif

void irdma_ieq_mpa_crc_ae(struct irdma_sc_dev *dev, struct irdma_sc_qp *qp);
bool irdma_vf_clear_to_send(struct irdma_sc_dev *dev);
void irdma_add_dev_ref(struct irdma_sc_dev *dev);
void irdma_put_dev_ref(struct irdma_sc_dev *dev);
enum irdma_status_code irdma_ieq_check_mpacrc(struct shash_desc *desc, void *addr,
					      u32 length, u32 value);
struct irdma_sc_qp *irdma_ieq_get_qp(struct irdma_sc_dev *dev, struct irdma_puda_buf *buf);
void irdma_ieq_update_tcpip_info(struct irdma_puda_buf *buf, u16 length, u32 seqnum);
void irdma_free_hash_desc(struct shash_desc *hash_desc);
enum irdma_status_code irdma_init_hash_desc(struct shash_desc **hash_desc);
enum irdma_status_code irdma_puda_get_tcpip_info(struct irdma_puda_completion_info *info,
						 struct irdma_puda_buf *buf);
enum irdma_status_code irdma_cqp_sds_cmd(struct irdma_sc_dev *dev,
					 struct irdma_update_sds_info *info);
enum irdma_status_code irdma_cqp_manage_hmc_fcn_cmd(struct irdma_sc_dev *dev,
						    struct irdma_hmc_fcn_info *hmcfcninfo,
						    u16 *pmf_idx);
enum irdma_status_code irdma_cqp_query_fpm_values_cmd(struct irdma_sc_dev *dev,
						      struct irdma_dma_mem *values_mem,
						      u8 hmc_fn_id);
enum irdma_status_code irdma_cqp_commit_fpm_values_cmd(struct irdma_sc_dev *dev,
						       struct irdma_dma_mem *values_mem,
						       u8 hmc_fn_id);
enum irdma_status_code irdma_alloc_query_fpm_buf(struct irdma_sc_dev *dev,
						 struct irdma_dma_mem *mem);
enum irdma_status_code irdma_cqp_manage_vf_pble_bp(struct irdma_sc_dev *dev,
						   struct irdma_manage_vf_pble_info *info);
void *irdma_remove_head(struct list_head *list);

void irdma_term_modify_qp(struct irdma_sc_qp *qp, u8 next_state, u8 term, u8 term_len);
void irdma_terminate_done(struct irdma_sc_qp *qp, int timeout_occurred);
void irdma_terminate_start_timer(struct irdma_sc_qp *qp);
void irdma_terminate_del_timer(struct irdma_sc_qp *qp);
enum irdma_status_code irdma_hw_manage_vf_pble_bp(struct irdma_device *iwdev,
						  struct irdma_manage_vf_pble_info *info,
						  bool wait);
struct irdma_sc_vsi;
void irdma_hw_stats_start_timer(struct irdma_sc_vsi *vsi);
void irdma_hw_stats_stop_timer(struct irdma_sc_vsi *vsi);
#define irdma_mmiowb() mmiowb()
void irdma_wr32(struct irdma_hw *hw, u32 reg, u32 value);
u32  irdma_rd32(struct irdma_hw *hw, u32 reg);
#ifdef VCHNL_SUPPORT
enum irdma_status_code irdma_vf_wait_vchnl_resp(struct irdma_sc_dev *dev,
						struct irdma_virtchnl_req *vchnl_req);
enum irdma_status_code irdma_vchnl_vf_get_resp(struct irdma_sc_dev *dev,
					       struct irdma_virtchnl_req *vchnl_req);
#endif
#endif	/* _IRDMA_OSDEP_H_ */
