#!/usr/bin/ruby

module C_KEYWORDS

	CNTL = {
		:for => '\bfor\b',
		:while => '\bwhile\b',
		:switch => '\bswitch\b',
		:return => '\breturn\b',
		:semi => ';',
	}

	BLOCK = {
		:open_p => '\(',
		:close_p => '\)',
		:open_b => '\{',
		:close_b => '\}',
	}

	TYPES = {
		:struct => '\bstruct\b',
		:enum => '\benum\b',
		:int => '\bint\b',
		:long => '\blong\b',
		:short => '\bshort\b',
		:unsigned => '\bunsigned\b',
		:u64 => '\bu64\b',
		:u32 => '\bu32\b',
		:u16 => '\bu16\b',
		:u8 => '\bu8\b',
		:void => '\bvoid\b',
		:char => '\bchar\b',
	}

	PREPROC = {
		:include => '^#include\b',
		:define => '^#define\b',
		:macro_cont => '\/\s*$',
		:ifdef => '^#ifdef\s+',
		:ifndef => '^#ifndef\s+',
		:if => '^#if\s+',
		:else => '^#else\s*',
		:endif => '^#endif\s*',
	}

	COMMENT = {
		:line => '\/\/',
		:open_b => '\/\*',
		:close_b => '\*\/',
	}

	PNTR = {
		:arrow => '\->',
		:star => '\*',
	}


end #C_KEYWORDS

module FUNC_TABLES

	OPS = {
		#irdma_qp_uk_ops
		"iw_qp_post_wr" => "irdma_qp_post_wr",
		"iw_qp_ring_push_db" => "irdma_qp_ring_push_db",
		"iw_rdma_write" => "irdma_rdma_write",
		"iw_rdma_read" => "irdma_rdma_read",
		"iw_send" => "irdma_send",
		"iw_inline_rdma_write" => "irdma_inline_rdma_write",
		"iw_inline_send" => "irdma_inline_send",
		"iw_stag_local_invalidate" => "irdma_stag_local_invalidate",
		"iw_mw_bind" => "irdma_mw_bind",
		"iw_post_receive" => "irdma_post_recieve",
		"iw_post_nop" => "irdma_nop",
		"iw_set_fragment" => "irdma_set_fragment",

		#irdma_cq_ops
		"iw_cq_request_notification" => "irdma_cq_request_notification",
		"iw_cq_poll_cmpl" => "irdma_cq_poll_cmpl",
		"iw_cq_post_entries" => "irdma_cq_post_entries",
		"iw_cq_clean" => "irdma_clean_cq",

		#irdma_device_uk_ops
		"iw_cq_uk_init" => "irdma_cq_uk_init",
		"iw_qp_uk_init" => "irdma_qp_uk_init",

		#irdma_uda_ops
		"init_ah" => "irdma_sc_init_ah",
		"create_ah" => "irdma_sc_create_ah",
		"modify_ah" => "irdma_sc_modify_ah",
		"destroy_ah" => "irdma_sc_destroy_ah",
		"manage_quad_hash_table_entry" => "irdma_manage_qhash_table_info",
		"mcast_grp_create" => "irdma_sc_create_mcast_grp",
		"mcast_grp_modify" => "irdma_sc_modify_mcast_grp",
		"mcast_grp_destroy" => "irdma_sc_destroy_mcast_grp",
		"mcast_grp_add" => "irdma_sc_add_mcast_grp",
		"mcast_grp_del" => "irdma_sc_del_mcast_grp",

		#irdma_irq_ops
		"irdma_en_irq" => "irdma_ena_irq",
		"irdma_dis_irq" => "irdma_disable_irq",
		"irdma_cfg_aeq" => "irdma_cfg_aeq",
		"irdma_cfg_ceq" => "irdma_cfg_ceq",

		#irdma_cqp_ops
		"cqp_init" => "irdma_sc_cqp_init",
		"cqp_create" => "irdma_sc_cqp_create",
		"cqp_post_sq" => "irdma_sc_cqp_post_sq",
		"cqp_get_next_send_wqe" => "irdma_sc_cqp_get_next_send_wqe",
		"cqp_destroy" => "irdma_sc_cqp_destroy",
		"poll_for_cqp_op_done" => "irdma_sc_poll_for_cqp_op_done",
		"check_cqp_progress" => "irdma_check_cqp_progress",

		#irdma_ccq_ops
		"ccq_init" => "irdma_sc_ccq_init",
		"ccq_create" => "irdma_sc_ccq_create",
		"ccq_destroy" => "irdma_sc_ccq_destroy",
		"ccq_create_done" => "irdma_sc_ccq_create_done",
		"ccq_get_cqe_info" => "irdma_sc_ccq_get_cqe_info",
		"ccq_arm" => "irdma_sc_ccq_arm",

		#irdma_ceq_ops
		"ceq_init" => "irdma_sc_ceq_init",
		"ceq_create" => "irdma_sc_ceq_create",
		"cceq_create_done" => "irdma_sc_cceq_create_done",
		"cceq_destroy_done" => "irdma_sc_cceq_destroy_done",
		"cceq_create" => "irdma_sc_ceq_cceq_create",
		"ceq_destroy" => "irdma_sc_ceq_destroy",
		"process_ceq" => "irdma_sc_process_ceq",

		#irdma_aeq_ops
		"aeq_init" => "irdma_sc_aeq_init",
		"aeq_create" => "irdma_sc_aeq_create",
		"aeq_destroy" => "irdma_sc_aeq_destroy",
		"get_next_aeqe" => "irdma_sc_get_next_aeqe",
		"repost_aeq_entries" => "irdma_sc_repost_aeq_entries",
		"aeq_create_done" => "irdma_sc_aeq_create_done",
		"aeq_destroy_done" => "irdma_sc_aeq_destroy_done",

		#irdma_pd_ops
		"pd_init" => "irdma_sc_pd_init",

		#irdma_priv_qp_ops
		"qp_init" => "irdma_sc_qp_init",
		"qp_create" => "irdma_sc_qp_create",
		"qp_modify" => "irdma_sc_qp_modify",
		"qp_destroy" => "irdma_sc_qp_destroy",
		"qp_flush_wqes" => "irdma_sc_qp_flush_wqes",
		"qp_upload_context" => "irdma_sc_qp_upload_context",
		"qp_setctx" => "irdma_sc_qp_setctx",
		"qp_setctx_roce" => "irdma_sc_qp_setctx_roce",
		"qp_send_lsmm" => "irdma_sc_send_lsmm",
		"qp_send_lsmm_nostag" => "irdma_sc_send_lsmm_nostag",
		"qp_send_rtt" => "irdma_sc_send_rtt",
		"iw_mr_fast_register" => "irdma_sc_mr_fast_register",

		#irdma_priv_cq_ops
		"cq_init" => "irdma_sc_cq_init",
		"cq_create" => "irdma_sc_cq_create",
		"cq_destroy" => "irdma_sc_cq_destroy",
		"cq_modify" => "irdma_sc_cq_modify",
		"cq_ack" => "irdma_sc_cq_ack",

		#irdma_mr_ops
		"alloc_stag" => "irdma_sc_alloc_stag",
		"mr_reg_non_shared" => "irdma_sc_mr_reg_non_shared",
		"mr_reg_shared" => "irdma_sc_mr_reg_shared",
		"dealloc_stag" => "irdma_sc_dealloc_stag",
		"query_stag" => "irdma_sc_query_stag",
		"mw_alloc" => "irdma_sc_mw_alloc",

		#irdma_cqp_misc_ops
		"manage_push_page" => "irdma_sc_manage_push_page",
		"manage_hmc_pm_func_table" => "irdma_sc_manage_hmc_pm_func_table",
		"commit_fpm_val" => "irdma_sc_commit_fpm_val",
		"query_fpm_val" => "irdma_sc_query_fpm_val",
		"static_hmc_pages_allocated" => "irdma_sc_static_hmc_pages_allocated",
		"add_arp_cache_entry" => "irdma_sc_add_arp_cache_entry",
		"del_arp_cache_entry" => "irdma_sc_del_arp_cache_entry",
		"query_arp_cache_entry" => "irdma_sc_query_arp_cache_entry",
		"manage_apbvt_entry" => "irdma_sc_manage_apbvt_entry",
		"manage_qhash_table_entry" => "irdma_sc_manage_qhash_table_entry",
		"cqp_nop" => "irdma_sc_cqp_nop",
		"commit_fpm_val_done" => "irdma_sc_commit_fpm_val_done",
		"query_fpm_val_done" => "irdma_sc_query_fpm_val_done",
		"manage_hmc_pm_func_table_done" => "irdma_sc_manage_hmc_pm_func_table_done",
		"update_suspend_qp" => "irdma_sc_suspend_qp",
		"update_resume_qp" => "irdma_sc_resume_qp",
		"manage_stats_instance" => "irdma_sc_manage_stats_inst",
		"gather_stats" => "irdma_sc_gather_stats",
		"manage_ws_node" => "irdma_sc_manage_ws_node",
		"set_up_map" => "irdma_sc_set_up_map",
		"query_rdma_features" => "irdma_sc_query_rdma_features",

		#irdma_hmc_ops
		"init_iw_hmc" => "irdma_sc_init_iw_hmc",
		"parse_fpm_query_buf" => "irdma_sc_parse_fpm_query_buf",
		"cfg_iw_fpm" => "irdma_sc_cfg_iw_fpm",
		"parse_fpm_commit_buf" => "irdma_sc_parse_fpm_commit_buf",
		"create_hmc_object" => "irdma_sc_create_hmc_obj",
		"del_hmc_object" => "irdma_sc_del_hmc_obj",

	}

end #FUNC_TABLES

class FSM
	attr_accessor :name, :pstate, :cstate, :nstate, :state_list, :debug, :fh, :function, :c_function, :func_name, :funcs, :c_funcs

	def setup(fh, name="fsm", debug=false)
		puts "setup: file_handle= #{fh}" unless !debug
		@fh = File.open(fh, "r")
		@name = name
		@debug = debug
		@pstate = @cstate = @nstate = :get_line
		@line = String.new
		@final_state = :done_parsing
		@tmp_lines = Array.new
		@tmp_c_lines = Array.new
		@function = Array.new
		@c_function = Array.new
		@func_name = String.new
		@funcs = Hash.new
		@c_funcs = Hash.new
	end #setup

	def run
		while @cstate != @final_state do
			run_step
			change_state
		end
	end #run

	def run_step
		case @cstate
			when :get_line then
				#will assume that file is compile friendly and does not have
				#EOF in the middle of a function definition
				if @line = @fh.gets
					puts ":get_line #{@line}" unless !@debug
					@tokens = @line.split(/\b/)
					if is_line_comment?
						@nstate = :get_line
					elsif is_start_block_comment?
						@nstate = :block_comment
					else
						@nstate = :begin_parse
					end
				else
					@nstate = :done_parsing
				end
			when :block_comment then
				@tokens = @fh.gets.split(/\b/)
				@line = @tokens.join
				if is_end_block_comment?
					@nstate = :get_line
				end
			when :begin_parse then
				remove_comments_from_tokens
				if @line =~ /^\s*$/
					@nstate = :get_line
				elsif is_preproc?
					if @line =~ /\\\s*$/
						while @line !~ /\\\s*$/
							@line = @fh.gets
						end
					end
					@nstate = :get_line
				elsif are_tokens_func_def?
					@nstate = :store_func_start
				else
					@nstate = :get_line
				end
			when :store_func_start then
				# for ProdCode class
				@tmp_lines.each do |x|
					@function << x
				end
				@tmp_lines.clear
				# for C style
				@tmp_c_lines.each do |x|
					@c_function << x
				end
				@tmp_c_lines.clear
				@nstate = :grab_func
			when :grab_func then
				@line = @fh.gets
				c_line = String.new(@line)
				@c_function << c_line;
				@tokens = @line.split(/\b/)
				# replace all function table calls with actual function call
				FUNC_TABLES::OPS.each do |key, val|
					@line =~ /(\()\S+(?:\.|->)#{key}\s*\(/
						conditional = $1
					@line.sub!(/\S+(?:\.|->)#{key}\s*\(/,"#{conditional}#{val}(")
				end
				@function << @line
				if is_end_of_func?
					@nstate = :get_line
					@funcs[@func_name] = @function
					@function = Array.new
					@c_funcs[@func_name] = @c_function
					@c_function = Array.new
				end
			when :done_parsing then
				@fh.close
				@nstate = :done_parsing
		end #case @cstate
	end #run_step

	def change_state
		puts "#{@cstate}->#{@nstate}" unless !@debug
		puts "\t#{@line}" unless !@debug
		@pstate = @cstate
		@cstate = @nstate
	end #change_state

	# assumes @line & @tokens set to current line
	def is_line_comment?
		if @line =~ /^\s*\/\//
			return true
		end
		return false
	end #is_line_comment?

	# assumes @line & @tokens set to current line
	def is_start_block_comment?
		if @line =~ /^\s*\/\*/
			return true
		end
		return false
	end #is_start_block_comment?

	# assumes @line & @tokens set to current line
	def is_end_block_comment?
		if @line =~ /\*\//
			return true
		end
		return false
	end #is_end_block_comment?

	# assumes @line is set correctly
	def is_preproc?
		return @line =~ /#{C_KEYWORDS::PREPROC[:include]}/ ||
		       @line =~ /#{C_KEYWORDS::PREPROC[:define]}/ ||
		       @line =~ /#{C_KEYWORDS::PREPROC[:if]}/ ||
		       @line =~ /#{C_KEYWORDS::PREPROC[:ifdef]}/ ||
		       @line =~ /#{C_KEYWORDS::PREPROC[:ifndef]}/ ||
		       @line =~ /#{C_KEYWORDS::PREPROC[:else]}/ ||
		       @line =~ /#{C_KEYWORDS::PREPROC[:endif]}/ 
	end #is_preproc

	# assumes @line & @tokens set to current line
	def remove_comments_from_tokens
		@line = @tokens.join
		if @line =~ /\/\//
			@tokens = @line.split(/\/\//)
			@tokens = @tokens[0].split(/\b/)
		end
	end #remove_comments_from_tokens

	# assumes @tokens set to current
	# will set @tmp_lines
	def are_tokens_func_def?
		@line = @tokens.join
		# if line has character at start of line
		# if line does not end in ';' nor '{' (statement/declaration/struct definition)
		if @line =~ /^\S/ && @line !~ /(?:;|\{)\s*$/
			# if line will continue onto next line
			# case: multiple line func start
			if @line =~ /(?:\w|,|\()\s*$/
				if @line =~ /^static\b/
					@line.sub!(/^static\b/, "virtual")
				else
					@line = "virtual " + @line
				end
				# remove 'inline'
				@line.sub!(/\binline\b/,"");
				while (@line !~ /^\{/) do
					if @line =~ /;$/
						@tmp_lines = Array.new
						return false
					end
					@tmp_lines << @line
					@tmp_c_lines << @line.sub(/^virtual /,"")
					@line = @fh.gets
				end
				@tmp_lines << @line
				@tmp_c_lines << @line.sub(/^virtual /,"")
			# handle case if function start is all on one line
			else @line =~ /\)\s*$/
				if @line =~ /^static\b/
					@line.sub!(/^static\b/, "virtual")
				else
					@line = "virtual " + @line
				end
				# remove 'inline'
				@line.sub!(/\binline\b/,"");
				@tmp_lines << @line
				@tmp_c_lines << @line.sub(/^virtual /,"")
			end
			#grab function handle
			@tmp_lines.each do |n|
				if (n =~ /\b(\w+)\s*\(/)
					@func_name = $1
				end
			end
			return true
		end
		return false
	end #are_tokens_func_def?

	# assumes @line & @tokens set to current line
	def is_end_of_func?
		if @line =~ /^\}/
			return true
		end
		return false
	end #is_end_of_func?

end #class fsm

# ---------------------- Start Script ------------------------------------


#global var will be filled in by clusters.rb
$clusters = Hash.new

# functions for parsing tb clusters and sub clusters
def get_dep_clusters(dep)
	file_array = Array.new
	file_array = $clusters[dep][:files]
	$clusters[dep][:deps].each do |f|
		file_array << get_dep_clusters(f)
	end
	return file_array.flatten
end

def get_dep_funcs(dep)
	func_array = Array.new
	func_array = $clusters[dep][:funcs]
	$clusters[dep][:deps].each do |f|
		func_array << get_dep_funcs(f)
	end
	return func_array.flatten
end

arg_cluster = ARGV[0]

#fill in global $clusters hash
load "clusters.rb"

#list of files to parse
files_to_search = Array.new

#add dep cluster files
files_to_search = get_dep_clusters(arg_cluster)

#remove duplicate files from accumulated from clusters
files_to_search.uniq!

#list of all funcs to find
funcs_list = Array.new

#add dep cluster funcs
funcs_list = get_dep_funcs(arg_cluster)

#remove duplicate funcs from accumulated from clusters
funcs_list.uniq!

#need some global storage for all funcs
all_lines_array = Array.new
fsm = FSM.new
base_path = `realpath ../../../icrdma`.chop.sub(/\/$/,"")
all_funcs_found = Hash.new
all_c_funcs_found = Hash.new

# Behavior of FSM
#	open the file for reading at the start of the file
#	need some type of state machine for controlling parsing
#		ignores comments
#		determines if a func definition or not
#		if func definition (may be multiple lines)
#			store lines of code including above line(s) in local storage
#			continue until end of function is reached
#				mv lines in local storage to global storage
#		repeat until end of file reached
#	close the file

files_to_search.each do |f|
	fsm.setup("#{base_path}/#{f}", "fsm", false)
	fsm.run
	all_funcs_found.merge!(fsm.funcs)
	all_c_funcs_found.merge!(fsm.c_funcs)
end


# output appropriate prodcode per cluster
out_file = File.open("prodcode.tmp", "w+")

funcs_list.each do |f|
	out_file.puts(all_funcs_found[f])
end

out_file.close

# output for c style
c_out_file = File.open("c_code.tmp", "w+")

funcs_list.each do |f|
	c_out_file.puts(all_c_funcs_found[f])
end

c_out_file.close
