#include "test_if.h"

void TEST_IF::randomize(void)
{
	rand_vars();
	rand_declare();
	rand_connect();
	randomize_rand_vars();
	post_randomize();
}

void TEST_IF::init(void)
{
	declare();
	connect();
}

void TEST_IF::rand_init(void)
{
	declare();
	connect();
	randomize();
}

void TEST_IF::rand_vars(void) {}
void TEST_IF::rand_declare(void) {}
void TEST_IF::rand_connect(void) {}
void TEST_IF::randomize_rand_vars(void) {}
void TEST_IF::post_randomize(void) {}

//--------------------------------------------------------------------

send_intf::send_intf(void)
{
	wr = new QP_SEND;
	constraints = new u64[SEND_MAX_CONST];
	memset(constraints, 0, SEND_MAX_CONST*sizeof(u64));
}

send_intf::~send_intf(void)
{
	delete wr;
	delete [] constraints;
}

void send_intf::set_debug(enum debug_level level)
{
	TEST_IF::set_debug(level);
	wr->set_debug(level);
}

void send_intf::declare(void)
{
	// Initialize 1st level vars and structs

	//clear local structs
	memset(&info, 0, sizeof(struct irdma_post_sq_info));
	memset(&dev, 0, sizeof(struct irdma_sc_dev));
	memset(&sc_qp, 0, sizeof(struct irdma_sc_qp));
	memset(&hw_attrs, 0, sizeof(struct irdma_hw_attrs));
	for (int i=0; i<IRDMA_MAX_REGS; i++)
		icrdma_regs[i] = 0;


	sq_array_size = default_sq_mem_size;
	initial_sq_head = 0;
	wr_size_bytes = 0;
	avail_space_in_block = 0;
	nops_size_bytes = 0;
	num_nops = 0;
	status = 0;
	mem_check = false;
	sq_array = aligned_alloc(256, sq_array_size *
				 sizeof(struct irdma_qp_quanta));
	push_wqe = aligned_alloc(256, 8 * sizeof(struct irdma_qp_quanta));
	sg_list = malloc(default_sge_list_size * sizeof(struct irdma_sge));
	sq_wrtrk_array = malloc(sq_array_size * sizeof(struct irdma_sq_uk_wr_trk_info));
	shadow_area = malloc(sizeof(u64*));
	wqe_alloc_db = malloc(sizeof(u32*));
	push_db = malloc(sizeof(u32*));

	// Clear Arrays
	for (int i=0; i < sq_array_size; i++) {
		for (int k=0; k<TB_64B_WORDS_PER_QUANTA; k++) {
			sq_array[i].elem[k] = 0ULL;
			if (i < 8)
				push_wqe[i].elem[k] = 0ULL;
		}
		sq_wrtrk_array[i].wrid = 0;
		sq_wrtrk_array[i].wr_len = 0;
		sq_wrtrk_array[i].quanta = 0;
	}
	if (debug >= DEBUG_HIGH)
		for (int i=0; i < sq_array_size; i++)
			for (int k=0; k<TB_64B_WORDS_PER_QUANTA; k++) {
				printf("sq_array[%d].elem[%d]=0x%llx\n", i, k, sq_array[i].elem[k]);
			}
	for (int j=0; j<default_sge_list_size; j++) {
		sg_list[j].tag_off = 0;
		sg_list[j].len = 0;
		sg_list[j].stag = 0;
	}
	*push_db = 0;

	// Initialize "hidden" variables
	*shadow_area = 0;
	*wqe_alloc_db = 0;
	sc_qp.qp_uk.sq_ring.size = default_sq_size;
	sc_qp.qp_uk.sq_ring.tail = 0;
	sc_qp.qp_uk.sq_ring.head = 0;
	sc_qp.qp_uk.initial_ring.size = default_sq_size;
	sc_qp.qp_uk.initial_ring.head = 0;
	sc_qp.qp_uk.initial_ring.tail = 0;
	initial_sq_head = sc_qp.qp_uk.sq_ring.head;
	sc_qp.qp_uk.max_sq_frag_cnt = TB_MAX_FRAG_COUNT;

	
}

void send_intf::connect(void)
{
	hw_attrs.max_hw_wq_size = IRDMA_QP_WQE_MAX_SIZE;
	hw_attrs.max_hw_sq_chunk = TB_MAX_HW_SQ_CHUNK_CPK;
	sc_qp.qp_uk.sq_base = sq_array;
	sc_qp.qp_uk.push_wqe = (__le64*) push_wqe;
	sc_qp.qp_uk.sq_wrtrk_array = sq_wrtrk_array;
	sc_qp.qp_uk.shadow_area = shadow_area;
	sc_qp.qp_uk.wqe_alloc_db = wqe_alloc_db;
	sc_qp.qp_uk.push_db = push_db;
	sc_qp.dev = &dev;
	sc_qp.qp_uk.hw_attrs = &hw_attrs;
	info.op.send.sg_list = sg_list;
}

/*
 * send_initf
 * Will setup send_intf to some default values such that any test of irdma_post_send
 * should run.
 * 
 * Values will be initialized to default values
 * Pointers will be set
 * Memory allocated to hold necessary structures and lists
 */

void send_intf::cleanup(void)
{
	free(sq_array);
	free(sg_list);
	free(sq_wrtrk_array);
	free(shadow_area);
	free(wqe_alloc_db);
	free(push_wqe);
	free(push_db);
}

void send_intf::rand_sq_state(void)
{
	sc_qp.qp_uk.sq_ring.head = rand() % sc_qp.qp_uk.sq_ring.size;
	sc_qp.qp_uk.sq_ring.tail = rand() % sc_qp.qp_uk.sq_ring.size;
	sc_qp.qp_uk.initial_ring.head = sc_qp.qp_uk.sq_ring.head;
	sc_qp.qp_uk.initial_ring.tail = sc_qp.qp_uk.sq_ring.tail;
	initial_sq_head = sc_qp.qp_uk.sq_ring.head;
	if (debug >= DEBUG_LOW)
		printf("TEST: head(%d), tail(%d), size(%d)\n",
		       sc_qp.qp_uk.sq_ring.head,
		       sc_qp.qp_uk.sq_ring.tail,
		       sc_qp.qp_uk.sq_ring.size);
}

void send_intf::rand_vars(void)
{
	if (constraints[SEND_RAND_SQ_STATE])
		rand_sq_state();

}

void send_intf::check_behavior(void)
{
	if (wr->fields[QPSQ_SEND_FIELD_ADDITIONAL_FRAG_COUNT].get_masked_data_64() + 1 > 14)
		RDMA_CHECK_EQUAL(IRDMA_ERR_INVALID_FRAG_COUNT, status,
				 "\tCHECK irdma_send status IRDMA_ERR_INVALID_FRAG_COUNT(-21)");
	else
		RDMA_CHECK_EQUAL(0, status, "\tCHECK irdma_send status SUCCESSFUL(0)");

	mem_check = wr->compare_wqe_words((u64*) sq_array,
					  initial_sq_head,
					  sc_qp.qp_uk.sq_ring.size,
					  32);

	RDMA_CHECK_EQUAL(true, mem_check, "\tCHECK QP SQ WR looks correct");

}

//--------------------------------------------------------------------

post_send_intf::post_send_intf(void)
{
	constraints = new u64[POST_SEND_MAX_CONST];
	memset(constraints, 0, POST_SEND_MAX_CONST*sizeof(u64));
}

post_send_intf::~post_send_intf(void)
{
	delete [] constraints;
}

//void post_send_intf::set_debug(enum debug_level level)
//{
//	TEST_IF::set_debug(level);
//}

void post_send_intf::declare(void)
{
	if (debug >= DEBUG_HIGH)
		printf("TB: %s\n", __func__);

	iwqp = new irdma_qp;
	rf = new irdma_pci_f;
	iwdev = new irdma_device;
	bad_wr = new ib_send_wr*;
	sq_array = aligned_alloc(256, default_sq_mem_size *
				 sizeof(struct irdma_qp_quanta));
	tb_sq_array = aligned_alloc(256, default_sq_mem_size *
				    sizeof(struct irdma_qp_quanta));
	push_wqe = aligned_alloc(256, 8 * sizeof(struct irdma_qp_quanta));
	sq_wrtrk_array = malloc(default_sq_mem_size * sizeof(struct irdma_sq_uk_wr_trk_info));
	shadow_area = malloc(sizeof(u64));
	wqe_alloc_db = malloc(sizeof(u32));

	//initialize values
	memset(iwqp, 0, sizeof(struct irdma_qp));
	memset(rf, 0, sizeof(struct irdma_pci_f));
	memset(iwdev, 0, sizeof(struct irdma_device));
	memset(sq_array, 0, default_sq_mem_size * sizeof(struct irdma_qp_quanta));
	memset(tb_sq_array, 0, default_sq_mem_size * sizeof(struct irdma_qp_quanta));
	memset(push_wqe, 0, 8 * sizeof(struct irdma_qp_quanta));
	memset(sq_wrtrk_array, 0, default_sq_mem_size * sizeof(struct irdma_sq_uk_wr_trk_info));
	memset(shadow_area, 0, sizeof(u64));
	memset(wqe_alloc_db, 0, sizeof(u32));
	memset(icrdma_regs, 0, IRDMA_MAX_REGS*sizeof(u64));
	
	rf->sc_dev.hw_attrs.max_hw_wq_size = IRDMA_QP_WQE_MAX_SIZE;
	rf->sc_dev.hw_attrs.max_hw_sq_chunk = TB_MAX_HW_SQ_CHUNK_CPK;
	tb_sqr.size = default_sq_size;
	tb_sqr.tail = 0;
	tb_sqr.head = 0;
	iwqp->sc_qp.qp_uk.sq_ring.size = tb_sqr.size;
	iwqp->sc_qp.qp_uk.sq_ring.tail = tb_sqr.tail;
	iwqp->sc_qp.qp_uk.sq_ring.head = tb_sqr.head;
	iwqp->sc_qp.qp_uk.initial_ring.size = default_sq_size;
	iwqp->sc_qp.qp_uk.initial_ring.head = iwqp->sc_qp.qp_uk.sq_ring.head;
	iwqp->sc_qp.qp_uk.initial_ring.tail = iwqp->sc_qp.qp_uk.sq_ring.tail;
	iwqp->sc_qp.qp_uk.max_sq_frag_cnt = TB_MAX_FRAG_COUNT;
	ret = 0;
	mem_check = false;
	initial_sq_head = tb_sqr.head;
	polarity = 0;
	check_db_ring = true;
}

void post_send_intf::connect(void)
{
	if (debug >= DEBUG_HIGH)
		printf("TB: %s\n", __func__);
	iwdev->rf = rf;
	ibqp = (struct ib_qp*) iwqp;
	iwqp->iwdev = iwdev;
	iwqp->sc_qp.qp_uk.sq_base = sq_array;
	iwqp->sc_qp.qp_uk.push_wqe = NULL;
	iwqp->sc_qp.qp_uk.push_wqe = (__le64*) push_wqe;
	iwqp->sc_qp.qp_uk.sq_wrtrk_array = sq_wrtrk_array;
	iwqp->sc_qp.qp_uk.shadow_area = shadow_area;
	iwqp->sc_qp.qp_uk.push_db = NULL;
	iwqp->sc_qp.qp_uk.hw_attrs = &rf->sc_dev.hw_attrs;
	iwqp->sc_qp.qp_uk.wqe_alloc_db = wqe_alloc_db;
}

void post_send_intf::cleanup(void)
{
	if (too_many_frags)
		delete [] too_many_frags;
	if (inline_data_too_large) {
		if (debug >= DEBUG_HIGH)
			printf("%s: deleting inline_data_too_large\n", __func__);
		delete [] inline_data_too_large;
	}
	if (sq_array)
		free(sq_array);
	if (tb_sq_array)
		free(tb_sq_array);
	if (push_wqe)
		free(push_wqe);
	if (sq_wrtrk_array)
		free(sq_wrtrk_array);
	if (shadow_area)
		free(shadow_area);
	if (wqe_alloc_db)
		free(wqe_alloc_db);
	if (ahs)
		delete [] ahs;
	if (push_db)
		free(push_db);
	if (bad_wr)
		delete bad_wr;
	if (iwqp)
		delete iwqp;
	if (rf)
		delete rf;
	if (iwdev)
		delete iwdev;

	if (mult_wr_ud)
		delete [] mult_wr_ud;

	if (mult_wr_rdma)
		delete [] mult_wr_rdma;

	if (mult_sg_list) {
		for (int i=0; i< num_wrs; i++)
			delete [] mult_sg_list[i];
		delete [] mult_sg_list;
	}

	if (inline_data) {
		for (int i=0; i<num_wrs; i++)
			delete [] inline_data[i];
		delete [] inline_data;
	}
	if (wr)
		delete [] wr;
}

enum ib_wr_opcode post_send_intf::rand_opcode(void)
{
/*      FIXME: need to support more opcodes
	u32 opcodes[10] = {IB_WR_RDMA_WRITE, IB_WR_RDMA_WRITE_WITH_IMM,
			   IB_WR_SEND, IB_WR_SEND_WITH_IMM, IB_WR_RDMA_READ,
			   IB_WR_SEND_WITH_INV, IB_WR_RDMA_READ_WITH_INV,
			   IB_WR_LOCAL_INV, IB_WR_REG_MR, IB_WR_RESERVED1}; //Add reserved for fun :)
*/
	u32 opcodes[3] = {IB_WR_SEND, IB_WR_SEND_WITH_IMM, IB_WR_SEND_WITH_INV};
	if (constraints[POST_SEND_SET_OP])
		return constraints[POST_SEND_OP_VALUE];
	else
		return opcodes[rand() % 3];
}

/*
 * post_send_intf::rand_vars
 *
 * Randomize non-dependent variables.
 */
void post_send_intf::rand_vars(void)
{
	u32 qp_types[3] = {IB_QPT_GSI, IB_QPT_RC, IB_QPT_UD};
	

	if (debug >= DEBUG_HIGH)
		printf("TB: %s\n", __func__);
	//Randomize ibqp
	
	//Randomize irdma_qp
	if (constraints[POST_SEND_NO_FLUSH])
		iwqp->flush_issued = 0;
	else
		iwqp->flush_issued = rand() % 2;
	iwqp->sig_all = rand() % 2;


	// randomize multiple WR
	if (constraints[POST_SEND_NUM_WR])
		num_wrs = constraints[POST_SEND_NUM_WR];
	else
		num_wrs = rand() % MAX_WRS;
	if (debug>=DEBUG_HIGH)
		printf("TB: num_wrs(%d)\n", num_wrs);


	// QP_TYPE & caps
	if (constraints[POST_SEND_QP_ROCE_UD]) {
		if (rand() % 2)
			iwqp->ibqp.qp_type = IB_QPT_UD;
		else
			iwqp->ibqp.qp_type = IB_QPT_GSI;
		iwqp->sc_qp.qp_uk.qp_caps = IRDMA_SEND_WITH_IMM;
	} else if (constraints[POST_SEND_QP_ROCE_RC]) {
		iwqp->ibqp.qp_type = IB_QPT_RC;
		iwqp->sc_qp.qp_uk.qp_caps = IRDMA_SEND_WITH_IMM | IRDMA_WRITE_WITH_IMM;
	} else if (constraints[POST_SEND_QP_IWARP_RC]) {
		iwqp->ibqp.qp_type = IB_QPT_RC;
		iwqp->sc_qp.qp_uk.qp_caps = IRDMA_WRITE_WITH_IMM;
	}
	iwqp->sc_qp.qp_uk.qp_id = rand();

	if (constraints[POST_SEND_NO_PUSH_MODE]) {
		iwqp->sc_qp.qp_uk.push_mode = false;
		push_mode_en = 0;
	} else {
		iwqp->sc_qp.qp_uk.push_mode = rand() % 2;
		push_mode_en = rand() % 2;
	}

	//inline_size_bytes = rand();
}

/*
 * post_send_intf::rand_declare
 *
 * declare/create random variables
 */
void post_send_intf::rand_declare(void)
{
	if (debug >= DEBUG_HIGH)
		printf("TB: %s\n", __func__);
	if (debug >= DEBUG_HIGH)
		printf("%s: num_wrs=%d\n", __func__, num_wrs);
	if (num_wrs > 0) {
		too_many_frags = new bool[num_wrs];
		inline_data_too_large = new bool[num_wrs];
		if (debug >= DEBUG_HIGH)
			printf("%s: created inline_data_too_large\n", __func__);
		mult_wr_ud = new ib_ud_wr[num_wrs];
		mult_wr_rdma = new ib_rdma_wr[num_wrs];
		mult_sg_list = new ib_sge*[num_wrs];
		inline_data = new u8*[num_wrs];
		wr = new QP_SEND[num_wrs];
		ahs = new irdma_ah[num_wrs];
		memset(ahs, 0, num_wrs*sizeof(struct irdma_ah));
		memset(too_many_frags, 0, num_wrs*sizeof(bool));
		memset(inline_data_too_large, 0, num_wrs*sizeof(bool));
		for (int i=0; i<num_wrs; i++) {
			//allocating 1 more for testing purposes
			mult_sg_list[i] = new ib_sge[NUM_SGES+1];
			inline_data[i] = new u8[TB_MAX_INLINE_LEN+1];
			memset(mult_sg_list[i], 0, (NUM_SGES+1)*sizeof(ib_sge));
			memset(inline_data[i], 0, (TB_MAX_INLINE_LEN+1)*sizeof(u8));
			if (debug >= DEBUG_HIGH)
				printf("TB: %s: debug(%d)\n", __func__, debug);
			wr[i].set_debug(debug);
		}
	}
	if (push_mode_en) {
		push_db = malloc(sizeof(u32));
		memset(push_db, 0, sizeof(u32));
	}

	//`inline_data = new u8[inline_size_bytes];
}

/*
 * post_send_intf::rand_connect
 *
 * Connect random variables
 */
void post_send_intf::rand_connect(void)
{
	ib_wr = NULL;

	if (debug >= DEBUG_HIGH)
		printf("TB: %s\n", __func__);
	if (!num_wrs)
		goto skip_wr_data;

	if (debug >= DEBUG_HIGH)
		printf("TB: %s: inline address(0x%llx)\n", __func__, &inline_data[0][0]);

	for (int j=0; j<num_wrs; j++) {
		mult_wr_ud[j].wr.sg_list = mult_sg_list[j];
		mult_wr_ud[j].ah = &ahs[j].ibah;
		mult_wr_rdma[j].wr.sg_list = mult_sg_list[j];
		if (j < num_wrs - 1) {
			mult_wr_ud[j].wr.next = &mult_wr_ud[j+1].wr;
			mult_wr_rdma[j].wr.next = &mult_wr_rdma[j+1].wr;
		}
	}
	if (iwqp->ibqp.qp_type == IB_QPT_UD ||
	    iwqp->ibqp.qp_type == IB_QPT_GSI)
		ib_wr = &mult_wr_ud[0].wr;
	else if (iwqp->ibqp.qp_type == IB_QPT_RC)
		ib_wr = &mult_wr_rdma[0].wr;
skip_wr_data:
		

	if (push_mode_en)
		iwqp->sc_qp.qp_uk.push_db = push_db;

}

/*
 * post_send_intf::rand_flags
 *
 * helper function to randomize flags for WR.
 * Returns flag value.
 */
int post_send_intf::rand_flags(void)
{
	int flags = rand();
	if (constraints[POST_SEND_NO_INLINE])
		flags &= ~IB_SEND_INLINE;
	else if (constraints[POST_SEND_INLINE])
		flags |= IB_SEND_INLINE;

	if (constraints[POST_SEND_NO_FENCE])
		flags &= ~IB_SEND_FENCE;
	else if (constraints[POST_SEND_FENCE])
		flags |= IB_SEND_FENCE;

	if (debug >= DEBUG_HIGH)
		printf("TB: %s: flags(0x%lx), inline(0x%lx)\n", __func__, flags, IB_SEND_INLINE);
	return flags;
}

/*
 * post_send_intf::setup_ib_wr
 *
 * @ib_wr:    pointer to current WR to setup
 * @next:     pointer to next WR if valid
 * @sg:       pointer to current sg list to associate with current WR
 * @inl_data: pointer to current inline data to associate with current WR
 *
 * Helper function to randomly setup struct ib_send_wr.
 * Since ib_send_wr is a common structure between UD and RC WRs, the setting
 * of this structure was functionalized.
 * Also sets some TB error flags for checking later
 */
void post_send_intf::setup_ib_wr(struct ib_send_wr *ib_wr, struct ib_send_wr *next,
				 struct ib_sge *sg, u8 *inl_data)
{
	int max_frags;

	ib_wr->wr_id = rand() << 32 | rand();
	if (!constraints[POST_SEND_ZERO_LEN]) {
		int num_sges = NUM_SGES;
		if (constraints[POST_SEND_NUM_SGES])
			num_sges = constraints[POST_SEND_NUM_SGES];
		ib_wr->num_sge = (rand() % num_sges)+1; //push a little more so that we get some fails
	} else {
		ib_wr->num_sge = 0;
	}
	if (constraints[POST_SEND_AVOID_ZERO_LEN])
		ib_wr->num_sge |= 1;
	ib_wr->opcode = rand_opcode();//opcodes[rand() % 10];
	ib_wr->send_flags = rand_flags();
	ib_wr->ex.imm_data = rand();

	ib_wr->next = next;

	if (ib_wr->send_flags & IB_SEND_INLINE) {
		ib_wr->num_sge = 0;
		sg[0].addr = &inl_data[0];

		if (!constraints[POST_SEND_ZERO_LEN])
			sg[0].length = rand() % (TB_MAX_INLINE_LEN + 1);
		else
			sg[0].length = 0;

		if (constraints[POST_SEND_AVOID_ZERO_LEN])
			sg[0].length |= 1;

		for (int i=0; i<(TB_MAX_INLINE_LEN+1); i++) {
			inl_data[i] = rand();
			if (debug>=DEBUG_HIGH)
				printf("inline_data[%d]=(0x%llx)\n", i, inline_data[i]);
		}
	} else {
		for (int i=0; i<ib_wr->num_sge; i++) {
			sg[i].addr = (u64) rand() << 32 | rand();
			if (!constraints[POST_SEND_ZERO_LEN])
				if (constraints[POST_SEND_ERRATA_MSG_SIZE])
					sg[i].length = rand() % (TB_MAX_MSG_SIZE/14);
				else if (constraints[POST_SEND_TEST_FIX_MSG_SIZE])
					sg[i].length = TB_MAX_MSG_SIZE + 1;
				else
					sg[i].length = rand(); //TODO: check this for SGEs?
			else
				sg[i].length = 0;
			if (constraints[POST_SEND_AVOID_ZERO_LEN])
				sg[i].length |= 1;
			sg[i].lkey = rand();
		}
	}
}

/*
 * post_send_intf::randomize_rand_vars
 *
 * Randomize variables which are dependent on other random variables.
 */
void post_send_intf::randomize_rand_vars(void)
{
	int max_frags;

	if (debug >= DEBUG_HIGH)
		printf("TB: %s: num_wrs(%d)\n", __func__, num_wrs);
	if (!num_wrs)
		return;

	for (int j=0; j<num_wrs; j++) {
		struct ib_send_wr *send_wr;
		struct ib_send_wr *next_wr = NULL;

		if (debug >= DEBUG_HIGH)
			printf("TB:%s: setting ib_wr num(%d)\n", __func__, j);

		if (iwqp->ibqp.qp_type == IB_QPT_UD ||
		    iwqp->ibqp.qp_type == IB_QPT_GSI) {
			send_wr = &mult_wr_ud[j].wr;
			if (j < num_wrs - 1)
				next_wr = &mult_wr_ud[j+1].wr;
		} else if (iwqp->ibqp.qp_type == IB_QPT_RC) {
			send_wr = &mult_wr_rdma[j].wr;
			if (j < num_wrs - 1)
				next_wr = &mult_wr_rdma[j+1].wr;
		} else
			printf("ERROR! qp_type(0x%llx) not supported!\n",
			       iwqp->ibqp.qp_type);

		setup_ib_wr(send_wr, next_wr, mult_sg_list[j], inline_data[j]);

		//set err flags
		if ((ib_wr->opcode == IB_WR_SEND_WITH_IMM) ||
		    (ib_wr->opcode == IB_WR_RDMA_WRITE_WITH_IMM))
			max_frags = TB_MAX_FRAG_COUNT - 1;
		else
			max_frags = TB_MAX_FRAG_COUNT;
		if (send_wr->num_sge > max_frags)
			too_many_frags[j] = true;

		if (mult_sg_list[j][0].length > TB_MAX_INLINE_LEN)
			inline_data_too_large[j] |= true;
	}
}

/*
 * post_send_intf::re_rand
 *
 * Used if a test wants to send multiple groups of WRs.
 * Call this to re-randomize a group of WRs to be sent
 * via irdma_post_send.
 *
 * This deletes previous classes which will be re-randomized.
 * Since not all classes need to be torn down or re-randomized,
 * this will happen "selectively".
 *
 * Will re-call:
 *   rand_connect - connect "new" vars
 *   randomize_rand_vars - randomize dependent vars
 *   post_randomize - fill QP_SEND WRs for TB
 */
void post_send_intf::re_rand(void)
{
	//delete old wr, sg_list and inline data
	if (too_many_frags)
		delete [] too_many_frags;
	if (inline_data_too_large) {
		if (debug >= DEBUG_HIGH)
			printf("%s: deleting inline_data_too_large\n", __func__);
		delete [] inline_data_too_large;
	}
	if (mult_wr_ud)
		delete [] mult_wr_ud;

	if (mult_wr_rdma)
		delete [] mult_wr_rdma;

	if (mult_sg_list) {
		for (int i=0; i< num_wrs; i++)
			delete [] mult_sg_list[i];
		delete [] mult_sg_list;
	}

	if (inline_data) {
		for (int i=0; i<num_wrs; i++)
			delete [] inline_data[i];
		delete [] inline_data;
	}
	if (wr)
		delete [] wr;

	//Randomize irdma_qp
	if (constraints[POST_SEND_NO_FLUSH])
		iwqp->flush_issued = 0;
	else
		iwqp->flush_issued = rand() % 2;

	iwqp->sig_all = rand() % 2;

	// randomize multiple WR
	if (constraints[POST_SEND_NUM_WR])
		num_wrs = constraints[POST_SEND_NUM_WR];
	else
		num_wrs = rand() % MAX_WRS;
	if (debug>=DEBUG_HIGH)
		printf("TB: num_wrs(%d)\n", num_wrs);

//	FIXME: doesn't handle db rings properly
//	iwqp->sc_qp.qp_uk.qp_id = rand();

	if (constraints[POST_SEND_NO_PUSH_MODE]) {
		iwqp->sc_qp.qp_uk.push_mode = false;
		push_mode_en = 0;
	} else {
		iwqp->sc_qp.qp_uk.push_mode = rand() % 2;
		push_mode_en = rand() % 2;
	}

	//inline_size_bytes = rand();

	if (num_wrs > 0) {
		too_many_frags = new bool[num_wrs];
		inline_data_too_large = new bool[num_wrs];
		if (debug >= DEBUG_HIGH)
			printf("%s: created inline_data_too_large\n", __func__);
		mult_wr_ud = new ib_ud_wr[num_wrs];
		mult_wr_rdma = new ib_rdma_wr[num_wrs];
		mult_sg_list = new ib_sge*[num_wrs];
		inline_data = new u8*[num_wrs];
		wr = new QP_SEND[num_wrs];
		memset(too_many_frags, 0, num_wrs*sizeof(bool));
		memset(inline_data_too_large, 0, num_wrs*sizeof(bool));

		for (int i=0; i<num_wrs; i++) {
			mult_sg_list[i] = new ib_sge[NUM_SGES+1];
			inline_data[i] = new u8[TB_MAX_INLINE_LEN+1];
			if (debug >= DEBUG_HIGH)
				printf("TB: %s: debug(%d)\n", __func__, debug);
			wr[i].set_debug(debug);
		}
	}

	rand_connect();
	randomize_rand_vars();

	// move initial head
	initial_sq_head = tb_sqr.head;//iwqp->sc_qp.qp_uk.sq_ring.head;
	if (debug >= DEBUG_HIGH)
		printf("TB: %s: init_sq_head(%d)\n", __func__, initial_sq_head);

	post_randomize();
}

/*
 * post_send_intf::post_randomize
 *
 * Using this function to set the QP_SEND WR based on
 * the TB's randomized structures.
 * Later, the TB will place the QP_SEND WR on the TB SQ Ring
 * to check against the actual WR placed in the SQ Ring.
 */
void post_send_intf::post_randomize(void)
{
	u32 sq_head = tb_sqr.head;
	u8 init_polarity = polarity;
	u32 next_head, prev_head = 0;
	if (debug >= DEBUG_HIGH)
		printf("TB: %s: sq_head(%d)\n", __func__, sq_head);
	for (int i=0; i<num_wrs; i++) {
		if (debug >= DEBUG_HIGH) {
			printf("TB: %s: before 1 sq_head(%d), prev_head(%d), tb_sq_head(%d)" \
			       " polarity(%d)\n", __func__,
			       iwqp->sc_qp.qp_uk.sq_ring.head, prev_head, tb_sqr.head, polarity);

			printf("TB: %s: before 2 sq_head(%d), prev_head(%d), polarity(%d)\n", __func__,
			       iwqp->sc_qp.qp_uk.sq_ring.head, prev_head, polarity);
		}

		if (iwqp->ibqp.qp_type == IB_QPT_UD || iwqp->ibqp.qp_type == IB_QPT_GSI)
			wr[i].fill_fields(&tb_sqr, iwqp, &mult_wr_ud[i].wr, polarity);
		else
			wr[i].fill_fields(&tb_sqr, iwqp, &mult_wr_rdma[i].wr, polarity);

		next_head = (tb_sqr.head + wr[i].num_quanta) % tb_sqr.size;

		if (debug >= DEBUG_HIGH) {
			printf("TB: %s: after 1 sq_head(%d), prev_head(%d), next_head(%d), polarity(%d)\n",
			       __func__, iwqp->sc_qp.qp_uk.sq_ring.head, prev_head, next_head, &polarity);
			printf("TB: %s: after 2 sq_head(%d), prev_head(%d), next_head(%d), polarity(%d)\n",
			       __func__, iwqp->sc_qp.qp_uk.sq_ring.head, prev_head, next_head, polarity);
		}

		tb_sqr.head = next_head;
	}
	tb_sqr.head = sq_head;
	polarity = init_polarity;
	if (debug >= DEBUG_HIGH)
		printf("TB: %s: tb_sq_head(%d), head(%d)\n", __func__,
		       tb_sqr.head,
		       sq_head);
}

bool post_send_intf::compare_qs(struct irdma_qp_quanta *q,
				struct irdma_qp_quanta *tbq,
				u32 size)
{
	bool same = true;
	for (int i=0; i<size; i++) {
		for (int j=0; j<4; j++) {
			bool check=true;
			check &= q[i].elem[j] == tbq[i].elem[j];
			if (debug >= DEBUG_LOW)
				if (check)
					printf("\tMEM CHECK: BYTE OFFSET(%d): act(0x%llx), exp(0x%llx)\n",
					       32*i+8*j, q[i].elem[j], tbq[i].elem[j]);
				else
					printf("\t!MEM CHECK: BYTE OFFSET(%d): act(0x%llx), exp(0x%llx)\n",
					       32*i+8*j, q[i].elem[j], tbq[i].elem[j]);
			same &= check;
		}
	}
	return same;
}

void breakpt(void)
{
	int x = 2;
}

/* 
 * check_behavior
 * Cases of Interest:
 *   1. Flush has been issued - ret should equal EINVAL (-22)
 *   2. Valid number of Fragments - If Frag number is invalid, return (-22)
 *      If no immediate data, num frags <= 14
 *      If immediate data, num frags <= 13
 *   3. SQ Full - If WR goes beyond available Quanta in SQ ENOMEM should be returned.
 *   4. Check Doorbell rung if Above checks pass, else check QP ID not found.
 *       4.1 non-push mode
 *   	     Check wqe_alloc_db for QP ID
 *   	 4.2 TODO:Check push_db for QP ID
 *   5. Check SQ data - Should match expected values
 */
void post_send_intf::check_behavior(void)
{
	u32 tot_size = 0, running_size = 0;
	u32 quanta_avail = 0;
	bool no_imm_support = false;
	bool invalid_frags = false;
	bool invalid_inline = false;
	bool passed[num_wrs];
	bool all_passed = true;
	// Used for memory checking
	u32 sq_head = initial_sq_head;
	u32 wqe_words_to_check = 0;
	u32 total_processed = 0;
	bool flip_polarity_after_nop = false;
	u32 next_head;

	for (int i=0; i<num_wrs; i++)
		tot_size += wr[i].num_quanta;

	quanta_avail = (tb_sqr.size +
			tb_sqr.tail -
			initial_sq_head - 1) %
			tb_sqr.size;

	if (debug >= DEBUG_HIGH)
		printf("TB: %s tot_size_q(%d), quanta_avail(%d)\n", __func__, 
		       tot_size, quanta_avail);

	//Check conditions per wr
	for (int i=0; i<num_wrs; i++) {
		passed[i] = false;
		running_size += wr[i].num_quanta;
		if (iwqp->ibqp.qp_type == IB_QPT_UD ||
		    iwqp->ibqp.qp_type == IB_QPT_GSI) {
			if (too_many_frags[i] && !(mult_wr_ud[i].wr.send_flags & IB_SEND_INLINE))
				invalid_frags = true;
			if (inline_data_too_large[i] && (mult_wr_ud[i].wr.send_flags & IB_SEND_INLINE))
				invalid_inline = true;
			if (mult_wr_ud[i].wr.opcode == IB_WR_SEND_WITH_IMM)
				if ((wr[i].fields[QPSQ_SEND_QP_TYPE].data != QP_TYPE_ROCEV2_UD) &&
				    (wr[i].fields[QPSQ_SEND_QP_TYPE].data != QP_TYPE_ROCEV2_RC))
					no_imm_support = true;
			else if (mult_wr_ud[i].wr.opcode == IB_WR_RDMA_WRITE_WITH_IMM)
				if ((wr[i].fields[QPSQ_SEND_QP_TYPE].data != QP_TYPE_IWARP) &&
				    (wr[i].fields[QPSQ_SEND_QP_TYPE].data != QP_TYPE_ROCEV2_RC))
					no_imm_support = true;
			if (!iwqp->flush_issued)
				if (!(invalid_frags || invalid_inline || no_imm_support))
					passed[i] = true;
			if (running_size > quanta_avail || iwqp->flush_issued)
				passed[i] = false;
			all_passed &= passed[i];
			if (debug >= DEBUG_HIGH) {
				printf("TB: %s: QP & WR[%d] attributes:" \
				       "opcode(%d), no_imm_support(%d), flush(%d)" \
				       " qp_type(%d)" \
				       "\n", __func__, i, mult_wr_ud[i].wr.opcode, no_imm_support,
				       iwqp->flush_issued, wr[i].fields[QPSQ_SEND_QP_TYPE].data);
			}
		} else if (iwqp->ibqp.qp_type == IB_QPT_RC) {
			if (too_many_frags[i] && !(mult_wr_rdma[i].wr.send_flags & IB_SEND_INLINE))
				invalid_frags = true;
			if (inline_data_too_large[i] && (mult_wr_rdma[i].wr.send_flags & IB_SEND_INLINE))
				invalid_inline = true;
			if (mult_wr_rdma[i].wr.opcode == IB_WR_SEND_WITH_IMM)
				if ((wr[i].fields[QPSQ_SEND_QP_TYPE].data != QP_TYPE_ROCEV2_UD) &&
				    (wr[i].fields[QPSQ_SEND_QP_TYPE].data != QP_TYPE_ROCEV2_RC))
					no_imm_support = true;
			else if (mult_wr_rdma[i].wr.opcode == IB_WR_RDMA_WRITE_WITH_IMM)
				if ((wr[i].fields[QPSQ_SEND_QP_TYPE].data != QP_TYPE_IWARP) &&
				    (wr[i].fields[QPSQ_SEND_QP_TYPE].data != QP_TYPE_ROCEV2_RC))
					no_imm_support = true;
			if (!iwqp->flush_issued)
				if (!(invalid_frags || invalid_inline || no_imm_support))
					passed[i] = true;
			if (running_size > quanta_avail || iwqp->flush_issued)
				passed[i] = false;
			all_passed &= passed[i];
			if (debug >= DEBUG_HIGH) {
				printf("TB: %s: QP & WR[%d] attributes:" \
				       "opcode(%d), no_imm_support(%d), flush(%d)" \
				       " qp_type(%d)" \
				       "\n", __func__, i, mult_wr_rdma[i].wr.opcode, no_imm_support,
				       iwqp->flush_issued, wr[i].fields[QPSQ_SEND_QP_TYPE].data);
			}
		}
	}

	/* Check Conditions 1-3 */
	if (iwqp->flush_issued)
		RDMA_CHECK_EQUAL(-22, ret, "\tCHECK EINVAL(-22) returned if flush issued");
	else if (no_imm_support && tot_size > quanta_avail)
		RDMA_CHECK_EQUAL(ret == -12 || ret == -22, true, "\tCHECK ENOMEM(-12) or EINVAL(-22) returned if SQ Full or Send w/ IMM not supported");
	else if (no_imm_support)
		RDMA_CHECK_EQUAL(-22, ret, "\tCHECK EINVAL(-22) returned if Send w/ IMM not supported");
	else if (invalid_frags && tot_size > quanta_avail)
		RDMA_CHECK_NOT_EQUAL(0, ret, "\tCHECK error(!0) returned, frags or SQ Full err");
	else if (invalid_frags)
		RDMA_CHECK_EQUAL(-22, ret, "\tCHECK EINVAL(-22) returned if too many frags");
	else if (invalid_inline && tot_size > quanta_avail)
		RDMA_CHECK_NOT_EQUAL(0, ret, "\tCHECK error(!0) returned, frags or SQ Full err");
	else if (invalid_inline)
		RDMA_CHECK_EQUAL(-22, ret, "\tCHECK EINVAL(-22) returned if too much inline data");
	else if (tot_size > quanta_avail)
		RDMA_CHECK_EQUAL(-12, ret, "\tCHECK ENOMEM(-12) returned if SQ Full");
	else //passing case
		RDMA_CHECK_EQUAL(0, ret, "\tCHECK all WR posted successfully(0)");

	if (debug >= DEBUG_HIGH)
		if (iwqp->sc_qp.qp_uk.push_db == NULL)
			printf("TB: passed(%d), qp_id(0x%llx), db(0x%llx), push_db(0x%llx)\n",
			       all_passed, iwqp->sc_qp.qp_uk.qp_id,
			       *iwqp->sc_qp.qp_uk.wqe_alloc_db,
			       0);
		else
			printf("TB: passed(%d), qp_id(0x%llx), db(0x%llx), push_db(0x%llx)\n",
			       all_passed, iwqp->sc_qp.qp_uk.qp_id,
			       *iwqp->sc_qp.qp_uk.wqe_alloc_db,
			       *iwqp->sc_qp.qp_uk.push_db);

	//TODO: Check that these are not mutually exclusive!
	if (!check_db_ring)
		goto check_mem;

	if (passed[0] == true) {
		if (iwqp->sc_qp.qp_uk.push_db != NULL) {
			RDMA_CHECK_EQUAL(true, 
					 (iwqp->sc_qp.qp_uk.qp_id ==
					 *iwqp->sc_qp.qp_uk.wqe_alloc_db) ||
					 (iwqp->sc_qp.qp_uk.qp_id ==
					 0xfffff & *iwqp->sc_qp.qp_uk.push_db),
					 "\tCHECK push doorbell rung");
		} else {
			RDMA_CHECK_EQUAL(true, 
					 (iwqp->sc_qp.qp_uk.qp_id ==
					 *iwqp->sc_qp.qp_uk.wqe_alloc_db),
					 "\tCHECK wqe doorbell rung");

		}
		check_db_ring = false;
	} else {
		if (iwqp->sc_qp.qp_uk.push_db != NULL) {
			RDMA_CHECK_EQUAL(true, 
					 (!push_db && iwqp->sc_qp.qp_uk.qp_id !=
					 *iwqp->sc_qp.qp_uk.wqe_alloc_db) &&
					 (push_db && iwqp->sc_qp.qp_uk.qp_id !=
					 0xfffff & *iwqp->sc_qp.qp_uk.push_db),
					 "\tCHECK push doorbell not rung");
		 } else {
			 RDMA_CHECK_EQUAL(true, 
					  (iwqp->sc_qp.qp_uk.qp_id !=
					  *iwqp->sc_qp.qp_uk.wqe_alloc_db),
					  "\tCHECK wqe doorbell not rung");

		 }
	}
	if (debug >= DEBUG_HIGH)
		printf("After DB check\n");

check_mem:
	for (int i=0; i<num_wrs; i++) {

		if (!passed[i])
			break;

		if (debug >= DEBUG_HIGH) {
			printf("TB: %s: num_quanta(%d), sq_head(%d)\n", __func__,
			       wr[i].num_quanta, sq_head);
		}
		wqe_words_to_check = wr[i].num_quanta * wr[i].min_wqe_words();

		total_processed += wr[i].num_quanta;

		if (total_processed > quanta_avail)
			break;

		if (tb_sqr.head == 0)
			if (polarity)
				polarity = 0;
			else
				polarity = 1;

		next_head = (tb_sqr.head + wr[i].num_quanta) % tb_sqr.size;

		if (next_head) //only do this when next_head is non-zero
			flip_polarity_after_nop = next_head < tb_sqr.head;

		wr[i].set_polarity(&polarity, flip_polarity_after_nop);

		wr[i].write_wr((u64*) tb_sq_array,
			       tb_sqr.head,//sq_head,
			       tb_sqr.size,//iwqp->sc_qp.qp_uk.sq_ring.size,
			       wqe_words_to_check);

		tb_sqr.head = (tb_sqr.head + wr[i].num_quanta) % tb_sqr.size;
		//sq_head += wr[i].num_quanta;
	}
	mem_check = compare_qs(sq_array, tb_sq_array, tb_sqr.size);
		        //iwqp->sc_qp.qp_uk.sq_ring.size);
	RDMA_CHECK_EQUAL(true, mem_check, "\tCHECK QP SQ WR looks correct");
}
