#include <stdio.h>
#include <stdlib.h>
#include <cstdlib>
#include <memory>
#include <stdalign.h>
#include <string>
#include <string.h>
#include "tb_base.h"
#include "cqp_fields.h"
#include "interface.h"
#include "wr.h"

#include "CppUTest/TestHarness.h"
#include "CppUTest/CommandLineTestRunner.h"
#include "CppUTestExt/MockSupport.h"


class TEST_IF : public TB_OBJ
{
	public:
	u64 *constraints;

	virtual void declare(void) = 0;
	virtual void connect(void) = 0;
	virtual void cleanup(void) = 0;
	virtual void rand_vars(void) = 0;
	virtual void check_behavior(void) = 0;
	virtual void rand_declare(void);
	virtual void rand_connect(void);
	virtual void randomize_rand_vars(void);
	virtual void post_randomize(void);
	void randomize(void);
	void init(void);
	void rand_init(void);
};

enum send_intf_const {
	SEND_RAND_SQ_STATE,
	SEND_QP_ROCE_UD,
	SEND_QP_ROCE_RC,
	SEND_QP_IWARP_RC,
	SEND_NUM_WR,
	SEND_OP_SEND,
	SEND_OP_SEND_SE,
	SEND_OP_SEND_SEI,
	SEND_NO_IMM,
	SEND_IMM,
	SEND_NO_INLINE,
	SEND_INLINE,
	SEND_NO_PUSH_MODE,
	SEND_MAX_CONST,
};
/*
 * send_intf
 * A catch-all class to hold all structures needed when testing
 * irdma_post_send.
 * Will also initialize data to default testing values
 */
class send_intf : public TEST_IF
{
	public:

	//make these different so we can potentially see overflows
	const u32 default_sq_mem_size = 32; //actually allocated
	const u32 default_sq_size = 16; //what test will use
	const u32 default_sge_list_size = 16;

	QP_SEND *wr;
	struct irdma_sc_dev dev;
	struct irdma_post_sq_info info;
	struct irdma_sc_qp sc_qp;
	struct irdma_hw_attrs hw_attrs;
	struct irdma_qp_quanta *sq_array;
	struct irdma_qp_quanta *push_wqe;
	struct irdma_sge *sg_list;
	struct irdma_sq_uk_wr_trk_info *sq_wrtrk_array;
	u64 *shadow_area;
	u32 *wqe_alloc_db;
	u32 *push_db;
	u64 icrdma_regs[IRDMA_MAX_REGS];
	u32 initial_sq_head;
	u16 sq_array_size;
	u16 wr_size_bytes;
	u16 avail_space_in_block;
	u16 nops_size_bytes;
	u16 num_nops;
	enum irdma_status_code status;
	bool mem_check;

	send_intf(void);
	~send_intf(void);
	virtual void set_debug(enum debug_level level);
	virtual void declare(void);
	virtual void connect(void);
	virtual void cleanup(void);
	virtual void rand_sq_state(void);
	virtual void rand_vars(void);
	virtual void check_behavior(void);
};


enum post_send_intf_const {
	POST_SEND_QP_ROCE_UD,
	POST_SEND_QP_ROCE_RC,
	POST_SEND_QP_IWARP_RC,
	POST_SEND_NUM_WR,
	POST_SEND_SET_OP,
	POST_SEND_OP_VALUE,
	POST_SEND_NO_INLINE,
	POST_SEND_INLINE,
	POST_SEND_NO_PUSH_MODE,
	POST_SEND_ZERO_LEN,
	POST_SEND_AVOID_ZERO_LEN,	//error for all post_send
	POST_SEND_NO_FENCE,		//error with UD send
	POST_SEND_FENCE,
	POST_SEND_NUM_SGES,
	POST_SEND_NO_FLUSH,
	POST_SEND_ERRATA_MSG_SIZE,
	POST_SEND_TEST_FIX_MSG_SIZE,
	POST_SEND_MAX_CONST,		//must be last
};

class post_send_intf : public TEST_IF
{
	public:
	//Related to structure sizes or number of structures
	//make these different so we can potentially see overflows
	const u32 default_sq_mem_size = 32; //actually allocated
	const u32 default_sq_size = 16; //what test will use
	const u16 NUM_SGES = 14;
	const u16 MAX_WRS  = 10;
	u16 num_wrs = 1;

	//TB structures and data
	QP_SEND *wr; //alloc done in constructor
	struct irdma_qp_quanta *tb_sq_array = NULL;
	struct irdma_ring tb_sqr;
	u8 polarity;
	u16 inline_size_bytes;

	//structures necessary for CODE
	struct irdma_qp *iwqp = NULL;
	struct irdma_pci_f *rf = NULL;
	struct irdma_device *iwdev = NULL;
	struct irdma_ah *ahs = NULL;
	struct ib_qp *ibqp = NULL;
	struct ib_send_wr *ib_wr = NULL;
	struct ib_send_wr **bad_wr = NULL;
	struct ib_ud_wr *mult_wr_ud = NULL; //initialized and connected in randomize
	struct ib_rdma_wr *mult_wr_rdma = NULL; //initialized and connected in randomize
	struct ib_sge **mult_sg_list = NULL; //initialized in randomize
	u8 **inline_data = NULL;
	struct irdma_qp_quanta *sq_array = NULL;
	struct irdma_qp_quanta *push_wqe = NULL;
	struct irdma_sq_uk_wr_trk_info *sq_wrtrk_array = NULL;
	u64 *shadow_area = NULL;
	u32 *wqe_alloc_db = NULL;
	u32 *push_db = NULL; //initialized in randomize
	u32 initial_sq_head;
	u64 icrdma_regs[IRDMA_MAX_REGS];

	//Used to determine checks
	bool *too_many_frags;
	bool *inline_data_too_large;
	u8 push_mode_en;
	bool check_db_ring;

	//Used for function returns
	bool mem_check;
	int ret;

	post_send_intf(void);
	~post_send_intf(void);
	virtual void declare(void);
	virtual void connect(void);
	virtual void cleanup(void);
	virtual void rand_vars(void);
	virtual void rand_declare(void);
	virtual void rand_connect(void);
	void setup_ib_wr(struct ib_send_wr *ib_wr, struct ib_send_wr *next,
			 struct ib_sge *sg, u8 *inl_data);
	virtual void randomize_rand_vars(void);
	virtual void post_randomize(void);
	virtual void check_behavior(void);
	enum ib_wr_opcode rand_opcode(void);
	int rand_flags(void);
	virtual void re_rand(void);
	virtual bool compare_qs(struct irdma_qp_quanta*,
				struct irdma_qp_quanta*, u32);
};
