
#include <stdio.h>
#include <stdlib.h>
#include <cstdlib>
#include <memory>
#include <stdalign.h>
#include <string>
#include <string.h> //this is needed for defs of memcpy and memset
#define __packed __attribute__ ((packed))
#include "tb_base.h"
#include "wqe.h"
#include "cqp_fields.h"
#include "wr.h"

#include "CppUTest/TestHarness.h"
#include "CppUTest/CommandLineTestRunner.h"
#include "CppUTestExt/MockSupport.h"

#include "interface.h"
#include "prodcode.cpp"
#include "test_if.h"

#include "c_code.c"
// ------------------------- DOA env setup group and test ------------------

/*
 * This Group and Test is primarily for TB ENV bring up
 * Makes sure env is connected correctly in scripts
 *  and Makefile compiles everything correctly
 * */

TEST_GROUP(env_setup)
{
	ProdCode *code;
	QP_SQ_WR *qp_sq_wr;
	/* setup should only initialize things, not allocate them */
	void setup()
	{
		long seed = 0;
		timespec ts;

		//Setup Randomization
		clock_gettime(CLOCK_REALTIME, &ts);
		seed = ts.tv_nsec;
		printf("\n\tSeed = 0x%llx\n", seed);
		srand(seed);
	}

	void teardown()
	{
		mock().checkExpectations();
		mock().clear();
		if (code)
			delete code;
		if (qp_sq_wr) {
			// this is needed as class declared is QP_SQ_WR
			// that desctructor does not delete fields
			delete [] qp_sq_wr->fields;
			delete qp_sq_wr;
		}
	}
};

TEST(env_setup, test_basic_env)
{
	qp_sq_wr = new QP_SEND;
	code = new ProdCode;
	enum irdma_status_code status;

	CHECK(true);

}

// -----------------------------------------------------------------------------


// -----------------------------------------------------------------------------

TEST_GROUP(irdma_send)
{
	ProdCode *code;
	send_intf *test_intf;
	/* setup should only initialize things, not allocate them */
	void setup()
	{
		long seed = 0;
		timespec ts;

		//Setup Randomization
		clock_gettime(CLOCK_REALTIME, &ts);
		seed = ts.tv_nsec;
		printf("\n\tSeed = 0x%llx\n", seed);
		srand(seed);
	}

	void teardown()
	{
		mock().checkExpectations();
		mock().clear();
		if (code)
			delete code;
		if (test_intf) {
			test_intf->cleanup();
			delete test_intf;
		}
	}
};

/*
 * send_basic_rc
 * Basic test of irdma_send
 * Starts Queues off at 0
 * Should take up 8 Quanta
 * No NOPs should be introduced
 *
 * NOTE: this utilizes a different testing interface than rest of CORE code tests
 * In SHARD code test environment, randomization was handled through WR, rather
 * than a test interface.
 * The test interface in this case consolidates all structures and contains place
 * for results of irdma_send to be stored.
 *
 */
TEST(irdma_send, send_basic_rc)
{
	code = new ProdCode;
	test_intf = new send_intf;

	test_intf->set_debug(DEBUG_LOW);

	test_intf->init();
	//constrain randomization
	test_intf->wr->set_setup_config(NO_INLINE);
	test_intf->wr->set_setup_config(NO_IMM_DATA);
	test_intf->wr->set_setup_config(RELIABLE_CONN);

	test_intf->wr->randomize();

	//setup tb interface
	test_intf->wr->fill_rdma_send_intf(&test_intf->sc_qp, &test_intf->info,
				test_intf->sg_list);

	test_intf->status = code->irdma_send(&test_intf->sc_qp.qp_uk,
					     &test_intf->info,
					     true);

	test_intf->check_behavior();
}

// ----------------------------- Outstanding Bugs ------------------------------
// irdma_post_send

TEST_GROUP(bugs_irdma_post_send)
{
	ProdCode *code;
	post_send_intf *test_intf;
	/* setup should only initialize things, not allocate them */
	void setup()
	{
		long seed = 0;
		timespec ts;

		//Setup Randomization
		clock_gettime(CLOCK_REALTIME, &ts);
		seed = ts.tv_nsec;
		printf("\n\tSeed = 0x%llx\n", seed);
		srand(seed);
	}

	void teardown()
	{
		mock().checkExpectations();
		mock().clear();
		if (code)
			delete code;
		if (test_intf) {
			test_intf->cleanup();
			delete test_intf;
		}
	}
};

/*
 * post_send_iwarp_rc_doa
 * DOA test for iWARP post_send
 * Randomizes number of SGEs and various command bits
 * Checks return code, SQ data, and db rung
 *
 * SQ head and tail at 0
 * No inline data
 * No immediate data
 * Only 1 WR
 * No push mode
 * Zero len Frags
 *   this will write a valid bit to an extra Frag following
 *   the 32 Byte WR. This should not happen, but has no ill
 *   effects to the HW.
 */
TEST(bugs_irdma_post_send, post_send_iwarp_rc_doa)
{
	code = new ProdCode;
	test_intf = new post_send_intf;
	bool passed = false;
	bool work_posted = false;

	//Set debug level and initialize test interface
	test_intf->set_debug(DEBUG_LOW);

	test_intf->init();

	//Randomization Constraints
	test_intf->constraints[POST_SEND_QP_IWARP_RC] = 1;
	test_intf->constraints[POST_SEND_NUM_WR] = 1;
	test_intf->constraints[POST_SEND_SET_OP] = 1;
	test_intf->constraints[POST_SEND_OP_VALUE] = IB_WR_SEND;
	test_intf->constraints[POST_SEND_NO_INLINE] = 1;
	test_intf->constraints[POST_SEND_NO_PUSH_MODE] = 1;
	test_intf->constraints[POST_SEND_NO_FLUSH] = 1;
	test_intf->constraints[POST_SEND_ZERO_LEN] = 1;

	test_intf->randomize();

	if (test_intf->debug >= DEBUG_HIGH)
		printf("TEST: 1 ret(%d)\n", test_intf->ret);
	test_intf->ret = code->irdma_post_send(test_intf->ibqp,
					       test_intf->ib_wr,
					       test_intf->bad_wr);
	if (test_intf->debug >= DEBUG_HIGH)
		printf("TEST: 2 ret(%d)\n", test_intf->ret);

	test_intf->check_behavior();
}

/*
 * post_send_roce_rc_doa
 * DOA test for ROCE post_send
 * Checks return code, SQ data, and db rung
 *
 * Read Fence Error Case
 *   Should be ignored for UD
 * SQ head and tail at 0
 * No inline data
 * No immediate data
 * Only 1 WR
 * No push mode
 * No zero len Frags, avoids test mismatch with SW
 *   this will write a valid bit to an extra Frag following
 *   the 32 Byte WR. This should not happen, but has no ill
 *   effects to the HW.
 */
TEST(bugs_irdma_post_send, post_send_roce_ud_doa)
{
	code = new ProdCode;
	test_intf = new post_send_intf;
	bool passed = false;
	bool work_posted = false;

	//Set debug level and initialize test interface
	test_intf->set_debug(DEBUG_LOW);

	test_intf->init();

	//Randomization Constraints
	test_intf->constraints[POST_SEND_QP_ROCE_UD] = 1;
	test_intf->constraints[POST_SEND_NUM_WR] = 1;
	test_intf->constraints[POST_SEND_SET_OP] = 1;
	test_intf->constraints[POST_SEND_OP_VALUE] = IB_WR_SEND;
	test_intf->constraints[POST_SEND_NO_INLINE] = 1;
	test_intf->constraints[POST_SEND_NO_PUSH_MODE] = 1;
	test_intf->constraints[POST_SEND_NO_FLUSH] = 1;
	test_intf->constraints[POST_SEND_AVOID_ZERO_LEN] = 1;
	test_intf->constraints[POST_SEND_FENCE] = 1;
	test_intf->constraints[POST_SEND_NUM_SGES] = 1; //make sure we always have a valid number

	test_intf->randomize();

	if (test_intf->debug >= DEBUG_HIGH)
		printf("TEST: 1 ret(%d)\n", test_intf->ret);
	test_intf->ret = code->irdma_post_send(test_intf->ibqp,
					       test_intf->ib_wr,
					       test_intf->bad_wr);
	if (test_intf->debug >= DEBUG_HIGH)
		printf("TEST: 2 ret(%d)\n", test_intf->ret);

	test_intf->check_behavior();
}

// ----------------------------- Test irdma_post_send --------------------------


TEST_GROUP(irdma_post_send)
{
	ProdCode *code;
	post_send_intf *test_intf;
	/* setup should only initialize things, not allocate them */
	void setup()
	{
		long seed = 0;
		timespec ts;

		//Setup Randomization
		clock_gettime(CLOCK_REALTIME, &ts);
		seed = ts.tv_nsec;
		printf("\n\tSeed = 0x%llx\n", seed);
		srand(seed);
	}

	void teardown()
	{
		mock().checkExpectations();
		mock().clear();
		if (code)
			delete code;
		if (test_intf) {
			test_intf->cleanup();
			delete test_intf;
		}
	}
};

/*
 * post_send_iwarp_rc_doa
 * DOA test for iWARP post_send
 * Randomizes number of SGEs and various command bits
 * Checks return code, SQ data, and db rung
 *
 * SQ head and tail at 0
 * No inline data
 * No immediate data
 * Only 1 WR
 * No push mode
 * No zero len Frags, avoids test mismatch with SW
 *   this will write a valid bit to an extra Frag following
 *   the 32 Byte WR. This should not happen, but has no ill
 *   effects to the HW.
 */
TEST(irdma_post_send, post_send_iwarp_rc_doa)
{
	code = new ProdCode;
	test_intf = new post_send_intf;
	bool passed = false;
	bool work_posted = false;

	//Set debug level and initialize test interface
	test_intf->set_debug(DEBUG_LOW);

	test_intf->init();

	//Randomization Constraints
	test_intf->constraints[POST_SEND_QP_IWARP_RC] = 1;
	test_intf->constraints[POST_SEND_NUM_WR] = 1;
	test_intf->constraints[POST_SEND_SET_OP] = 1;
	test_intf->constraints[POST_SEND_OP_VALUE] = IB_WR_SEND;
	test_intf->constraints[POST_SEND_NO_INLINE] = 1;
	test_intf->constraints[POST_SEND_NO_PUSH_MODE] = 1;
	test_intf->constraints[POST_SEND_AVOID_ZERO_LEN] = 1;

	test_intf->randomize();

	if (test_intf->debug >= DEBUG_HIGH)
		printf("TEST: 1 ret(%d)\n", test_intf->ret);
	test_intf->ret = code->irdma_post_send(test_intf->ibqp,
					       test_intf->ib_wr,
					       test_intf->bad_wr);
	if (test_intf->debug >= DEBUG_HIGH)
		printf("TEST: 2 ret(%d)\n", test_intf->ret);

	test_intf->check_behavior();
}

TEST(irdma_post_send, post_send_iwarp_rc_sge_imm)
{
	code = new ProdCode;
	test_intf = new post_send_intf;
	bool passed = false;
	bool work_posted = false;

	//Set debug level and initialize test interface
	test_intf->set_debug(DEBUG_LOW);

	test_intf->init();

	//Randomization Constraints
	test_intf->constraints[POST_SEND_QP_IWARP_RC] = 1;
	test_intf->constraints[POST_SEND_NUM_WR] = 1;
	test_intf->constraints[POST_SEND_SET_OP] = 1;
	test_intf->constraints[POST_SEND_OP_VALUE] = IB_WR_SEND_WITH_IMM;
	test_intf->constraints[POST_SEND_NO_INLINE] = 1;
	test_intf->constraints[POST_SEND_NO_PUSH_MODE] = 1;
	test_intf->constraints[POST_SEND_AVOID_ZERO_LEN] = 1;

	test_intf->randomize();

	if (test_intf->debug >= DEBUG_HIGH)
		printf("TEST: 1 ret(%d)\n", test_intf->ret);
	test_intf->ret = code->irdma_post_send(test_intf->ibqp,
					       test_intf->ib_wr,
					       test_intf->bad_wr);
	if (test_intf->debug >= DEBUG_HIGH)
		printf("TEST: 2 ret(%d)\n", test_intf->ret);

	test_intf->check_behavior();
}

TEST(irdma_post_send, post_send_iwarp_rc_write)
{
	code = new ProdCode;
	test_intf = new post_send_intf;
	bool passed = false;
	bool work_posted = false;

	//Set debug level and initialize test interface
	test_intf->set_debug(DEBUG_HIGH);

	test_intf->init();

	//Randomization Constraints
	test_intf->constraints[POST_SEND_QP_IWARP_RC] = 1;
	test_intf->constraints[POST_SEND_NUM_WR] = 1;
	test_intf->constraints[POST_SEND_SET_OP] = 1;
	test_intf->constraints[POST_SEND_OP_VALUE] = IB_WR_RDMA_WRITE;
	test_intf->constraints[POST_SEND_NO_INLINE] = 1;
	test_intf->constraints[POST_SEND_NO_PUSH_MODE] = 1;
	test_intf->constraints[POST_SEND_AVOID_ZERO_LEN] = 1;
	test_intf->constraints[POST_SEND_ERRATA_MSG_SIZE] = 1;

	test_intf->randomize();

	if (test_intf->debug >= DEBUG_HIGH)
		printf("TEST: 1 ret(%d)\n", test_intf->ret);
	test_intf->ret = code->irdma_post_send(test_intf->ibqp,
					       test_intf->ib_wr,
					       test_intf->bad_wr);
	if (test_intf->debug >= DEBUG_HIGH)
		printf("TEST: 2 ret(%d)\n", test_intf->ret);

	test_intf->check_behavior();
}

TEST(irdma_post_send, post_send_roce_rc_sge_imm)
{
	code = new ProdCode;
	test_intf = new post_send_intf;
	bool passed = false;
	bool work_posted = false;

	//Set debug level and initialize test interface
	test_intf->set_debug(DEBUG_HIGH);

	test_intf->init();

	//Randomization Constraints
	test_intf->constraints[POST_SEND_QP_ROCE_RC] = 1;
	test_intf->constraints[POST_SEND_NUM_WR] = 1;
	test_intf->constraints[POST_SEND_SET_OP] = 1;
	test_intf->constraints[POST_SEND_OP_VALUE] = IB_WR_SEND_WITH_IMM;
	test_intf->constraints[POST_SEND_NO_INLINE] = 1;
	test_intf->constraints[POST_SEND_NO_PUSH_MODE] = 1;
	test_intf->constraints[POST_SEND_AVOID_ZERO_LEN] = 1;

	test_intf->randomize();

	if (test_intf->debug >= DEBUG_HIGH)
		printf("TEST: 1 ret(%d)\n", test_intf->ret);
	test_intf->ret = code->irdma_post_send(test_intf->ibqp,
					       test_intf->ib_wr,
					       test_intf->bad_wr);
	if (test_intf->debug >= DEBUG_HIGH)
		printf("TEST: 2 ret(%d)\n", test_intf->ret);

	test_intf->check_behavior();
}

/*
 * post_send_roce_rc_doa
 * DOA test for ROCE post_send
 * Checks return code, SQ data, and db rung
 *
 * SQ head and tail at 0
 * No inline data
 * No immediate data
 * Only 1 WR
 * No push mode
 * No zero len Frags, avoids test mismatch with SW
 *   this will write a valid bit to an extra Frag following
 *   the 32 Byte WR. This should not happen, but has no ill
 *   effects to the HW.
 */
TEST(irdma_post_send, post_send_roce_ud_doa)
{
	code = new ProdCode;
	test_intf = new post_send_intf;
	bool passed = false;
	bool work_posted = false;

	//Set debug level and initialize test interface
	test_intf->set_debug(DEBUG_LOW);

	test_intf->init();

	//Randomization Constraints
	test_intf->constraints[POST_SEND_QP_ROCE_UD] = 1;
	test_intf->constraints[POST_SEND_NUM_WR] = 1;
	test_intf->constraints[POST_SEND_SET_OP] = 1;
	test_intf->constraints[POST_SEND_OP_VALUE] = IB_WR_SEND;
	test_intf->constraints[POST_SEND_NO_INLINE] = 1;
	test_intf->constraints[POST_SEND_NO_PUSH_MODE] = 1;
	test_intf->constraints[POST_SEND_AVOID_ZERO_LEN] = 1;
	test_intf->constraints[POST_SEND_NO_FENCE] = 1;

	test_intf->randomize();

	if (test_intf->debug >= DEBUG_HIGH)
		printf("TEST: 1 ret(%d)\n", test_intf->ret);
	test_intf->ret = code->irdma_post_send(test_intf->ibqp,
					       test_intf->ib_wr,
					       test_intf->bad_wr);
	if (test_intf->debug >= DEBUG_HIGH)
		printf("TEST: 2 ret(%d)\n", test_intf->ret);

	test_intf->check_behavior();
}

TEST(irdma_post_send, post_send_roce_ud_sge_imm)
{
	code = new ProdCode;
	test_intf = new post_send_intf;
	bool passed = false;
	bool work_posted = false;

	//Set debug level and initialize test interface
	test_intf->set_debug(DEBUG_HIGH);

	test_intf->init();

	//Randomization Constraints
	test_intf->constraints[POST_SEND_QP_ROCE_UD] = 1;
	test_intf->constraints[POST_SEND_NUM_WR] = 1;
	test_intf->constraints[POST_SEND_SET_OP] = 1;
	test_intf->constraints[POST_SEND_OP_VALUE] = IB_WR_SEND_WITH_IMM;
	test_intf->constraints[POST_SEND_NO_INLINE] = 1;
	test_intf->constraints[POST_SEND_NO_PUSH_MODE] = 1;
	test_intf->constraints[POST_SEND_AVOID_ZERO_LEN] = 1;
	test_intf->constraints[POST_SEND_NO_FENCE] = 1;

	test_intf->randomize();

	if (test_intf->debug >= DEBUG_HIGH)
		printf("TEST: 1 ret(%d)\n", test_intf->ret);
	test_intf->ret = code->irdma_post_send(test_intf->ibqp,
					       test_intf->ib_wr,
					       test_intf->bad_wr);
	if (test_intf->debug >= DEBUG_HIGH)
		printf("TEST: 2 ret(%d)\n", test_intf->ret);

	test_intf->check_behavior();
}

/*
 * post_send_iwarp_rc_inline_doa
 * Will randomize the number of inline data bytes
 * Checks return code, SQ data, and db rung
 *
 * IWARP RC connection
 * 1 WR
 * Send OP
 * No push mode
 */
TEST(irdma_post_send, post_send_iwarp_rc_inline_doa)
{
	code = new ProdCode;
	test_intf = new post_send_intf;
	bool passed = false;
	bool work_posted = false;

	//Set debug level and initialize test interface
	test_intf->set_debug(DEBUG_LOW);

	test_intf->init();

	//Randomization Constraints
	test_intf->constraints[POST_SEND_QP_IWARP_RC] = 1;
	test_intf->constraints[POST_SEND_NUM_WR] = 1;
	test_intf->constraints[POST_SEND_SET_OP] = 1;
	test_intf->constraints[POST_SEND_OP_VALUE] = IB_WR_SEND;
	test_intf->constraints[POST_SEND_INLINE] = 1;
	test_intf->constraints[POST_SEND_NO_PUSH_MODE] = 1;

	test_intf->randomize();

	if (test_intf->debug >= DEBUG_HIGH)
		printf("TEST: 1 ret(%d)\n", test_intf->ret);
	test_intf->ret = code->irdma_post_send(test_intf->ibqp,
					       test_intf->ib_wr,
					       test_intf->bad_wr);
	if (test_intf->debug >= DEBUG_HIGH)
		printf("TEST: 2 ret(%d)\n", test_intf->ret);

	test_intf->check_behavior();
}

TEST(irdma_post_send, post_send_roce_rc_inline_imm)
{
	code = new ProdCode;
	test_intf = new post_send_intf;
	bool passed = false;
	bool work_posted = false;

	//Set debug level and initialize test interface
	test_intf->set_debug(DEBUG_LOW);

	test_intf->init();

	//Randomization Constraints
	test_intf->constraints[POST_SEND_QP_ROCE_RC] = 1;
	test_intf->constraints[POST_SEND_NUM_WR] = 1;
	test_intf->constraints[POST_SEND_SET_OP] = 1;
	test_intf->constraints[POST_SEND_OP_VALUE] = IB_WR_SEND_WITH_IMM;
	test_intf->constraints[POST_SEND_INLINE] = 1;
	test_intf->constraints[POST_SEND_NO_PUSH_MODE] = 1;

	test_intf->randomize();

	if (test_intf->debug >= DEBUG_HIGH)
		printf("TEST: 1 ret(%d)\n", test_intf->ret);
	test_intf->ret = code->irdma_post_send(test_intf->ibqp,
					       test_intf->ib_wr,
					       test_intf->bad_wr);
	if (test_intf->debug >= DEBUG_HIGH)
		printf("TEST: 2 ret(%d)\n", test_intf->ret);

	test_intf->check_behavior();
}

/*
 * post_send_roce_ud_inline_doa
 * Will randomize the number of inline data bytes
 * Checks return code, SQ data, and db rung
 *
 * roce ud connection
 * 1 WR
 * Send OP
 * No push mode
 */
TEST(irdma_post_send, post_send_roce_ud_inline_doa)
{
	code = new ProdCode;
	test_intf = new post_send_intf;
	bool passed = false;
	bool work_posted = false;

	//Set debug level and initialize test interface
	test_intf->set_debug(DEBUG_LOW);

	test_intf->init();

	//Randomization Constraints
	test_intf->constraints[POST_SEND_QP_ROCE_UD] = 1;
	test_intf->constraints[POST_SEND_NUM_WR] = 1;
	test_intf->constraints[POST_SEND_SET_OP] = 1;
	test_intf->constraints[POST_SEND_OP_VALUE] = IB_WR_SEND;
	test_intf->constraints[POST_SEND_INLINE] = 1;
	test_intf->constraints[POST_SEND_NO_PUSH_MODE] = 1;
	test_intf->constraints[POST_SEND_AVOID_ZERO_LEN] = 1;
	test_intf->constraints[POST_SEND_NO_FENCE] = 1;

	test_intf->randomize();

	if (test_intf->debug >= DEBUG_HIGH)
		printf("TEST: 1 ret(%d)\n", test_intf->ret);
	test_intf->ret = code->irdma_post_send(test_intf->ibqp,
					       test_intf->ib_wr,
					       test_intf->bad_wr);
	if (test_intf->debug >= DEBUG_HIGH)
		printf("TEST: 2 ret(%d)\n", test_intf->ret);

	test_intf->check_behavior();
}

/*
 * post_send_iwarp_rc_mult
 * Tests iWARP post_send with multiple WR
 * Randomizes number of SGEs and various command bits
 * Checks return code, SQ data, and db rung
 *
 * SQ head and tail at 0
 * No inline data
 * No immediate data
 * Multiple WR
 * No push mode
 * No zero len Frags, avoids test mismatch with SW
 *   this will write a valid bit to an extra Frag following
 *   the 32 Byte WR. This should not happen, but has no ill
 *   effects to the HW.
 */
TEST(irdma_post_send, post_send_iwarp_rc_mult)
{
	code = new ProdCode;
	test_intf = new post_send_intf;
	bool passed = false;
	bool work_posted = false;

	//Set debug level and initialize test interface
	test_intf->set_debug(DEBUG_LOW);

	test_intf->init();

	//Randomization Constraints
	test_intf->constraints[POST_SEND_QP_IWARP_RC] = 1;
	test_intf->constraints[POST_SEND_NUM_WR] = 3;
	test_intf->constraints[POST_SEND_SET_OP] = 1;
	test_intf->constraints[POST_SEND_OP_VALUE] = IB_WR_SEND;
	test_intf->constraints[POST_SEND_NO_INLINE] = 1;
	test_intf->constraints[POST_SEND_NO_PUSH_MODE] = 1;
	test_intf->constraints[POST_SEND_AVOID_ZERO_LEN] = 1;

	test_intf->randomize();

	if (test_intf->debug >= DEBUG_HIGH)
		printf("TEST: 1 ret(%d)\n", test_intf->ret);
	test_intf->ret = code->irdma_post_send(test_intf->ibqp,
					       test_intf->ib_wr,
					       test_intf->bad_wr);
	if (test_intf->debug >= DEBUG_HIGH)
		printf("TEST: 2 ret(%d)\n", test_intf->ret);

	test_intf->check_behavior();
}

/*
 * post_send_roce_ud_mult
 * Tests RoCe post_send with multiple WR
 * Randomizes number of SGEs and various command bits
 * Checks return code, SQ data, and db rung
 *
 * SQ head and tail at 0
 * No inline data
 * No immediate data
 * Multiple WR
 * No push mode
 * No zero len Frags, avoids test mismatch with SW
 *   this will write a valid bit to an extra Frag following
 *   the 32 Byte WR. This should not happen, but has no ill
 *   effects to the HW.
 */
TEST(irdma_post_send, post_send_roce_ud_mult)
{
	code = new ProdCode;
	test_intf = new post_send_intf;
	bool passed = false;
	bool work_posted = false;

	//Set debug level and initialize test interface
	test_intf->set_debug(DEBUG_HIGH);

	test_intf->init();

	//Randomization Constraints
	test_intf->constraints[POST_SEND_QP_ROCE_UD] = 1;
	test_intf->constraints[POST_SEND_NUM_WR] = 3;
	test_intf->constraints[POST_SEND_SET_OP] = 1;
	test_intf->constraints[POST_SEND_OP_VALUE] = IB_WR_SEND;
	test_intf->constraints[POST_SEND_NO_INLINE] = 1;
	test_intf->constraints[POST_SEND_NO_PUSH_MODE] = 1;
	test_intf->constraints[POST_SEND_AVOID_ZERO_LEN] = 1;
	test_intf->constraints[POST_SEND_NO_FENCE] = 1;

	test_intf->randomize();

	if (test_intf->debug >= DEBUG_HIGH)
		printf("TEST: 1 ret(%d)\n", test_intf->ret);
	test_intf->ret = code->irdma_post_send(test_intf->ibqp,
					       test_intf->ib_wr,
					       test_intf->bad_wr);
	if (test_intf->debug >= DEBUG_HIGH)
		printf("TEST: 2 ret(%d)\n", test_intf->ret);

	test_intf->check_behavior();
}

/*
 * post_send_iwarp_rc_mult_mult
 * Tests iWARP post_send with multiple WR, multiple times
 * Randomizes number of SGEs and various command bits
 * Checks return code, SQ data, and db rung
 * Won't move sq_tail and will send more data
 *
 * SQ head and tail at 0
 * No inline data
 * No immediate data
 * Multiple WR
 * No push mode
 * No zero len Frags, avoids test mismatch with SW
 *   this will write a valid bit to an extra Frag following
 *   the 32 Byte WR. This should not happen, but has no ill
 *   effects to the HW.
 */
TEST(irdma_post_send, post_send_iwarp_rc_mult_mult)
{
	code = new ProdCode;
	test_intf = new post_send_intf;
	bool passed = false;
	bool work_posted = false;

	//Set debug level and initialize test interface
	test_intf->set_debug(DEBUG_HIGH);

	test_intf->init();

	//Randomization Constraints
	test_intf->constraints[POST_SEND_QP_IWARP_RC] = 1;
	test_intf->constraints[POST_SEND_NUM_WR] = 2;
	test_intf->constraints[POST_SEND_SET_OP] = 1;
	test_intf->constraints[POST_SEND_OP_VALUE] = IB_WR_SEND;
	test_intf->constraints[POST_SEND_NO_INLINE] = 1;
	test_intf->constraints[POST_SEND_NO_PUSH_MODE] = 1;
	test_intf->constraints[POST_SEND_AVOID_ZERO_LEN] = 1;

	test_intf->randomize();

	if (test_intf->debug >= DEBUG_HIGH)
		printf("TEST: 1 ret(%d)\n", test_intf->ret);
	test_intf->ret = code->irdma_post_send(test_intf->ibqp,
					       test_intf->ib_wr,
					       test_intf->bad_wr);
	if (test_intf->debug >= DEBUG_HIGH)
		printf("TEST: 2 ret(%d)\n", test_intf->ret);

	test_intf->check_behavior();

	test_intf->re_rand();

	if (test_intf->debug >= DEBUG_HIGH)
		printf("TEST: 3 ret(%d)\n", test_intf->ret);
	test_intf->ret = code->irdma_post_send(test_intf->ibqp,
					       test_intf->ib_wr,
					       test_intf->bad_wr);
	if (test_intf->debug >= DEBUG_HIGH)
		printf("TEST: 4 ret(%d)\n", test_intf->ret);

	// may fail to see that db was not written as might have old id
	// may be able to get around this by re-randomizing the qp_id
	//   would just be for this test as not yet available
	test_intf->check_behavior();
}

/*
 * post_send_roce_ud_mult_mult
 * Tests RoCE post_send with multiple WR, multiple times
 * Randomizes number of SGEs and various command bits
 * Checks return code, SQ data, and db rung
 * Won't move sq_tail and will send more data
 *
 * SQ head and tail at 0
 * No inline data
 * No immediate data
 * Multiple WR
 * No push mode
 * No zero len Frags, avoids test mismatch with SW
 *   this will write a valid bit to an extra Frag following
 *   the 32 Byte WR. This should not happen, but has no ill
 *   effects to the HW.
 */
TEST(irdma_post_send, post_send_roce_ud_mult_mult)
{
	code = new ProdCode;
	test_intf = new post_send_intf;
	bool passed = false;
	bool work_posted = false;

	//Set debug level and initialize test interface
	test_intf->set_debug(DEBUG_LOW);

	test_intf->init();

	//Randomization Constraints
	test_intf->constraints[POST_SEND_QP_ROCE_UD] = 1;
	test_intf->constraints[POST_SEND_NUM_WR] = 2;
	test_intf->constraints[POST_SEND_SET_OP] = 1;
	test_intf->constraints[POST_SEND_OP_VALUE] = IB_WR_SEND;
	test_intf->constraints[POST_SEND_INLINE] = 1;
	test_intf->constraints[POST_SEND_NO_PUSH_MODE] = 1;
	test_intf->constraints[POST_SEND_AVOID_ZERO_LEN] = 1;
	test_intf->constraints[POST_SEND_NO_FENCE] = 1;

	test_intf->randomize();

	if (test_intf->debug >= DEBUG_HIGH)
		printf("TEST: 1 ret(%d)\n", test_intf->ret);
	test_intf->ret = code->irdma_post_send(test_intf->ibqp,
					       test_intf->ib_wr,
					       test_intf->bad_wr);
	if (test_intf->debug >= DEBUG_HIGH)
		printf("TEST: 2 ret(%d)\n", test_intf->ret);

	test_intf->check_behavior();

	test_intf->re_rand();

	if (test_intf->debug >= DEBUG_HIGH)
		printf("TEST: 3 ret(%d)\n", test_intf->ret);
	test_intf->ret = code->irdma_post_send(test_intf->ibqp,
					       test_intf->ib_wr,
					       test_intf->bad_wr);
	if (test_intf->debug >= DEBUG_HIGH)
		printf("TEST: 4 ret(%d)\n", test_intf->ret);

	// may fail to see that db was not written as might have old id
	// may be able to get around this by re-randomizing the qp_id
	//   would just be for this test as not yet available
	test_intf->check_behavior();
}

TEST(irdma_post_send, post_send_iwarp_rc_mult_mult_rand_op)
{
	code = new ProdCode;
	test_intf = new post_send_intf;
	bool passed = false;
	bool work_posted = false;

	//Set debug level and initialize test interface
	test_intf->set_debug(DEBUG_HIGH);

	test_intf->init();

	//Randomization Constraints
	test_intf->constraints[POST_SEND_QP_IWARP_RC] = 1;
	test_intf->constraints[POST_SEND_NUM_WR] = 2;
//	test_intf->constraints[POST_SEND_OP_VALUE] = IB_WR_SEND;
//	test_intf->constraints[POST_SEND_NO_INLINE] = 1;
	test_intf->constraints[POST_SEND_NO_PUSH_MODE] = 1;
	test_intf->constraints[POST_SEND_AVOID_ZERO_LEN] = 1;

	test_intf->randomize();

	if (test_intf->debug >= DEBUG_HIGH)
		printf("TEST: 1 ret(%d)\n", test_intf->ret);
	test_intf->ret = code->irdma_post_send(test_intf->ibqp,
					       test_intf->ib_wr,
					       test_intf->bad_wr);
	if (test_intf->debug >= DEBUG_HIGH)
		printf("TEST: 2 ret(%d)\n", test_intf->ret);

	test_intf->check_behavior();

	test_intf->re_rand();

	if (test_intf->debug >= DEBUG_HIGH)
		printf("TEST: 3 ret(%d)\n", test_intf->ret);
	test_intf->ret = code->irdma_post_send(test_intf->ibqp,
					       test_intf->ib_wr,
					       test_intf->bad_wr);
	if (test_intf->debug >= DEBUG_HIGH)
		printf("TEST: 4 ret(%d)\n", test_intf->ret);

	// may fail to see that db was not written as might have old id
	// may be able to get around this by re-randomizing the qp_id
	//   would just be for this test as not yet available
	test_intf->check_behavior();
}

TEST(irdma_post_send, post_send_roce_rc_mult_mult_rand_op)
{
	code = new ProdCode;
	test_intf = new post_send_intf;
	bool passed = false;
	bool work_posted = false;

	//Set debug level and initialize test interface
	test_intf->set_debug(DEBUG_HIGH);

	test_intf->init();

	//Randomization Constraints
	test_intf->constraints[POST_SEND_QP_ROCE_RC] = 1;
	test_intf->constraints[POST_SEND_NUM_WR] = 2;
	test_intf->constraints[POST_SEND_NO_PUSH_MODE] = 1;
	test_intf->constraints[POST_SEND_AVOID_ZERO_LEN] = 1;

	test_intf->randomize();

	if (test_intf->debug >= DEBUG_HIGH)
		printf("TEST: 1 ret(%d)\n", test_intf->ret);
	test_intf->ret = code->irdma_post_send(test_intf->ibqp,
					       test_intf->ib_wr,
					       test_intf->bad_wr);
	if (test_intf->debug >= DEBUG_HIGH)
		printf("TEST: 2 ret(%d)\n", test_intf->ret);

	test_intf->check_behavior();

	test_intf->re_rand();

	if (test_intf->debug >= DEBUG_HIGH)
		printf("TEST: 3 ret(%d)\n", test_intf->ret);
	test_intf->ret = code->irdma_post_send(test_intf->ibqp,
					       test_intf->ib_wr,
					       test_intf->bad_wr);
	if (test_intf->debug >= DEBUG_HIGH)
		printf("TEST: 4 ret(%d)\n", test_intf->ret);

	// may fail to see that db was not written as might have old id
	// may be able to get around this by re-randomizing the qp_id
	//   would just be for this test as not yet available
	test_intf->check_behavior();
}

/*
 * post_send_wrap_sq
 * test for iWARP post_send to wrap sq
 * Randomizes number of SGEs and various command bits
 * Checks return code, SQ data, and db rung
 *
 * SQ head and tail at 0
 * No inline data
 * No immediate data
 * Only 1 WR
 * No push mode
 * No zero len Frags, avoids test mismatch with SW
 *   this will write a valid bit to an extra Frag following
 *   the 32 Byte WR. This should not happen, but has no ill
 *   effects to the HW.
 */
TEST(irdma_post_send, post_send_wrap_sq)
{
	code = new ProdCode;
	test_intf = new post_send_intf;
	bool passed = false;
	bool work_posted = false;

	//Set debug level and initialize test interface
	test_intf->set_debug(DEBUG_LOW);

	test_intf->init();

	//Randomization Constraints
	test_intf->constraints[POST_SEND_QP_IWARP_RC] = 1;
	test_intf->constraints[POST_SEND_NUM_WR] = 1;
	test_intf->constraints[POST_SEND_NUM_SGES] = 1;
	test_intf->constraints[POST_SEND_SET_OP] = 1;
	test_intf->constraints[POST_SEND_OP_VALUE] = IB_WR_SEND;
	test_intf->constraints[POST_SEND_NO_INLINE] = 1;
	test_intf->constraints[POST_SEND_NO_PUSH_MODE] = 1;
	test_intf->constraints[POST_SEND_AVOID_ZERO_LEN] = 1;
	test_intf->constraints[POST_SEND_NO_FLUSH] = 1;

	test_intf->randomize();

	for (int i=0; i<(3+(2*test_intf->iwqp->sc_qp.qp_uk.sq_ring.size)); i++) {

		if (test_intf->debug >= DEBUG_HIGH)
			printf("TEST: 1 ret(%d)\n", test_intf->ret);
		test_intf->ret = code->irdma_post_send(test_intf->ibqp,
						       test_intf->ib_wr,
						       test_intf->bad_wr);
		if (test_intf->debug >= DEBUG_HIGH)
			printf("TEST: 2 ret(%d)\n", test_intf->ret);

		test_intf->check_behavior();

		test_intf->iwqp->sc_qp.qp_uk.sq_ring.tail =
		  test_intf->iwqp->sc_qp.qp_uk.sq_ring.head;

		test_intf->tb_sqr.tail = test_intf->tb_sqr.head;

		test_intf->re_rand();
	}
}

TEST(irdma_post_send, post_send_wrap_sq_rand_quanta)
{
	code = new ProdCode;
	test_intf = new post_send_intf;
	bool passed = false;
	bool work_posted = false;

	//Set debug level and initialize test interface
	test_intf->set_debug(DEBUG_LOW);

	test_intf->init();

	//Randomization Constraints
	test_intf->constraints[POST_SEND_QP_IWARP_RC] = 1;
	test_intf->constraints[POST_SEND_NUM_WR] = 1;
	test_intf->constraints[POST_SEND_SET_OP] = 1;
	test_intf->constraints[POST_SEND_OP_VALUE] = IB_WR_SEND;
	test_intf->constraints[POST_SEND_NO_INLINE] = 1;
	test_intf->constraints[POST_SEND_NO_PUSH_MODE] = 1;
	test_intf->constraints[POST_SEND_AVOID_ZERO_LEN] = 1;
	test_intf->constraints[POST_SEND_NO_FLUSH] = 1;

	test_intf->randomize();

	for (int i=0; i<(3+(2*test_intf->iwqp->sc_qp.qp_uk.sq_ring.size)); i++) {

		if (test_intf->debug >= DEBUG_HIGH)
			printf("TEST: 1 ret(%d)\n", test_intf->ret);
		test_intf->ret = code->irdma_post_send(test_intf->ibqp,
						       test_intf->ib_wr,
						       test_intf->bad_wr);
		if (test_intf->debug >= DEBUG_HIGH)
			printf("TEST: 2 ret(%d)\n", test_intf->ret);

		test_intf->check_behavior();

		test_intf->iwqp->sc_qp.qp_uk.sq_ring.tail =
		  test_intf->iwqp->sc_qp.qp_uk.sq_ring.head;

		test_intf->tb_sqr.tail = test_intf->tb_sqr.head;

		test_intf->re_rand();
	}
}

//---------------------------------------------------------------------

TEST_GROUP(cpp_test_code)
{
	ProdCode *code;
	/* setup should only initialize things, not allocate them */
	void setup()
	{
		long seed = 0;
		timespec ts;

		//Setup Randomization
		clock_gettime(CLOCK_REALTIME, &ts);
		seed = ts.tv_nsec;
		printf("\n\tSeed = 0x%llx\n", seed);
		srand(seed);
	}

	void teardown()
	{
		mock().checkExpectations();
		mock().clear();
		if (code)
			delete code;
	}
};

/*
 * test_code_1
 * see that pointers are being used correctly
 * 1) create 2D array to hold inline data per WR
 * 2) give address of single WR inline data to addr
 * 3) iterate through data and assign to field data
 */
TEST(cpp_test_code, test_code_1)
{
	u16 num_wrs = 9;
	u16 num_data_bytes = 9; //want to be more than 64 bytes
	u8 **array;
	u64 **array64;
	u64 addr;
	u64 addr64;
	u8 *ptr;
	u64 *ptr64;
	u16 ptr_idx = 0;
	u64 field_1;
	u64 field_2;
	u64 field64_1;
	u64 field64_2;
	code = new ProdCode;

	// dynamically create array
	array = new u8*[num_wrs];
	for (int i=0; i<num_wrs; i++)
		array[i] = new u8[num_data_bytes];

	array64 = new u64*[num_wrs];
	for (int i=0; i<num_wrs; i++)
		array64[i] = new u64[num_data_bytes];
	// connect array by assigning address to addr var
	addr = &array[0][0];
	addr64 = &array64[0][0];

	// randomize data held in array
	for (int i=0; i<num_wrs; i++)
		for(int j=0; j<num_data_bytes; j++) {
			array[i][j] = rand(); //want to see if overwrite occurs
			printf("array[%d][%d]=0x%x\n", i, j, array[i][j]);
		}

	for (int i=0; i<num_wrs; i++)
		for(int j=0; j<num_data_bytes; j++) {
			array64[i][j] = (u64) rand() << 32 | rand(); //want to see if overwrite occurs
			printf("array64[%d][%d]=0x%llx\n", i, j, array64[i][j]);
		}
	// now try to place data in field via addr var
	ptr = (u8*) addr;

	field_1 = (u64) ptr[ptr_idx];
	ptr_idx += 1;
	field_2 = (u64) ptr[ptr_idx];
	printf("field_1=0x%llx\n", field_1);
	printf("field_2=0x%llx\n", field_2);

	ptr_idx=0;
	ptr64 = (u64*) addr64;
	int i = 1;

	field64_1 = *ptr64;
	ptr_idx += 1;
	ptr64 = (u64*) (addr64 + ptr_idx);
	field64_2 = *ptr64;
	printf("field64_1=0x%llx\n", field64_1);
	printf("field64_2=0x%llx\n", field64_2);

	// clean up dynamically allocated array
	for (int i=0; i<num_wrs; i++)
		delete [] array[i];
	delete [] array;

	for (int i=0; i<num_wrs; i++)
		delete [] array64[i];
	delete [] array64;
}

struct anon_wr {
	union {
		struct ib_rdma_wr rdma_wr;
		struct ib_atomic_wr atomic_wr;
		struct ib_ud_wr ud_wr;
	} wr_type;
};

TEST(cpp_test_code, test_code_2)
{
	struct anon_wr wrs[10];

	printf("size of anon_wr(%d)\n", sizeof(struct anon_wr));
	printf("size of anon_wr[10](%d)\n", 10*sizeof(struct anon_wr));
	printf("addr of anon_wr[0](%p)\n", wrs);
	printf("addr of anon_wr[1](%p)\n", &wrs[1]);
}
// -----------------------------------------------------------------------------

int main(int ac, char **av)
{
	return CommandLineTestRunner::RunAllTests(ac, av);
}
