/* Copyright (c) 2015 - 2023 Intel Corporation */
#ifndef LICENSE_REPLACED
/* For licensing information, see the file 'LICENSE' in the root folder */
#endif /* LICENSE_REPLACED */
#ifndef IRDMA_VERBS_H
#define IRDMA_VERBS_H

#ifndef EXTERNAL_RELEASE
/* PORT TBD: this file should likely be duplicated since it is largely ib_ specific? */

/* Not clear if we really need 4 entries per MR. Only 1 entry seems used to store
 * starting phy pg addr
 */
#endif
#define IRDMA_MAX_SAVED_PHY_PGADDR	4
#define IRDMA_FLUSH_DELAY_MS		20

#define IRDMA_PKEY_TBL_SZ		1
#define IRDMA_DEFAULT_PKEY		0xFFFF

#ifdef MEV_SUPPORT
#define IRDMA_QPS_PER_PUSH_PAGE 16
#define IRDMA_PUSH_WIN_SIZE 256

#endif
#ifndef EXTERNAL_RELEASE
/* Extra page for DB shadow area folded in WQ memory allocated in user-space */
#endif
#define IRDMA_SHADOW_PGCNT		1

#define iwdev_to_idev(iwdev)	(&(iwdev)->rf->sc_dev)

struct irdma_ucontext {
#ifndef HPM_ESXI
	struct ib_ucontext ibucontext;
#endif
	struct irdma_device *iwdev;
#ifdef RDMA_MMAP_DB_SUPPORT
	struct rdma_user_mmap_entry *db_mmap_entry;
#else
	struct irdma_user_mmap_entry *db_mmap_entry;
	DECLARE_HASHTABLE(mmap_hash_tbl, 6);
	spinlock_t mmap_tbl_lock; /* protect mmap hash table entries */
#endif
	struct list_head cq_reg_mem_list;
	spinlock_t cq_reg_mem_list_lock; /* protect CQ memory list */
	struct list_head qp_reg_mem_list;
	spinlock_t qp_reg_mem_list_lock; /* protect QP memory list */
#ifdef MEV_SUPPORT
	struct list_head srq_reg_mem_list;
	spinlock_t srq_reg_mem_list_lock; /* protect SRQ memory list */
#endif
#ifdef DEBUG_SUPPORT
#ifdef CONFIG_DEBUG_FS
	struct list_head uctx_list;
#endif
#endif
#ifndef UPSTREAM_RELEASE
	/* FIXME: Move to kcompat ideally. Used < 4.20.0 for old diassasscoaite flow */
	struct list_head vma_list;
	struct mutex vma_list_mutex; /* protect the vma_list */
#endif
	int abi_ver;
	bool legacy_mode:1;
	bool use_raw_attrs:1;
};

struct irdma_pd {
	struct ib_pd ibpd;
	struct irdma_sc_pd sc_pd;
#ifdef MEV_SUPPORT
	struct mutex push_alloc_mutex; /* protect push page alloc within a PD*/
	DECLARE_BITMAP(push_offset_bmap, IRDMA_QPS_PER_PUSH_PAGE);
	u32 push_idx;
	u16 qs_handle;
#endif
#ifdef UD_QS_ASSIGN_WA
	struct list_head udqp_list;
	spinlock_t udqp_list_lock;
#endif
};

union irdma_sockaddr {
#ifdef HPM_ESXI
	struct sockaddr saddr;
#endif
	struct sockaddr_in saddr_in;
	struct sockaddr_in6 saddr_in6;
};

struct irdma_av {
	u8 macaddr[16];
#ifdef HPM_ESXI
	struct ib_ah_attr attrs;
#else
	struct rdma_ah_attr attrs;
#endif /* HPM_ESXI */
	union irdma_sockaddr sgid_addr;
	union irdma_sockaddr dgid_addr;
	u8 net_type;
};

struct irdma_ah {
	struct ib_ah ibah;
	struct irdma_sc_ah sc_ah;
	struct irdma_pd *pd;
	struct irdma_av av;
	u8 sgid_index;
	union ib_gid dgid;
#ifdef AH_REUSE
	struct hlist_node list;
	refcount_t refcnt;
	struct irdma_ah *parent_ah;	/* AH from cached list */
#endif
};

struct irdma_hmc_pble {
	union {
		u32 idx;
		dma_addr_t addr;
	};
};

struct irdma_cq_mr {
	struct irdma_hmc_pble cq_pbl;
	dma_addr_t shadow;
	bool split;
};

#ifdef MEV_SUPPORT
struct irdma_srq_mr {
	struct irdma_hmc_pble srq_pbl;
	dma_addr_t shadow;
};

#endif
struct irdma_qp_mr {
	struct irdma_hmc_pble sq_pbl;
	struct irdma_hmc_pble rq_pbl;
	dma_addr_t shadow;
#ifdef MEV_SUPPORT
	dma_addr_t rq_pa;
#endif
	struct page *sq_page;
};

struct irdma_cq_buf {
	struct irdma_dma_mem kmem_buf;
	struct irdma_cq_uk cq_uk;
	struct irdma_hw *hw;
	struct list_head list;
	struct work_struct work;
};

struct irdma_pbl {
	struct list_head list;
	union {
		struct irdma_qp_mr qp_mr;
		struct irdma_cq_mr cq_mr;
#ifdef MEV_SUPPORT
		struct irdma_srq_mr srq_mr;
#endif
	};

	bool pbl_allocated:1;
	bool on_list:1;
#ifndef HPM_ESXI
	u64 user_base;
#else
	union {
		u64 user_base;
		vmk_uint64 *frwrPageList; // Stack allocated page lis
	};
#endif
	struct irdma_pble_alloc pble_alloc;
	struct irdma_mr *iwmr;
};

struct irdma_mr {
	union {
		struct ib_mr ibmr;
		struct ib_mw ibmw;
	};
#ifndef HPM_ESXI
	struct ib_umem *region;
#else
	union {
		struct irdma_pd *iwpd;
		vmk_RDMAFastRegPageList frwr;
	};
#endif
	int access;
	u8 is_hwreg;
	u16 type;
	u32 page_cnt;
	u64 page_size;
#ifndef UPSTREAM_RELEASE
	u64 page_msk;
#endif
	u32 npages;
	u32 stag;
	u64 len;
	u64 pgaddrmem[IRDMA_MAX_SAVED_PHY_PGADDR];
#ifdef DEBUG_SUPPORT
#ifdef CONFIG_DEBUG_FS
	u64 level0_pa;
#endif
#endif
	struct irdma_pbl iwpbl;
};

#ifdef MEV_SUPPORT
struct irdma_srq {
	struct ib_srq ibsrq;
	struct irdma_sc_srq sc_srq;
	struct irdma_dma_mem kmem;
#if !defined(HPM_ESXI) && defined(MEV_SRQ_TABLE)
	struct completion free_srq;
#endif /* !defined(HPM_ESXI) && defined(MEV_SRQ_TABLE) */
	u64 *srq_wrid_mem;
	refcount_t refcnt;
	spinlock_t lock; /* for poll srq */
	struct irdma_pbl *iwpbl;
	struct irdma_sge *sg_list;
	u16 srq_head;
	u32 srq_num;
	u32 max_wr;
	bool user_mode:1;
};

#endif
struct irdma_cq {
	struct ib_cq ibcq;
	struct irdma_sc_cq sc_cq;
	u16 cq_head;
	u16 cq_size;
#ifndef EXTERNAL_RELEASE
	/* Maybe cq_id ?? */
#endif
	u16 cq_num;
	bool user_mode;
	atomic_t armed;
	enum irdma_cmpl_notify last_notify;
	u32 polled_cmpls;
	u32 cq_mem_size;
#ifndef EXTERNAL_RELEASE
	/* Kmode sq ring */
#endif
	struct irdma_dma_mem kmem;
	struct irdma_dma_mem kmem_shadow;
#ifndef HPM_ESXI
	struct completion free_cq;
#endif
	refcount_t refcnt;
	spinlock_t lock; /* for poll cq */
	struct irdma_pbl *iwpbl;
	struct irdma_pbl *iwpbl_shadow;
	struct list_head resize_list;
	struct irdma_cq_poll_info cur_cqe;
	struct list_head cmpl_generated;
};

#ifndef HPM_ESXI
struct irdma_cmpl_gen {
	struct list_head list;
	struct irdma_cq_poll_info cpi;
};

#endif
struct disconn_work {
	struct work_struct work;
	struct irdma_qp *iwqp;
};

struct if_notify_work {
	struct work_struct work;
	struct irdma_device *iwdev;
	u32 ipaddr[4];
	u16 vlan_id;
	bool ipv4:1;
	bool ifup:1;
};

struct iw_cm_id;

struct irdma_qp_kmode {
#ifndef EXTERNAL_RELEASE
	/* kmod sq,rq, and shadow */
#endif
	struct irdma_dma_mem dma_mem;
#ifndef EXTERNAL_RELEASE
	/* This is for wrid arrays */
#endif
#if defined(CPK_CQE_WA) || defined(CPK_CHECK_MISSING_SQ_CMPL)
	u32 *sig_trk_mem;
#endif
	struct irdma_sq_uk_wr_trk_info *sq_wrid_mem;
	u64 *rq_wrid_mem;
};

struct irdma_qp {
	struct ib_qp ibqp;
	struct irdma_sc_qp sc_qp;
	struct irdma_device *iwdev;
	struct irdma_cq *iwscq;
	struct irdma_cq *iwrcq;
	struct irdma_pd *iwpd;
#ifdef RDMA_MMAP_DB_SUPPORT
	struct rdma_user_mmap_entry *push_wqe_mmap_entry;
	struct rdma_user_mmap_entry *push_db_mmap_entry;
#else
	struct irdma_user_mmap_entry *push_wqe_mmap_entry;
	struct irdma_user_mmap_entry *push_db_mmap_entry;
#endif
	struct irdma_qp_host_ctx_info ctx_info;
	union {
		struct irdma_iwarp_offload_info iwarp_info;
		struct irdma_roce_offload_info roce_info;
	};

	union {
		struct irdma_tcp_offload_info tcp_info;
		struct irdma_udp_offload_info udp_info;
	};

	struct irdma_ah roce_ah;
	struct list_head teardown_entry;
#ifdef UD_QS_ASSIGN_WA
	struct list_head ud_list_elem;
#endif
	refcount_t refcnt;
	struct iw_cm_id *cm_id;
	struct irdma_cm_node *cm_node;
#ifndef HPM_ESXI
	struct delayed_work dwork_flush;
	struct ib_mr *lsmm_mr;
#else
	struct irdma_mr *lsmm_mr;
#endif
	atomic_t hw_mod_qp_pend;
	enum ib_qp_state ibqp_state;
	u32 qp_mem_size;
	u32 last_aeq;
	int max_send_wr;
	int max_recv_wr;
#ifdef HPM_ESXI
	u32 qp_num;
	u32 qp_vmk_type;
#endif
	atomic_t close_timer_started;
	spinlock_t lock; /* serialize posting WRs to SQ/RQ */
#ifdef HPM_FBSD
	spinlock_t dwork_flush_lock; /* protect mod_delayed_work */
#endif
	struct irdma_qp_context *iwqp_context;
	void *pbl_vbase;
	dma_addr_t pbl_pbase;
	struct page *page;
	u8 iwarp_state;
	u16 term_sq_flush_code;
	u16 term_rq_flush_code;
	u8 hw_iwarp_state;
	u8 hw_tcp_state;
	struct irdma_qp_kmode kqp;
	struct irdma_dma_mem host_ctx;
#ifdef TERM_SUPPORT
	struct timer_list terminate_timer;
#endif
	struct irdma_pbl *iwpbl;
	struct irdma_sge *sg_list;
	struct irdma_dma_mem q2_ctx_mem;
	struct irdma_dma_mem ietf_mem;
#ifndef HPM_ESXI
	struct completion free_qp;
#else
	atomic_t disconn_counter;
	vmk_WorldID disconn_wrld;
	vmk_WorldID rfl_wrld;
#endif
	wait_queue_head_t waitq;
	wait_queue_head_t mod_qp_waitq;
	u8 rts_ae_rcvd;
#if defined(MEV_SUPPORT) && defined(MEV_HVL)
	struct irdma_mr *iwmr;
#endif
	bool active_conn:1;
	bool user_mode:1;
	bool hte_added:1;
	bool flush_issued:1;
	bool sig_all:1;
	bool pau_mode:1;
	bool suspend_pending:1;
#ifdef HPM_ESXI
	bool force_lpb:1;
#endif
};

#ifdef UD_QS_ASSIGN_WA
struct irdma_udqs_work {
	struct work_struct work;
	struct irdma_qp *iwqp;
	u8 user_prio;
	bool qs_change:1;
};

#endif
enum irdma_mmap_flag {
	IRDMA_MMAP_IO_NC,
	IRDMA_MMAP_IO_WC,
};

struct irdma_user_mmap_entry {
#ifdef RDMA_MMAP_DB_SUPPORT
	struct rdma_user_mmap_entry rdma_entry;
#else
	struct irdma_ucontext *ucontext;
	struct hlist_node hlist;
	u64 pgoff_key; /* Used to compute offset (in bytes) returned to user libc's mmap */
#endif
	u64 bar_offset;
	u8 mmap_flag;
};

#if defined(MEV_SUPPORT) && defined(MEV_HVL)
/* Adding CRT specific extensions for CRT protocol family.
 **/
enum crt_mtu {
	CRT_MTU_256  = 1,
	CRT_MTU_512  = 2,
	CRT_MTU_1024 = 3,
	CRT_MTU_2048 = 4,
	CRT_MTU_4096 = 5,
	CRT_MTU_8192 = 6
};

static inline int crt_mtu_enum_to_int(enum crt_mtu mtu)
{
	switch (mtu) {
	case CRT_MTU_256:  return  256;
	case CRT_MTU_512:  return  512;
	case CRT_MTU_1024: return 1024;
	case CRT_MTU_2048: return 2048;
	case CRT_MTU_4096: return 4096;
	case CRT_MTU_8192: return 8192;
	default:	   return -1;
	}
}

static inline enum crt_mtu crt_mtu_int_to_enum(int mtu)
{
	if (mtu >= 8192)
		return CRT_MTU_8192;
	if (mtu >= 4096)
		return CRT_MTU_4096;
	if (mtu >= 2048)
		return CRT_MTU_2048;
	if (mtu >= 1024)
		return CRT_MTU_1024;
	if (mtu >= 512)
		return CRT_MTU_512;
	return CRT_MTU_256;
}

#endif
static inline u16 irdma_fw_major_ver(struct irdma_sc_dev *dev)
{
	return (u16)FIELD_GET(IRDMA_FW_VER_MAJOR, dev->feature_info[IRDMA_FEATURE_FW_INFO]);
}

static inline u16 irdma_fw_minor_ver(struct irdma_sc_dev *dev)
{
	return (u16)FIELD_GET(IRDMA_FW_VER_MINOR, dev->feature_info[IRDMA_FEATURE_FW_INFO]);
}

#ifndef HPM_ESXI
static inline void set_ib_wc_op_sq(struct irdma_cq_poll_info *cq_poll_info,
				   struct ib_wc *entry)
{
	struct irdma_sc_qp *qp;

	switch (cq_poll_info->op_type) {
	case IRDMA_OP_TYPE_RDMA_WRITE:
	case IRDMA_OP_TYPE_RDMA_WRITE_SOL:
		entry->opcode = IB_WC_RDMA_WRITE;
		break;
	case IRDMA_OP_TYPE_RDMA_READ_INV_STAG:
	case IRDMA_OP_TYPE_RDMA_READ:
		entry->opcode = IB_WC_RDMA_READ;
		break;
	case IRDMA_OP_TYPE_SEND_SOL:
	case IRDMA_OP_TYPE_SEND_SOL_INV:
	case IRDMA_OP_TYPE_SEND_INV:
	case IRDMA_OP_TYPE_SEND:
		entry->opcode = IB_WC_SEND;
		break;
	case IRDMA_OP_TYPE_FAST_REG_NSMR:
		entry->opcode = IB_WC_REG_MR;
		break;
#ifdef MEV_SUPPORT
	case IRDMA_OP_TYPE_ATOMIC_COMPARE_AND_SWAP:
		entry->opcode = IB_WC_COMP_SWAP;
		break;
	case IRDMA_OP_TYPE_ATOMIC_FETCH_AND_ADD:
		entry->opcode = IB_WC_FETCH_ADD;
		break;
#endif
	case IRDMA_OP_TYPE_INV_STAG:
		entry->opcode = IB_WC_LOCAL_INV;
		break;
	default:
		qp = cq_poll_info->qp_handle;
		ibdev_err(to_ibdev(qp->dev), "Invalid opcode = %d in CQE\n",
			  cq_poll_info->op_type);
		entry->status = IB_WC_GENERAL_ERR;
	}
}

#ifdef MEV_SUPPORT
static inline void set_ib_wc_op_rq_gen_3(struct irdma_cq_poll_info *cq_poll_info,
					 struct ib_wc *entry)
{
	switch (cq_poll_info->op_type) {
	case IRDMA_OP_TYPE_RDMA_WRITE:
	case IRDMA_OP_TYPE_RDMA_WRITE_SOL:
		entry->opcode = IB_WC_RECV_RDMA_WITH_IMM;
		break;
	default:
#ifndef EXTERNAL_RELEASE
		/* op_type may be unexpected, assume RECV unless it's a WRITE. */
#endif
		entry->opcode = IB_WC_RECV;
	}
}

#endif /* MEV_SUPPORT */
static inline void set_ib_wc_op_rq(struct irdma_cq_poll_info *cq_poll_info,
				   struct ib_wc *entry, bool send_imm_support)
{
	/**
	 * iWARP does not support sendImm, so the presence of Imm data
	 * must be WriteImm.
	 */
	if (!send_imm_support) {
		entry->opcode = cq_poll_info->imm_valid ?
				IB_WC_RECV_RDMA_WITH_IMM :
				IB_WC_RECV;
		return;
	}
#ifndef EXTERNAL_RELEASE
	/* On CVL the HW writes the IB opcode from the received packet */
#endif
	switch (cq_poll_info->op_type) {
	case IB_OPCODE_RDMA_WRITE_ONLY_WITH_IMMEDIATE:
	case IB_OPCODE_RDMA_WRITE_LAST_WITH_IMMEDIATE:
		entry->opcode = IB_WC_RECV_RDMA_WITH_IMM;
		break;
	default:
		entry->opcode = IB_WC_RECV;
	}
}

#endif /* HPM_ESXI */
/**
 * irdma_mcast_mac_v4 - Get the multicast MAC for an IP address
 * @ip_addr: IPv4 address
 * @mac: pointer to result MAC address
 *
 */
static inline void irdma_mcast_mac_v4(u32 *ip_addr, u8 *mac)
{
	u8 *ip = (u8 *)ip_addr;
	unsigned char mac4[ETH_ALEN] = {0x01, 0x00, 0x5E, ip[2] & 0x7F, ip[1],
					ip[0]};

	ether_addr_copy(mac, mac4);
}

/**
 * irdma_mcast_mac_v6 - Get the multicast MAC for an IP address
 * @ip_addr: IPv6 address
 * @mac: pointer to result MAC address
 *
 */
static inline void irdma_mcast_mac_v6(u32 *ip_addr, u8 *mac)
{
	u8 *ip = (u8 *)ip_addr;
	unsigned char mac6[ETH_ALEN] = {0x33, 0x33, ip[3], ip[2], ip[1], ip[0]};

	ether_addr_copy(mac, mac6);
}

#ifdef KMODE_PUSH_PAGE
static inline void irdma_remove_kmode_push_pages(struct irdma_qp *iwqp)
{

	if (iwqp->sc_qp.qp_uk.push_wqe)
		iounmap(iwqp->sc_qp.qp_uk.push_wqe);

	if (iwqp->sc_qp.qp_uk.push_db)
		iounmap(iwqp->sc_qp.qp_uk.push_db);
}
#endif /* KMODE_PUSH_PAGE */

#ifndef UPSTREAM_RELEASE
#ifdef ALLOC_HW_STATS_STRUCT_V2
extern const struct rdma_stat_desc irdma_hw_stat_descs[];

#endif /* ALLOC_HW_STATS_STRUCT_V2 */
#ifdef RDMA_MMAP_DB_SUPPORT
struct rdma_user_mmap_entry*
irdma_user_mmap_entry_insert(struct irdma_ucontext *ucontext, u64 bar_offset,
			     enum irdma_mmap_flag mmap_flag, u64 *mmap_offset);
#else
struct irdma_user_mmap_entry *
irdma_user_mmap_entry_add_hash(struct irdma_ucontext *ucontext, u64 bar_offset,
			       enum irdma_mmap_flag mmap_flag, u64 *mmap_offset);
void irdma_user_mmap_entry_del_hash(struct irdma_user_mmap_entry *entry);
#endif /* RDMA_MMAP_DB_SUPPORT */
#endif /* UPSTREAM_RELEASE */
#ifndef HPM_ESXI
int irdma_ib_register_device(struct irdma_device *iwdev);
void irdma_ib_unregister_device(struct irdma_device *iwdev);
#endif
void irdma_ib_qp_event(struct irdma_qp *iwqp, enum irdma_qp_event_type event);
#ifndef HPM_ESXI
void irdma_generate_flush_completions(struct irdma_qp *iwqp);
void irdma_remove_cmpls_list(struct irdma_cq *iwcq);
int irdma_generated_cmpls(struct irdma_cq *iwcq, struct irdma_cq_poll_info *cq_poll_info);
void irdma_sched_qp_flush_work(struct irdma_qp *iwqp);
#ifndef UPSTREAM_RELEASE
void irdma_flush_worker(struct work_struct *work);
#endif
#endif /* HPM_ESXI */
#if defined(MEV_SUPPORT) && defined(MEV_HVL)
struct ib_mr *wa_reg_phys_mr(struct ib_pd *pd);
#endif
#endif /* IRDMA_VERBS_H */
